/*
** clock.c - Test for programmable clock generators; user can specify a
**           clock generator to program, and the application (approximately)
**           measures the actual frequency.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**
** (C) Copyright 2008-2009 Alpha Data
**
** Cycle counters map to clock generators on a per-model basis as
** follows:
**
** Model            Cycle counter index      Clock generator index
** -------------------------------------------------------------------
**
** ADM-XRC          LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**
** ADM-XRC-P        LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**
** ADM-XRC-II-Lite  LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**
** ADM-XRC-II       LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**
** ADM-XPL          LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   LCLK * 2 (0)
**                  REFCLK pin    (3)   =>   N/A (fixed @ 125 MHz)
**
** ADM-XP           LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   LCLK * 2 (0)
**                  REFCLK pin    (3)   =>   N/A (fixed @ 125 MHz)
**
** ADP-XPI          LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   LCLK * 2 (0)
**                  REFCLK pin    (3)   =>   N/A (fixed @ 125 MHz)
**
** ADP-WRC-II       LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**
** ADP-DRC-II       LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**                  DDRCLK pin    (3)   =>   N/A (fixed @ 66 MHz)
**
** ADM-XRC-4LX      LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**                  REFCLK pin    (3)   =>   N/A (fixed @ 200 MHz)
**
** ADM-XRC-4SX      LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**                  REFCLK pin    (3)   =>   N/A (fixed @ 200 MHz)
**
** ADM-XRC-4FX      LCLK pin      (1)   =>   LCLK (0)
**                  MCLKA pin     (2)   =>   MCLK (1)
**                  MCLKB pin     (3)   =>   MCLK (1)
**                  REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                  XRMCLK pin    (5)   =>   N/A (depends on I/O pin)
**                  XRMMGTREF pin (6)   =>   N/A (depends on I/O pin)
**                  PCIE250A pin  (7)   =>   N/A (fixed @ 250 MHz)
**
** ADPE-XRC-4FX     LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**                  REFCLK200 pin (3)   =>   N/A (fixed @ 200 MHz)
**
** ADM-XRC-5LX      LCLK pin      (1)   =>   LCLK (0)
**                  MCLK pin      (2)   =>   MCLK (1)
**                  REFCLK200 pin (3)   =>   N/A (fixed @ 200 MHz)
**                  XRMCLK pin    (4)   =>   N/A (depends on I/O pin)
**
** ADM-XRC-5T1      LCLK pin      (1)   =>   LCLK (0)
**                  MCLKA pin     (2)   =>   MCLK (1)
**                  MCLKB pin     (3)   =>   MCLK (1)
**                  REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                  XRMCLK pin    (5)   =>   N/A (depends on I/O pin)
**                  XRMMGTREF pin (6)   =>   N/A (depends on I/O pin)
**                  PCIE100A pin  (7)   =>   N/A (fixed @ 100 MHz)
**
** ADM-XRC-5T2 or   LCLK pin      (1)   =>   LCLK (0)
** ADM-XRC-5T2-ADV  MCLKA pin     (2)   =>   MCLK (1)
**                  MCLKB pin     (3)   =>   MCLK (1)
**                  REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                  XRMCLK pin    (5)   =>   N/A (depends on I/O pin)
**                  XRMMGTREF pin (6)   =>   N/A (depends on I/O pin)
**                  PCIE100A pin  (7)   =>   N/A (fixed @ 100 MHz)
**
** ADM-XRC-5TZ      LCLK pin      (1)   =>   LCLK (0)
**                  MCLKA pin     (2)   =>   MCLK (1)
**                  MCLKB pin     (3)   =>   MCLK (1)
**                  REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                  XRMCLK pin    (5)   =>   N/A (depends on I/O pin)
**                  XRMMGTREF pin (6)   =>   N/A (depends on I/O pin)
**                  PCIE100A pin  (7)   =>   N/A (fixed @ 100 MHz)
**
** ADM-XRC-5T-DA1   LCLK pin      (1)   =>   LCLK (0)
**                  MCLKA pin     (2)   =>   MCLK (1)
**                  MCLKB pin     (3)   =>   MCLK (1)
**                  REFCLK200 pin (4)   =>   N/A (fixed @ 200 MHz)
**                  PCIE100A pin  (5)   =>   N/A (fixed @ 100 MHz)
**
** Note that this application does not attempt to compensate for any
** innacuracy in reported measurements resulting from:
**
**   (a) limited system timer resolution, and
**
**   (b) preemptive scheduling resulting delaying the reads of a cycle
**       counters due to the system being busy executing other processes
**       or threads.
**
** This application is intended to provide a ball-park measurement of
** clock frequency, rather than a precise measurement.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>

#include <admxrc2.h>

#include "../common/common.h"
#include "../common/platform.h"

#define RESET_COUNT_REG_INDEX       (0)
#define READ_COUNT_REG_INDEX        (1)
#define READY_REG_INDEX             (2)
#define COUNT_VAL_REG_INDEX(n)      (32 + (n))

typedef int BOOL;

/*
** Returns TRUE if ready flag for the specified clock input is high
** within 2 milliseconds (give or take the system timer resolution.
*/
static BOOL
waitReady(
  volatile uint32_t* pFpgaSpace,
  unsigned int clockInput)
{
  /* Cycle counter indices are zero-based in FPGA */
  unsigned int clockInputIndex = clockInput - 1;
  unsigned int i;
  uint32_t val32;

  for (i = 0; i < 10000; i++) {
    val32 = pFpgaSpace[READY_REG_INDEX];
    val32 = os_le32_to_cpu(val32);
    if (val32 & (0x1U << clockInputIndex)) {
      return TRUE;
    }
  }
  return FALSE;
}

/*
** Performs a read of a cycle counter.
**
** Returns TRUE if timeout NOT detected, i.e. the clock for that cycle
** counter appears to be running.
*/
static BOOL
readCounter(
  volatile uint32_t* pFpgaSpace,
  unsigned int clockInput,
  uint32_t* pCountValue)
{
  /* Cycle counter indices are zero-based in FPGA */
  unsigned int clockInputIndex = clockInput - 1;
  uint32_t val32;

  if (!waitReady(pFpgaSpace, clockInput)) {
    return FALSE;
  }
  val32 = 0x1U << clockInputIndex;
  pFpgaSpace[READ_COUNT_REG_INDEX] = os_cpu_to_le32(val32);
  if (!waitReady(pFpgaSpace, clockInput)) {
    return FALSE;
  }
  val32 = pFpgaSpace[COUNT_VAL_REG_INDEX(clockInputIndex)];
  *pCountValue = os_le32_to_cpu(val32);
  return TRUE;
}

static const char*
getClockInfo(
  ADMXRC2_CARD_INFO cardInfo,
  unsigned int clockInput,
  BOOL* bProgrammable,
  unsigned int* clockGen)
{
  if (clockInput == 1) {
    *bProgrammable = TRUE;
    *clockGen = 0;
    return "LCLK";
  } else if (clockInput == 2) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC:
    case ADMXRC2_BOARD_ADMXRC_P:
    case ADMXRC2_BOARD_ADMXRC2_LITE:
    case ADMXRC2_BOARD_ADMXRC2:
    case ADMXRC2_BOARD_ADPWRC2:
    case ADMXRC2_BOARD_ADPDRC2:
    case ADMXRC2_BOARD_ADMXRC4LS:
    case ADMXRC2_BOARD_ADMXRC4LX:
    case ADMXRC2_BOARD_ADMXRC4SX:
    case ADMXRC2_BOARD_ADPEXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5LX:
      *bProgrammable = TRUE;
      *clockGen = 1;
      return "MCLK";
    case ADMXRC2_BOARD_ADMXPL:
    case ADMXRC2_BOARD_ADMXP:
    case ADMXRC2_BOARD_ADPXPI:
      *bProgrammable = FALSE;
      *clockGen = 0;
      return "MCLK";
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5TDA1:
      *bProgrammable = TRUE;
      *clockGen = 1;
      return "MCLKA";
    default:
      *bProgrammable = FALSE;
      return "unknown";
    }
  } else if (clockInput == 3) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXPL:
    case ADMXRC2_BOARD_ADMXP:
    case ADMXRC2_BOARD_ADPXPI:
    case ADMXRC2_BOARD_ADMXRC4LX:
    case ADMXRC2_BOARD_ADMXRC4SX:
    case ADMXRC2_BOARD_ADMXRC5LX:
    case ADMXRC2_BOARD_ADPEXRC4FX:
      *bProgrammable = FALSE;
      return "REFCLK";
    case ADMXRC2_BOARD_ADPDRC2:
      *bProgrammable = FALSE;
      return "DDRCLK";
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5TDA1:
      *bProgrammable = TRUE;
      *clockGen = 1;
      return "MCLKB";
    default:
      *bProgrammable = FALSE;
      return "unknown";
    }
  } else if (clockInput == 4) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5TDA1:
      *bProgrammable = FALSE;
      return "REFCLK";
    case ADMXRC2_BOARD_ADMXRC5LX:
      *bProgrammable = FALSE;
      return "XRM_CLKIN";
    default:
      *bProgrammable = FALSE;
      return "unknown";
    }
  } else if (clockInput == 5) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
      *bProgrammable = FALSE;
      return "XRM_CLKIN";
    case ADMXRC2_BOARD_ADMXRC5TDA1:
      *bProgrammable = FALSE;
      return "PCIE100A";
    default:
      *bProgrammable = FALSE;
      return "unknown";
    }
  } else if (clockInput == 6) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
      *bProgrammable = FALSE;
      return "XRM_MGTREF";
    default:
      *bProgrammable = FALSE;
      return "unknown";
    }
  } else if (clockInput == 7) {
    switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC4FX:
      *bProgrammable = FALSE;
      return "PCIE250A";
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
      *bProgrammable = FALSE;
      return "PCIE100A";
    default:
      *bProgrammable = FALSE;
      return "unknown";
    }
  } else {
    *bProgrammable = FALSE;
    return "unknown";
  }
}

int main(int argc, char *argv[])
{
  int ret = 0;
  ADMXRC2_STATUS status;
  ADMXRC2_HANDLE card = ADMXRC2_HANDLE_INVALID_VALUE;
  ADMXRC2_CARD_INFO cardInfo;
  ADMXRC2_SPACE_INFO spInfo;
  volatile uint32_t* pFpgaSpace;
  const char* filename;
  Option options[3];
  OptionType paramType[2];
  char* paramInfo[2];
  char* paramMeaning[2];
  Arguments arguments;
  unsigned int clockInput;
  unsigned int numClockInput;
  uint32_t value1, value2;
  uint32_t delta;
  BOOL bMeasure;
  BOOL bSuccess;
  
  /* Set up expected options */
  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  options[2].key = "measure";
  options[2].type = Option_boolean;
  options[2].def.booleanVal = TRUE;
  options[2].meaning = "TRUE to measure frequency";
  arguments.nOption = 3;
  arguments.option = options;
  
  paramType[0] = Option_uint;
  paramInfo[0] = "clockInput";
  paramMeaning[0] = "Index of clock generator to test";
  paramType[1] = Option_double;
  paramInfo[1] = "frequency";
  paramMeaning[1] = "Requested frequency (MHz)";
  arguments.nParamType = 2;
  arguments.minNParam = 0;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  
  /* Parse command line */
  parse_command_line("clock",
		     argc,
		     argv,
		     &arguments);
  
  if (options[0].specified && options[1].specified) {
    printf("Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }
  
  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;
    
    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with ID %lu: %s\n",
	     (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;
    
    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with index %lu: %s\n",
	     (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }
  
  bMeasure = options[2].value.booleanVal;

  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  
  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
    numClockInput = 2; /* LCLK, MCLK */
    break;
    
  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
    numClockInput = 3; /* LCLK, MCLK, REFCLK */
    break;
    
  case ADMXRC2_BOARD_ADPDRC2:
    numClockInput = 3; /* LCLK, MCLK, DDRCLK */
    break;
    
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADPEXRC4FX:
    numClockInput = 3; /* LCLK, MCLK, REFCLK */
    break;
    
  case ADMXRC2_BOARD_ADMXRC5LX:
    numClockInput = 4; /* LCLK, MCLK, REFCLK, XRMCLK */
    break;
    
  case ADMXRC2_BOARD_ADMXRC4FX:
    numClockInput = 7; /* LCLK, MCLKA, MCLKB, REFCLK, XRMCLK, XRMMGTREF, PCIE250A */
    break;
    
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
    numClockInput = 7; /* LCLK, MCLKA, MCLKB, REFCLK, XRMCLK, XRMMGTREF, PCIE100A */
    break;

  case ADMXRC2_BOARD_ADMXRC5TDA1:
    numClockInput = 5; /* LCLK, MCLKA, MCLKB, REFCLK, PCIE100A */
    break;
    
  default:
    printf("*** This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRC-II-Lite\n");
    printf("\tADM-XRC-II\n");
    printf("\tADM-XPL\n");
    printf("\tADM-XP\n");
    printf("\tADP-WRC-II\n");
    printf("\tADP-DRC-II\n");
    printf("\tADP-XPI\n");
    printf("\tADM-XRC-4LX\n");
    printf("\tADM-XRC-4SX\n");
    printf("\tADM-XRC-4FX / ADM-XMC-4FX\n");
    printf("\tADPE-XRC-4FX\n");
    printf("\tADM-XRC-5LX\n");
    printf("\tADM-XRC-5T1\n");
    printf("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV\n");
    printf("\tADM-XRC-5TZ\n");
    printf("\tADM-XRC-5T-DA1\n");
    ret = -1;
    goto done;
  }
  
  /* Get the address of FPGA space */
  status = ADMXRC2_GetSpaceInfo(card, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get space 0 info: %s\n",
	   ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  pFpgaSpace = (volatile uint32_t*) spInfo.VirtualBase;
  if (pFpgaSpace == NULL) {
    printf("*** Failed to retrieve pointer to FPGA space\n");
    return -1;
  }
  
  if (arguments.nParam >= 1) {
    const char* pClockName;
    unsigned int clockGen;
    BOOL bProgrammable;
    
    clockInput = (unsigned int)arguments.param[0].uintVal;
    
    if (clockInput < 1 || clockInput > 32) {
      printf("*** Clock index of %lu is invalid; valid range is 1..32\n",
	     (unsigned long)clockInput);
      ret = -1;
      goto done;
    }
    if (clockInput > numClockInput) {
      printf("*** Clock index of %lu is invalid; maximum value for this card is %lu.\n",
	     (unsigned long)clockInput, (unsigned long)numClockInput);
      ret = -1;
      goto done;
    }
    
    pClockName = getClockInfo(cardInfo, clockInput, &bProgrammable, &clockGen);
    
    if (bMeasure) {
      filename = get_fpga_filename(&cardInfo, "clock", &arguments);
      status = ADMXRC2_ConfigureFromFile(card, filename);
      if (status != ADMXRC2_SUCCESS) {
	printf ("Failed to load the bitstream '%s': %s\n",
		filename, ADMXRC2_GetStatusString(status));
	ret = -1;
	goto done;
      }
    }

    if (arguments.nParam >= 2) {
      double reqFreq, progFreq;
      
      if (!bProgrammable) {
	printf("*** Clock input %lu is not programmable.\n", (unsigned long)clockInput);
	ret = -1;
	goto done;
      }
      /* Set specified clock generator to requested frequency */
      reqFreq = arguments.param[1].doubleVal * 1000000.0;
      printf("Programming clock generator %lu for %.1f MHz...\n",
	     (unsigned long)clockGen, reqFreq / 1.0e6);
      status = ADMXRC2_SetClockRate(card, clockGen, reqFreq, &progFreq);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to program clock generator %lu for %.1f MHz: %s\n",
	       (unsigned long)clockGen, reqFreq / 1.0e6,
	       ADMXRC2_GetStatusString(status));
	ret = -1;
	goto done;
      }
      printf("Actual programmed frequency = %.1f Hz\n", progFreq);
    }

    if (bMeasure) {
      printf("Measuring frequency of clock input %lu (%s)...\n",
	     (unsigned long)clockInput, pClockName);
      
      /* Zero all cycle counters */
      pFpgaSpace[RESET_COUNT_REG_INDEX] = os_cpu_to_le32(0xffffffffU);
      pFpgaSpace[RESET_COUNT_REG_INDEX];
      
      /* Now read the specified cycle counter twice 1 second apart */
      sleep_ms(1);
      bSuccess = readCounter(pFpgaSpace, clockInput, &value1);
      if (bSuccess) {
	sleep_ms(1000);
	bSuccess = readCounter(pFpgaSpace, clockInput, &value2);
	if (bSuccess) {
	  delta = value2 - value1;
	  printf("Initial counter value = %lu\n", (unsigned long)value1);
	  printf("Final counter value   = %lu, delta = %lu\n",
		 (unsigned long)value2, (unsigned long)delta);
	}
      }
      if (!bSuccess) {
	printf("Clock input %lu (%s) appears to be stopped.\n",
	       (unsigned long)clockInput, pClockName);
      }
    }
  } else {
    unsigned int i;
    const char* pClockName;
    BOOL bProgrammable;
    unsigned int clockGen;
    
    printf("Clock pins available on specified card:\n\n");
    printf("Clock input      Programmable   Clock generator index\n\n");
    for (i = 1; i <= numClockInput; i++) {
      pClockName = getClockInfo(cardInfo, i, &bProgrammable, &clockGen);
      if (bProgrammable) {
	printf("%-2lu %-12s  %-12s   %-lu\n",
	       (unsigned long)i, pClockName, bProgrammable ? "Yes" : "No", (unsigned long)clockGen);
      } else {
	printf("%-2lu %-12s  %-12s   N/A\n",
	       (unsigned long)i, pClockName, bProgrammable ? "Yes" : "No");
      }
    }
  }
  
 done:
  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }
  
  return ret;
}
