/*
** dll.c - Example demonstrating use of Virtex DLLs to obtain
**         a multiple of local bus clock frequency.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV
**   o ADM-XRC-5TZ
**
** (C) Copyright 2001-2008 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <unistd.h>

#include <admxrc2.h>

#include "../common/common.h"
#include "../common/platform.h"

int main(int argc, char *argv[])
{
  int                ret = 0;
  ADMXRC2_STATUS     status;
  ADMXRC2_HANDLE     card = ADMXRC2_HANDLE_INVALID_VALUE;
  ADMXRC2_CARD_INFO  cardInfo;
  ADMXRC2_SPACE_INFO spInfo;
  volatile uint32_t* fpgaSpace;
  int                i;
  unsigned long      counter, prev;
  const char*        filename;
  double             lfreq;
  double             mfreq = 33.0e6;
  double             alfreq, amfreq;
  Arguments          arguments;
  Option             options[2];
  OptionType         paramType[1];
  char*              paramInfo[1];
  char*              paramMeaning[1];
  int                bSetMclk;

  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  arguments.nOption = 2;
  arguments.option = options;

  paramType[0] = Option_double;
  paramInfo[0] = "f_lclk";
  paramMeaning[0] = "LCLK frequency (MHz)";
  arguments.nParamType = 1;
  arguments.minNParam = 1;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  parse_command_line("dll", argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    printf("Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with ID %ld: %s\n",
	     (unsigned long) cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  lfreq = arguments.param[0].doubleVal * 1000000.0;

  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
    bSetMclk = TRUE;
    if (lfreq < 25.0e6) {
      printf("DLLs require an input clock frequency of at least 25MHz\n");
      ret = -1;
      goto done;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    bSetMclk = TRUE;
    if (lfreq < 24.0e6) {
      printf("DCMs require an input clock frequency of at least 24MHz\n");
      ret = -1;
      goto done;
    }
    break;

  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADPEXRC4FX:
  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADCPXRC4LX:
  case ADMXRC2_BOARD_ADMAMC5A2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
    bSetMclk = FALSE;
    if (lfreq < 24.0e6) {
      printf("DCMs require an input clock frequency of at least 24MHz\n");
      ret = -1;
      goto done;
    }
    break;


  default:
    printf("This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRC-II-Lite\n");
    printf("\tADM-XRC-II\n");
    printf("\tADM-XPL\n");
    printf("\tADM-XP\n");
    printf("\tADP-WRC-II\n");
    printf("\tADP-DRC-II\n");
    printf("\tADP-XPI\n");
    printf("\tADM-XRC-4LX\n");
    printf("\tADM-XRC-4SX\n");
    printf("\tADM-XRC-4FX\n");
    printf("\tADPE-XRC-4FX\n");
    printf("\tADM-XRC-5LX\n");
    printf("\tADM-XRC-5T1\n");
    printf("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV\n");
    printf("\tADM-XRC-5TZ\n");
    ret = -1;
    goto done;
  }

  if (lfreq > 40.1e6) {
    printf("Maximum local bus clock frequency in this example is 40MHz\n");
    ret = -1;
    goto done;
  }

  status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, lfreq, &alfreq);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to set LCLK to %.1fMHz: %s\n",
	   lfreq / 1.0e6, ADMXRC2_GetStatusString(status));
    return -1;
  }

  if (bSetMclk) {
    status = ADMXRC2_SetClockRate(card, 1, mfreq, &amfreq);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to set clock 1 to %.1fMHz: %s\n",
	     mfreq / 1.0e6, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    amfreq = alfreq * 2.0;
  }

  printf("LCLK = %fHz, MCLK = %fHz\n", alfreq, amfreq);

  filename = get_fpga_filename(&cardInfo, "dll", &arguments);
  status = ADMXRC2_ConfigureFromFile(card, filename);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to load the bitstream '%s': %s\n",
	   filename, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /*
  ** Allow time for any DCMs/DLLs to be reset and then achieve
  ** lock. This is required for the engineering silicon (ES)
  ** Virtex-II devices.
  */
  sleep(1);

  /* Get the address of FPGA space */
  status = ADMXRC2_GetSpaceInfo(card, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get space 0 info: %s\n",
	   ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  fpgaSpace = (volatile uint32_t*) spInfo.VirtualBase;
	
  /* Reset the 2x cycle counter */
  fpgaSpace[0] = os_cpu_to_le32(0);

  /* Read the 2x cycle counter at 1s intervals, and display its value */
  prev = os_le32_to_cpu(fpgaSpace[0]);
  for (i = 0; i < 3; i++) {
    sleep(1);
    counter = os_le32_to_cpu(fpgaSpace[0]);
    printf("Counter = %ld, Delta = %ld\n", counter, counter - prev);
    prev = counter;
  }

 done:
  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }
	
  return ret;
}
