/*
** dma.c - Example demonstrating multithreaded DMA transfers
**         to and from FPGA, using demand mode DMA.
**
** Data is transferred using demand mode DMA to the FPGA, stored
** in a FIFO and transferred out again also using demand mode DMA.
** Also demonstrated is configuring the FPGA using DMA.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-WRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**
** (C) Copyright 2000-2009 Alpha Data
**
**
** NOTES ON OPERATION
**
** This application is designed to show that multiple threads can
** use the same DMA channel without error and without affecting
** data integrity.
**
** 1. Threading
**
** A sender DMA thread is started up. This threads writes data
** into the FIFO in the FPGA.
**
** A receiver DMA thread is started up, which reads data out
** of the FIFO in the FPGA and checks that it is correct.
**
** 2. DMA mode bits
**
** The DMA transfers are performed with the DEMAND mode bit set.
** Thus, the DMA transfers are regulated by the ability of the FPGA
** to accept or supply data.
**
** Other mode bits used are:
**
**    - FIXEDLOCAL:  the FIFO input and output ports are 32-bit wide
**                   registers on the local buss. Thus a DMA transfer
**                   must not increment the local bus address.
**    - BURSTENABLE: to achieve good throughput, it is necessary to
**                   enable burst transfers on the local bus.
**    - USEBTERM:    the FPGA can terminate a local bus burst by
**                   asserting BTERM.
**    - USEREADY:    the FPGA uses LREADYI# to control when data
**                   transfer can occur within a local bus burst.
**
** 3. DMA driven FPGA configuration
**
** The -dmacfg option disables DMA configuration of the FPGA (enabled by
** default) so that it uses programmed I/O.
**
** 4. Data integrity checking
**
** This program checks that the data received by the sucker thread
** is correct. Changing the PERFORM_CHECK macro from 1 to 0 may
** result in an increase in terms in data throughput.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <pthread.h>
#include <semaphore.h>
#include <unistd.h>

#include <admxrc2.h>

#include "../common/common.h"
#include "../common/platform.h"

#define PERFORM_CHECK       (1)         /* 1 => Check data received */
#define INTERACTIVE         (1)         /* 1 => Press enter to display counters */

static ADMXRC2_HANDLE     card = ADMXRC2_HANDLE_INVALID_VALUE;
static uint32_t           mode;

static BOOLEAN            readySemaValid = FALSE;
static sem_t              readySema;

static BOOLEAN            terminate;
static BOOLEAN            terminateReceiver;

static pthread_t          receiver, sender;
static double             got, sent;

static unsigned long      bufSize;
static ADMXRC2_DMADESC    sendbufHandle;
BOOLEAN                   sendbufValid = FALSE;
static uint8_t*           sendbuf = NULL;
static ADMXRC2_DMADESC    recvbufHandle;
BOOLEAN                   recvbufValid = FALSE;
static uint8_t*           recvbuf = NULL;

static volatile uint32_t* fpgaSpace;

static void*
senderThread(void* arg)
{
  ADMXRC2_STATUS status;
  void*          event = NULL;
  unsigned long  j;
  uint32_t*      buf = (uint32_t*) sendbuf;
  unsigned long  offs = 0;
  unsigned long  val;

  /* Set up the data */
  val = (unsigned long) 0;
  val |= val << 8;
  val |= val << 16;
  for (j = 0; j < (bufSize >> 2); j++) {
    buf[j] = val++;
  }

  while (1) {
    if (terminate) {
      terminateReceiver = TRUE;
      sem_post(&readySema);
      break;
    }

    sem_post(&readySema);

    /*
    ** Program the PCI-to-local transfer counter.
    */
    fpgaSpace[0] = os_cpu_to_le32(bufSize);
    fpgaSpace[0];

    /*
    ** Perform the DMA transfer.
    */
    status = ADMXRC2_DoDMA(card,
			   sendbufHandle,
			   offs,
			   bufSize,
			   0x40,
			   ADMXRC2_PCITOLOCAL,
			   0,
			   mode,
			   0,
			   NULL,
			   event);
    if (status != ADMXRC2_SUCCESS) {
      printf("Sender thread: failed to do DMA: %s\n", ADMXRC2_GetStatusString(status));
      break;
    }

    sent += bufSize;
  }

  if (0) printf("sender exiting\n");
  pthread_exit(NULL);
}

static void*
receiverThread(void* arg)
{
  ADMXRC2_STATUS status;
  unsigned long  j;
  void*          event;
  uint32_t*      buf = (uint32_t*) recvbuf;
  unsigned long  offs = 0;
  unsigned long  n = 0;
  uint32_t       exp, act;
  int            nerr;

  while (1) {
    sem_wait(&readySema);
    if (terminateReceiver) {
      break;
    }

    /*
    ** Program the local-to-PCI transfer counter.
    */
    fpgaSpace[1] = os_cpu_to_le32(bufSize);
    fpgaSpace[1];

    /*
    ** Perform the DMA transfer.
    */
    status = ADMXRC2_DoDMA(card,
			   recvbufHandle,
			   offs,
			   bufSize,
			   0x80,
			   ADMXRC2_LOCALTOPCI,
			   1,
			   mode,
			   0,
			   NULL,
			   event);
    if (status != ADMXRC2_SUCCESS) {
      printf("Receiver thread: failed to do DMA: %s\n", ADMXRC2_GetStatusString(status));
      break;
    }

#if PERFORM_CHECK
    /* Check the data */
    nerr = 0;
    exp = buf[0];
    for (j = 1; j < (bufSize >> 2); j++) {
      exp++;
      act = buf[j];
      if (act != exp) {
	if (nerr < 10) {
	  printf("Receiver thread: error at block %ld index %ld, exp/act=0x%8.8lx/0x%8.8lx\n",
		 (unsigned long) n, (unsigned long) j,
		 (unsigned long) exp, (unsigned long) act);
	} else if (nerr == 10) {
	  printf("Receiver thread: [more errors]\n");
	}
	nerr++;
      }
    }
#endif

    got += bufSize;
    n++;
  }

  if (0) printf("receiver exiting\n");
  pthread_exit(NULL);
}

int main(int argc, char *argv[])
{
  int                ret = 0;
  ADMXRC2_STATUS     status;
  ADMXRC2_CARD_INFO  cardInfo;
  ADMXRC2_SPACE_INFO spaceInfo;
  struct timeval     start;
  void*              event = NULL;
  const char*        filename;
  Arguments          arguments;
  Option             options[6];
  float              clock_freq;
  int                res;
  int                bUse64bit;
  int                bDmaConfiguration;

  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  options[2].key = "size";
  options[2].type = Option_uint;
  options[2].def.uintVal = 64*1024;
  options[2].meaning = "send/receive buffer size (bytes)";
  options[3].key = "lclk";
  options[3].type = Option_float;
  options[3].def.floatVal = 33.0;
  options[3].meaning = "local bus clock frequency (MHz)";
  options[4].key = "64";
  options[4].type = Option_boolean;
  options[4].def.booleanVal = FALSE;
  options[4].meaning = "TRUE => run local bus in 64 bit mode";
  options[5].key = "dmacfg";
  options[5].type = Option_boolean;
  options[5].def.booleanVal = TRUE;
  options[5].meaning = "TRUE => configure FPGA using DMA";
  arguments.nOption = 6;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  parse_command_line("dma", argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    printf("Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with ID %ld: %s\n",
	     (unsigned long) cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  /* Allocate memory for the DMA data buffers */
  bufSize = options[2].value.uintVal;
  recvbuf = (uint8_t*) ADMXRC2_Malloc(bufSize);
  if (recvbuf == NULL) {
    printf("*** Failed to allocate receive buffer\n");
    ret = -1;
    goto done;
  }
  sendbuf = (uint8_t*) ADMXRC2_Malloc(bufSize);
  if (sendbuf == NULL) {
    printf("*** Failed to allocate send buffer\n");
    ret = -1;
    goto done;
  }

  clock_freq = options[3].value.floatVal * 1.0e6f;
  bUse64bit = options[4].value.booleanVal;

  /* Get card information, for device fitted */
  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPDRC2: 
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
    if (bUse64bit) {
      printf("*** The selected card cannot be run in 64-bit mode.\n");
      ret = -1;
      goto done;
    }
    break;

  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADPEXRC4FX:
  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADCPXRC4LX:
  case ADMXRC2_BOARD_ADMAMC5A2:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
    break;

  default:
    printf("This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRC-II-Lite\n");
    printf("\tADM-XRC-II\n");
    printf("\tADM-XPL\n");
    printf("\tADM-XP\n");
    printf("\tADP-WRC-II\n");
    printf("\tADP-DRC-II\n");
    printf("\tADP-XPI\n");
    printf("\tADM-XRC-4LX\n");
    printf("\tADM-XRC-4SX\n");
    printf("\tADM-XRC-4FX\n");
    printf("\tADPE-XRC-4FX\n");
    printf("\tADM-XRC-5LX\n");
    printf("\tADM-XRC-5T1\n");
    printf("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV\n");
    printf("\tADM-XRC-5TZ\n");
    printf("\tADM-XRC-5T-DA1\n");
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    if (options[5].specified && options[5].value.booleanVal) {
      printf("*** The selected board does not support DMA FPGA configuration\n");
      ret = -1;
      goto done;
    }
    bDmaConfiguration = FALSE;
    break;
  default:
    bDmaConfiguration = options[5].value.booleanVal;
    break;
  }
  
  status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, clock_freq, NULL);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to set LCLK to %.1fMHz: %s\n",
	   clock_freq / 1.0e6f, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  status = ADMXRC2_GetSpaceInfo(card, 0, &spaceInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get information about FPGA space\n");
    ret = -1;
    goto done;
  }
  fpgaSpace = (volatile uint32_t*) spaceInfo.VirtualBase;

  if (bUse64bit) {
    uint32_t config;
    
    config = ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
    status = ADMXRC2_SetSpaceConfig(card, 0, config);
    if (status != ADMXRC2_SUCCESS) {
      printf ("*** Failed to set space configuration: %s\n",
	      ADMXRC2_GetStatusString(status));
      return -1;
    }
  }

  if (bUse64bit) {
    filename = get_fpga_filename(&cardInfo, "ddma64", &arguments);
  } else {
    filename = get_fpga_filename(&cardInfo, "ddma", &arguments);
  }
  if (bDmaConfiguration) {
    /* Load the FPGA using DMA and measure the configuration time */
    const int      num_rep = 10;
    int            i;
    ADMXRC2_IMAGE  image;
    unsigned long  size;
    struct timeval t, now;

    status = ADMXRC2_LoadBitstream(card, filename, &image, &size);
    if (status != ADMXRC2_SUCCESS) {
      printf ("Failed to load FPGA file '%s': %s\n",
	      filename, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
    
    t = get_time();
    for (i = 0; i < num_rep; i++) {
      status = ADMXRC2_ConfigureFromBufferDMA(card,
					      image,
					      size,
					      ADMXRC2_DMACHAN_ANY,
					      event);
      if (status != ADMXRC2_SUCCESS) {
	ADMXRC2_UnloadBitstream(image);
	printf("Failed to configure FPGA from buffer with DMA: %s\n",
	       ADMXRC2_GetStatusString(status));
	ret = -1;
	goto done;
      }
    }
    now = get_time();
    t = time_subtract(&now, &t);

    printf("FPGA configuration time = %.1fms\n",
	     time_to_double(&t) * 1000.0 / ((double) num_rep));
      
    ADMXRC2_UnloadBitstream(image);
  } else {
    /* Load the FPGA using programmed I/O */
    status = ADMXRC2_ConfigureFromFile(card, filename);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to configure FPGA from buffer: %s\n",
	     ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  /*
  ** Build the DMA mode word. We want bursting, demand-mode,
  ** use ready, use bterm and fixed local bus addressing.
  */
  mode = ADMXRC2_BuildDMAModeWord(cardInfo.BoardType,
				  bUse64bit ? ADMXRC2_IOWIDTH_64 : ADMXRC2_IOWIDTH_32,
				  0,
				  ADMXRC2_DMAMODE_FIXEDLOCAL |
				  ADMXRC2_DMAMODE_USEREADY |
				  ADMXRC2_DMAMODE_USEBTERM |
				  ADMXRC2_DMAMODE_BURSTENABLE |
				  ADMXRC2_DMAMODE_DEMAND);

  /*
  ** A semaphore is required for the send thread to notify the receive thread
  ** that the receive thread can proceed with its DMA.
  */
  res = sem_init(&readySema, 0, 0);
  if (res) {
    printf("*** Failed to initialise semaphore\n");
    ret = -1;
    goto done;
  }
  readySemaValid = TRUE;

  /* Get a DMA descriptor for the buffer used for DMA transfers to the card */
  status = ADMXRC2_SetupDMA(card, sendbuf, bufSize, 0, &sendbufHandle);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to set up send DMA buffer: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  sendbufValid = TRUE;

  /* Get a DMA descriptor for the buffer used for DMA transfers from the card */
  status = ADMXRC2_SetupDMA(card, recvbuf, bufSize, 0, &recvbufHandle);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to set up receive DMA buffer: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  recvbufValid = TRUE;

  /*
  ** Start up the sender and receiver DMA threads.
  */
  terminate = FALSE;
  terminateReceiver = FALSE;
  printf("Starting thread(s)...\n");
  start = get_time();
  sent = 0.0;
  res = pthread_create(&sender, NULL, senderThread, NULL);
  if (res) {
    printf("*** Failed to create sender thread.\n");
    ret = -1;
    goto done;
  }
  got = 0.0;
  res = pthread_create(&receiver, NULL, receiverThread, NULL);
  if (res) {
    printf("*** Failed to create receiver thread.\n");
    ret = -1;
    goto done;
  }

  sleep(1);

  /* Display average transfer rates for ever or until user enters 'q' */
  while (1) {
    int		   c;
    double	   s, g, t;
    struct timeval now, delta;
    double         fDelta;

    now = get_time();
    delta = time_subtract(&now, &start);

    s = sent / 1048576.0f;
    g = got / 1048576.0f;
    t = s + g;

    printf("Sent %.2fMB, received %.2fMB, total %.2fMB\n", s, g, t);
    fDelta = time_to_double(&delta);
    if (fDelta > 0.1) {
      printf("Send rate %.2fMB/s, receive rate %.2fMB/s, total rate %.2fMB/s\n",
	     s / fDelta,
	     g / fDelta,
	     t / fDelta);
    }

#if INTERACTIVE
    printf("Enter 'q' to exit, anything else to view rates:\n");
    c = getchar();
    if (c == 'q') {
      break;
    }
#else
    sleep(1);
#endif
  }

  /*
  ** Terminate the threads cleanly.
  */
  printf("Terminating threads...\n");
  terminate = TRUE;
  pthread_join(sender, NULL);
  pthread_join(receiver, NULL);

 done:
  /* Free the DMA descriptors */
  if (sendbufValid) {
    status = ADMXRC2_UnsetupDMA(card, sendbufHandle);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to delete DMA send buffer: %s\n", ADMXRC2_GetStatusString(status));
    }
  }
  if (recvbufValid) {
    status = ADMXRC2_UnsetupDMA(card, recvbufHandle);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to delete DMA receive buffer: %s\n", ADMXRC2_GetStatusString(status));
    }
  }

  /* Free the Win32 Semaphore */
  if (readySemaValid) {
    sem_destroy(&readySema);
  }

  /* Close the device */
  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }
	
  /* Free the memory used for the DMA data buffers */
  if (recvbuf != NULL) {
    ADMXRC2_Free(recvbuf);
  }
  if (sendbuf != NULL) {
    ADMXRC2_Free(sendbuf);
  }

  return ret;
}
