/*
** eptest.c - Utility to read and write the configuration EEPROM on an
**            ADM-XRC series card.
**
** Warning: do not use this utility to write to the EEPROM unless
**          you know what you are doing!
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**
** (C) Copyright 2000-2009 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>

#include <admxrc2.h>

#include "../common/common.h"
#include "../common/platform.h"

/*
** Return TRUE if 'loc' is an EEPROM location containing
** adapter configuration information that could render the
** card unusable if incorrectly programmed.
*/
static BOOLEAN
protected_location(
  unsigned long      loc,
  ADMXRC2_BOARD_TYPE board)
{
  switch (board) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
    return (loc < 0x2BU);
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    return (loc < 0x2DU);
  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADPEXRC4FX:
  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMAMC5A2:
  case ADMXRC2_BOARD_ADCPXRC4LX:
  case ADMXRC2_BOARD_ADCBBP:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
    return FALSE;
  default:
    return TRUE;
  }

  /* Don't allow programming if board type is invalid */
  return TRUE;
}

int main(int argc, char *argv[])
{
  int               ret = 0;
  ADMXRC2_HANDLE    card = (ADMXRC2_HANDLE_INVALID_VALUE);
  ADMXRC2_STATUS    status;
  ADMXRC2_CARD_INFO cardInfo;
  Option            options[4];
  Arguments         arguments;
  OptionType        paramType[2];
  char*             paramInfo[2];
  char*             paramMeaning[2];
  int               width;

  /* Set up expected options */
  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  options[2].key = "f";
  options[2].type = Option_boolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = "If TRUE, don't confirm when writing";
  options[3].key = "unlock";
  options[3].type = Option_boolean;
  options[3].def.booleanVal = FALSE;
  options[3].meaning = "If TRUE, allow write to protected locations";
  arguments.nOption = sizeof(options) / sizeof(Option);
  arguments.option = options;

  paramType[0] = Option_hex;
  paramInfo[0] = "location";
  paramMeaning[0] = "EEPROM location";
  paramType[1] = Option_hex;
  paramInfo[1] = "data";
  paramMeaning[1] = "Data value to write";
  arguments.nParamType = sizeof(paramType) / sizeof(OptionType);
  arguments.minNParam = 0;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  parse_command_line("eptest",
		     argc,
		     argv,
		     &arguments);

  if (options[0].specified && options[1].specified) {
    printf("Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with ID %lu: %s\n",
	     (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with index %lu: %s\n",
	     (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get card info: %s\n",	ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    width = 2;
    break;

  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADPEXRC4FX:
  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADCPXRC4LX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMAMC5A2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADCBBP:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
    width = 4;
    break;

  default:
    printf("This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRC-II-Lite\n");
    printf("\tADM-XRC-II\n");
    printf("\tADM-XPL\n");
    printf("\tADM-XP\n");
    printf("\tADP-WRC-II\n");
    printf("\tADP-DRC-II\n");
    printf("\tADP-XPI\n");
    printf("\tADM-XRC-4LX\n");
    printf("\tADM-XRC-4SX\n");
    printf("\tADM-XRC-4FX\n");
    printf("\tADPE-XRC-4FX\n");
    printf("\tADM-XRC-5LX\n");
    printf("\tADM-XRC-5T1\n");
    printf("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV\n");
    printf("\tADM-XRC-5TZ\n");
    printf("\tADM-XRC-5T-DA1\n");
    ret = -1;
    goto done;
  }

  if (arguments.nParam < 1) {
    /* Dump all of configuration ROM */
    unsigned long index;
    uint32_t      data;

    printf("Selected card ID is %lu(0x%lX)\n\n",
	   (unsigned long)cardInfo.CardID, (unsigned long) cardInfo.CardID);
    if (width == 2) {
      for (index = 0; index < 0x20; index++) {
	ADMXRC2_ReadConfig(card, index, &data);
	printf("   [0x%02lX] = 0x%04lX\t\t", index, (unsigned long) data);
	ADMXRC2_ReadConfig(card, index + 0x20, &data);
	printf("[0x%02lX] = 0x%04lX\n", index + 0x20, (unsigned long) data);
      }
    } else {
      for (index = 0; index < 0x20; index++) {
	ADMXRC2_ReadConfig(card, index, &data);
	printf("   [0x%02lX] = 0x%08lX\t\t", index, (unsigned long) data);
	ADMXRC2_ReadConfig(card, index + 0x20, &data);
	printf("[0x%02lX] = 0x%08lX\n", index + 0x20, (unsigned long) data);
      }
    }
  } else if (arguments.nParam < 2) {
    /* Display specified location of configuration ROM */
    uint32_t data;

    printf("Selected card ID is %lu(0x%lx)\n",
	   (unsigned long)cardInfo.CardID, (unsigned long) cardInfo.CardID);
    ADMXRC2_ReadConfig((unsigned long) card, (unsigned long) arguments.param[0].hexVal, &data);
    if (width == 2) {
      printf("[0x%0lX] = 0x%4lx\n",
	     (unsigned long) arguments.param[0].hexVal, (unsigned long) data);
    } else {
      printf("[0x%0lX] = 0x%08lx\n",
	     (unsigned long) arguments.param[0].hexVal, (unsigned long) data);
    }
  } else {
    /* Write to specified location of configuration ROM */
    char	c;
    int		res;

    if (!options[2].value.booleanVal) {
      printf("Selected card ID is %lu(0x%lx)\n",
	     (unsigned long)cardInfo.CardID, (unsigned long) cardInfo.CardID);
      printf("Warning: this will write the value 0x%lx to EEPROM location 0x%lx\n",
	     (unsigned long)arguments.param[1].hexVal,
	     (unsigned long)arguments.param[0].hexVal);
      printf("Are you sure you want to continue (y/n)? ");
      res = scanf("%c", &c);
      if (!res || res == EOF) {
	c = 'n';
      }
    } else {
      c = 'y';
    }

    if (protected_location(arguments.param[0].hexVal, cardInfo.BoardType) &&
	!options[3].value.booleanVal) {
      printf("Permission denied to protected location.\n");
    } else {
      if (c == 'y' || c == 'Y') {
	ADMXRC2_WriteConfig(card, arguments.param[0].hexVal, arguments.param[1].hexVal);
      } else {
	printf("Aborted.\n");
      }
    }
  }

 done:
  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }

  return ret;
}
