/*
** flash.c - utility to program, erase and verify the Flash memory on
**           an ADM-XRC series card.
**
** Warning: Although this utility performs checks in order to ensure
**          that you do not attempt to program the Flash memory with a
**          bitstream for the wrong type of FPGA, ensure that you do
**          not program the flash with an invalid bitstream. Doing so
**          and powering up the FPGA for an extended period of time
**          can damage devices on the card.
**
** Utility program for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**
** (C) Copyright 2003-2009 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>

#include <admxrc2.h>
#include <admxrc2/flash.h>

#include "../common/common.h"
#include "../common/platform.h"

static ADMXRC2_HANDLE		card = ADMXRC2_HANDLE_INVALID_VALUE;
static ADMXRC2_FLASH_INFO	flashInfo;

static void
myStrlwr(char* s)
{
  char c;

  while ((c = *s) != '\0') {
    *s++ = tolower(c);
  }
}

int main(int argc, char *argv[])
{
  int               ret = 0;
  ADMXRC2_STATUS    status;
  ADMXRC2_CARD_INFO cardInfo;
  ADMXRC2_IMAGE     image = NULL;
  unsigned long     imageSize;
  unsigned char*    flash = NULL;
  char*             command;
  Arguments         arguments;
  Option            options[4];
  OptionType        paramType[4];
  char*             paramInfo[4];
  char*             paramMeaning[4];
  BOOLEAN           doChkblank, doErase, doProgram, doReadback, doVerify;
  unsigned long     bootFlagAddr;
  unsigned long     imageAddr, eraseAddr, eraseLimit;
  int               nError = 0;
  int               bBootFlag32Bit;
  int               bFailsafeImage;
  unsigned int      targetIndex;

  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  options[2].key = "failsafe";
  options[2].type = Option_boolean;
  options[2].def.booleanVal = FALSE;
  options[2].meaning = "program failsafe image";
  options[3].key = "target";
  options[3].type = Option_uint;
  options[3].def.uintVal = 0;
  options[3].meaning = "specify target FPGA image";
  arguments.nOption = 4;
  arguments.option = options;
  paramType[0] = Option_string;
  paramInfo[0] = "command";
  paramMeaning[0] = "One of erase, program, verify, chkblank or readback";
  paramType[1] = Option_path;
  paramInfo[1] = "file";
  paramMeaning[1] = "Bitstream filename";
  paramType[2] = Option_uint;
  paramInfo[2] = "offset";
  paramMeaning[2] = "Offset for readback (readback command only)";
  paramType[3] = Option_uint;
  paramInfo[3] = "length";
  paramMeaning[3] = "Length for readback (readback command only)";
  arguments.nParamType = sizeof(paramType) / sizeof(OptionType);
  arguments.minNParam = 1;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;

  parse_command_line("flash",
		     argc,
		     argv,
		     &arguments);

  command = arguments.param[0].stringVal;
  myStrlwr(command);
  doProgram = strcmp(command, "program") == 0;
  doVerify = strcmp(command, "verify") == 0;
  doErase = strcmp(command, "erase") == 0;
  doChkblank = strcmp(command, "chkblank") == 0;
  doReadback = strcmp(command, "readback") == 0;
  if (!doProgram && !doVerify && !doErase && !doChkblank && !doReadback) {
    printf("*** Please specify a valid command. Use ftest /? to get help.\n");
    ret = -1;
    goto done;
  }

  if (doProgram || doVerify) {
    if (arguments.nParam < 2) {
      printf("*** Must specify a file for programming or verification\n");
      ret = -1;
      goto done;
    }
  }

  if (doReadback) {
    if (arguments.nParam < 4) {
      printf("*** Must specify a filename, offset and length for readback\n");
      ret = -1;
      goto done;
    }
  }
  
  if (options[0].specified && options[1].specified) {
    printf("*** Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with ID %ld: %s\n",
	     (unsigned long) cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  bFailsafeImage = options[2].value.booleanVal;
  targetIndex = options[3].value.uintVal;

  ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADPEXRC4FX:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADCBBP:
    if (bFailsafeImage) {
      printf("*** +failsafe option is not supported for this model.\n");
      ret = -1;
      goto done;
    }
    break;

  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMAMC5A2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
    break;

  default:
    printf("*** This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRC-II\n");
    printf("\tADM-XPL\n");
    printf("\tADM-XP\n");
    printf("\tADP-WRC-II\n");
    printf("\tADP-DRC-II\n");
    printf("\tADP-XPI\n");
    printf("\tADM-XRC-4LX\n");
    printf("\tADM-XRC-4SX\n");
    printf("\tADM-XRC-4FX\n");
    printf("\tADPE-XRC-4FX\n");
    printf("\tADM-XRC-5LX\n");
    printf("\tADM-XRC-5T1\n");
    printf("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV\n");
    printf("\tADM-XRC-5TZ\n");
    printf("\tADM-XRC-5T-DA1\n");
    ret = -1;
    goto done;
  }

  status = ADMXRC2_GetFlashInfo(card, &flashInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to get Flash information: %s\n",
	   ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
    bootFlagAddr = flashInfo.UseableStart + 0x8000U;
    imageAddr = 0x8001U;
    eraseAddr = flashInfo.UseableStart;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    bBootFlag32Bit = FALSE;
    break;

  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADPEXRC4FX:
    bootFlagAddr = flashInfo.UseableStart;
    imageAddr = flashInfo.UseableStart + 0x20000U;
    eraseAddr = flashInfo.UseableStart;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    bBootFlag32Bit = FALSE;
    break;

  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    bootFlagAddr = flashInfo.UseableStart + 0x20000U;
    imageAddr = flashInfo.UseableStart + 0x20004U;
    eraseAddr = flashInfo.UseableStart + 0x20000U;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    bBootFlag32Bit = FALSE;
    break;

  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
    if (bFailsafeImage) {
      bootFlagAddr = flashInfo.UseableStart + 0xC00000U;
      imageAddr = flashInfo.UseableStart + 0xC20000U;
      eraseAddr = flashInfo.UseableStart + 0xC00000U;
      eraseLimit = flashInfo.UseableStart + 0x1800000U;
    } else {
      bootFlagAddr = flashInfo.UseableStart;
      imageAddr = flashInfo.UseableStart + 0x20000U;
      eraseAddr = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0xC00000U;
    }
    bBootFlag32Bit = TRUE;
    break;

  case ADMXRC2_BOARD_ADMAMC5A2:
    if (bFailsafeImage) {
      bootFlagAddr = flashInfo.UseableStart + 0x1000000U;
      imageAddr = flashInfo.UseableStart + 0x1020000U;
      eraseAddr = flashInfo.UseableStart + 0x1000000U;
      eraseLimit = flashInfo.UseableStart + 0x2000000U;
    } else {
      bootFlagAddr = flashInfo.UseableStart;
      imageAddr = flashInfo.UseableStart + 0x20000U;
      eraseAddr = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0x1000000U;
    }
    bBootFlag32Bit = TRUE;
    break;

  case ADMXRC2_BOARD_ADCBBP:
    if (targetIndex == 0) {
      if (bFailsafeImage) {
	bootFlagAddr = flashInfo.UseableStart + 0x00800000U;
	imageAddr = flashInfo.UseableStart + 0x00820000U;
	eraseAddr = flashInfo.UseableStart + 0x00800000U;
	eraseLimit = flashInfo.UseableStart + 0x01000000U;
      } else {
	bootFlagAddr = flashInfo.UseableStart;
	imageAddr = flashInfo.UseableStart + 0x00020000U;
	eraseAddr = flashInfo.UseableStart;
	eraseLimit = flashInfo.UseableStart + 0x00800000U;
      }
    } else if (targetIndex == 1) {
      if (bFailsafeImage) {
	bootFlagAddr = flashInfo.UseableStart + 0x01E00000U;
	imageAddr = flashInfo.UseableStart + 0x01E20000U;
	eraseAddr = flashInfo.UseableStart + 0x01E00000U;
	eraseLimit = flashInfo.UseableStart + 0x02C00000U;
      } else {
	bootFlagAddr = flashInfo.UseableStart + 0x01000000U;
	imageAddr = flashInfo.UseableStart + 0x01020000U;
	eraseAddr = flashInfo.UseableStart + 0x01000000U;
	eraseLimit = flashInfo.UseableStart + 0x01E00000U;
      }
    } else {
      printf("*** Illegal value '%lu' for -target option (must be 0 or 1 for this model)\n", (unsigned long)targetIndex);
      ret = -1;
      goto done;
    }
    bBootFlag32Bit = TRUE;
    break;

  default:
    assert(FALSE);
    break;
  }

  /* Set local bus clock to nominal frequency */
  status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, 33.0e6, NULL);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to set LCLK to %.1fMHz: %s\n",
	   33.0, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  
  if (doProgram || doVerify) {
    status = ADMXRC2_LoadBitstream(card,
				   arguments.param[1].pathVal,
				   &image,
				   &imageSize);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to load bitstream '%s': %s\n",
	     arguments.param[1].pathVal, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    image = malloc(flashInfo.UseableLength);
    if (image == NULL) {
      printf("*** Failed to allocate 0x%lx bytes for image\n",
	     (unsigned long) flashInfo.UseableLength);
      ret = -1;
      goto done;
    }
    memset(image, 0xff, flashInfo.UseableLength);
  }

  printf("Flash type is %s, %ld(0x%lx) kB\n",
	 flashInfo.DeviceName,
	 (unsigned long) flashInfo.DeviceSize / 1024,
	 (unsigned long) flashInfo.DeviceSize / 1024);
  printf("Useable region is 0x%lx-0x%lx\n",
	 (unsigned long) flashInfo.UseableStart,
	 (unsigned long) flashInfo.UseableStart + flashInfo.UseableLength - 1);
  if (doProgram || doVerify) {
    printf("Bitstream length is %ld(0x%lx) bytes\n",
	   (unsigned long) imageSize, (unsigned long) imageSize);
  }
  printf("\n");

  if (doErase || doProgram) {
    unsigned long i;
    ADMXRC2_FLASHBLOCK_INFO blockInfo;

    printf("Erasing Flash from 0x%lx-0x%lx...\n",
	   (unsigned long)eraseAddr, (unsigned long)eraseLimit);
    i = eraseAddr;
    while (i < eraseLimit) {
      status = ADMXRC2_GetFlashBlockInfo(card, i, &blockInfo);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to get Flash block info for location 0x%lx: %s\n",
	       (unsigned long) i, ADMXRC2_GetStatusString(status));
	nError++;
	break;
      }

      printf("0x%lx\r", i);
      fflush(stdout);

      status = ADMXRC2_EraseFlash(card, blockInfo.Start, blockInfo.Length);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to erase Flash at 0x%lx-0x%lx: %s\n",
	       (unsigned long) blockInfo.Start,
	       (unsigned long) blockInfo.Start + blockInfo.Length - 1,
	       ADMXRC2_GetStatusString(status));
	nError++;
	break;
      }

      i += blockInfo.Length;
    }

    if (nError) {
      printf("*** Erase failed - aborting.\n");
      ret = -1;
      goto done;
    }
  }
		
  if (doProgram) {
    uint8_t*                data;
    unsigned long           i, remaining, chunk;
    ADMXRC2_FLASHBLOCK_INFO blockInfo;

    /* Write the bitstream into the flash */
    if (!nError) {
      printf("Writing bitstream image at 0x%lx-0x%lx...\n",
	     imageAddr, imageAddr + imageSize - 1);

      i = imageAddr;
      remaining = imageSize;
      data = (uint8_t*) image;
      while (remaining) {
	status = ADMXRC2_GetFlashBlockInfo(card, i, &blockInfo);
	if (status != ADMXRC2_SUCCESS) {
	  printf("*** Failed to get Flash block info for location 0x%lx: %s\n",
		 (unsigned long) i, ADMXRC2_GetStatusString(status));
	  nError++;
	  break;
	}

	chunk = blockInfo.Length - (i - blockInfo.Start);
	if (chunk > remaining) {
	  chunk = remaining;
	}
	printf("0x%lx\r", i);
	fflush(stdout);

	status = ADMXRC2_WriteFlash(card, i, chunk, data);
	if (status != ADMXRC2_SUCCESS) {
	  printf("*** Failed to write Flash at 0x%lx-0x%lx: %s\n",
		 i, i + chunk - 1, ADMXRC2_GetStatusString(status));
	  nError++;
	  break;
	}

	i += chunk;
	remaining -= chunk;
	data += chunk;
      }
    }

    if (nError) {
      printf("*** Programming failed - aborting\n");
      ret = -1;
      goto done;
    }
  }
	
  if (!doReadback) {
    ADMXRC2_FLASHBLOCK_INFO blockInfo;
    unsigned long           verifyStart;
    unsigned long           verifyLimit;
    unsigned long           i, j, remaining;
    unsigned long           pos;
    unsigned long           chunk;
    uint8_t*                buf;
    uint8_t*                img = (uint8_t*) image;

    if (doProgram || doVerify) {
      verifyStart = imageAddr;
      verifyLimit = imageAddr + imageSize;
      printf("Verifying data from 0x%lx-0x%lx...\n",
	     verifyStart, verifyLimit - 1);
    } else {
      verifyStart = eraseAddr;
      verifyLimit = eraseLimit;
      printf("Blank checking from 0x%lx-0x%lx...\n",
	     verifyStart, verifyLimit - 1);
    }

    remaining = verifyLimit - verifyStart;
    flash = malloc(remaining);
    if (flash == NULL) {
      printf("*** Failed to allocate memory to read back flash\n");
      ret = -1;
      goto done;
    }

    buf = flash;
    pos = verifyStart;
    i = 0;
    while (pos < verifyLimit) {
      status = ADMXRC2_GetFlashBlockInfo(card, pos, &blockInfo);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to get Flash block info for location 0x%lx: %s\n",
	       (unsigned long) i, ADMXRC2_GetStatusString(status));
	nError++;
	break;
      }

      chunk = blockInfo.Length - (pos - blockInfo.Start);
      if (chunk > remaining) {
	chunk = remaining;
      }
      printf("0x%lx\r", pos);
      fflush(stdout);

      status = ADMXRC2_ReadFlash(card, pos, chunk, buf);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to read back Flash memory at 0x%lx-0x%lx: %s\n",
	       i, i + chunk - 1, ADMXRC2_GetStatusString(status));
	nError++;
	break;
      }

      for (j = 0; j < chunk; j++) {
	if (img[i + j] != buf[j]) {
	  if (nError < 10) {
	    printf("*** Error at Flash location 0x%lx: expected=0x%2.2lx, got=0x%2.2lx\n",
		   (unsigned long) (pos + j), (unsigned long) img[i + j], (unsigned long) buf[j]);
	  } else if (nError == 10) {
	    printf("*** (more errors)\n");
	  }
	  nError++;
	}
      }

      i += chunk;
      remaining -= chunk;
      pos += chunk;
      buf += chunk;
    }

    if (nError) {
      printf("*** Flash %s failed with %ld data error(s)\n",
	     doChkblank ? "blank check" : "verification", (unsigned long) nError);
      goto done;
    } else {
      if (doProgram || doVerify) {
	printf("Flash is correct - bitstream verified.\n");
      } else {
	printf("Flash is correct - no bitstream present.\n");
      }
    }

    if (doProgram) {
      if (bBootFlag32Bit) {
	uint32_t val32;

	/*
	** Set the flag that causes the FPGA to be loaded from flash,
	** and bitstream length.
	*/
	val32 = (imageSize - 1) << 8;
	printf("Setting load-on-reset flag at 0x%lx-0x%lx to 0x%8.8lx...\n",
	       (unsigned long) bootFlagAddr,
	       (unsigned long) bootFlagAddr + sizeof(val32) - 1,
	       (unsigned long) val32);
	val32 = os_cpu_to_le32(val32); /* Boot flag is stored little-endian */
	status = ADMXRC2_WriteFlash(card, bootFlagAddr, sizeof(val32), &val32);
	if (status != ADMXRC2_SUCCESS) {
	  printf("*** Failed to set load-on-reset flag: %s\n", ADMXRC2_GetStatusString(status));
	  nError++;
	}
      } else {
	uint8_t val8;

				/* Set the flag that causes the FPGA to be loaded from flash */
	val8 = 0x0U;
	printf("Setting load-on-powerup flag at 0x%lx to 0x%2.2lx...\n",
	       (unsigned long) bootFlagAddr, (unsigned long) val8);
	status = ADMXRC2_WriteFlash(card, bootFlagAddr, 0x1U, &val8);
	if (status != ADMXRC2_SUCCESS) {
	  printf("*** Failed to set load-on-reset flag: %s\n", ADMXRC2_GetStatusString(status));
	  nError++;
	}
      }
    }

    if ((doProgram || doVerify) && !nError) {
      if (bBootFlag32Bit) {
	uint32_t expected, bootFlag;

	expected = (imageSize - 1) << 8;
	status = ADMXRC2_ReadFlash(card, bootFlagAddr, sizeof(bootFlag), &bootFlag);
	if (status != ADMXRC2_SUCCESS) {
	  printf("*** Failed to read load-on-reset flag: %s\n",
		 ADMXRC2_GetStatusString(status));
	  nError++;
	} else {
	  bootFlag = os_le32_to_cpu(bootFlag); /* Boot flag is stored little-endian */
	  if (bootFlag != expected) {
	    printf("*** Load-on-reset flag is 0x%8.8lx, should be 0x%8.8lx\n",
		   (unsigned long) bootFlag, (unsigned long) expected);
	    nError++;
	  }
	}
      } else {
	uint8_t expected, bootFlag;

	expected = 0x0U;
	status = ADMXRC2_ReadFlash(card, bootFlagAddr, 0x1U, &bootFlag);
	if (status != ADMXRC2_SUCCESS) {
	  printf("*** Failed to read back load-on-powerup flag: %s\n",
		 ADMXRC2_GetStatusString(status));
	  nError++;
	} else {
	  if (bootFlag != 0x0U) {
	    printf("*** Load-on-powerup flag is 0x%2lx, should be 0x%2lx\n",
		   (unsigned long) bootFlag, (unsigned long) expected);
	    nError++;
	  }
	}
      }

      if (nError) {
	printf("*** The bitstream may not be loaded on power up.\n");
	goto done;
      } else {
	printf("The bitstream will be loaded on power up.\n");
      }
    }
  } else {
    ADMXRC2_FLASHBLOCK_INFO blockInfo;
    unsigned long readStart;
    unsigned long readLength;
    unsigned long i, remaining;
    unsigned long pos;
    unsigned long chunk;
    size_t n;
    uint8_t* buf;
    FILE* file;

    readStart = arguments.param[2].uintVal;
    readLength = arguments.param[3].uintVal;
    printf("Reading data from 0x%lx-0x%lx...\n",
	   readStart, readStart + readLength - 1);

    flash = malloc(readLength);
    if (flash == NULL) {
      printf("*** Failed to allocate memory to read back flash\n");
      ret = -1;
      goto done;
    }

    buf = flash;
    remaining = readLength;
    pos = readStart;
    i = 0;
    while (i < readLength) {
      status = ADMXRC2_GetFlashBlockInfo(card, pos, &blockInfo);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to get Flash block info for location 0x%lx: %s\n",
	       (unsigned long) i, ADMXRC2_GetStatusString(status));
	nError++;
	break;
      }

      chunk = blockInfo.Length - (pos - blockInfo.Start);
      if (chunk > remaining) {
	chunk = remaining;
      }
      printf("0x%lx\r", pos);

      status = ADMXRC2_ReadFlash(card, pos, chunk, buf);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to read back Flash memory at 0x%lx-0x%lx: %s\n",
	       i, i + chunk - 1, ADMXRC2_GetStatusString(status));
	nError++;
	break;
      }

      if (nError) {
	break;
      }

      i += chunk;
      remaining -= chunk;
      pos += chunk;
      buf += chunk;
    }

    file = fopen(arguments.param[1].pathVal, "wb");
    if (file == NULL) {
      ret = -1;
      printf("*** Failed to open file '%s' for writing\n", arguments.param[1].pathVal);
      goto done;
    }
    n = fwrite(flash, 1, readLength, file);
    if (n != readLength) {
      ret = -1;
      printf("*** Failed to write all data for file '%s' (%ld vs %ld)\n",
	     arguments.param[1].pathVal, (long) n, (long) readLength);
      goto done;
    }
    fflush(file);
    fclose(file);
  }

 done:
  if (flash != NULL) {
    free(flash);
  }

  if (image != NULL) {
    if (doProgram || doVerify) {
      ADMXRC2_UnloadBitstream(image);
    } else {
      free(image);
    }
  }
	
  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }
   
  return ret;
}
