/*
** itest.c - example demonstrating interrupt wait functionality in
**           the ADM-XRC API via the ADMXRC2_WaitForInterrupt function.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**
** (C) Copyright 2000-2009 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <pthread.h>

#include <admxrc2.h>

#include "../common/common.h"
#include "../common/platform.h"

#define USE_TIMEOUT (1)

/* FPGA registers */
#define FPGA_IMASK	(0x0U)
#define FPGA_ISTAT	(0x4U)
#define FPGA_IARM	(0x8U)
#define FPGA_TEST	(0xcU)

static ADMXRC2_HANDLE card = ADMXRC2_HANDLE_INVALID_VALUE;

static unsigned long  signalled;

static pthread_t      int_thread;
static BOOLEAN        int_running = FALSE;
static BOOLEAN        terminate = FALSE;

static void*
interruptThread(void* arg)
{
  ADMXRC2_STATUS status;
  uint32_t       istat;
  uint32_t       dummy;
  uint32_t       val32;
  int            bError = 0;

  printf("Interrupt thread started.\n");
	
  while (1) {
#if USE_TIMEOUT
    unsigned long timeout = 10000000; /* 10 seconds */

    status = ADMXRC2_WaitForInterrupt(card, &timeout, 0, NULL);
#else
    status = ADMXRC2_WaitForInterrupt(card, NULL, 0, NULL);
#endif
    switch (status) {
    case ADMXRC2_SUCCESS:
#if USE_TIMEOUT
      printf("Remaining time: %ld us\n", timeout);
#endif
      break;
    case ADMXRC2_CANCELLED:
      break;
    case ADMXRC2_TIMEOUT:
      break;
    default:
      printf("*** Interrupt thread failed to wait for interrupt: %s\n",
	     ADMXRC2_GetStatusString(status));
      bError = TRUE;
      break;
    }

    if (bError || terminate) {
      break;
    }

    /* Get the pending FPGA interrupts */
    ADMXRC2_Read(card,
		 ADMXRC2_IOWIDTH_32,
		 ADMXRC2_IOLOCAL,
		 FPGA_ISTAT,
		 &val32,
		 sizeof(val32));
    istat = os_le32_to_cpu(val32);
    if (istat) {
      signalled++;
    }

    /* Clear the pending FPGA interrupts */
    val32 = os_cpu_to_le32(istat);
    ADMXRC2_Write(card,
		  ADMXRC2_IOWIDTH_32,
		  ADMXRC2_IOLOCAL,
		  FPGA_ISTAT,
		  &val32,
		  sizeof(val32));

    /* Rearm the FINTI_L line by forcing it high for 1 cycle */
    ADMXRC2_Write(card,
		  ADMXRC2_IOWIDTH_32,
		  ADMXRC2_IOLOCAL,
		  FPGA_IARM,
		  &dummy,
		  sizeof(dummy));
  }

  return (void*) 0;
}

int main(int argc, char *argv[])
{
  const double      clock_freq = 33.0e6;
  int               ret = 0;
  ADMXRC2_STATUS    status;
  ADMXRC2_CARD_INFO cardInfo;
  int               nint;
  const char*       filename;
  Arguments         arguments;
  Option            options[2];
  uint32_t          val32;
  int               bTerminate;

  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  arguments.nOption = 2;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  parse_command_line("itest", argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    printf("*** Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS)
      {
	printf("*** Failed to open card with ID %ld: %s\n",
	       (unsigned long) cardID, ADMXRC2_GetStatusString(status));
	ret = -1;
	goto done;
      }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to get card information: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
	
  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADPEXRC4FX:
  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADCPXRC4LX:
  case ADMXRC2_BOARD_ADMAMC5A2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
    break;

  default:
    printf("*** This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRC-II-Lite\n");
    printf("\tADM-XRC-II\n");
    printf("\tADM-XPL\n");
    printf("\tADM-XP\n");
    printf("\tADP-WRC-II\n");
    printf("\tADP-DRC-II\n");
    printf("\tADP-XPI\n");
    printf("\tADM-XRC-4LX\n");
    printf("\tADM-XRC-4SX\n");
    printf("\tADM-XRC-4FX\n");
    printf("\tADPE-XRC-4FX\n");
    printf("\tADM-XRC-5LX\n");
    printf("\tADM-XRC-5T1\n");
    printf("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV\n");
    printf("\tADM-XRC-5TZ\n");
    printf("\tADM-XRC-5T-DA1\n");
    ret = -1;
    goto done;
  }

  /* Set local bus clock to nominal frequency */
  status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, clock_freq, NULL);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to set LCLK to %.1fMHz: %s\n",
	   clock_freq / 1.0e6f, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  filename = get_fpga_filename(&cardInfo, "itest", &arguments);
  status = ADMXRC2_ConfigureFromFile(card, filename);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to load the bitstream '%s': %s\n",
	   filename, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /* Start up the thread that waits for interrupts. */
  signalled = 0;
  terminate = FALSE;
  if (pthread_create(&int_thread, NULL, interruptThread, NULL)) {
    printf("*** Failed to start interrupt thread\n");
    ret = -1;
    goto done;
  }
  int_running = TRUE;

  /* Enable interrupts in the FPGA */
  val32 = os_cpu_to_le32(0x0U);
  status = ADMXRC2_Write(card,
			 ADMXRC2_IOWIDTH_32,
			 ADMXRC2_IOLOCAL,
			 FPGA_IMASK,
			 &val32,
			 sizeof(val32));
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Write to FPGA_IMASK failed: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /*
  ** Allow the user to generate interrupts by pressing enter.
  ** Ctrl-Z exits this loop.
  */
  nint = 0;
  bTerminate = FALSE;
  while (1) {
    int      c;

    printf("Enter 'q' to quit, or hit enter to generate an interrupt:\n");
    while (1) {
      c = getchar();
      if (c == EOF || c == 'q' || c == 'Q') {
	bTerminate = TRUE;
	break;
      }
      if (c == '\n') {
	break;
      }
    }

    if (bTerminate) {
      break;
    }

    val32 = os_cpu_to_le32(0x1U);
    status = ADMXRC2_Write(card,
			   ADMXRC2_IOWIDTH_32,
			   ADMXRC2_IOLOCAL,
			   FPGA_TEST,
			   &val32,
			   sizeof(val32));
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Write to FPGA_TEST failed: %s\n", ADMXRC2_GetStatusString(status));
    }
    nint++;
  }

  printf("Generated %ld interrupts\n", (unsigned long) nint);
  printf("Interrupt thread saw %ld interrupt(s)\n", (unsigned long) signalled);

 done:
  if (int_running) {
    /* Force the interrupt thread to terminate */
    terminate = TRUE;
    ADMXRC2_CancelWaitForInterrupt(card);
    pthread_join(int_thread, NULL);
  }

  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }
	
  return ret;
}
