/*
** master.c - Example demonstrating direct master cycles initiated
**            by the FPGA.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**
** (C) Copyright 2000-2006 Alpha Data
**
**
** NOTES ON OPERATION OF THIS EXAMPLE
**
** 1. Local bus direct master window
**
**   We configure the PCI9080/9656 to give us an 8MB window which is
**   direct mapped onto an 8MB section of the 32-bit PCI address space.
**
**   To allow the FPGA to access the full 32-bit PCI memory address
**   space, the PLX.DMPBAM register is used as a paging register. The
**   32-bit PCI address is composed of the lower 21 bits of the local
**   bus address that is output by the FPGA, and the upper 9 bits of
**   PLX.DMPBAM, as follows:
**
**     PCI_address[1:0]   = 0
**     PCI_address[22:2]  = LA[22:2]
**     PCI_address[31:23] = PLX.DMPBAM[31:23]
**
** 2. PCI9080/9656 local bus configuration register settings
**
**     We set PLX.DMRR              to 0xff800000
**     We set PLX.DMLBAM            to 0xff000000
**     We maintain PLX.DMPBAM[15:0] at 0x0001
**     We maintain PLX.DMPBAM[31:16] as a paging register
**
**   This means that the local bus to PCI window for direct master
**   access is 8MB in size, from 0xff000000 to 0xff7fffff on the
**   local bus.
**
** 3. LA[] usage
**
**   In earlier models in the ADM-XRC range, not all of LA[31:2] is
**   connected between the FPGA and the PCI9080. In later models
**   such as the ADM-XRCII-Lite and ADM-XRCII, all bits of LA[31:2]
**   are connected between the FPGA and the PCI9080/PCI9656.
**
**     - In the ADM-XRC, LA[23:2] are connected between the FPGA and
**       the PCI9080. LA[31:24] are pulled up. Thus, the PCI9080 always
**       sees addresses in the range 0xff000000 to 0xfffffffc when the
**       FPGA is the local bus master.
**
**     - In the ADM-XRCII-Lite, all of LA[31:2] are connected between the
**       FPGA and the PCI9080. Thus the PCI9080 sees addresses in the
**       range 0x00000000 to 0xfffffffc when the FPGA is the local bus
**       master.
**
**     - In the ADM-XRCII, all of LA[31:2] are connected between the FPGA
**       and the PCI9656. Thus the PCI9656 sees addresses in the range
**       0x00000000 to 0xfffffffc when the FPGA is the local bus master.
**
** 4. Access to PCI9080/9656 registers by the FPGA
**
**   To access the local bus configuration registers in the PCI9080/9656,
**   the FPGA must drive an address where LA[23:12] == 12'hfff. The
**   control logic on the card will then assert the S0# signal to the
**   PCI9080/9656. This causes the PCI9080/9656 to treat the access as a
**   read or write of its registers, from the local bus side.
**
**   Accesses to the PCI9080/9656's registers by a direct master capable
**   FPGA takes precedence over the local bus to PCI window. In other
**   words, if the local bus to PCI window were programmed to overlap the
**   range that results in PCI9080/9656 register access, the result would
**   be a PLX register access rather than a PCI access.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <ctype.h>

#include <admxrc2.h>

#include "../common/common.h"
#include "../common/platform.h"

#define BUFFSIZE (16)

#define DIRECT_PCI_WINDOW_SHIFT (23U)
#define DIRECT_PCI_WINDOW_SIZE  (1U << DIRECT_PCI_WINDOW_SHIFT)
#define DIRECT_PCI_WINDOW_MASK  (DIRECT_PCI_WINDOW_SIZE - 1)

/* PLX register offsets (from PCI side) */
#define PLX_PCI_DMRR		(0x1cU)
#define PLX_PCI_DMLBAM		(0x20U)
#define PLX_PCI_DMPBAM		(0x28U)

/* PLX register offsets (from local bus side) */
#define PLX_LOCAL_DMPBAM	(0xa8U)

static ADMXRC2_HANDLE     card = ADMXRC2_HANDLE_INVALID_VALUE;

static uint32_t*          user_buffer = NULL;
static ADMXRC2_BUFFERMAP  map;
static uint32_t           pages_pci[2];
static ADMXRC2_DMADESC    desc;
static BOOLEAN            descValid = FALSE;

static volatile uint32_t* fpga_space;

static void
dump_map(
  ADMXRC2_BUFFERMAP* map)
{
  uint32_t i;
	
  printf("PageSpan = 0x%8.8lx ByteSpan=0x%8.8lx InitOffset=0x%8.8lx\n"
	 "PageLength=0x%8.8lx PageBits=%ld\n",
	 (unsigned long) map->PagesSpanned,
	 (unsigned long) map->BytesSpanned,
	 (unsigned long) map->InitOffset,
	 (unsigned long) map->PageLength,
	 (unsigned long) map->PageBits);
  for (i = 0; i < map->PagesSpanned; i++) {
    printf("PagesPci[% 3ld] = 0x%8.8lx\n",
	   (unsigned long) i,
	   (unsigned long) map->PagesPci[i]);
  }
}

static char
menu(void)
{
  static BOOLEAN wait_nl = FALSE;
  int            c;

  printf("Enter command, (i)nit buffer, (r)read, (s)how buffer, (w)rite or (q)uit: ");
  while (1) {
    c = getchar();
    if (c == EOF) {
      c = 'q';
      break;
    } else {
      if (c == '\n') {
	wait_nl = FALSE;
      } else {
	if (!wait_nl) {
	  if (isprint(c)) {
	    wait_nl = TRUE;
	    break;
	  }
	}
      }
    }
  }

  return tolower((char) c);
}

static void
init_buffer(void)
{
  uint32_t i;

  user_buffer[0] = 0xdeadbeef;
  user_buffer[1] = 0xcafeface;
  for (i = 2; i < BUFFSIZE; i++) {
    user_buffer[i] = i;
  }
}

static void
show_buffer(void)
{
  uint32_t i, j;

  printf("  --> contents of user buffer:\n");
  for (i = 0; i < BUFFSIZE; i += 4) {
    printf("      index %2.2lx: ", (unsigned long) i);
    for (j = 0; j < 4; j++) {
      printf("0x%8.8lx ", (unsigned long) (user_buffer[i + j]));
    }
    printf("\n");
  }
  printf("\n");
}

static uint32_t
get_pci_addr(uint32_t offs)
{
  uint32_t pg_idx, pg_offs;
  uint32_t pci;

  offs += map.InitOffset;
  pg_idx = offs >> map.PageBits;
  pg_offs = offs % (map.PageLength - 1);

  pci = pages_pci[pg_idx] + pg_offs;

  return pci;
}

static void
set_DMPBAM(uint32_t pci_addr)
{
  uint32_t page;

  page = pci_addr & ~DIRECT_PCI_WINDOW_MASK;

  /* Supply data to write to DMPBAM */
  fpga_space[1] = os_cpu_to_le32(page | 0x1U);
  fpga_space[1];

  /* Select write operation */
  fpga_space[2] = os_cpu_to_le32(0x1);
  fpga_space[2];

  /* Write PLX.DMPBAM */
  fpga_space[0] = os_cpu_to_le32(0xfff000U | PLX_LOCAL_DMPBAM);
  fpga_space[0];
}

static void
do_read(
  uint32_t  index,
  uint32_t* data)
{
  uint32_t pci_addr;
  uint32_t val;
  uint32_t offs;

  offs = index * sizeof(uint32_t);

  pci_addr = get_pci_addr(offs);

  /* Set up paging bits of PLX.DMPBAM */
  set_DMPBAM(pci_addr);

  /* Select read operation */
  fpga_space[2] = os_cpu_to_le32(0x0U);
  fpga_space[2];

  /* Supply address & do the read from the direct master window */
  val = pci_addr & (DIRECT_PCI_WINDOW_MASK & ~0x3U);
  fpga_space[0] = os_cpu_to_le32(val);
  fpga_space[0];

  /*
  ** Read back the data read by the FPGA. Note that we don't
  ** perform endian-conversion, because when the FPGA reads it
  ** from host memory, the 32-bit value will already be in
  ** native endian format!
  */
  *data = fpga_space[3];
}

static void
do_write(
  uint32_t index,
  uint32_t data)
{
  uint32_t pci_addr;
  uint32_t val;
  uint32_t offs;

  offs = index * sizeof(uint32_t);

  pci_addr = get_pci_addr(offs);

  /* Set up paging bits of PLX.DMPBAM */
  set_DMPBAM(pci_addr);

  /*
  ** Supply data to write. Note that we don't perform endian-
  ** conversion, because when the FPGA writes it to host memory,
  ** the FPGA will not perform endian-conversion. Hence we should
  ** supply the value here in native endian format!
  */
  fpga_space[1] = data;
  fpga_space[1];

  /* Select write operation */
  fpga_space[2] = os_cpu_to_le32(0x1);
  fpga_space[2];

  /* Supply address & do the write to the direct master window */
  val = pci_addr & (DIRECT_PCI_WINDOW_SIZE - 1) & ~0x3U;
  fpga_space[0] = os_cpu_to_le32(val);
  fpga_space[0];
}

int
main(int argc, char *argv[])
{
  const double       clock_freq = 33.0e6;
  int                ret = 0;
  ADMXRC2_STATUS     status;
  ADMXRC2_CARD_INFO  cardInfo;
  ADMXRC2_SPACE_INFO spInfo;
  BOOLEAN            quit;
  const char*        filename;
  char               choice;
  Arguments          arguments;
  Option             options[2];
  uint32_t           val32;

  /* Set up expected options */
  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  arguments.nOption = 2;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  /* Parse command line */
  parse_command_line("master", argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    printf("Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with ID %ld: %s\n",
	     (unsigned long) cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2_LITE:
  case ADMXRC2_BOARD_ADMXRC2:
    break;

  default:
    printf("This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRCII-Lite\n");
    printf("\tADM-XRCII\n");
    ret = -1;
    goto done;
  }

  status = ADMXRC2_GetSpaceInfo(card, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get space 0 info: %s\n",
	   ADMXRC2_GetStatusString(status));
    return -1;
  }
  fpga_space = (volatile uint32_t*) spInfo.VirtualBase;

  /* Set local bus clock to nominal frequency */
  status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, clock_freq, NULL);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to set LCLK to %.1fMHz: %s\n",
	   clock_freq / 1.0e6f, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  filename = get_fpga_filename(&cardInfo, "master", &arguments);
  status = ADMXRC2_ConfigureFromFile(card, filename);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to load the bitstream '%s': %s\n",
	    filename, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /* Allocate the user buffer */
  user_buffer = (uint32_t*) ADMXRC2_Malloc(sizeof(uint32_t) * BUFFSIZE);
  if (user_buffer == NULL) {
    printf("Failed to allocate user buffer\n");
    ret = -1;
    goto done;
  }

  /* Lock down the user buffer and obtain a handle to it */
  status = ADMXRC2_SetupDMA(card, user_buffer, sizeof(uint32_t) * BUFFSIZE, 0, &desc);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to lock user buffer: %s\n",
	    ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /*
  ** Obtain a scatter-gather map of the user buffer. We don't
  ** expect the user buffer to span more than 2 pages
  */
  map.MaxPages = 2;
  map.PagesPci = pages_pci;
  status = ADMXRC2_MapDirectMaster(card, desc, 0, sizeof(uint32_t) * BUFFSIZE, &map);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to map user buffer: %s\n",
	    ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /*
  ** Initialise the PCI9080/9656 for Direct Master access.
  **
  ** We set PLX.DMRR   to 0xff800000
  ** We set PLX.DMLBAM to 0xff000000
  **
  ** See notes at top of file for more information.
  */
  val32 = 0xff800000U;
  status = ADMXRC2_Write(card,
			 ADMXRC2_IOWIDTH_32,
			 ADMXRC2_IOADAPTER,
			 PLX_PCI_DMRR,
			 &val32,
			 sizeof(val32));
  if (status != ADMXRC2_SUCCESS) {
    printf ("Failed to set up PLX.DMRR register: %s\n",
	    ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  val32 = 0xff000000U;
  status = ADMXRC2_Write(card,
			 ADMXRC2_IOWIDTH_32,
			 ADMXRC2_IOADAPTER,
			 PLX_PCI_DMLBAM,
			 &val32,
			 sizeof(val32));
  if (status != ADMXRC2_SUCCESS) {
    printf ("Failed to set up PLX.DMLBAM register: %s\n",
	    ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /* Initialise the user buffer to some recognisable values */
  init_buffer();

  printf("ADM-XRC Series Direct Master demonstration\n\n");

  show_buffer();

  quit = FALSE;
  while (!quit) {
    choice = menu();
    switch (choice) {
    case 'd':
      {
	dump_map(&map);
	break;
      }
    case 'i':
      {
	init_buffer();
	printf("  --> reinitialised buffer.\n\n");
	show_buffer();
	break;
      }

    case 'r':
      {
        unsigned long idx;
	uint32_t      val;

	printf("Enter buffer index (hex): ");
	scanf("%lx", &idx);
	if (idx >= BUFFSIZE)
	  {
	    printf("*** invalid index: 0x%lx\n", (unsigned long) idx);
	    break;
	  }
	do_read(idx, &val);
	printf("  --> FPGA read 0x%8.8lx at index 0x%lx\n\n",
	       (unsigned long) val, (unsigned long) idx);
	break;
      }

    case 's':
      {
	show_buffer();
	break;
      }

    case 'q':
      {
	quit = TRUE;
	break;
      }

    case 'w':
      {
	unsigned long idx, val;

	printf("Enter buffer index (hex): ");
	scanf("%lx", &idx);
	if (idx >= BUFFSIZE) {
	  printf("*** invalid index: 0x%lx\n", (unsigned long) idx);
	  break;
	}
	printf("Enter value to write (hex): ");
	scanf("%lx", &val);
	do_write(idx, (uint32_t) val);
	printf("  --> FPGA wrote 0x%8.8lx to index 0x%lx\n\n",
	       (unsigned long) val, (unsigned long) idx);
	break;
      }

    default:
      {
	printf("Invalid option.\n");
	break;
      }
    }
  }

 done:
  if (descValid) {
    ADMXRC2_UnsetupDMA(card, desc);
  }

  if (user_buffer != NULL) {
    ADMXRC2_Free(user_buffer);
  }

  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }

  return ret;
}
