/*
** memoryf.c - Example demonstrating how to perform chip-driven memory test
**             on an Alpha Data reconfigurable computing card.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XP
**   o ADM-XPL
**   o ADP-DRC-II
**   o ADP-WRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**
** (C) Copyright 2008-2009 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>

#include <admxrc2.h>

#include "../common/common.h"
#include "../common/platform.h"

typedef int boolean_t;
typedef struct timeval timestamp_t;

/* If nonzero, host reads memory during memory test to try to provoke some kind of error */
#define HOST_READ_ACTIVITY (1)

/*
** Register indices and bit definitions
*/

#define BANK_REG                        (0x0)               /* Bank select register */
#define BANK_REG_BANKMASK               (0xfU << 0)

#define PAGE_REG                        (0x1)               /* Page register */
#define PAGE_REG_PAGEMASK               (0x7ffU << 0)

#define MEMCTL_REG                      (0x2)               /* Memory control register */
#define MEMCTL_REG_MEMRST               (0x1U << 0)

#define STATUS_REG                      (0x4)               /* Status register */
#define STATUS_REG_LCLKLOCKED           (0x1U << 0)
#define STATUS_REG_LCLKSTICKY           (0x1U << 1)
#define STATUS_REG_SHIFT_LOCKED         (8)
#define STATUS_REG_SHIFT_STICKY         (16)

#define MEMSTAT_REG                     (0x6)               /* Memory status register */
#define MEMSTAT_REG_SHIFT_TRAINED       (0)

#define MODEx_REG(n)                    (0x10 + (n))        /* Bank 'n' mode register */
#define MODEx_REG_DDR2SSRAM_BLMASK      (0x3U << 0)         /* DDR2 SSRAM burst length mask */
#define MODEx_REG_DDR2SSRAM_BL2         (0x0U << 0)         /* DDR2 SSRAM burst length 2 */
#define MODEx_REG_DDR2SSRAM_BL4         (0x1U << 0)         /* DDR2 SSRAM burst length 4 */
#define MODEx_REG_DDR2SSRAM_DLLON       (0x0U << 2)         /* DDR2 SSRAM DLL enable */
#define MODEx_REG_DDR2SSRAM_DLLOFF      (0x1U << 2)         /* DDR2 SSRAM DLL disable */

#define MODEx_REG_ZBTSSRAM_FLOWTHRU     (0x0U << 0)         /* ZBT SSRAM pipelined mode disable */
#define MODEx_REG_ZBTSSRAM_PIPELINE     (0x1U << 0)         /* ZBT SSRAM pipelined mode enable */

#define MODEx_REG_DDRSDRAM_UNBUFFERED   (0x0U << 0)
#define MODEx_REG_DDRSDRAM_REGISTERED   (0x1U << 0)
#define MODEx_REG_DDRSDRAM_X8X16        (0x0U << 1)
#define MODEx_REG_DDRSDRAM_X4           (0x1U << 1)
#define MODEx_REG_DDRSDRAM_ROWBITS_12   (0x0U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_13   (0x1U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_14   (0x2U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_15   (0x3U << 2)
#define MODEx_REG_DDRSDRAM_COLBITS_M4   (0x0U << 4)         /* col bits = row bits - 4 */
#define MODEx_REG_DDRSDRAM_COLBITS_M3   (0x1U << 4)         /* col bits = row bits - 3 */
#define MODEx_REG_DDRSDRAM_COLBITS_M2   (0x2U << 4)         /* col bits = row bits - 2 */
#define MODEx_REG_DDRSDRAM_COLBITS_M1   (0x3U << 4)         /* col bits = row bits - 1 */
#define MODEx_REG_DDRSDRAM_BANKBITS_0   (0x0U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_1   (0x1U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_2   (0x2U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_3   (0x3U << 6)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_1  (0x0U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_2  (0x1U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_4  (0x2U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_8  (0x3U << 8)

#define SPD_REG                         (0x5)
#define SPD_REG_DO                      (0x1U << 31)
#define SPD_REG_BUSY                    (0x1U << 31)
#define SPD_REG_SHIFT_ADDR              (0)
#define SPD_REG_SHIFT_WRDATA            (8)
#define SPD_REG_SHIFT_RDDATA            (16)
#define SPD_REG_ERR                     (0x1U << 24)
#define SPD_REG_SHIFT_SLOT              (25)

#define USER_REG(x)                     (0x40U | (x))

#define MIN(x, y) ((x) < (y) ? (x) : (y))

/* Max. number of memory banks supported by this application */
#define MAX_BANK (16)

/* Max. logical width of a memory bank, in bytes */
#define MAX_BANK_WIDTH (64)

struct _Globals {
  void* pFpgaSpace;
  volatile uint32_t* pFpgaRegs;
  volatile uint32_t* pMemory;
  ADMXRC2_HANDLE hCard;
  uint32_t bankTestMask;
  boolean_t bBankDdr[MAX_BANK];
  unsigned int bankWidth[MAX_BANK];
  uint64_t bankSize[MAX_BANK];
  boolean_t bUse64Bit;
};

typedef struct _Globals Globals;

static boolean_t
readSpdByte(
  Globals* pGlob,
  unsigned int slot,
  unsigned int index,
  uint8_t* value)
{
  volatile uint32_t* pFpgaRegs = pGlob->pFpgaRegs;
  uint32_t val32;
  uint32_t stat32;
  boolean_t bFail;
  
  val32 = (index & 0xffU) << SPD_REG_SHIFT_ADDR;
  val32 |= (slot & 0x7U) << SPD_REG_SHIFT_SLOT;
  pFpgaRegs[SPD_REG] = os_cpu_to_le32(val32);
  val32 |= SPD_REG_DO;
  pFpgaRegs[SPD_REG] = os_cpu_to_le32(val32);
  stat32 = pFpgaRegs[SPD_REG];
  while (1) {
    stat32 = pFpgaRegs[SPD_REG];
    stat32 = os_le32_to_cpu(stat32);
    if (!(stat32 & SPD_REG_BUSY)) {
      break;
    }
  }
  if (stat32 & SPD_REG_ERR) {
    bFail = TRUE;
  } else {
    bFail = FALSE;
  }
  *value = (uint8_t) (stat32 >> SPD_REG_SHIFT_RDDATA);
  return bFail;
}

static boolean_t
readSpd(
  Globals* pGlob,
  unsigned int slot,
  uint8_t data[128])
{
  unsigned int i;
  boolean_t bFail;
  
  for (i = 0; i < 128; i++) {
    bFail = readSpdByte(pGlob, slot, i, &data[i]);
    if (bFail || (i == 0 && data[i] == 0xffU)) {
      break;
    }
  }
  return (!bFail && data[0] != 0xffU);
}

static void
selectBank(
  Globals* pGlob,
  unsigned int bank)
{
  uint32_t val32 = (uint32_t)bank & BANK_REG_BANKMASK;

  pGlob->pFpgaRegs[BANK_REG] = os_cpu_to_le32(val32);
  pGlob->pFpgaRegs[BANK_REG];
}

/*
** Check that local bus clock DCM has achieved lock.
*/
static boolean_t
waitForLclkDcm(
  Globals* pGlob)
{
  uint32_t status;
  uint32_t val32;

  /*
  ** Delay for 500ms to allow DLLs/DCMs time to lock, and then verify
  ** that they are locked as expected.
  */
  sleep_ms(500);
  status = pGlob->pFpgaRegs[STATUS_REG];
  status = os_le32_to_cpu(status);
  printf("STATUS=0x%8.8lx\n", (unsigned long)status);
  if (!(status & STATUS_REG_LCLKLOCKED)) {
    printf("*** LCLK DCM not locked\n");
    return FALSE;
  }
  /* Clear "sticky" loss of lock bit */
  val32 = STATUS_REG_LCLKSTICKY;
  pGlob->pFpgaRegs[STATUS_REG] = os_cpu_to_le32(val32);
  pGlob->pFpgaRegs[STATUS_REG];
  
  return TRUE;
}

/*
** Check that DCMs/PLLs in memory clock domain have achieved lock, and that
** any memory banks that are self-training have trained themselves without
** error.
*/
static boolean_t
checkForMemoryLock(
  Globals* pGlob,
  unsigned int numLockFlags)
{
  uint32_t status, memstat, mask, val32;

  mask = (1 << numLockFlags) - 1;
  status = pGlob->pFpgaRegs[STATUS_REG];
  status = os_le32_to_cpu(status);
  memstat = pGlob->pFpgaRegs[MEMSTAT_REG];
  memstat = os_le32_to_cpu(memstat);
  printf("STATUS=0x%8.8lx MEMSTAT=0x%8.8lx\n",
	 (unsigned long)status, (unsigned long)memstat);
  val32 = status >> STATUS_REG_SHIFT_LOCKED;
  val32 &= mask;
  if (val32 != mask) {
    printf("*** Not all memory lock flags asserted: expected=0x%lx actual=0x%lx\n",
	   (unsigned long)mask, (unsigned long)val32);
    return FALSE;
  }
  val32 = memstat >> MEMSTAT_REG_SHIFT_TRAINED;
  val32 &= pGlob->bankTestMask;
  if (val32 != pGlob->bankTestMask) {
    printf("*** Not all memory banks trained: expected=0x%lx actual=0x%lx\n",
	   (unsigned long)pGlob->bankTestMask, (unsigned long)val32);
    return FALSE;
  }
  /* Clear sticky loss-of-lock / loss-of-trained bits */
  val32 = 0xffU << STATUS_REG_SHIFT_STICKY;
  pGlob->pFpgaRegs[STATUS_REG] = os_cpu_to_le32(val32);
  pGlob->pFpgaRegs[STATUS_REG];
  
  return TRUE;
}

/*
** Check that DLLs/DCMs/PLLs in memory clock domain did not lose lock during
** the test, and that memory banks did not transition to untrained.
*/
static boolean_t
checkStickyStatus(
  Globals* pGlob,
  unsigned int numLockFlags)
{
  uint32_t status, memstat, mask, val32;
  
  mask = (1 << numLockFlags) - 1;
  status = pGlob->pFpgaRegs[STATUS_REG];
  status = os_le32_to_cpu(status);
  memstat = pGlob->pFpgaRegs[MEMSTAT_REG];
  memstat = os_le32_to_cpu(memstat);
  printf("STATUS=0x%8.8lx MEMSTAT=0x%8.8lx\n",
	 (unsigned long)status, (unsigned long)memstat);
  val32 = status >> STATUS_REG_SHIFT_STICKY;
  val32 &= mask;
  if (val32) {
    printf("*** Loss of lock occurred: flags=0x%lx\n", (unsigned long)val32);
    return FALSE;
  }
  
  return TRUE;
}

static boolean_t
getDimmMode(
  ADMXRC2_CARD_INFO* pCardInfo,
  ADMXRC2_BANK_INFO* pBankInfo,
  Globals* pGlob,
  unsigned int bank,
  uint32_t* pMode,
  boolean_t* x4)
{
  uint8_t spd[128];
  boolean_t bPresent;
  unsigned long size;
  unsigned int rows, cols, banks, pbanks;
  unsigned int devWidth;
  uint8_t clMask;
  uint8_t modAttr;
  boolean_t bRegistered;
  uint32_t mode = 0;
  
  bPresent = readSpd(pGlob, bank, spd);
  if (bPresent) {
    rows = spd[3];      /* Number of bits used for row address on devices */
    cols = spd[4];      /* Number of bits used for column address on devices */
    pbanks = spd[5];    /* Number of physical banks on DIMM */
    devWidth = spd[13]; /* DQ width of devices */
    banks = spd[17];    /* Number of banks on devices */
    clMask = spd[18];   /* CAS latencies supported by devices */
    modAttr = spd[21];  /* DIMM module attributes */
    switch (devWidth) {
    case 4:
    case 8:
    case 16:
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): unknown device data width (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)devWidth);
      return FALSE;
    }
    if (banks < 1 || banks > 8) {
      printf(" *** Bank %ld (DDR SDRAM): unsupported device banks value (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)banks);
      return FALSE;
    }
    if (!(clMask & 0x8U)) {
      printf(" +++ Bank %ld (DDR SDRAM): CAS latency 2.5 not supported by devices, not testing\n",
	     (unsigned long)bank);
      return FALSE;
    }
    switch (modAttr) {
    case 0x20: /* Unbuffered */
      bRegistered = FALSE;
      break;
    case 0x26: /* Registered */
      bRegistered = TRUE;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid DIMM module attributes value (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)modAttr);
      return FALSE;
    }
    switch (rows) {
    case 12:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_12;
      break;
    case 13:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
      break;
    case 14:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_14;
      break;
    case 15:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_15;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid number of row bits (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)rows);
      return FALSE;
    }
    switch (cols + 4 - rows) {
    case 0:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
      break;
    case 1:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M2;
      break;
    case 3:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M1;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid number of column bits (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)cols);
      return FALSE;
    }
    switch (banks) {
    case 1:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_0;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_1;
      break;
    case 4:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
      break;
    case 8:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid number of banks (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)banks);
      return FALSE;
    }
    switch (pbanks) {
    case 1:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_2;
      break;
    case 4:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_4;
      break;
    case 8:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_8;
      break;
    default:
      printf(" +++ Bank %ld (DDR SDRAM): invalid number of physical banks (0x%2.2lx), not testing\n",
	     (unsigned long)bank, (unsigned long)pbanks);
      return FALSE;
    }
    if (bRegistered) {
      mode |= 0x1U;
    }
    if (devWidth == 4) {
      mode |= 0x2U;
    }
    pBankInfo[bank].Size = (1 << (rows + cols - 1)) * banks * pbanks;
    size = (pBankInfo[bank].Size >> 20) * (pBankInfo[bank].Width >> 3);
    printf(" Bank %ld (DDR SDRAM): x%ld %s rows=%ld cols=%ld banks=%ld physBanks=%ld %ld MiB\n",
	   (unsigned long)bank,
	   (unsigned long)devWidth,
	   bRegistered ? "registered" : "unbuffered",
	   (unsigned long)rows, (unsigned long)cols,
	   (unsigned long)banks, (unsigned long)pbanks, (unsigned long)size);
  } else {
    printf(" +++ Bank %ld (DDR SDRAM): SPD data not found, not testing\n",
	   (unsigned long)bank);
    return FALSE;
  }

  /* Bank SPD seems OK */
  *pMode = mode;
  *x4 = (devWidth == 4);
  return TRUE;
}

static void
setMemoryPortConfiguration(
  Globals* pGlob,
  uint32_t* modes)
{
  unsigned int i;

  printf("Configuring memory ports...\n");
  for (i = 0; i < MAX_BANK; i++) {
    if (!(pGlob->bankTestMask & (0x1UL << i))) {
      continue;
    }
    pGlob->pFpgaRegs[MODEx_REG(i)] = os_cpu_to_le32(modes[i]);
  }
}

static void
scanMemoryPorts(
  ADMXRC2_CARD_INFO* pCardInfo,
  ADMXRC2_BANK_INFO* pBankInfo,
  Globals* pGlob,
  uint32_t* modes)
{
  unsigned int i;
  boolean_t dimmX4[4];
  
  /* Initialize memory bank config registers where necessary */
  printf("Scanning memory ports...\n");
  for (i = 0; i < pCardInfo->NumRAMBank; i++) {
    if (!(pGlob->bankTestMask & (0x1U << i))) {
      continue;
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SRAM_DDR2) {
      /* Select BL4 for DDR-II SSRAM and ensure DLL is enabled */
      modes[i] = MODEx_REG_DDR2SSRAM_BL4 | MODEx_REG_DDR2SSRAM_DLLON;
      pGlob->bBankDdr[i] = TRUE;
      printf(" Bank %ld (DDR-II SRAM): burst length 4, DLL enabled\n",
	     (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_ZBTP) {
      /* Select pipelined mode for ZBT SSRAM */
      modes[i] = MODEx_REG_ZBTSSRAM_PIPELINE;
      pGlob->bBankDdr[i] = FALSE;
      printf(" Bank %ld (ZBT SSRAM): pipelined mode\n",
	     (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_ZBTFT) {
      /* Select flowthrough mode for ZBT SSRAM */
      modes[i] = MODEx_REG_ZBTSSRAM_FLOWTHRU;
      pGlob->bBankDdr[i] = FALSE;
      printf(" Bank %ld (ZBT SSRAM): flowthrough mode\n",
	     (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR) {
      uint32_t mode;
      
      /* Select registered, row bits, col bits, bank bits and phys bank bits for DDR SDRAM */
      switch (pCardInfo->BoardType) {
      case ADMXRC2_BOARD_ADMXPL:
      case ADMXRC2_BOARD_ADMXP:
	mode = 0;
	if (0x800000 == pBankInfo[i].Size) {
	  /* unbufferred, row bits 13, col bits 9, bank bits 2, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR SDRAM): unbuffered, row 13, col 9, bank 2, phys banks 1\n",
		 (unsigned long)i);
	} else if (0x1000000 == pBankInfo[i].Size) {
	  /* unregistered, row bits 13, col bits 10, bank bits 2, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR SDRAM): unbuffered, row 13, col 10, bank 2, phys banks 1\n",
		 (unsigned long)i);
	} else if (0x2000000 == pBankInfo[i].Size) {
	  /* unregistered, row bits 13, col bits 11, bank bits 2, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M2;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR SDRAM): unbuffered, row 13, col  11, bank 2, phys banks 1\n",
		 (unsigned long)i);
	} else {
	  pGlob->bankTestMask &= ~(0x1U << i);
	  printf(" +++ Bank %ld (DDR SDRAM): invalid size (0x%lx words), not testing\n",
		 (unsigned long)i, (unsigned long)pBankInfo[i].Size);
	}
	modes[i] = mode;
	pGlob->bBankDdr[i] = TRUE;
	break;
	
      case ADMXRC2_BOARD_ADPXPI:
	if (!getDimmMode(pCardInfo, pBankInfo, pGlob, i, &mode, &dimmX4[i])) {
	  pGlob->bankTestMask &= ~(0x1U << i);
	}
	modes[i] = mode;
	pGlob->bBankDdr[i] = TRUE;
	break;

      default:
	assert(FALSE);
	break;
      }
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR2) {
      uint32_t mode;
      
      /* Select registered, row bits, col bits, bank bits and phys bank bits for DDR SDRAM */
      switch (pCardInfo->BoardType) {
      case ADMXRC2_BOARD_ADMXRC4FX:
      case ADMXRC2_BOARD_ADPEXRC4FX:
      case ADMXRC2_BOARD_ADMXRC5LX:
      case ADMXRC2_BOARD_ADMXRC5T1:
      case ADMXRC2_BOARD_ADMXRC5T2:
      case ADMXRC2_BOARD_ADMXRC5T2ADV:
      case ADMXRC2_BOARD_ADMXRC5TDA1:
	mode = 0;
	if (0x1000000 == pBankInfo[i].Size) {
	  /* unbufferred, row bits 13, col bits 10, bank bits 3, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR-II SDRAM): unbuffered, row 13, col 10, bank 3, phys banks 1\n",
		 (unsigned long)i);
	} else if (0x2000000 == pBankInfo[i].Size) {
	  /* unbufferred, row bits 14, col bits 10, bank bits 3, phys banks 1 */
	  mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
	  mode |= MODEx_REG_DDRSDRAM_ROWBITS_14;
	  mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
	  mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
	  mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
	  printf(" Bank %ld (DDR-II SDRAM): unbuffered, row 14, col 10, bank 3, phys banks 1\n",
		 (unsigned long)i);
	} else {
	  pGlob->bankTestMask &= ~(0x1U << i);
	  printf(" +++ Bank %ld (DDR-II SDRAM): invalid size (0x%lx words), not testing\n",
		 (unsigned long)i, (unsigned long)pBankInfo[i].Size);
	}
	modes[i] = mode;
	pGlob->bBankDdr[i] = TRUE;
	break;
      default:
	assert(FALSE);
	break;
      }
    }
  }
}

int
main(
  int argc,
  char *argv[])
{
  Globals             glob;
  int                 ret = 0;
  ADMXRC2_STATUS      status;
  ADMXRC2_CARD_INFO   cardInfo;
  ADMXRC2_SPACE_INFO  spInfo;
  ADMXRC2_BANK_INFO   bankInfo[MAX_BANK];
  const char*         filename;
  char*               design;
  Option              options[8];
  Arguments           arguments;
  unsigned int        i, rep;
  unsigned int        numLockFlags;
  uint64_t            memorySize;
  boolean_t           bSetLclk, bSetMclk, bAllow64Bit;
  boolean_t           bUseModule;
  uint32_t            modes[MAX_BANK];
  char*               moduleName;
  double              lclkFreq, actualLclkFreq;
  double              mclkFreq, actualMclkFreq;
  unsigned long       nrep;

  /***********************************************************************
  ** 1. Initialize stuff so that we know where we are if we have to     **
  **    clean up if some later step fails.                              **
  ***********************************************************************/

  glob.hCard = ADMXRC2_HANDLE_INVALID_VALUE;

  /***********************************************************************
  ** 2. Set up the data structure that defines command-line syntax.     **
  ***********************************************************************/

  options[0].key = "banks";
  options[0].type = Option_hex;
  options[0].def.hexVal = 0xffffU;
  options[0].meaning = "bitmask of banks to test";
  options[1].key = "card";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "ID of card to use";
  options[2].key = "index";
  options[2].type = Option_uint;
  options[2].def.uintVal = 0;
  options[2].meaning = "index of card to use";
  options[3].key = "repeat";
  options[3].type = Option_uint;
  options[3].def.uintVal = 1;
  options[3].meaning = "number of repetitions";
  options[4].key = "64";
  options[4].type = Option_boolean;
  options[4].def.booleanVal = FALSE;
  options[4].meaning = "TRUE => use 64 bit local bus";
  options[5].key = "lclk";
  options[5].type = Option_float;
  options[5].def.floatVal = 33.0;
  options[5].meaning = "local bus clock frequency, in MHz";
  options[6].key = "mclk";
  options[6].type = Option_float;
  options[6].def.floatVal = 33.0;
  options[6].meaning = "MCLK frequency, in MHz";
  options[7].key = "module";
  options[7].type = Option_string;
  options[7].def.stringVal = "<none>";
  options[7].meaning = "target specified XRM module";
  arguments.nOption = 8;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;
    
  /***********************************************************************
  ** 3. Parse the command line                                          **
  ***********************************************************************/

  parse_command_line("memory",
		     argc,
		     argv,
		     &arguments);
    
  if (options[1].specified && options[2].specified) {
    printf("*** Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }
    
  nrep = options[3].value.uintVal;
  glob.bUse64Bit = options[4].value.booleanVal;
  moduleName = options[7].value.stringVal;
  bUseModule = strcmp(moduleName, "<none>") != 0;

  /***********************************************************************
  ** 4. Open the selected card                                          **
  ***********************************************************************/

  if (!options[2].specified) {
    ADMXRC2_CARDID cardID = options[1].value.uintVal;
        
    status = ADMXRC2_OpenCard(cardID, &glob.hCard);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with ID %ld: %s\n",
	     (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[2].value.uintVal;
        
    status = ADMXRC2_OpenCardByIndex(index, &glob.hCard);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to open card with index %ld: %s\n",
	     (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }
    
  /***********************************************************************
  ** 5. Get card information for the selected card. This will be used   **
  **    later to determine default clock frequencies and other model-   **
  **    specific settings.                                              **
  ***********************************************************************/

  status = ADMXRC2_GetCardInfo(glob.hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    return -1;
  }

  /***********************************************************************
  ** 6. Figure out model-specific settings from card type.              **
  ***********************************************************************/

  bSetLclk = TRUE; /* Falsified later if necessary */
  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 33.33;
    mclkFreq = 40.0;
    numLockFlags = 2;
    break;

  case ADMXRC2_BOARD_ADMXRC_P:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 33.33;
    mclkFreq = 40.0;
    numLockFlags = 1;
    break;

  case ADMXRC2_BOARD_ADMXRC2_LITE:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 40.0;
    mclkFreq = 66.67;
    numLockFlags = 3;
    break;

  case ADMXRC2_BOARD_ADMXRC2:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 50.0;
    mclkFreq = 66.67;
    numLockFlags = 3;
    break;

  case ADMXRC2_BOARD_ADMXPL:
    bAllow64Bit = TRUE;
    bSetMclk = FALSE;
    lclkFreq = 66.67;
    mclkFreq = 2.0 * lclkFreq;
    numLockFlags = 2;
    break;
        
  case ADMXRC2_BOARD_ADMXP:
    bAllow64Bit = TRUE;
    bSetMclk = FALSE;
    lclkFreq = 66.67;
    mclkFreq = 2.0 * lclkFreq;
    numLockFlags = 1;
    break;
        
  case ADMXRC2_BOARD_ADPXPI:
    bAllow64Bit = TRUE;
    bSetMclk = FALSE;
    lclkFreq = 66.67;
    mclkFreq = 2.0 * lclkFreq;
    numLockFlags = 1;
    break;

  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 66.0;
    mclkFreq = 150.0; /* 150MHz is max 1X output frequency for DCM in low frequency mode */
    numLockFlags = 2;
    break;

  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADPEXRC4FX:
    bAllow64Bit = TRUE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 200.0MHz MCLK => 100.0MHz in memory clock domain, 200.0MHz DDR-II SDRAM clock */
    mclkFreq = 200.0;
    numLockFlags = 2;
    break;

  case ADMXRC2_BOARD_ADMXRC5LX:
    bAllow64Bit = TRUE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
    mclkFreq = 266.67;
    numLockFlags = 2;
    break;
      
  case ADMXRC2_BOARD_ADMXRC5T1:
    bAllow64Bit = TRUE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
    mclkFreq = 266.67;
    numLockFlags = 2;
    break;

  case ADMXRC2_BOARD_ADCPXRC4LX:
    bAllow64Bit = FALSE;
    bSetMclk = FALSE;
    lclkFreq = 50.0;
    mclkFreq = 200.0;
    numLockFlags = 3;
    break;

  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
    bAllow64Bit = TRUE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
    mclkFreq = 266.67;
    numLockFlags = 2;
    break;

  case ADMXRC2_BOARD_ADMXRC5TZ:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 166.67MHz MCLK => 166.67MHz in memory clock domain, 166.67MHz ZBT SSRAM clock */
    mclkFreq = 166.67;
    numLockFlags = 2;
    break;

  case ADMXRC2_BOARD_ADMXRC5TDA1:
    bAllow64Bit = FALSE;
    bSetMclk = TRUE;
    lclkFreq = 66.67;
    /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
    mclkFreq = 266.67;
    numLockFlags = 2;
    break;

  default:
    printf("*** This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRC-II-Lite\n");
    printf("\tADM-XRC-II\n");
    printf("\tADM-XPL\n");
    printf("\tADM-XP\n");
    printf("\tADP-WRC-II\n");
    printf("\tADP-DRC-II\n");
    printf("\tADP-XPI\n");
    printf("\tADM-XRC-4LX\n");
    printf("\tADM-XRC-4SX\n");
    printf("\tADM-XRC-4FX / ADM-XMC-4FX\n");
    printf("\tADPE-XRC-4FX\n");
    printf("\tADM-XRC-5LX\n");
    printf("\tADM-XRC-5T1\n");
    printf("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV\n");
    printf("\tADM-XRC-5TZ\n");
    printf("\tADM-XRC-5T-DA1\n");
    ret = -1;
    goto done;
  }

  if (glob.bUse64Bit && !bAllow64Bit) {
    printf("*** 64-bit local bus is not supported on the specified card.\n");
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 7. Get the addresses for the FPGA registers and the memory access  **
  **    window, in our virtual address space.                           **
  ***********************************************************************/

  status = ADMXRC2_GetSpaceInfo(glob.hCard, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to get space 0 info: %s\n",
	   ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  glob.pFpgaSpace = spInfo.VirtualBase;
  if (glob.pFpgaSpace == NULL) {
    printf("*** Failed to retrieve pointer to FPGA space\n");
    ret = -1;
    goto done;
  }
  glob.pFpgaRegs = (volatile uint32_t*)glob.pFpgaSpace;
  glob.pMemory = (volatile uint32_t*)((uint8_t*)glob.pFpgaSpace + 0x200000U);

  /***********************************************************************
  ** 8. Get memory bank information.                                    **
  **                                                                    **
  **    To determine the byte width of a memory bank, we first divide   **
  **    by 9. If there is no remainder, then that bank has parity bits, **
  **    and we have calculated the byte width. If there is a remainder, **
  **    then we assume the bank has no parity bits, in which case we    **
  **    should be able to calculate the byte width by dividing by 8.    **
  **                                                                    **
  **    For the ADM-XRC-II-Lite, we fudge the data a little. Each pair  **
  **    of 18-bit wide banks is combined to get a logical 32-bit wide   **
  **    bank.                                                           **
  ***********************************************************************/
  
  /* Get bank information */
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    if (cardInfo.RAMBanksFitted & (0x1UL << i)) {
      status = ADMXRC2_GetBankInfo(glob.hCard, i, &bankInfo[i]);
      if (status != ADMXRC2_SUCCESS) {
	printf("*** Failed to get bank %ld info: %s\n", (unsigned long)i,
	       ADMXRC2_GetStatusString(status));
	return -1;
      }

      if (bankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR2) {
	/*
	** The DDR-II SSRAM memory interface for Virtex 4/5 has a user-
	** interface that is 4 times as wide as the physical interface
	** to the memory device.
	*/
	bankInfo[i].Size >>= 1;
	bankInfo[i].Width <<= 1;
      }

      if (bankInfo[i].Type & ADMXRC2_RAM_SRAM_DDR2) {
	switch (cardInfo.BoardType) {
	case ADMXRC2_BOARD_ADMXRC5T1:
	case ADMXRC2_BOARD_ADMXRC5T2:
	case ADMXRC2_BOARD_ADMXRC5T2ADV:
	case ADMXRC2_BOARD_ADMXRC5TDA1:
	  /*
	  ** The high speed DDR-II SSRAM memory interface for Virtex 4/5
	  ** has a user-interface that is 4 times as wide as the
	  ** physical interface to the memory device.
	  */
	  bankInfo[i].Size >>= 1;
	  bankInfo[i].Width <<= 1;
	  break;
	default:
	  break;
	}
      }
    }
  }

  /*
  ** For the ADM-XRC-II-Lite, fix up the memory bank information to make
  ** it look like we have 32-bit wide banks (we combine two 18-bit physical
  ** banks into one logical bank).
  */
  if (cardInfo.BoardType == ADMXRC2_BOARD_ADMXRC2_LITE) {
    uint32_t banksFitted = 0x0U;
        
    cardInfo.NumRAMBank = 2;
    for (i = 0; i < cardInfo.NumRAMBank; i++) {
      unsigned int j = i * 2;
      boolean_t bFittedA, bFittedB;
      boolean_t bSame;

      bFittedA = (cardInfo.RAMBanksFitted >> j) != 0;
      bFittedB = (cardInfo.RAMBanksFitted >> (j + 1)) != 0;
      if (bFittedA && bFittedB) {
	bSame = (bankInfo[j].Width == bankInfo[j + 1].Width) &&
	  (bankInfo[j].Type == bankInfo[j + 1].Type);
	if (bSame) {
	  bankInfo[i].Size = MIN(bankInfo[j].Size, bankInfo[j + 1].Size);
	  bankInfo[i].Type = bankInfo[j].Type & bankInfo[j + 1].Type;
	  bankInfo[i].Fitted = bankInfo[j].Fitted & bankInfo[j + 1].Fitted;
	  banksFitted |= (0x1U << i);
	} else {
	  /* This logical bank has been deemed invalid */
	  printf("+++ Can't pair banks %ld and %ld due to different configuration\n",
		 (unsigned long)j, (unsigned long)(j + 1));
	}
      }
    }
    cardInfo.RAMBanksFitted = banksFitted;
  }

  /*
  ** Establish the bank test mask. This is the set of banks that are we
  ** have been instructed to test AND are fitted. This value may be
  ** further modified during the further stages of this application in
  ** order to remove banks that are deemed invalid.
  */
  glob.bankTestMask = options[0].value.hexVal & cardInfo.RAMBanksFitted;

  /* Process bank information */
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    uint32_t typeMask = 0;
        
    typeMask |= ADMXRC2_RAM_ZBTFT;
    typeMask |= ADMXRC2_RAM_ZBTP;
    typeMask |= ADMXRC2_RAM_SRAM_DDR2;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR2;

    if (!(glob.bankTestMask & (0x1UL << i))) {
      /* Ignore banks that we're instructed not to test */
      continue;
    }
        
    if (!(bankInfo[i].Type & typeMask)) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }
        
    if ((bankInfo[i].Width % 9) == 0) {
      /* Bank has parity bits - don't use to calculate bank size */
      glob.bankWidth[i] = bankInfo[i].Width / 9;
    } else {
      /* Bank doesn't have parity bits */
      glob.bankWidth[i] = bankInfo[i].Width / 8;
    }
  }
    
  /***********************************************************************
  **  9. Program the frequencies for LCLK and MCLK. The defaults are    **
  **     obtained from the board type, but may be overridden on the     **
  **     command-line.                                                  **
  ***********************************************************************/

  if (options[5].specified) {
    lclkFreq = options[5].value.floatVal;
  }

  if (options[6].specified) {
    mclkFreq = options[6].value.floatVal;
  }

  printf("Setting LCLK frequency to %.1fMHz\n", lclkFreq);
  status = ADMXRC2_SetClockRate(glob.hCard, ADMXRC2_CLOCK_LCLK, lclkFreq * 1.0e6, &actualLclkFreq);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to set LCLK to %.1fMHz: %s\n",
	   lclkFreq, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
    
  if (bSetMclk) {
    printf("Setting MCLK frequency to %.1fMHz\n", mclkFreq);
    status = ADMXRC2_SetClockRate(glob.hCard, 1, mclkFreq * 1.0e6, &actualMclkFreq);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to set clock 1 to %.1fMHz: %s\n",
	     mclkFreq, ADMXRC2_GetStatusString(status));
      return -1;
    }
  }

  /***********************************************************************
  ** 10. Configure the FPGA with the "memory" bitstream.                **
  ***********************************************************************/

  design = glob.bUse64Bit ? "memory64" : "memory";
  filename = get_fpga_filename_dir(&cardInfo, design, design, &arguments);
  printf("Configuring FPGA with '%s'...\n", filename);
  status = ADMXRC2_ConfigureFromFile(glob.hCard, filename);
  if (status != ADMXRC2_SUCCESS) {
    printf("*** Failed to load the bitstream '%s': %s\n",
	   filename, ADMXRC2_GetStatusString(status));
    return -1;
  }

  /***********************************************************************
  ** 11. After configuring the FPGA, it's necessary to wait a while     **
  **     to enable any DLLs, DCMs and PLLs that depend upon LCLK to     **
  **     achieve lock. The delay is a conservative 500ms, because we    **
  **     make no assumptions about whether or not an ES (engineering    **
  **     silicon) FPGA is being used.                                   **
  ***********************************************************************/

  if (!waitForLclkDcm(&glob)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 12. If running in 64-bit mode was requested on the command-line,   **
  **     configure Direct Slave space 0 for a 64-bit local bus. Also    **
  **     generate the DMA mode word for DMA using a 64-bit local bus.   **
  ***********************************************************************/

  if (glob.bUse64Bit) {
    uint32_t config = 0;

    config |= ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
    status = ADMXRC2_SetSpaceConfig(glob.hCard, 0, config);
    if (status != ADMXRC2_SUCCESS) {
      printf("*** Failed to set space configuration: %s\n",
	     ADMXRC2_GetStatusString(status));
      return -1;
    }
  }

  /***********************************************************************
  ** 13. Here we use the memory bank information read in step 8 to      **
  **     determine the proper values for the MODE registers.            **
  ***********************************************************************/

  scanMemoryPorts(
		  &cardInfo,
		  bankInfo,
		  &glob,
		  modes);

  /***********************************************************************
  ** 14. Establish the MODE registers, based on what was determined     **
  **     in step 13.                                                    **
  ***********************************************************************/

  setMemoryPortConfiguration(&glob, modes);

  /***********************************************************************
  ** 15. Reset the memory subsystem, and wait for DLLs, DCMs, PLL and   **
  **     IDELAYCTRLs to achieve lock, as well as allowing any self-     **
  **     training memory interfaces sufficient time to train            **
  *     themselves.                                                    **
   ***********************************************************************/

  /*
  ** Reset the memory subsystem to ensure that memory banks can
  ** train themselves etc.
  */
  printf("Resetting memory subsystem...\n");
  glob.pFpgaRegs[MEMCTL_REG] = os_cpu_to_le32(0x1U);
  glob.pFpgaRegs[MEMCTL_REG];
  sleep_ms(1);
  glob.pFpgaRegs[MEMCTL_REG] = os_cpu_to_le32(0x0U);
  glob.pFpgaRegs[MEMCTL_REG];
  sleep_ms(500);

  /* Check for memory DLL/DCM/PLL locking and training */
  if (!checkForMemoryLock(&glob, numLockFlags)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 16. Figure out total size (in bytes) of memory banks that are to   **
  **     be tested.                                                     **
  ***********************************************************************/

  memorySize = 0;
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    uint32_t typeMask = 0;
        
    typeMask |= ADMXRC2_RAM_ZBTFT;
    typeMask |= ADMXRC2_RAM_ZBTP;
    typeMask |= ADMXRC2_RAM_SRAM_DDR2;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR2;
        
    if (!(glob.bankTestMask & (0x1UL << i))) {
      /* Ignore banks that we're instructed not to test */
      continue;
    }
        
    if (!bankInfo[i].Fitted) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }
        
    if (!(bankInfo[i].Type & typeMask)) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }
        
    glob.bankSize[i] = glob.bankWidth[i] * bankInfo[i].Size;
    memorySize += (glob.bankSize[i] >> 10);
  }

  /* If there's nothing to do, abort! */
  if (!glob.bankTestMask) {
    printf("*** Nothing to test - aborting.\n");
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 17. This is where the memory tests begin. There are four different **
  **     tests:                                                         **
  **                                                                    **
  **      a) 55 and AA pattern, for detecting stuck / shorted data bits **
  **      b) Own address, for detecing faulty address lines             **
  **      c) Own address, bit reversed, for detecting faulty address    **
  **         lines                                                      **
  **      d) Randomly generated data, for detecting pattern-sensitive   **
  **         faults.                                                    **
  ***********************************************************************/

  printf("Total memory size to test : %ld kiB (0x%lx kiB)\n",
	 (unsigned long)memorySize, (unsigned long)memorySize);

  for (rep = 0; rep < nrep; rep++) {
    static const char str[] = "|/-\\";
    int strIndex = 0;
    uint32_t val32;
    uint16_t bstat, berror;
    uint32_t eaddr;
    uint8_t ephase;
    timestamp_t start, now, tick;
    timestamp_t tick_increment = double_to_time(0.1);
    timestamp_t elapsed;

    /* Set test start address and length for each bank */
    for (i = 0; i < 16; i++) {
      if (glob.bankTestMask & (0x1U << i)) {
	glob.pFpgaRegs[USER_REG(i)] = os_cpu_to_le32(0);
	glob.pFpgaRegs[USER_REG(16 + i)] = os_cpu_to_le32(bankInfo[i].Size - 1);
	glob.pFpgaRegs[USER_REG(16 + i)];
      }
    }
 
    /* Kick off the tests for all selected banks */
    printf("Commanding tests to begin (bank mask 0x%lx)...\n",
	   (unsigned long)glob.bankTestMask);
    glob.pFpgaRegs[USER_REG(63)] = os_cpu_to_le32(glob.bankTestMask);
    glob.pFpgaRegs[USER_REG(63)];
    sleep_ms(1);

    start = tick = get_time();
    while (1) {
      now = get_time();
      if (time_after(&now, &tick)) {
	val32 = glob.pFpgaRegs[USER_REG(63)];
	val32 = os_le32_to_cpu(val32);
	bstat = (uint16_t)val32;
	berror = (uint16_t)(val32 >> 16U);
	if ((bstat & glob.bankTestMask) == (0xffffU & glob.bankTestMask)) {
	  break;
	}
                
	printf("%c Waiting for test completion (BSTAT=0x%8.8lx BERROR=0x%4.4lx)...\r",
	       str[strIndex], (unsigned long)bstat, (unsigned long)berror);
	strIndex = (strIndex + 1) & 0x3U;
	tick = time_add(&tick, &tick_increment);
      }

#if HOST_READ_ACTIVITY
      /*
      ** Read memory banks while we wait for completion, to see if we can provoke
      ** an error of any kind. Note that this will slow down the memory test
      ** considerably, since the local bus interface is given priority access to
      ** the memory.
      */
      for (i = 0; i < 16; i++) {
	if (glob.bankTestMask & (0x1U << i)) {
	  selectBank(&glob, i);
	  val32 = glob.pMemory[0];
	}
      }
#endif
    }
    now = get_time();
    elapsed = time_subtract(&now, &start);
    printf("\nDuration %ld ms\n", (unsigned long)(time_to_double(&elapsed) * 1000.0));
        
    val32 = glob.pFpgaRegs[USER_REG(63)];
    val32 = os_le32_to_cpu(val32);
    bstat = (uint16_t)val32;
    berror = (uint16_t)(val32 >> 16U);
    printf("All selected banks done - BSTAT=0x%4.4lx BERROR=0x%4.4lx\n",
	   (unsigned long)bstat, (unsigned long)berror);

    for (i = 0; i < 16; i++) {
      if (glob.bankTestMask & (0x1U << i)) {
	if (berror & (0x1U << i)) {
	  eaddr = glob.pFpgaRegs[USER_REG(32 + i)];
	  eaddr = os_le32_to_cpu(eaddr);
	  val32 = glob.pFpgaRegs[USER_REG(48 + (i >> 2))];
	  val32 = os_le32_to_cpu(val32);
	  ephase = (uint8_t)(val32 >> (i << 3));
	  printf("*** Bank %lu failed - EADDR[%lu]=0x%8.8lx EPHASE[%lu]=%2.2lx\n",
		 (unsigned long)i,
		 (unsigned long)i, (unsigned long)val32,
		 (unsigned long)i, (unsigned long)ephase);
	} else {
	  printf("Bank %lu OK\n", (unsigned long)i);
	}
      }
    }

    if (berror & glob.bankTestMask) {
      break;
    }
  }

  /***********************************************************************
  ** 18. Here we check that none of the sticky flags in STATUS indicate **
  **     that a DLL/DCM/PLL lost lock, or that a memory bank's          **
  **     "trained" status transitioned to zero.                         **
  ***********************************************************************/

  if (!checkStickyStatus(&glob, numLockFlags)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 19. Clean up.                                                      **
  ***********************************************************************/

 done:
  if (glob.hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(glob.hCard);
  }
    
  return ret;
}
