/*
** reario.c - loader for the REARIO FPGA
**
** Example for:
**
**   o ADM-XRC-P
**   o ADM-XRC-II
**
** (C) Copyright 2000-2006 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>

#include <admxrc2.h>

#include "../common/common.h"
#include "../common/platform.h"

int main(int argc, char *argv[])
{
  const double      clock_freq = 33.0e6;
  int               ret = 0;
  ADMXRC2_STATUS    status;
  ADMXRC2_HANDLE    card = ADMXRC2_HANDLE_INVALID_VALUE;
  ADMXRC2_CARD_INFO cardInfo;
  const char*       filename;
  Arguments         arguments;
  Option            options[2];

  options[0].key = "card";
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = "ID of card to use";
  options[1].key = "index";
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = "index of card to use";
  arguments.nOption = 2;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  parse_command_line("reario", argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    printf("Do not specify both /card and /index\n");
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with ID %ld: %s\n",
	     (unsigned long) cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &card);
    if (status != ADMXRC2_SUCCESS) {
      printf("Failed to open card with index %ld: %s\n",
	     (unsigned long) index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  status = ADMXRC2_GetCardInfo(card, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2:
    break;

  default:
    printf("This example must be run on one of the following cards:\n\n");
    printf("\tADM-XRC-P\n");
    printf("\tADM-XRCII\n");
    ret = -1;
    goto done;
  }

  /* Set local bus clock to nominal frequency */
  status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, clock_freq, NULL);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to set LCLK to %.1fMHz: %s\n",
	   clock_freq / 1.0e6f, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  filename = get_fpga_filename(&cardInfo, "reario", &arguments);
  status = ADMXRC2_ConfigureFromFile(card, filename);
  if (status != ADMXRC2_SUCCESS) {
    printf("Failed to load the bitstream '%s': %s\n",
	   filename, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  printf("REARIO FPGA loaded. Press return to exit.\n");
  getchar();

 done:
  if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(card);
  }
	
  return ret;
}
