/*
** plxddsm.v - PLX local bus demand mode DMA state machine
**
** Modules defined:
**
**    plxddsm               PLX local bus demand mode DMA state machine
**
** The plxddsm module defines the state machine and associated logic that
** controls a single PLX DMA channel in demand mode.
**
** One cycle of address decoding is permitted.
*/

`timescale 1ns / 100ps

module plxddsm(
   clk,        // Local bus clock
   rst,        // Asynchronous reset, should reduce to FPGA global reset
   sr,         // Synchronous reset (should never be asserted mid-burst)
   qlads,      // LADS# signal qualified by address, FHOLDA and/or other decoding
   ldack,      // LDACK# signal from local bus, active high
   lblast,     // LBLAST# signal from local bus, active high
   lwrite,     // LWRITE signal from local bus
   ldreq,      // To local bus LDREQ# signal, active high
   lready,     // To local bus LREADY# signal, active high
   lready_oe,  // Enable for LREADY# on local bus
   lbterm,     // To local bus LBTERM# signal, active high
   lbterm_oe,  // Enable for LBTERM# on local bus
   ld_oe,      // Output enable for LD pins
   request,    // Pulse to request cycle
   burst,      // Burst cycle requested / end burst
   transfer,   // Data transfer on this clock edge, when asserted
   decode,     // Application should do further address decoding, when asserted
   write);     // Burst is a write, when asserted

//
//	State machine one-hot states
//
parameter   [3:0]  s_idle        = 4'b1,
                   s_req         = 4'b10,
                   s_decode      = 4'b100,
                   s_xfer        = 4'b1000;

input              clk,
                   rst,
                   sr,
                   qlads,
                   ldack,
                   lblast,
                   lwrite,
                   request,
                   burst;

output             ld_oe,
                   ldreq,
                   lready,
                   lready_oe,
                   lbterm,
                   lbterm_oe,
                   transfer,
                   decode,
                   write;

reg      [3:0]     state; 
reg      [3:0]     n_state; 

reg                transfer, n_transfer;
reg                lready, n_lready;
reg                lready_oe, n_lready_oe;
reg                lbterm, n_lbterm;
reg                lbterm_oe;
reg                ld_oe, n_ld_oe;
reg                write, n_write;
reg                burst_q, n_burst_q;
reg                decode, n_decode;

wire               ldreq;

assign             ldreq = ((state & s_req) && ~ldack) ||
                           ((state & s_req) && ldack && burst_q && burst) ||
                           (((state & s_xfer) || (state & s_decode)) && burst_q && burst);

//
// Define the synchronous state machine transition
//
always @ (posedge clk or posedge rst)
begin
   if (rst) begin
      state <= s_idle;
      transfer <= 1'b0;
      lready <= 1'b0;
      lready_oe <= 1'b0;
      lbterm <= 1'b0;
      lbterm_oe <= 1'b0;
      ld_oe <= 1'b0;
      write <= 1'b0;
      decode <= 1'b0;
      burst_q <= 1'b0;
   end else begin
      if (sr) begin
         state <= s_idle;
         transfer <= 1'b0;
         lready <= 1'b0;
         lready_oe <= 1'b0;
         lbterm <= 1'b0;
         lbterm_oe <= 1'b0;
         ld_oe <= 1'b0;
         write <= 1'b0;
         decode <= 1'b0;
         burst_q <= 1'b0;
      end else begin
         state <= n_state;
         transfer <= n_transfer;
         lready <= n_lready;
         lready_oe <= n_lready_oe;
         lbterm <= n_lbterm;
         lbterm_oe <= n_lready_oe;
         ld_oe <= n_ld_oe;
         write <= n_write;
         decode <= n_decode;
         burst_q <= n_burst_q;
      end
   end
end

//
// Determine next state for the machine
//
always @ (state      or
          qlads      or
          lblast     or
          lwrite     or
          ldack      or
          lbterm     or
          ld_oe      or
          lready     or
          lready_oe  or
          request    or
          transfer   or
          decode     or
          burst_q    or
          burst      or
          write)
begin

   n_state = state;
   n_burst_q = burst_q;
   n_decode = decode;
   n_lbterm = lbterm;
   n_ld_oe = ld_oe;
   n_lready = lready;
   n_lready_oe = lready_oe;
   n_transfer = transfer;
   n_write = write;

   case (state)
      s_idle : begin
         n_transfer = 1'b0;
         n_burst_q = burst;
         n_decode = 1'b0;
         n_lready = 1'b0;
         n_lbterm = 1'b0;
         n_ld_oe  = 1'b0;
         n_lready_oe = 1'b0;
         n_write = 1'b0;
         if (request) begin
            n_state = s_req;
         end else begin
            n_state = s_idle;
         end
      end

      s_req : begin
         n_transfer = 1'b0;
         n_burst_q = burst_q & burst;
         n_lbterm = 1'b0;
         n_ld_oe = 1'b0;
         n_lready = 1'b0;
         n_lready_oe = 1'b0;
         if (qlads && ldack) begin
            n_state = s_decode;
            n_decode = 1'b1;
            n_write = lwrite;
         end else begin
            n_state = s_req;
            n_decode = 1'b0;
            n_write = write;
         end 
      end

      s_decode : begin
         n_state = s_xfer;
         n_burst_q = burst_q & burst;
         n_decode = 1'b0;
         n_lbterm = 1'b0;
         n_ld_oe = ~lwrite;
         n_lready = 1'b1;
         n_lready_oe = 1'b1;
         n_transfer = 1'b1;
         n_write = write;
      end

      s_xfer : begin
         n_burst_q = burst_q & burst;
         n_decode = 1'b0;
         n_lbterm = 1'b0;
         n_lready_oe = 1'b1;
         n_write = write;
         if (lblast) begin
            n_state = s_idle;
            n_ld_oe = 1'b0;
            n_lready = 1'b0;
            n_transfer = 1'b0;
         end else begin
            n_state = s_xfer;
            n_ld_oe = ~lwrite;
            n_lready = 1'b1;
            n_transfer = 1'b1;
         end
      end
   endcase
end

endmodule
