/*
** plxdmsm.v - PLX local bus direct master state machine
**
** Modules defined:
**
**    plxdmsm              PLX local bus direct master state machine
**
** This state machine is designed to initiate direct master transfers on
** the PLX local bus.
*/

`timescale 1ns / 100ps

module plxdmsm(
    clk,         // Local bus clock
    rst,         // Asynchronous reset, should reduce to global reset
    sr,          // Synchronous reset, must be asserted when bus idle
    request,     // Pulse to request a direct-master cycle
    write,       // Assert coincident with 'request' for a write cycle
    stop,        // Pulse to terminate burst
    transfer,    // Asserted 1 clock after data transfer takes place
    done,        // Asserted when machine is idle
    lready,      // Local bus LREADYO# signal (PCI9080) or LREADY# signal (PCI9656)
    lbterm,      // Local bus LBTERMO# signal (PCI9080) or LBTERM# signal (PCI9656)
    fholda,      // From local bus arbiter, FPGA's bus grant signal
    fhold,       // To local bus arbiter, FPGA's bus request signal
    la_oe,       // Enable for local bus LA signal
    lads,        // To local bus LADS# signal, active high
    lads_oe,     // Enable for local bus LADS# signal
    lbe_oe,      // Enable for local bus LBE# signal
    lblast,      // To local bus LBLAST# signal, active high
    lblast_oe,   // Enable for local bus LBLAST# signal
    ld_oe,       // Enable for local bus LD signal
    lwrite,      // To local bus LWRITE signal
    lwrite_oe);  // Enable for local bus LWRITE signal

//
// State machine one-hot states
//
parameter   [3:0]  s_idle        = 4'b1,
                   s_arb         = 4'b10,
                   s_addr        = 4'b100,
                   s_xfer        = 4'b1000;

input              clk,
                   rst,
                   sr,
                   request,
                   stop,
                   write,
                   lready,
                   lbterm,
                   fholda;

output             fhold,
                   la_oe,
                   lads,
                   lads_oe,
                   lbe_oe,
                   lblast,
                   lblast_oe,
                   ld_oe,
                   lwrite,
                   lwrite_oe,
                   transfer,
                   done;

wire               clk, rst, sr;

wire               fholda;
wire               lready;

reg                la_oe, n_la_oe;
reg                lads, n_lads;
reg                lads_oe, n_lads_oe;
reg                lbe_oe, n_lbe_oe;
reg                lblast, n_lblast;
reg                lblast_oe, n_lblast_oe;
reg                ld_oe, n_ld_oe;
reg                lwrite, n_lwrite;
reg                lwrite_oe, n_lwrite_oe;
reg                fhold, n_fhold;

wire               write;
wire               stop;
wire               request;
reg                done, n_done;
reg                transfer, n_transfer;

reg         [3:0]  state, n_state; 

reg                write_q, n_write_q;
reg                stopping, n_stopping;

//
// Define the synchronous state machine transition
//
always @ (posedge clk or posedge rst) begin
   if (rst) begin
      state <= s_idle;
      la_oe <= 1'b0;
      lads <= 1'b0;
      lads_oe <= 1'b0;
      lbe_oe <= 1'b0;
      lblast <= 1'b0;
      lblast_oe <= 1'b0;
      ld_oe <= 1'b0;
      lwrite <= 1'b0;
      lwrite_oe <= 1'b0;
      fhold <= 1'b0;
      done <= 1'b1;
      stopping <= 1'b0;
      transfer <= 1'b0;
      write_q <= 1'b0;
   end else begin
      if (sr) begin
         state <= s_idle;
         la_oe <= 1'b0;
         lads <= 1'b0;
         lads_oe <= 1'b0;
         lbe_oe <= 1'b0;
         lblast <= 1'b0;
         lblast_oe <= 1'b0;
         ld_oe <= 1'b0;
         lwrite <= 1'b0;
         lwrite_oe <= 1'b0;
         fhold <= 1'b0;
         done <= 1'b1;
         stopping <= 1'b0;
         transfer <= 1'b0;
         write_q <= 1'b0;
      end else begin
         state <= n_state;
         la_oe <= n_la_oe;
         lads <= n_lads;
         lads_oe <= n_lads_oe;
         lbe_oe <= n_lbe_oe;
         lblast <= n_lblast;
         lblast_oe <= n_lblast_oe;
         ld_oe <= n_ld_oe;
         lwrite <= n_lwrite;
         lwrite_oe <= n_lwrite_oe;
         fhold <= n_fhold;
         done <= n_done;
         stopping <= n_stopping;
         transfer <= n_transfer;
         write_q <= n_write_q;
      end
   end
end

//
// Determine next state for the main machine
//
always @ (state      or
          request    or
          stop       or
          write_q    or
          stopping   or
          fholda     or
          lready     or
          lbterm     or
          lblast     or
          fhold      or
          ld_oe      or
          lads       or
          lads_oe    or
          la_oe      or
          lbe_oe     or
          lblast_oe  or
          lwrite     or
          lwrite_oe  or
          transfer   or
          done       or
          write)
begin

   n_state = state;
   n_done = done;
   n_fhold = fhold;
   n_stopping = stopping;
   n_transfer = transfer;
   n_write_q = write_q;
   n_lads = lads;
   n_lads_oe = lads_oe;
   n_la_oe = la_oe;
   n_lbe_oe = lbe_oe;
   n_lblast = lblast;
   n_lblast_oe = lblast_oe;
   n_ld_oe = ld_oe;
   n_lwrite = lwrite;
   n_lwrite_oe = lwrite_oe;

   case (state)
      s_idle: begin
         n_ld_oe = 1'b0;
         n_transfer = 1'b0;
         if (request) begin
            n_state = s_arb;
            n_fhold = 1'b1;
            n_lads = 1'b0;
            n_lads_oe = 1'b0;
            n_la_oe = 1'b0;
            n_lbe_oe = 1'b0;
            n_lblast = 1'b0;
            n_lblast_oe = 1'b0;
            n_lwrite = 1'b0;
            n_lwrite_oe = 1'b0;
            n_done = 1'b0;
            n_stopping = stop;
            n_write_q = write;
         end else begin
            n_state = s_idle;
            n_fhold = 1'b0;
            n_lads = 1'b0;
            n_lads_oe = 1'b0;
            n_la_oe = 1'b0;
            n_lbe_oe = 1'b0;
            n_lblast = 1'b0;
            n_lblast_oe = 1'b0;
            n_lwrite = 1'b0;
            n_lwrite_oe = 1'b0;
            n_done = 1'b1;
            n_stopping = 1'b0;
            n_write_q = 1'b0;
         end
      end

      s_arb: begin
         n_done = 1'b0;
         n_fhold = 1'b1;
         n_ld_oe = 1'b0;
         n_stopping = stop | stopping;
         n_transfer = 1'b0;
         n_write_q = write_q;
         if (fholda) begin
            n_state = s_addr;
            n_lads = 1'b1;
            n_lads_oe = 1'b1;
            n_la_oe = 1'b1;
            n_lbe_oe = 1'b1;
            n_lblast = 1'b0;
            n_lblast_oe = 1'b1;
            n_lwrite = write_q;
            n_lwrite_oe = 1'b1;
         end else begin
            n_state = s_arb;
            n_lads = 1'b0;
            n_lads_oe = 1'b0;
            n_la_oe = 1'b0;
            n_lbe_oe = 1'b0;
            n_lblast = 1'b0;
            n_lblast_oe = 1'b0;
            n_lwrite = 1'b0;
            n_lwrite_oe = 1'b0;
         end
      end

      s_addr: begin
         n_state = s_xfer;
         n_done = 1'b0;
         n_fhold = 1'b1;
         n_lads = 1'b0;
         n_lads_oe = 1'b1;
         n_la_oe = 1'b1;
         n_lbe_oe = 1'b1;
         n_lblast = stop | stopping;
         n_lblast_oe = 1'b1;
         n_ld_oe = write_q;
         n_lwrite = write_q;
         n_lwrite_oe = 1'b1;
         n_stopping = stop | stopping;
         n_transfer = 1'b0;
         n_write_q = write_q;
      end

      s_xfer: begin
         n_fhold = 1'b1;
         n_stopping = stop | stopping;
         n_write_q = write_q;
         n_lads = 1'b0;
         n_lads_oe = 1'b0;
         if ((lblast && lready) || lbterm) begin
            n_state = s_idle;
            n_done = 1'b1;
            n_la_oe = 1'b0;
            n_lbe_oe = 1'b0;
            n_lblast = 1'b0;
            n_lblast_oe = 1'b1;
            n_ld_oe = 1'b0;
            n_lwrite = write_q;
            n_lwrite_oe = 1'b1;
            n_transfer = 1'b1;
         end else begin
            n_state = s_xfer;
            n_done = 1'b0;
            n_la_oe = 1'b1;
            n_lbe_oe = 1'b1;
            n_lblast = stop | stopping;
            n_lblast_oe = 1'b1;
            n_ld_oe = write_q;
            n_lwrite = write_q;
            n_lwrite_oe = 1'b1;
            n_transfer = lready;
         end
      end
   endcase
end

endmodule
