/*
** plxdssm.v - PLX local bus direct slave state machine
**
** Modules defined:
**
**     plxdssm           PLX local bus direct slave state machine
**
** This state machine is designed to respond to direct slave transfers or
** programmed DMA transfers, NOT demand mode DMA transfers.
**
** One cycle of address decoding is permitted.
*/

`timescale 1ns / 100ps

module plxdssm(
    clk,         // Local bus clock
    rst,         // Asynchronous reset, should reduce to FPGA global reset
    sr,          // Synchronous reset (should never be asserted mid-burst)
    qlads,       // LADS# signal qualified by address, FHOLDA and/or other decoding
    lblast,      // Local bus LBLAST# signal, active high
    lwrite,      // Local bus LWRITE signal, active high
    ld_oe,       // Output enable for local bus data pins
    lready,      // To local bus LREADY# signal
    lready_oe,   // Enable for local bus LREADYI# signal
    lbterm,      // To local bus LBTERM# signal
    lbterm_oe,   // Enable for local bus LBTERM# signal
    transfer,    // Data transfer on this clock edge
    decode,      // Application should do further address decoding
    write,       // 'lwrite' latched on 'qlads'
    ready,       // Apply pulse when ready
    stop);       // Pulse to terminate burst

//
// State machine one-hot states
//
parameter   [3:0]  s_idle        = 4'b1,
                   s_decode      = 4'b10,
                   s_wait        = 4'b100,
                   s_xfer        = 4'b1000;

input              clk,
                   rst,
                   sr,
                   qlads,
                   lblast,
                   lwrite,
                   ready,
                   stop;

output             transfer,
                   lready,
                   lready_oe,
                   lbterm,
                   lbterm_oe,
                   ld_oe,
                   decode,
                   write;

reg         [3:0]  state; 
reg         [3:0]  n_state; 

reg                transfer, n_transfer;
reg                ld_oe, n_ld_oe;
reg                lready, n_lready;
reg                lready_oe, n_lready_oe;
reg                lbterm, n_lbterm;
reg                lbterm_oe;
reg                decode, n_decode;
reg                stopping, n_stopping;
reg                write, n_write;

//
// Define the synchronous state machine transition
//
always @ (posedge clk or posedge rst) begin
   if (rst) begin
      state <= s_idle;
      transfer <= 1'b0;
      lready <= 1'b0;
      lready_oe <= 1'b0;
      lbterm <= 1'b0;
      lbterm_oe <= 1'b0;
      decode <= 1'b0;
      write <= 1'b0;
      ld_oe <= 1'b0;
      stopping <= 1'b0;
   end else begin
      if (sr) begin
         state <= s_idle;
         transfer <= 1'b0;
         lready <= 1'b0;
         lready_oe <= 1'b0;
         lbterm <= 1'b0;
         lbterm_oe <= 1'b0;
         decode <= 1'b0;
         write <= 1'b0;
         ld_oe <= 1'b0;
         stopping <= 1'b0;
      end else begin
         state <= n_state;
         transfer <= n_transfer;
         lready <= n_lready;
         lready_oe <= n_lready_oe;
         lbterm <= n_lbterm;
         lbterm_oe <= n_lready_oe;
         decode <= n_decode;
         write <= n_write;
         ld_oe <= n_ld_oe;
         stopping <= n_stopping;
      end
   end
end

//
// Determine next state for the machine
//
always @ (state      or
          lblast     or
          qlads      or
          lbterm     or
          lwrite     or
          ready      or
          stop       or
          decode     or
          ld_oe      or
          lready     or
          lready_oe  or
          transfer   or
          stopping   or
          write)
begin

   n_state = state;
   n_decode = decode;
   n_lbterm = lbterm;
   n_ld_oe = ld_oe;
   n_lready = lready;
   n_lready_oe = lready_oe;
   n_transfer = transfer;
   n_stopping = stopping;
   n_write = write;

   case (state)
      s_idle: begin
         n_transfer = 1'b0;
         n_lready = 1'b0;
         n_lready_oe = 1'b0;
         n_lbterm = 1'b0;
         n_ld_oe = 1'b0;
         n_stopping = 1'b0;
         if (qlads) begin
            n_state = s_decode;
            n_decode = 1'b1;
            n_write = lwrite;
         end else begin
            n_state = s_idle;
            n_decode = 1'b0;
            n_write = write;
         end
      end

      s_decode: begin
         n_decode = 1'b0;
         n_ld_oe = ~write;
         n_lready_oe = 1'b1;
         n_write = write;
         if (ready) begin
            n_state = s_xfer;
            n_lready = 1'b1;
            n_lbterm = stop;
            n_transfer = 1'b1;
            n_stopping = 1'b0;
         end else begin
            n_state = s_wait;
            n_lready = 1'b0;
            n_lbterm = 1'b0;
            n_transfer = 1'b0;
            n_stopping = stop | stopping;
         end
      end

      s_wait: begin
         n_decode = 1'b0;
         n_ld_oe = ~write;
         n_lready_oe = 1'b1;
         n_write = write;
         if (ready) begin
            n_state = s_xfer;
            n_lready = 1'b1;
            n_lbterm = stop | stopping;
            n_transfer = 1'b1;
            n_stopping = 1'b0;
         end else begin
            n_state = s_wait;
            n_lready = 1'b0;
            n_lbterm = 1'b0;
            n_transfer = 1'b0;
            n_stopping = stop | stopping;
         end
      end

      s_xfer: begin
         n_decode = 1'b0;
         n_lready_oe = 1'b1;
         n_write = write;
         n_stopping = 1'b0;
         if (lblast || lbterm) begin
            n_state = s_idle;
            n_lready = 1'b0;
            n_lbterm = 1'b0;
            n_transfer = 1'b0;
            n_ld_oe = 1'b0;
         end else begin
            n_state = s_xfer;
            n_lready = 1'b1;
            n_lbterm = stop;
            n_transfer = 1'b1;
            n_ld_oe = ~write;
         end
      end
   endcase
end

endmodule
