/*
** sfifo.vhd - Synchronous 32-bit wide FIFO
**
** Modules defined:
**
**    sfifo                 512 word by 32-bit FIFO using BlockRAMs
*/

`timescale 1ns / 100ps

module sfifo(
   clk,      // Clock
   rst,      // Reduces to global reset
   sr,       // Synchronous reset
   d,        // Data in
   q,        // Data out
   w,        // FIFO write command
   we,       // Write-side empty flag
   wnf,      // Write-side nearly-full flag
   wf,       // Write-side full flag
   r,        // FIFO read command
   re,       // Read-side empty flag
   rne,      // Read side nearly-empty flag
   rf,       // Read side full flag
   status);  // Status vector (for debugging)

parameter n_full  = 4; // FIFO is nearly-full when n_full or less free locations
parameter n_empty = 4; // FIFO is nearly-empty when n_empty or less used locations

input             clk;
input             rst;
input             sr;
input  [31:0]     d;
output [31:0]     q;
input             w;
output            we;
output            wnf;
output            wf;
input             r;
output            re;
output            rne;
output            rf;
output [34:0]     status;

wire [31:0]       q;

reg [8:0]         da;
reg [8:0]         qa;
reg [9:0]         size;
reg               we, wnf, wf;
reg               re, rne, rf;
reg               r1e;
reg               wq;

wire              push = wq && re;
wire              enb = push || (r && !r1e);
wire              wea = w && !wf;
wire [34:0]       status = { qa, da, size, r1e, rf, rne, re, wf, wnf, we };

always @ (posedge clk or posedge rst)
begin
    if (rst) begin
        size <= 0;
        we <= 1;
        wnf <= 0;
        wf <= 0;
        re <= 1;
        rne <= 1;
        r1e <= 1;
        rf <= 0;
        da <= 0;
        qa <= 0;
        wq <= 0;
    end else begin
        if (sr) begin
            size <= 0;
            we <= 1;
            wnf <= 0;
            wf <= 0;
            re <= 1;
            rne <= 1;
            r1e <= 1;
            rf <= 0;
            da <= 0;
            qa <= 0;
            wq <= 0;
        end else begin
            /*
            ** Write side
            */
            if (w && !wf) begin
                da <= da + 1;
            end
            if (w && !r) begin
                we <= 0;
            end else if (!w && r && (size == 1)) begin
                we <= 1;
            end
            if (w && !r && (size == 511)) begin
                wf <= 1;
            end else if (!w && r && wf) begin
                wf <= 0;
            end
            if (w && !r && (size == (512 - n_full - 1))) begin
                wnf <= 1;
            end else if (!w && r && (size == (512 - n_full))) begin
                wnf <= 0;
            end

            /*
            ** Register the qualified write signal into 'wq' - this is used
            ** to push the data through when the FIFO read empty flag becomes
            ** empty (see the 'push' signal above).
            */
            wq <= w && !wf;
      
            /*
            ** Read side
            */
            if (enb) begin
                qa <= qa + 1;
            end
            if (r && !w) begin
                rf <= 0;
            end else if (!r && w && (size == 511)) begin
                rf <= 1;
            end
            if (r && r1e) begin
                re <= 1;
            end else if (!r && wq && re) begin
                re <= 0;
            end
            if (r && !w && (size == 2)) begin
                r1e <= 1;
            end else if (!r && w && (size == 1)) begin
                r1e <= 0;
            end
            if (r && !w && (size == (n_empty + 1))) begin
                rne <= 1;
            end else if (!r && w && (size == n_empty)) begin
                rne <= 0;
            end

            /*
            ** FIFO level counter
            */
            if (w && !r && !wf) begin
                size <= size + 1;
            end else if (r && !w && !re) begin
                size <= size - 1;
            end
        end
    end
end

sfiforam sfiforam0(
    .rst  (rst),
    .clk  (clk),
    .sr   (1'b0),
    .da   (da),
    .qa   (qa),
    .d    (d),
    .q    (q),
    .r    (enb),
    .w    (wea));

endmodule
