/*
** ddma-xpl.v - example demand-mode DMA design
**
** (C) Copyright Alpha Data Parallel Systems Ltd. 2001-2005
**
** Example program for:
**
**   o ADM-XPL
**   o ADM-XP
**
** Demand mode DMA is used to read data from the host via PLX DMA channel 0,
** into a FIFO. The data read from the FIFO and written back to the host via
** PLX DMA channel 1, also in demand-mode.
*/

`timescale 1ns / 100ps

module ddma(
   lclk,
   lreset_l, 
   lwrite,
   lads_l,
   lblast_l,
   lbterm_l,
   lad,
   lready_l, 
   lbe_l,
   fholda,
   ldreq_l,
   ldack_l);

input         lclk;
wire          lclk;
input         lreset_l;
wire          lreset_l;
input         lwrite;
wire          lwrite;
input         lads_l;
wire          lads_l;
input         lblast_l;
wire          lblast_l;
output        lbterm_l;
wire          lbterm_l;
inout [31:0]  lad;
wire  [31:0]  lad;
output        lready_l;
wire          lready_l;
input [3:0]   lbe_l;
wire  [3:0]   lbe_l;
input         fholda;
wire          fholda;
output [1:0]  ldreq_l;
wire   [1:0]  ldreq_l;
input  [1:0]  ldack_l;
wire   [1:0]  ldack_l;

wire          rst;
wire          lads_i;
wire          lblast_i;
wire  [3:0]   lbe_i;
wire          lwrite_i;
wire  [1:0]   ldack_i;
wire  [31:0]  ld_i;
wire  [23:2]  la_i;
wire  [1:0]   ldreq_o;
wire  [31:0]  ld_o;
wire          ld_oe;
wire          lbterm_o;
wire          lbterm_oe;
wire          lready_o;
wire          lready_oe;

reg   [23:2]  la_q;

wire          qlads_dd;
wire          qlads_ds;
wire          write_0;
wire          write_1;
wire          write_ds;
wire          xfer_0;
wire          xfer_1;
wire          xfer_ds;
wire          lrdy_0;
wire          lrdy_1;
wire          lrdy_ds;
wire          lrdy_oe_0;
wire          lrdy_oe_1;
wire          lrdy_oe_ds;
wire          lbt_0;
wire          lbt_1;
wire          lbt_ds;
wire          lbt_oe_0;
wire          lbt_oe_1;
wire          lbt_oe_ds;

wire          fifo_e;
wire          fifo_ne;
wire          fifo_read;
wire          fifo_f;
wire          fifo_nf;
wire          fifo_write;
wire [34:0]   fifo_status;

wire          request_0;
wire          request_1;
wire          burst_0;
wire          burst_1;

reg  [29:0]   icount;
reg           icount0;
reg           icount4;
reg           wrote_icount;
reg  [29:0]   ocount;
reg           ocount0;
reg           ocount4;
reg           wrote_ocount;

//
// Convert the inputs to active high.
//
assign rst        = ~lreset_l;
assign lads_i     = ~lads_l;
assign lblast_i   = ~lblast_l;
assign lbe_i      = ~lbe_l;
assign lwrite_i   = lwrite;
assign ld_i       = lad;
assign la_i       = lad[23:2];
assign ldack_i    = ~ldack_l;

//
// Output active low signals.
//
assign ldreq_l    = ~ldreq_o;

//
// Generate qualified versions of 'lads_l'.
//
// For direct slave access, this is when 'lads_l' is asserted AND the FPGA
// is being addressed AND the FPGA is not the local bus master AND the
// cycle is not demand-mode DMA.
//
// For demand-mode DMA access, this is when 'lads_l' is asserted AND the
// FPGA is being addressed and the FPGA is not the local bus master. The
// demand-mode DMA acknowledge signals are checked within the 'plxddsm'
// module.
//
assign qlads_ds = lads_i && !la_i[23] && !fholda && !ldack_i;
assign qlads_dd = lads_i && !la_i[23] && !fholda;

//
// 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
// float, because the control logic on the card might also drive it.
//
// The individual signals from 3 sources are ORed together here:
//
//   o bterm from DMA channel 0 demand-mode DMA state machine
//   o bterm from DMA channel 1 demand-mode DMA state machine
//   o bterm from direct slave state machine
//
assign lbterm_o   = lbt_0 | lbt_1 | lbt_ds;
assign lbterm_oe  = lbt_oe_0 | lbt_oe_1 | lbt_oe_ds;
assign lbterm_l   = lbterm_oe ? ~lbterm_o : 1'bZ;

//
// 'lready_l' should only be driven when the FPGA is addressed; otherwise
// float because the control logic on the card might also drive it.
//
// The individual signals from 3 sources are ORed together here:
//
//   o ready from DMA channel 0 demand-mode DMA state machine
//   o ready from DMA channel 1 demand-mode DMA state machine
//   o ready from direct slave state machine
//
assign lready_o  = lrdy_0 | lrdy_1 | lrdy_ds;
assign lready_oe = lrdy_oe_0 | lrdy_oe_1 | lrdy_oe_ds;
assign lready_l  = lready_oe ? ~lready_o : 1'bZ;

//
// Drive the local data bus on a read.
//
assign lad = ld_oe ? ld_o : 32'hZZZZZZZZ;

//
// Latch 'lad' on assertion of 'lads_l', in order to improve timing. There
// is no need to increment the address, because we don't allow bursting for
// register addresses, and DMA transfers that burst are expected to use
// constant local addressing.
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        la_q <= 0;
    end else begin
        if (lads_i) begin
            la_q <= la_i;
        end
    end
end

//
// This process implements the DMA transfer count registers and flags for
// each direction.
//
//    icount          number of words remaining in PCI-to-local DMA transfer
//    icount0         '1' => 'icount' is zero
//    icount4         '1' => 'icount' is less than or equal to 4
//    ocount          number of words remaining in local-to-PCI DMA transfer
//    ocount0         '1' => 'ocount' is zero
//    ocount4         '1' => 'ocount' is less than or equal to 4
//
// Prior to the host kicking off a demand-mode DMA transfer, the host must
// program the transfer count registers must be programmed with the length
// of the DMA transfer being performed.
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        icount  <= 0;
        icount4 <= 1'b1;
        icount0 <= 1'b1;
        wrote_icount <= 1'b0;
        ocount  <= 0;
        ocount4 <= 1'b1;
        ocount0 <= 1'b1;
        wrote_ocount <= 1'b0;
    end else begin
        if (xfer_ds && write_ds && !la_q[2]) begin
            if (lbe_i[0]) begin
                icount[5:0] <= ld_i[7:2];
            end
            if (lbe_i[1]) begin
                icount[13:6] <= ld_i[15:8];
            end
            if (lbe_i[2]) begin
                icount[21:14] <= ld_i[23:16];
            end
            if (lbe_i[3]) begin
                icount[29:22] <= ld_i[31:24];
            end
        end else begin
            if (fifo_write) begin
                icount <= icount - 1;
            end
        end

        if (xfer_ds && write_ds && la_q[2]) begin
            if (lbe_i[0]) begin
                ocount[5:0] <= ld_i[7:2];
            end
            if (lbe_i[1]) begin
                ocount[13:6] <= ld_i[15:8];
            end
            if (lbe_i[2]) begin
                ocount[21:14] <= ld_i[23:16];
            end
            if (lbe_i[3]) begin
                ocount[29:22] <= ld_i[31:24];
            end
        end else begin
            if (fifo_read) begin
                ocount <= ocount - 1;
            end
        end

        wrote_icount <= xfer_ds && write_ds && !la_q[2];
        if (wrote_icount) begin
            if (~|icount[29:4] && icount[3:0] <= 4) begin
                icount4 <= 1'b1;
            end else begin
                icount4 <= 1'b0;
            end
            if (~|icount) begin
                icount0 <= 1'b1;
            end else begin
                icount0 <= 1'b0;
            end
        end else begin
            if (fifo_write) begin
                if (icount == 5) begin
                    icount4 <= 1'b1;
                end
                if (icount == 1) begin
                    icount0 <= 1'b1;
                end
            end
        end

        wrote_ocount <= xfer_ds && write_ds && la_q[2];
        if (wrote_ocount) begin
            if (~|ocount[29:4] && ocount[3:0] <= 4) begin
                ocount4 <= 1'b1;
            end else begin
                ocount4 <= 1'b0;
            end
            if (~|ocount) begin
                ocount0 <= 1'b1;
            end else begin
                ocount0 <= 1'b0;
            end
        end else begin
            if (fifo_read) begin
                if (ocount == 5) begin
                    ocount4 <= 1'b1;
                end
                if (ocount == 1) begin
                    ocount0 <= 1'b1;
                end
            end
        end
    end
end

//
// For PCI->local direction, we can wait until the FIFO is not nearly-full
// before requesting a demand-mode DMA burst. We need to pause if the FIFO
// becomes nearly-full.
//
assign request_0  = !fifo_nf && !icount0;
assign burst_0    = !fifo_nf;

//
// For the local->PCI direction, we wait until the FIFO is not empty
// before requesting a demand-mode DMA burst. We need to pause if the FIFO
// becomes empty.
//
// NOTE: the bug that afflicts demand-mode DMA on the PCI9656 does not
//       exist in the ADM-XPL; hence the simpler code here (although the
//       code used for the ADM-XRCII version would also work with the
//       ADM-XPL).
//
assign request_1  = !fifo_e && !ocount0;
assign burst_1    = !fifo_ne;

//
// Writing into the FIFO only happens when data is written on
// PLX DMA channel 0.
//
assign fifo_write = xfer_0 && write_0;

//
// Reading from the FIFO only happens when data is read on PLX
// DMA channel 1.
//
assign fifo_read  = xfer_1 && !write_1;

//
// Instantiate the demand-mode DMA state machine for DMA channel 0.
// Monitors the local bus for channel 0 demand-mode DMA cycles and
// responds appropriately.
//
plxddsm
plxddsm0(
   .clk        (lclk),
   .rst        (rst),
   .sr         (1'b0),
   .qlads      (qlads_dd),
   .ldack      (ldack_i[0]),
   .lblast     (lblast_i),
   .lwrite     (lwrite_i),
   .ld_oe      (),
   .lready     (lrdy_0),
   .lready_oe  (lrdy_oe_0),
   .lbterm     (lbt_0),
   .lbterm_oe  (lbt_oe_0),
   .ldreq      (ldreq_o[0]),
   .transfer   (xfer_0),
   .decode     (),
   .request    (request_0),
   .burst      (burst_0),
   .write      (write_0));

//
// Instantiate the demand-mode DMA state machine for DMA channel 1.
// Monitors the local bus for channel 1 demand-mode DMA cycles and
// responds appropriately.
//
plxddsm
plxddsm1(
   .clk        (lclk),
   .rst        (rst),
   .sr         (1'b0),
   .qlads      (qlads_dd),
   .ldack      (ldack_i[1]),
   .lblast     (lblast_i),
   .lwrite     (lwrite_i),
   .ld_oe      (ld_oe),
   .lready     (lrdy_1),
   .lready_oe  (lrdy_oe_1),
   .lbterm     (lbt_1),
   .lbterm_oe  (lbt_oe_1),
   .ldreq      (ldreq_o[1]),
   .transfer   (xfer_1),
   .decode     (),
   .request    (request_1),
   .burst      (burst_1),
   .write      (write_1));

//
// Instantiate the direct slave state machine; monitors the local bus for
// direct slave cycles and responds appropriately.
//
plxdssm
slave0(
   .clk        (lclk),
   .rst        (rst),
   .sr         (1'b0),
   .qlads      (qlads_ds),
   .lblast     (lblast_i),
   .lwrite     (lwrite_i),
   .ld_oe      (),
   .lready     (lrdy_ds),
   .lready_oe  (lrdy_oe_ds),
   .lbterm     (lbt_ds),
   .lbterm_oe  (lbt_oe_ds),
   .transfer   (xfer_ds),
   .decode     (),
   .write      (write_ds),
   .ready      (1'b1),
   .stop       (1'b1));

//
// The value of 4 for the parameters 'n_full' and 'n_empty' means
// that nearly-full => 4 or less free locations in the FIFO, and
// nearly-empty => 4 or less words present the FIFO.
//    
sfifo #(4, 4)
sfifo0(
   .clk        (lclk),
   .rst        (rst),
   .sr         (1'b0),
   .d          (ld_i),
   .q          (ld_o),
   .w          (fifo_write),
   .we         (),
   .wnf        (fifo_nf),
   .wf         (fifo_f),
   .r          (fifo_read),
   .re         (fifo_e),
   .rne        (fifo_ne),
   .rf         (),
   .status     (fifo_status));

endmodule
