/*
** dll-xpl.v - Example FPGA design demonstrating clock multiplication
**
** (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2005
**
** Example program for:
**
**   o ADM-XPL
**   o ADM-XP
**
** The local bus clock is input and multiplied by 2 using a Virtex DCM.
**
** The 2x clock is used to clock a free running counter, which may be read
** via by direct slave reads on the local bus (anywhere in the FPGA space).
*/

`timescale 1ns / 100ps

module dll(
   lclk,
   lreset_l, 
   lwrite,
   lads_l,
   lblast_l,
   lbterm_l,
   lad,
   lready_l, 
   lbe_l,
   fholda,
   mclk
);

input         lclk;
wire          lclk;
input         lreset_l;
wire          lreset_l;
input         lwrite;
wire          lwrite;
input         lads_l;
wire          lads_l;
input         lblast_l;
wire          lblast_l;
output        lbterm_l;
wire          lbterm_l;
inout [31:0]  lad;
wire  [31:0]  lad;
output        lready_l;
wire          lready_l;
input [3:0]   lbe_l;
wire  [3:0]   lbe_l;
input         fholda;
wire          fholda;
input         mclk;
wire          mclk;

wire          rst;
wire          lads_i;
wire          lblast_i;
wire          lwrite_i;
wire [3:0]    lbe_i;
wire [23:2]   la_i;
wire [31:0]   ld_i;
    
wire          lbterm_o;
wire          lbterm_oe;
wire          lready_o;
wire          lready_oe;
wire [31:0]   ld_o;
wire          ld_oe;
    
wire          clk;
wire          multclk;
wire          locked;
    
wire          qlads;

wire          ds_xfer;
wire          ds_decode;
wire          ds_write;

reg  [31:0]   count;
reg  [31:0]   count_q;
reg           oe_reg0;

integer       i;


//
// Convert the inputs to active high.
//	
assign rst      = ~lreset_l;
assign lblast_i = ~lblast_l;
assign lads_i   = ~lads_l;
assign lwrite_i = lwrite;
assign la_i     = lad[23:2];
assign ld_i     = lad;
assign lbe_i    = ~lbe_l;
    
//
// Generate a qualified version of 'lads_l', which is
// asserted when the FPGA is addressed AND the FPGA is
// not the local bus master.
//
assign qlads = lads_i && !la_i[23] && !fholda;

//
// 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
// float, because the control logic on the card might also drive it.
//
assign lbterm_l = lbterm_oe ? ~lbterm_o : 1'bZ;
    
//
// 'lready_l' should only be driven when the FPGA is addressed; otherwise
// float because the control logic on the card might also drive it.
//
assign lready_l = lready_oe ? ~lready_o : 1'bZ;
    
//
// Drive the local data bus on a read.
//
assign lad = ld_oe ? ld_o : 32'hZZZZZZZZ;
    
//
// Generate the 'oe_reg0' signal, for enabling the cycle count register onto
// the internal data bus.
//
// 'oe_reg0' is set to 1 when the 'ds_decode' pulse occurs and the local bus
// cycle is a read.
//
// 'oe_reg0' is cleared to 0 when the local bus cycle ends (signified by
// assertion of 'lblast_l').
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        oe_reg0 <= 1'b0;
    end else begin
        if (ds_xfer && (lblast_i || lbterm_o)) begin
            oe_reg0 <= 1'b0;
        end else begin
            if (ds_decode && !ds_write) begin
                oe_reg0 <= 1'b1;
            end
        end
    end
end

//
// Generate the internal output data bus - in this case trivially.
//
assign ld_o = oe_reg0 ? count_q : 32'hZZZZZZZZ;

//
// Cycle counter, clocked by LCLK*2
//
always @ (posedge rst or posedge multclk)
begin
    if (rst) begin
        count <= 0;
    end else begin
        if (ds_xfer && ds_write) begin
            for (i = 0; i <= 3; i = i + 1) begin
                if (lbe_i[0]) begin
                    count[7:0] <= ld_i[7:0];
                end
                if (lbe_i[1]) begin
                    count[15:8] <= ld_i[15:8];
                end
                if (lbe_i[2]) begin
                    count[23:16] <= ld_i[23:16];
                end
                if (lbe_i[3]) begin
                    count[31:24] <= ld_i[31:24];
                end
            end
        end else begin
            count <= count + 1;
        end
    end
end

//
// Resample the LCLK*2 cycle counter to LCLK. Should be safe with
// regard to metastability because the two clock domains are
// phase-locked.
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        count_q <= 0;
    end else begin
        count_q <= count;
    end
end

//
// Instantiate the clocks module, which generates LCLK*2.
//
clocks
clocks0(
   .rst     (rst),
   .lclk    (lclk),
   .mclk    (mclk),
   .clk     (clk),
   .multclk (multclk),
   .locked  (locked));
    
//
// Instantiate the direct slave state machine; monitors the local bus for
// direct slave cycles and responds appropriately.
//
plxdssm
slave0(
   .clk        (clk),
   .rst        (rst),
   .sr         (1'b0),
   .qlads      (qlads),
   .lblast     (lblast_i),
   .lwrite     (lwrite_i),
   .ld_oe      (ld_oe),
   .lready     (lready_o),
   .lready_oe  (lready_oe),
   .lbterm     (lbterm_o),
   .lbterm_oe  (lbterm_oe),
   .transfer   (ds_xfer),
   .decode     (ds_decode),
   .write      (ds_write),
   .ready      (1'b1),
   .stop       (1'b1));

endmodule
