/*
** itest.v - FPGA design demonstrating interrupt generation
**
** (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2004
**
** Example program for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRCII-Lite
**   o ADM-XRCII
**
** NOTES ON OPERATION
**
** The logic in this FPGA design works in the way described
** in the application note an-xrc06.pdf, in the doc/ directory
** of this SDK.
**
** 32 sources of interrupt are provided, each of which can be
** triggered by writing a 1 to a corresponding bit of a write-
** only register called TEST. When a source of interrupt has
** is triggered, the corresponding bit of ISTAT becomes 1.
**
** Each source of interrupt may be masked (disabled) by the
** corresponding bit of IMASK being 1. The IMASK register resets
** to 0xffffffff, ie. all interrupts disabled.
**
** A write to byte 0 of the IARM register forces the FINTI_L signal
** high (deasserted) for a single cycle, to "rearm" the FINTI_L signal.
** The reason for including the IARM register is discussed in
** an-xrc06.pdf .
*/

`timescale 1ns / 100ps

module itest(
   lclk,
   lreset_l, 
   lwrite,
   lads_l,
   lblast_l,
   lbterm_l,
   ld,
   la,
   lready_l, 
   lbe_l,
   fholda,
   finti_l
);

input         lclk;
wire          lclk;
input         lreset_l;
wire          lreset_l;
input         lwrite;
wire          lwrite;
input         lads_l;
wire          lads_l;
input         lblast_l;
wire          lblast_l;
output        lbterm_l;
wire          lbterm_l;
inout [31:0]  ld;
wire  [31:0]  ld;
input [23:2]  la;
wire  [23:2]  la;
output        lready_l;
wire          lready_l;
input [3:0]   lbe_l;
wire  [3:0]   lbe_l;
input         fholda;
wire          fholda;
output        finti_l;
wire          finti_l;

wire          rst;
wire          lads_i;
wire          lblast_i;
wire [3:0]    lbe_i;
wire          lwrite_i;
wire [31:0]   ld_i;
wire [23:2]   la_i;
wire          lbterm_o;
wire          lbterm_oe;
wire          lready_o;
wire          lready_oe;
wire [31:0]   ld_o;
wire          ld_oe;
    
reg           finti;
    
reg  [23:2]   la_q;

wire          qlads;

wire          ds_xfer;
wire          ds_decode;
wire          ds_write;
    
reg           oe_imask;
reg           oe_istat;
reg           oe_count;
reg  [31:0]   istat;
reg  [31:0]   imask;
reg           iarm;
reg  [31:0]   count;
   
//
// Convert the inputs to active high.
//	
assign rst      = ~lreset_l;
assign lblast_i = ~lblast_l;
assign lads_i   = ~lads_l;
assign lwrite_i = lwrite;
assign la_i     = la;
assign ld_i     = ld;
assign lbe_i    = ~lbe_l;

//
// Generate a qualified version of 'lads_l', which is
// asserted when the FPGA is addressed AND the FPGA is
// not the local bus master.
//
assign qlads = lads_i && !la_i[23] && !fholda;

//
// Latch the local bus address on the 'lads_l' pulse.
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        la_q <= 0;
    end else begin
        if (lads_i) begin
            la_q <= la_i;
        end
    end
end
    
//
// 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
// float, because the control logic on the card might also drive it.
//
assign lbterm_l = lbterm_oe ? ~lbterm_o : 1'bZ;
    
//
// 'lready_l' should only be driven when the FPGA is addressed; otherwise
// float because the control logic on the card might also drive it.
//
assign lready_l = lready_oe ? ~lready_o : 1'bZ;
    
//
// Drive the local data bus on a read.
//
assign ld = ld_oe ? ld_o : 32'hZZZZZZZZ;
    
//
// Generate the 'oe_xxx' signals, for enabling the registers onto the
// internal data bus.
//
// 'oe_imask' is set to 1 when the 'ds_decode' pulse occurs and the local bus
// cycle is a read and 'la[4:2]' is '000'.
//
// 'oe_istat' is set to 1 when the 'ds_decode' pulse occurs and the local bus
// cycle is a read and 'la[4:2]' is '001'.
//
// 'oe_count' is set to 1 when the 'ds_decode' pulse occurs and the local bus
// cycle is a read and 'la[4:2]' is '100'.
//
// 'oe_istat', 'oe_imask' and 'oe_count' are all cleared to 0 when the local
// bus cycle ends (signified by assertion of 'lblast_l').
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        oe_imask <= 1'b0;
        oe_istat <= 1'b0;
        oe_count <= 1'b0;
    end else begin
        if (ds_xfer && (lblast_i || lbterm_o)) begin
            oe_imask <= 1'b0;
            oe_istat <= 1'b0;
            oe_count <= 1'b0;
        end else begin
            if (ds_decode && !ds_write) begin
                if (la_q[4:2]  == 3'b000) begin
                    oe_imask <= 1;
                end else begin
                    oe_imask <= 0;
                end
                if (la_q[4:2]  == 3'b001) begin
                    oe_istat <= 1;
                end else begin
                    oe_istat <= 0;
                end
                if (la_q[4:2]  == 3'b100) begin
                    oe_count <= 1;
                end else begin
                    oe_count <= 0;
                end
            end
        end
    end
end

//
// Generate the internal output data bus.
//
// This synthesises to an internal tristate bus, unless the synthesis
// tool optimises it into a multiplexor.
//
assign ld_o = oe_istat ? istat : 32'hZZZZZZZZ;
assign ld_o = oe_imask ? imask : 32'hZZZZZZZZ;
assign ld_o = oe_count ? count : 32'hZZZZZZZZ;

//
// generate 'finti_l' (FPGA interrupt)
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        finti <= 1'b0;
    end else begin
        if (!(~imask & istat) || iarm) begin
            finti <= 1'b0;
        end else begin
            finti <= 1'b1;
        end
    end
end

assign finti_l = ~finti;

//
// Implement the registers.
//
// Register     Local addr      Type            Function
// ======================================================================
// IMASK        0x0             R/W             Interrupt mask
// ISTAT        0x4             R/W1C           Interrupt status
// IARM         0x8             WO/RAX          Interrupt rearm strobe
// TEST         0xC             WO              Sets ISTAT for testing
// COUNT        0x10            R/W             LCLK cycle counter
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        imask <= 32'hFFFFFFFF;
        istat <= 0;
        iarm  <= 0;
        count <= 0;
    end else begin
        if (ds_xfer && ds_write && la_q[4:2] == 3'b000) begin
            // Write to IMASK
            if (lbe_i[0]) begin
                imask[7:0] <= ld_i[7:0];
            end
            if (lbe_i[1]) begin
                imask[15:8] <= ld_i[15:8];
            end
            if (lbe_i[2]) begin
                imask[23:16] <= ld_i[23:16];
            end
            if (lbe_i[3]) begin
                imask[31:24] <= ld_i[31:24];
            end
        end
            
        if (ds_xfer && ds_write && la_q[4:2] == 3'b001) begin
            // Write to ISTAT - 1s clear corresponding bit
            if (lbe_i[0]) begin
                istat[7:0] <= istat[7:0] & ~ld_i[7:0];
            end
            if (lbe_i[1]) begin
                istat[15:8] <= istat[15:8] & ~ld_i[15:8];
            end
            if (lbe_i[2]) begin
                istat[23:16] <= istat[23:16] & ~ld_i[23:16];
            end
            if (lbe_i[3]) begin
                istat[31:24] <= istat[31:24] & ~ld_i[31:24];
            end
        end

        // Write to IARM
        iarm <= ds_xfer && ds_write && la_q[4:2] == 3'b010 && lbe_i[0];

        if (ds_xfer && ds_write && la_q[4:2] == 3'b011) begin
            // Write to TEST
            if (lbe_i[0]) begin
                istat[7:0] <= istat[7:0] | ld_i[7:0];
            end
            if (lbe_i[1]) begin
                istat[15:8] <= istat[15:8] | ld_i[15:8];
            end
            if (lbe_i[2]) begin
                istat[23:16] <= istat[23:16] | ld_i[23:16];
            end
            if (lbe_i[3]) begin
                istat[31:24] <= istat[31:24] | ld_i[31:24];
            end
        end

        // Cycle counter
        if (!istat[0]) begin
            // Allow writes to cycle counter when ISTAT[0] is clear.
            if (ds_xfer && ds_write && la_q[4:2] == 3'b100) begin
                // Write to COUNT
                if (lbe_i[0]) begin
                    count[7:0] <= ld_i[7:0];
                end
                if (lbe_i[1]) begin
                    count[15:8] <= ld_i[15:8];
                end
                if (lbe_i[2]) begin
                    count[23:16] <= ld_i[23:16];
                end
                if (lbe_i[3]) begin
                    count[31:24] <= ld_i[31:24];
                end
            end
        else
            // Increment the cycle counter when ISTAT[0] is 1.
            count <= count + 1;
        end
    end
end

//
// Instantiate the direct slave state machine; monitors the local bus for
// direct slave cycles and responds appropriately.
//
plxdssm plxdssm0(
    .clk        (lclk),
    .rst        (rst),
    .sr         (1'b0),
    .qlads      (qlads),
    .lblast     (lblast_i),
    .lwrite     (lwrite_i),
    .ld_oe      (ld_oe),
    .lready     (lready_o),
    .lready_oe  (lready_oe),
    .lbterm     (lbterm_o),
    .lbterm_oe  (lbterm_oe),
    .transfer   (ds_xfer),
    .decode     (ds_decode),
    .write      (ds_write),
    .ready      (1'b1),
    .stop       (1'b1));
    
endmodule
