/*
** master-xrc2l.v - Direct Master example FPGA design
**
** (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2004
**
** Example program for:
**
**   o ADM-XRCII-Lite
**
** NOTES ON OPERATION
** ==================
**
** LA[31:2] are connected to the FPGA.
**
** This design implements the following registers, accessible by
** direct slave tranfers:
**
**   Offset Type  Name     Function
**   ===================================================================
**
**   0x0    WO    ADDR     Local bus address for direct master cycles.
**
**   0x4    WO    WDATA    Data to be written in next direct master
**                         write cycle.
**
**   0x8    WO    CFG      Selects what type of direct master cycle to
**                         perform; bit fields as follows:
**
**                         [0] WRITE   1 => perform a write
**                                     0 => perform a read
**
**   0xC    RO    RDATA    Data read in last direct master read cycle.
**
**   0x10   RO    STAT     Status register; bit fields as follows:
**
**                         [0] IDLE    1 => no direct master cycle
**                                          in progress
**                                     0 => direct master cycle in
**                                          progress
*/

`timescale 1ns / 100ps

module master(
    lclk,
    lreset_l,
    lwrite,
    lads_l,
    lblast_l,
    lbterm_l,
    ld,
    la,
    lreadyi_l,
    lbe_l,
    fhold,
    fholda,
    lreadyo_l,
    lbtermo_l);

parameter la_top = 31;

input                lclk;
input                lreset_l;
inout                lwrite;
inout                lads_l;
inout                lblast_l;
output               lbterm_l;
inout  [31:0]        ld;
inout  [la_top:2]    la;
output               lreadyi_l; 
inout  [3:0]         lbe_l;
output               fhold; 
input                fholda; 
input                lreadyo_l;
input                lbtermo_l;

wire                 rst;
wire                 lads_i;
wire                 lblast_i;
wire   [3:0]         lbe_i;
wire                 lreadyi_i;
wire                 lbtermo_i;
wire                 lwrite_i;
    
wire   [31:0]        ld_i;
reg    [31:0]        ld_i_q;
wire   [la_top:2]    la_i;
    
wire                 lads_o;
wire                 lads_oe;
wire   [3:0]         lbe_o;
wire                 lbe_oe;
wire                 lblast_o;
wire                 lblast_oe;
wire                 lbterm_o;
wire                 lbterm_oe;
wire                 lreadyi_o;
wire                 lreadyi_oe;
wire                 lreadyo_i;
wire                 lwrite_o;
wire                 lwrite_oe;
    
wire                 la_oe;
wire   [la_top:2]    la_o;
wire                 ld_oe;
wire   [31:0]        ld_o;
wire   [31:0]        ld_o_ds;
wire   [31:0]        ld_o_dm;
    
reg    [la_top:2]    la_q;
    
wire                 ds_qlads;
wire                 ds_xfer;
wire                 ds_decode;
wire                 ds_write;
wire                 ds_ld_oe;
    
reg    [31:0]        reg_addr;
reg    [31:0]        reg_wdata;
reg    [31:0]        reg_cfg;
reg                  oe_rdata;
reg    [31:0]        reg_rdata;
reg                  oe_stat;
wire   [31:0]        reg_stat;

reg                  dm_req;
wire                 dm_xfer;
wire                 dm_ld_oe;
wire                 dm_done;

// -------------------------------------------------------------------
// Common
// -------------------------------------------------------------------

//    
// Convert inputs to active high.
//
assign rst  = ~lreset_l;
assign ld_i = ld;

//
// Register local bus data on the way in.
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        ld_i_q <= 0;
    end else begin
        ld_i_q <= ld_i;
    end
end

//
// Drive the local data bus when responding to a direct slave read cycle
// or direct master write cycle.
//
assign ld_oe = ds_ld_oe || dm_ld_oe;
assign ld    = ld_oe ? ld_o : 32'hZZZZZZZZ;

//
// Generate 'ld_o', the data for the LD pins.
//
assign ld_o = ds_ld_oe ? ld_o_ds : ld_o_dm;

// -------------------------------------------------------------------
// Direct Slave
// -------------------------------------------------------------------

//
// Convert inputs to active high.
//
assign la_i     = la;
assign lads_i   = ~lads_l;
assign lblast_i = ~lblast_l;
assign lbe_i    = ~lbe_l;
assign lwrite_i = lwrite;

//
// 'ds_qlads' is 'lads_l' qualified as a direct slave access to the FPGA.
//
assign ds_qlads = lads_i && !la_i[23] && !fholda;

//
// Latch local bus address on direct slave transfers.
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        la_q <= 0;
    end else begin
        if (lads_i) begin
            la_q <= la_i;
        end
    end
end

//
// Drive the bus when responding to a direct slave cycle.
//
assign lbterm_l  = lbterm_oe ? ~lbterm_o : 1'bZ;
assign lreadyi_l = lreadyi_oe ? ~lreadyi_o : 1'bZ;

//
// Implement the registers that are writable via direct slave cycles.
//
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        dm_req    <= 1'b0;
        reg_addr  <= 0;
        reg_wdata <= 0;
        reg_cfg   <= 0;
    end else begin
        if (ds_xfer && ds_write && la_q[4:2] == 3'h0) begin
            //
            // Write to ADDR register
            //
            if (lbe_i[0]) begin
                reg_addr[7:2] <= ld_i[7:2];
            end
            if (lbe_i[1]) begin
                reg_addr[15:8] <= ld_i[15:8];
            end
            if (lbe_i[2]) begin
                reg_addr[23:16] <= ld_i[23:16];
            end
            if (lbe_i[3]) begin
                reg_addr[31:24] <= ld_i[31:24];
            end
        end

        //
        // Writing to byte 3 of ADDR register also triggers a
        // direct master cycle.
        //
        dm_req <= ds_xfer && ds_write && la_q[4:2] == 3'h0 && lbe_i[3];

        if (ds_xfer && ds_write && la_q[4:2] == 3'h1) begin
            //
            // Write to WDATA register.
            //
            if (lbe_i[0]) begin
                reg_wdata[7:0] <= ld_i[7:0];
            end
            if (lbe_i[1]) begin
                reg_wdata[15:8] <= ld_i[15:8];
            end
            if (lbe_i[2]) begin
                reg_wdata[23:16] <= ld_i[23:16];
            end
            if (lbe_i[3]) begin
                reg_wdata[31:24] <= ld_i[31:24];
            end
        end
            
        if (ds_xfer && ds_write && la_q[4:2] == 3'h2) begin
            //
            // Write to CFG register
            //
            if (lbe_i[0]) begin
                reg_cfg[0] <= ld_i[0];
            end
        end
    end
end

//
// Implement the registers that are readable via direct slave cycles.
//

always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        reg_rdata <= 32'hDEADBEEF;
    end else begin
        if (dm_xfer) begin
            reg_rdata <= ld_i_q;
        end
    end
end

assign reg_stat = { dm_done };

//
// Generate the output enables (for the internal data bus in the
// FPGA) for the registers that are readable via direct slave cycles.
//    
always @ (posedge rst or posedge lclk)
begin
    if (rst) begin
        oe_rdata <= 1'b0;
        oe_stat  <= 1'b0;
    end else begin
        if (ds_xfer && (lblast_i || lbterm_o)) begin
            oe_rdata <= 1'b0;
            oe_stat  <= 1'b0;
        end else begin
            if (ds_decode && !ds_write) begin
                if (la_q[4:2] == 3'h3) begin
                    oe_rdata <= 1'b1;
                end

                if (la_q[4:2] == 3'h4) begin
                    oe_stat <= 1'b1;
                end
            end
        end
    end
end

//
// Generate 'ld_o_ds', the data for the LD pins when current bus cycle
// is a direct slave transfer.
//
assign ld_o_ds = oe_rdata ? reg_rdata : reg_stat;

//
// Instantiate the direct slave state machine; monitors the local bus for
// direct slave cycles and responds appropriately.
//
plxdssm plxdssm0(
    .clk        (lclk),
    .rst        (rst),
    .sr         (1'b0),
    .qlads      (ds_qlads),
    .lblast     (lblast_i),
    .lwrite     (lwrite_i),
    .ld_oe      (ds_ld_oe),
    .lready     (lreadyi_o),
    .lready_oe  (lreadyi_oe),
    .lbterm     (lbterm_o),
    .lbterm_oe  (lbterm_oe),
    .transfer   (ds_xfer),
    .decode     (ds_decode),
    .write      (ds_write),
    .ready      (1'b1),
    .stop       (1'b1)); 

// -------------------------------------------------------------------
// Direct Master
// -------------------------------------------------------------------

//
// Convert inputs to active high.
//
assign lreadyo_i = ~lreadyo_l;
assign lbtermo_i = ~lbtermo_l;

//
// Drive the bus when performing a direct master cycle.
//
assign la       = la_oe ? la_o : 32'hZZZZZZZZ;
assign lbe_l    = lbe_oe ? ~lbe_o : 4'hZ;
assign lads_l   = lads_oe ? ~lads_o : 1'bZ;
assign lblast_l = lblast_oe ? ~lblast_o : 1'bZ;
assign lwrite   = lwrite_oe ? lwrite_o : 1'bZ;

//
// Generate the address and byte enables to be output onto the local bus.
//
assign la_o  = { 8'hFF, reg_addr[23:2] };
assign lbe_o = 4'hF;

//
// There is currently only one source of data for direct master writes.
//
assign ld_o_dm = reg_wdata;

//
// Instantiate the direct master state machine; performs local bus
// cycles as a master.
//
plxdmsm plxdmsm0(
    .clk        (lclk),
    .rst        (rst),
    .sr         (1'b0),
    .request    (dm_req),
    .stop       (1'b1),
    .transfer   (dm_xfer),
    .done       (dm_done),
    .write      (reg_cfg[0]),
    .lready     (lreadyo_i),
    .lbterm     (lbtermo_i),
    .fholda     (fholda),
    .fhold      (fhold),
    .la_oe      (la_oe),
    .lads       (lads_o),
    .lads_oe    (lads_oe),
    .lbe_oe     (lbe_oe),
    .lblast     (lblast_o),
    .lblast_oe  (lblast_oe),
    .ld_oe      (dm_ld_oe),
    .lwrite     (lwrite_o),
    .lwrite_oe  (lwrite_oe));

endmodule
