//
// clocks_dll2.v - clock generator module for ZBT design, using
//                 Virtex/Virtex-E/Virtex-EM DLLs
//
//                 Generates 2 SSRAM clocks.
//
// (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2003
//
// Example program for: ADM-XRC
//

`timescale 1ns / 100ps

module clocks(
    lreset_l,
    rst,
    lclk,
    mclk,
    clk,
    ramclko,
    ramclki,
    locked);

parameter num_clock = 2;

input                   lreset_l;
output                  rst;
input                   lclk;
input                   mclk;
input  [num_clock-1:0]  ramclki;
output                  clk;
output [num_clock-1:0]  ramclko;
output [31:0]           locked;

wire                    rst;
wire                    lclk;
wire                    mclk;
wire   [num_clock-1:0]  ramclki;
wire                    clk;
wire   [num_clock-1:0]  ramclko;
wire   [31:0]           locked;

wire                    lclk_ibufg;
wire                    lclk0;
wire                    locked_lclk;
wire   [num_clock-1:0]  locked_ramclk;
wire   [7:0]            ramclki_ibufg;

reg                     dllrst;
reg    [3:0]            rstcnt;

wire                    mclk_ibufg;
wire                    mclk_bufg;

//
// Generate the 'locked' vector. 1 bit per DLL/DCM, where 1
// indicates 'lock achieved'.
//
assign locked   = { locked_ramclk, locked_lclk };

//
// Use a global buffer to distribute the reset. This avoids the
// problem of PAR being unable to pack registers into IOBs because
// various different buffered copies of the reset signal are used.
//
BUFG ibufg_reset(
    .I        (~lreset_l),
    .O        (rst));

//
// Input MCLK
//

IBUFG ibufg_mclk(
    .I        (mclk),
    .O        (mclk_ibufg));

BUFG bufg_mclk(
    .I        (mclk_ibufg),
    .O        (mclk_bufg));

//
// Generate reset signal to DLLs/DCMs
//
always @ (posedge mclk_bufg or posedge rst) begin
    if (rst) begin
        dllrst <= 1'b1;
        rstcnt <= 0;
    end else begin
        if (dllrst) begin
            rstcnt <= rstcnt + 1;
        end
        if (&rstcnt) begin
            dllrst <= 1'b0;
        end
    end
end
    
//
// Input the local bus clock
//

IBUFG ibufg_lclk(
    .I      (lclk),
    .O      (lclk_ibufg));

CLKDLL dll_lclk(    
    .CLKIN  (lclk_ibufg),
    .CLKFB  (clk),
    .RST    (dllrst),
    .CLK0   (lclk0),
    .LOCKED (locked_lclk));
/* synopsys attribute
DUTY_CYCLE_CORRECTION "TRUE"
STARTUP_WAIT          "FALSE"
*/
// synthesis attribute DUTY_CYCLE_CORRECTION of dll_lclk is "TRUE";
// synthesis attribute STARTUP_WAIT          of dll_lclk is "TRUE";

BUFG bufg_lclk(
    .I      (lclk0),
    .O      (clk));
    
//
// Generate the SSRAM clocks
//

IBUFG ibufg0(
    .I      (ramclki[0]),
    .O      (ramclki_ibufg[0]));

CLKDLL dll0(
    .CLKIN  (clk),
    .CLKFB  (ramclki_ibufg[0]),
    .RST    (~locked_lclk),
    .CLK0   (ramclko[0]),
    .LOCKED (locked_ramclk[0]));
/* synopsys attribute
DUTY_CYCLE_CORRECTION "TRUE"
STARTUP_WAIT          "FALSE"
*/
// synthesis attribute DUTY_CYCLE_CORRECTION of dll0 is "TRUE";
// synthesis attribute STARTUP_WAIT          of dll0 is "TRUE";

IBUFG ibufg1(
    .I      (ramclki[1]),
    .O      (ramclki_ibufg[1]));

CLKDLL dll1(
    .CLKIN  (clk),
    .CLKFB  (ramclki_ibufg[1]),
    .RST    (~locked_lclk),
    .CLK0   (ramclko[1]),
    .LOCKED (locked_ramclk[1]));
/* synopsys attribute
DUTY_CYCLE_CORRECTION "TRUE"
STARTUP_WAIT          "FALSE"
*/
// synthesis attribute DUTY_CYCLE_CORRECTION of dll1 is "TRUE";
// synthesis attribute STARTUP_WAIT          of dll1 is "TRUE";

endmodule
