//
// zbt_main.v - Local bus interface giving host access to ZBT SSRAM
//              on an ADM-XRC series card.
//
// (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2004
//
// This module is a parameterisable 32-bit local bus interface to ZBT
// SSRAM. The parameters are as follows:
//
//   num_bank     - number of logical banks of SSRAM
//   num_clock    - number of SSRAM clocks to output
//   addr_width   - (maximum) number of address bits to the SSRAM
//   ctl_width    - number of control bits to the SSRAM
//   data_width   - number of data bits per physical SSRAM bank
//   bank_group   - number of physical SSRAM banks to combine in
//                  order to produce a logical SSRAM bank
//
// When the physical SSRAM banks on the card are less than 32 bits
// wide, two or more banks (specified by 'bank_group') may be combined
// in order to produce a single logical bank. Thus, the product of
// 'bank_group' and 'data_width' must equal 32.
//

`timescale 1ns / 100ps

module zbt_main(
    lclk,
    mclk,
    ramclko,
    ramclki,
    lreset_l,
    lwrite,
    lads_l,
    lblast_l,
    lbterm_l,
    la_i,
    ld_i,
    ld_o,
    oe_ld,
    lready_l, 
    lbe_l,
    fholda,
    ra0,
    rc0,
    rd0,
    ra1,
    rc1,
    rd1,
    ra2,
    rc2,
    rd2,
    ra3,
    rc3,
    rd3,
    ra4,
    rc4,
    rd4,
    ra5,
    rc5,
    rd5,
    ra6,
    rc6,
    rd6,
    ra7,
    rc7,
    rd7,
    glob_lbo_l,
    glob_flothru_l);

parameter num_bank   = 4;
parameter num_clock  = 1;
parameter addr_width = 19;
parameter ctl_width  = 9;
parameter data_width = 32;
parameter bank_group = 1;

input                              lclk;
input                              mclk;
output [num_clock-1:0]             ramclko;
input  [num_clock-1:0]             ramclki;
input                              lreset_l;
input                              lwrite;
input                              lads_l;
input                              lblast_l;
output                             lbterm_l;
input  [23:2]                      la_i;
input  [31:0]                      ld_i;
output [31:0]                      ld_o;
output                             oe_ld;
output                             lready_l;
input  [3:0]                       lbe_l;
input                              fholda;
output [addr_width*bank_group-1:0] ra0;
output [ctl_width *bank_group-1:0] rc0;
inout  [data_width*bank_group-1:0] rd0;
output [addr_width*bank_group-1:0] ra1;
output [ctl_width *bank_group-1:0] rc1;
inout  [data_width*bank_group-1:0] rd1;
output [addr_width*bank_group-1:0] ra2;
output [ctl_width *bank_group-1:0] rc2;
inout  [data_width*bank_group-1:0] rd2;
output [addr_width*bank_group-1:0] ra3;
output [ctl_width *bank_group-1:0] rc3;
inout  [data_width*bank_group-1:0] rd3;
output [addr_width*bank_group-1:0] ra4;
output [ctl_width *bank_group-1:0] rc4;
inout  [data_width*bank_group-1:0] rd4;
output [addr_width*bank_group-1:0] ra5;
output [ctl_width *bank_group-1:0] rc5;
inout  [data_width*bank_group-1:0] rd5;
output [addr_width*bank_group-1:0] ra6;
output [ctl_width *bank_group-1:0] rc6;
inout  [data_width*bank_group-1:0] rd6;
output [addr_width*bank_group-1:0] ra7;
output [ctl_width *bank_group-1:0] rc7;
inout  [data_width*bank_group-1:0] rd7;
output                             glob_lbo_l;
output                             glob_flothru_l;

integer                            i;

//
// We support up to 8 logical banks of SSRAM.
//
parameter max_bank = 8;

//
// We have a 2MB window in local bus space where the SSRAM appears.
// This window is augmented with a page register.
//
parameter page_order = 19;
parameter page_size  = 1 << page_order; // in longwords

//
// The minumum number of address bits in a logical SSRAM bank.
//
parameter min_bank_order = 17;
parameter min_bank_size  = 1 << min_bank_order; // in SSRAM words

//
// The number of address bits and maximum size (in longwords) of a
// logical SSRAM bank.
//
parameter bank_order = addr_width;
parameter bank_size  = 1 << bank_order; // in SSRAM words

//
// We terminate bursts at a page boundary or the smallest possible bank
// boundary, whichever occurs more frequently.
//
parameter bterm_bound     = (min_bank_size < page_size) ? min_bank_size : page_size; // in longwords
parameter bterm_order     = (min_bank_order < bank_order) ? min_bank_order : bank_order;
parameter bterm_bound_vec = bterm_bound - 1;

//
// Get widths of address, control and data busses to/from a logical SSRAM bank
//
parameter log_addr_width  = addr_width * bank_group;
parameter log_ctl_width   = ctl_width * bank_group;
parameter log_data_width  = data_width * bank_group;

reg    [log_addr_width-1:0]        ra0;
reg    [log_ctl_width-1:0]         rc0;
wire   [log_data_width-1:0]        rd0;
reg    [log_addr_width-1:0]        ra1;
reg    [log_ctl_width-1:0]         rc1;
wire   [log_data_width-1:0]        rd1;
reg    [log_addr_width-1:0]        ra2;
reg    [log_ctl_width-1:0]         rc2;
wire   [log_data_width-1:0]        rd2;
reg    [log_addr_width-1:0]        ra3;
reg    [log_ctl_width-1:0]         rc3;
wire   [log_data_width-1:0]        rd3;
reg    [log_addr_width-1:0]        ra4;
reg    [log_ctl_width-1:0]         rc4;
wire   [log_data_width-1:0]        rd4;
reg    [log_addr_width-1:0]        ra5;
reg    [log_ctl_width-1:0]         rc5;
wire   [log_data_width-1:0]        rd5;
reg    [log_addr_width-1:0]        ra6;
reg    [log_ctl_width-1:0]         rc6;
wire   [log_data_width-1:0]        rd6;
reg    [log_addr_width-1:0]        ra7;
reg    [log_ctl_width-1:0]         rc7;
wire   [log_data_width-1:0]        rd7;

reg  [31:0]                        ld_o;

wire                               rst;
wire                               clk;
wire [31:0]                        locked;

wire                               lads_i;
wire                               lblast_i;
wire                               qlads;
wire                               lwrite_i;
wire                               lready_o;
wire                               lready_oe;
wire                               lbterm_o;
wire                               lbterm_oe;
wire [3:0]                         lbe_i;

reg  [23:2]                        la_q;

wire [31:0]                        id;

reg                                sram_rd_ready;
wire                               sram_wr_ready;
reg                                regs_rd_ready;
wire                               regs_wr_ready;
wire                               sram_rd_stop;
wire                               sram_wr_stop;
wire                               regs_rd_stop;
wire                               regs_wr_stop;

wire                               ds_xfer;
wire                               ds_decode;
wire                               ds_write;
wire                               ds_ready;
wire                               ds_stop;
    
reg                                oe_page_reg;
reg                                oe_pipeline_reg;
reg                                oe_size_reg;
reg                                oe_info_reg;
reg                                oe_status_reg;
reg  [7:0]                         page_reg;
reg  [0:0]                         pipeline_reg;
reg  [1:0]                         size_reg;
wire [31:0]                        info_reg;
reg  [31:0]                        status_reg;

reg  [page_order-1:0]              page_addr;
wire                               inc_page_addr;
wire [page_order+8-1:0]            ssram_addr;

wire [page_order+8-1:0]            decode_addr;
reg  [2:0]                         banksel_addr;
reg  [max_bank-1:0]                banksel;

reg                                read_enable;

reg                                zbt_read;
reg  [max_bank-1:0]                zbt_readv;
reg                                zbt_write;
reg  [max_bank-1:0]                zbt_writev;
reg  [max_bank-1:0]                oe_zbt_q;

wire [31:0]                        zbt_q_port0, zbt_q_port1, zbt_q_port2, zbt_q_port3;
wire [31:0]                        zbt_q_port4, zbt_q_port5, zbt_q_port6, zbt_q_port7;
wire [log_addr_width-1:0]          zbt_t_ra_port0, zbt_t_ra_port1, zbt_t_ra_port2, zbt_t_ra_port3;
wire [log_addr_width-1:0]          zbt_t_ra_port4, zbt_t_ra_port5, zbt_t_ra_port6, zbt_t_ra_port7;
wire [log_addr_width-1:0]          zbt_ra_o_port0, zbt_ra_o_port1, zbt_ra_o_port2, zbt_ra_o_port3;
wire [log_addr_width-1:0]          zbt_ra_o_port4, zbt_ra_o_port5, zbt_ra_o_port6, zbt_ra_o_port7;
wire [log_ctl_width-1:0]           zbt_t_rc_port0, zbt_t_rc_port1, zbt_t_rc_port2, zbt_t_rc_port3;
wire [log_ctl_width-1:0]           zbt_t_rc_port4, zbt_t_rc_port5, zbt_t_rc_port6, zbt_t_rc_port7;
wire [log_ctl_width-1:0]           zbt_rc_o_port0, zbt_rc_o_port1, zbt_rc_o_port2, zbt_rc_o_port3;
wire [log_ctl_width-1:0]           zbt_rc_o_port4, zbt_rc_o_port5, zbt_rc_o_port6, zbt_rc_o_port7;
wire [log_data_width-1:0]          zbt_t_rd_port0, zbt_t_rd_port1, zbt_t_rd_port2, zbt_t_rd_port3;
wire [log_data_width-1:0]          zbt_t_rd_port4, zbt_t_rd_port5, zbt_t_rd_port6, zbt_t_rd_port7;
wire [log_data_width-1:0]          zbt_rd_o_port0, zbt_rd_o_port1, zbt_rd_o_port2, zbt_rd_o_port3;
wire [log_data_width-1:0]          zbt_rd_o_port4, zbt_rd_o_port5, zbt_rd_o_port6, zbt_rd_o_port7;
wire [max_bank-1:0]                zbt_evalid_port;

wire [31:0]                        zbt_q0, zbt_q1, zbt_q2, zbt_q3;
wire [31:0]                        zbt_q4, zbt_q5, zbt_q6, zbt_q7;
wire [log_addr_width-1:0]          zbt_t_ra0, zbt_t_ra1, zbt_t_ra2, zbt_t_ra3;
wire [log_addr_width-1:0]          zbt_t_ra4, zbt_t_ra5, zbt_t_ra6, zbt_t_ra7;
wire [log_addr_width-1:0]          zbt_ra_o0, zbt_ra_o1, zbt_ra_o2, zbt_ra_o3;
wire [log_addr_width-1:0]          zbt_ra_o4, zbt_ra_o5, zbt_ra_o6, zbt_ra_o7;
wire [log_ctl_width-1:0]           zbt_t_rc0, zbt_t_rc1, zbt_t_rc2, zbt_t_rc3;
wire [log_ctl_width-1:0]           zbt_t_rc4, zbt_t_rc5, zbt_t_rc6, zbt_t_rc7;
wire [log_ctl_width-1:0]           zbt_rc_o0, zbt_rc_o1, zbt_rc_o2, zbt_rc_o3;
wire [log_ctl_width-1:0]           zbt_rc_o4, zbt_rc_o5, zbt_rc_o6, zbt_rc_o7;
wire [log_data_width-1:0]          zbt_t_rd0, zbt_t_rd1, zbt_t_rd2, zbt_t_rd3;
wire [log_data_width-1:0]          zbt_t_rd4, zbt_t_rd5, zbt_t_rd6, zbt_t_rd7;
wire [log_data_width-1:0]          zbt_rd_o0, zbt_rd_o1, zbt_rd_o2, zbt_rd_o3;
wire [log_data_width-1:0]          zbt_rd_o4, zbt_rd_o5, zbt_rd_o6, zbt_rd_o7;
wire [log_data_width-1:0]          zbt_rd_i0, zbt_rd_i1, zbt_rd_i2, zbt_rd_i3;
wire [log_data_width-1:0]          zbt_rd_i4, zbt_rd_i5, zbt_rd_i6, zbt_rd_i7;
wire [max_bank-1:0]                zbt_evalid;

// =============================================================================
// Local bus interface
// =============================================================================

//
//	Convert the inputs to active high.
//	
assign lads_i   = ~lads_l;
assign lblast_i = ~lblast_l;
assign lwrite_i = lwrite;
assign lbe_i    = ~lbe_l;

//
// Generate the local bus output data
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        ld_o <= 0;
    end else begin
        ld_o <= id;
    end
end

//
// 'la_q' mirrors 'la' during the burst. To improve timing, rather than using
// 'la' direct from the local bus, we latch it at the beginning of the burst and
// increment it during the burst.
//
// Note that 'la' cannot be used as the address output to the SSRAMs, as
// there are potentially several cycles of latency between data on the local
// bus and the same data being read/written from/to a particular SSRAM bank.
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        la_q <= 0;
    end else begin
        if (lads_i) begin
            la_q <= la_i;
        end else begin
            if (ds_xfer) begin
                la_q <= la_q + 1;
            end
        end
    end
end

//
// Generate the 'ds_ready' signal to the direct slave state machine.
//
// This is generated according to the region being addressed by the current burst,
// and also whether a read or write is being performed:
//
//   o For a read from the SSRAM region, this is generated from the 'zbt_evalid'
//     vector with 1 cycle of delay.
//   o For a read from the registers region, this is generated from 'ds_decode'
//     with 1 cycle of delay.
//   o For a write to the SSRAM region, this is simply 'ds_decode'.
//   o For a write to the register region, this is simply 'ds_decode'.
//
// Since 'id' is registered before it is output on the local bus, reads require
//  1 cycle of delay before 'ds_ready' is asserted, which is why 'regs_rd_ready'
// and 'sram_rd_ready' are implemented as registers.
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        regs_rd_ready <= 1'b0;
        sram_rd_ready <= 1'b0;
    end else begin
        regs_rd_ready <= ds_decode && !ds_write && !la_q[21];
        sram_rd_ready <= read_enable && |(zbt_evalid & banksel);
    end
end
assign regs_wr_ready = ds_decode && ds_write && !la_q[21];
assign sram_wr_ready = ds_decode && ds_write && la_q[21];
assign ds_ready      = sram_rd_ready || sram_wr_ready || regs_rd_ready || regs_wr_ready;
            
//
// Generate the 'ds_stop' signal to the direct slave state machine.
//
// Terminates a burst if the any of the following conditions are met:
//
//   o burst is addressing the register region (first case)
//   o burst is addressing the SSRAM region AND a page or bank boundary
//     has been reached (last two cases)
//
assign ds_stop = (ds_decode && !la_q[21]) ||
                 (ds_decode && la_q[21] && la_q[bterm_order+1:2] == bterm_bound_vec) ||
                 (ds_xfer && la_q[21] && la_q[bterm_order+1:2] == (bterm_bound_vec - 1));

//
// Generate a qualified address strobe, ie. 'lads_l' qualified by address
// decoding for the FPGA space.
//
assign qlads = lads_i && !la_i[23] && !fholda;

//
// 'lbterm_l' should only be driven when the FPGA is addressed in direct
// slave mode, since other local bus masters on the card might also drive it.
//
assign lbterm_l = lbterm_oe ? ~lbterm_o : 1'bZ;

//
// 'lready_l' should only be driven when the FPGA is addressed in direct
// slave mode, since other local bus masters on the card might also drive it.
//
assign lready_l = lready_oe ? ~lready_o : 1'bZ;

//
// Instantiate the state machine for responding to direct slave transfers.
//
plxdssm plxdssm0(
    .clk        (clk),
    .rst        (rst),
    .sr         (1'b0),
    .qlads      (qlads),
    .lblast     (lblast_i),
    .lwrite     (lwrite_i),
    .ld_oe      (oe_ld),
    .lready     (lready_o),
    .lready_oe  (lready_oe),
    .lbterm     (lbterm_o),
    .lbterm_oe  (lbterm_oe),
    .transfer   (ds_xfer),
    .decode     (ds_decode),
    .write      (ds_write),
    .ready      (ds_ready),
    .stop       (ds_stop));

// =============================================================================
// FPGA Registers
// =============================================================================

//
// Generate output enable signals for registers
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        oe_zbt_q    <= 8'h00;
        oe_page_reg <= 1'b0;
        oe_size_reg <= 1'b0;
        oe_pipeline_reg <= 1'b0;
        oe_info_reg <= 1'b0;
        oe_status_reg <= 1'b0;
    end else begin
        if (ds_decode) begin
            for (i = 0; i < max_bank; i = i + 1) begin
                if (la_q[22:21] == 2'h1 && !ds_write && banksel[i]) begin
                    oe_zbt_q[i] <= 1'b1;
                end
            end

            if (la_q[22:21] == 2'h0 && !ds_write && la_q[4:2] == 3'h0) begin
                oe_page_reg <= 1'b1;
            end

            if (la_q[22:21] == 2'h0 && !ds_write && la_q[4:2] == 3'h1) begin
                oe_pipeline_reg <= 1'b1;
            end

            if (la_q[22:21] == 2'h0 && !ds_write && la_q[4:2] == 3'h2) begin
                oe_size_reg <= 1'b1;
            end

            if (la_q[22:21] == 2'h0 && !ds_write && la_q[4:2] == 3'h4) begin
                oe_info_reg <= 1'b1;
            end

            if (la_q[22:21] == 2'h0 && !ds_write && la_q[4:2] == 3'h5) begin
                oe_status_reg <= 1'b1;
            end
        end else begin
            if (ds_xfer && (lblast_i || lbterm_o)) begin
                oe_zbt_q        <= 8'h00;
                oe_page_reg     <= 1'b0;
                oe_pipeline_reg <= 1'b0;
                oe_size_reg     <= 1'b0;
                oe_info_reg     <= 1'b0;
                oe_status_reg   <= 1'b0;
            end
        end
    end
end

//
// Implement the read only 'info_reg' register
//
assign info_reg[31:24] = num_bank;
assign info_reg[23:0]  = bank_size;

//
// Implement the read only 'status_reg' register
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        status_reg <= 0;
    end else begin
        status_reg <= locked;
    end
end

//
// Implement the 'page', 'pipeline' and 'size' registers
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        page_reg     <= 0;
        pipeline_reg <= 0;
        size_reg     <= 1;
    end else begin
        if (ds_xfer && ds_write) begin
            if (la_q[22:21] == 2'h0 && la_q[4:2] == 3'h0) begin
                if (lbe_i[0]) begin
                    page_reg <= ld_i[7:0];
                end
            end

            if (la_q[22:21] == 2'h0 && la_q[4:2] == 3'h1) begin
                if (lbe_i[0]) begin
                    pipeline_reg <= ld_i[0:0];
                end
            end

            if (la_q[22:21] == 2'h0 && la_q[4:2] == 3'h2) begin
                if (lbe_i[0]) begin
                    size_reg <= ld_i[1:0];
                end
            end
        end
    end
end

//
// Drive the internal data bus 'id' on reads. At most one of the registers or SSRAM
// banks may be enabled onto the internal data bus.
//
assign id = oe_zbt_q[0] ? zbt_q0 : 32'hZZZZZZZZ;
assign id = oe_zbt_q[1] ? zbt_q1 : 32'hZZZZZZZZ;
assign id = oe_zbt_q[2] ? zbt_q2 : 32'hZZZZZZZZ;
assign id = oe_zbt_q[3] ? zbt_q3 : 32'hZZZZZZZZ;
assign id = oe_zbt_q[4] ? zbt_q4 : 32'hZZZZZZZZ;
assign id = oe_zbt_q[5] ? zbt_q5 : 32'hZZZZZZZZ;
assign id = oe_zbt_q[6] ? zbt_q6 : 32'hZZZZZZZZ;
assign id = oe_zbt_q[7] ? zbt_q7 : 32'hZZZZZZZZ;
assign id = oe_pipeline_reg ? pipeline_reg : 32'hZZZZZZZZ;
assign id = oe_page_reg ? page_reg : 32'hZZZZZZZZ;
assign id = oe_size_reg ? size_reg : 32'hZZZZZZZZ;
assign id = oe_info_reg ? info_reg : 32'hZZZZZZZZ;
assign id = oe_status_reg ? status_reg : 32'hZZZZZZZZ;

// =============================================================================
// SSRAM interface
// =============================================================================

//
// The 'page_addr' value must be incremented with every time:
//
//   o A word is read from the SSRAM
//   o A word is written to the SSRAM
//
assign inc_page_addr = zbt_read | zbt_write;

//
// Generate 'page_addr', which is the SSRAM address within the current page.
//
//   o At the beginning of the burst, the local bus address is latched into 'page_addr'
//   o Each time a word is read from or written to the SSRAM ('inc_page_addr'),
//     'page_addr' is incremented
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        page_addr <= 0;
    end else begin
        if (ds_decode && la_q[22:21] == 2'h1) begin
            page_addr <= la_q[page_order+1:2];
        end else begin
            if (inc_page_addr) begin
                page_addr <= page_addr + 1;
            end
        end
    end
end

//
// Generate the address that is output to the SSRAMs, by augmenting 'page_addr'
// with the value of the PAGE register.
//
// Note that we ensure that we never burst across a page or bank boundary, by
// asserting 'lbterm_l' at appropriate times (see generation of 'lbterm_l' above).
//
assign ssram_addr = { page_reg, page_addr };

//
// Generate 'decode_addr', from the page register and the local bus address.
// This value is used during address decoding to identify the bank being accessed.
//
assign decode_addr = { page_reg, la_i[page_order+1:2] };

//
// Pick out the bits of 'decode_addr' that identify the bank being accessed. Which
// particular bits are used depends upon the value of the SIZE register.
//
always @ (decode_addr or size_reg)
begin
    case (size_reg[1:0])
        0: begin
            // SSRAMs are 128k by n
            banksel_addr <= decode_addr[19:17];
        end

        1: begin
            // SSRAMs are 256k by n
            banksel_addr <= decode_addr[20:18];
        end

        2: begin
            // SSRAMs are 512k by n
            banksel_addr <= decode_addr[21:19];
        end

        3: begin
            // SSRAMs are 1M by n
            banksel_addr <= decode_addr[22:20];
        end
    endcase
end
         
//
// Generate the one-hot 'banksel' vector, from 'banksel_addr', which
// indicates whether or not a particular bank is being accessed.
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        banksel <= 0;
    end else begin
        if (lads_i) begin
            for (i = 0; i < max_bank; i = i + 1) begin
                if (banksel_addr == i) begin
                    banksel[i] <= 1'b1;
                end else begin
                    banksel[i] <= 1'b0;
                end
            end
        end
    end
end

//
// Generate the 'reading' signal, which is asserted on reads from cycle
// following the 'lads_l' pulse until the end of the current burst.
//
// This signal is used to tell a selected zbt_port module to begin reading.
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        read_enable <= 1'b0;
    end else begin
        if (ds_xfer && (lblast_i || lbterm_o)) begin
            read_enable <= 1'b0;
        end else begin
            if (ds_decode && !ds_write && la_q[22:21] == 2'h1) begin
                read_enable <= 1'b1;
            end
        end
    end
end

//
// Generate the one-hot 'zbt_writev' and 'zbt_readv' vectors to
// the zbt_port modules.
//
always @ (
    ds_xfer or
    ds_write or
    banksel or
    read_enable or
    la_q)
begin
    zbt_write <= ds_xfer && ds_write;
    for (i = 0; i < max_bank; i = i + 1) begin
        zbt_writev[i] <= ds_xfer && ds_write && banksel[i] && la_q[22:21] == 2'b01;
    end
end

always @ (posedge clk or posedge rst)
begin
    if (rst) begin
        zbt_readv <= 0;
        zbt_read  <= 1'b0;
    end else begin
        for (i = 0; i < max_bank; i = i + 1) begin
            zbt_readv[i]  <= read_enable && banksel[i];
        end

        zbt_read <= read_enable;
    end
end

//
// Instantiate the ZBT ports (bank interfaces)
//

zbt_port #(addr_width, data_width, ctl_width, bank_group) zbt_port0(
    .clk      (clk),
    .rst      (rst),
    .sr       (1'b0),
    .pipeline (pipeline_reg[0]),
    .read     (zbt_readv[0]),
    .write    (zbt_writev[0]),
    .a        (ssram_addr[bank_order-1:0]),
    .d        (ld_i),
    .be       (lbe_i),
    .q        (zbt_q_port0),
    .evalid   (zbt_evalid_port[0]),
    .t_ra     (zbt_t_ra_port0),
    .ra_o     (zbt_ra_o_port0),
    .t_rc     (zbt_t_rc_port0),
    .rc_o     (zbt_rc_o_port0),
    .t_rd     (zbt_t_rd_port0),
    .rd_o     (zbt_rd_o_port0),
    .rd_i     (zbt_rd_i0));

zbt_port #(addr_width, data_width, ctl_width, bank_group) zbt_port1(
    .clk      (clk),
    .rst      (rst),
    .sr       (1'b0),
    .pipeline (pipeline_reg[0]),
    .read     (zbt_readv[1]),
    .write    (zbt_writev[1]),
    .a        (ssram_addr[bank_order-1:0]),
    .d        (ld_i),
    .be       (lbe_i),
    .q        (zbt_q_port1),
    .evalid   (zbt_evalid_port[1]),
    .t_ra     (zbt_t_ra_port1),
    .ra_o     (zbt_ra_o_port1),
    .t_rc     (zbt_t_rc_port1),
    .rc_o     (zbt_rc_o_port1),
    .t_rd     (zbt_t_rd_port1),
    .rd_o     (zbt_rd_o_port1),
    .rd_i     (zbt_rd_i1));

zbt_port #(addr_width, data_width, ctl_width, bank_group) zbt_port2(
    .clk      (clk),
    .rst      (rst),
    .sr       (1'b0),
    .pipeline (pipeline_reg[0]),
    .read     (zbt_readv[2]),
    .write    (zbt_writev[2]),
    .a        (ssram_addr[bank_order-1:0]),
    .d        (ld_i),
    .be       (lbe_i),
    .q        (zbt_q_port2),
    .evalid   (zbt_evalid_port[2]),
    .t_ra     (zbt_t_ra_port2),
    .ra_o     (zbt_ra_o_port2),
    .t_rc     (zbt_t_rc_port2),
    .rc_o     (zbt_rc_o_port2),
    .t_rd     (zbt_t_rd_port2),
    .rd_o     (zbt_rd_o_port2),
    .rd_i     (zbt_rd_i2));

zbt_port #(addr_width, data_width, ctl_width, bank_group) zbt_port3(
    .clk      (clk),
    .rst      (rst),
    .sr       (1'b0),
    .pipeline (pipeline_reg[0]),
    .read     (zbt_readv[3]),
    .write    (zbt_writev[3]),
    .a        (ssram_addr[bank_order-1:0]),
    .d        (ld_i),
    .be       (lbe_i),
    .q        (zbt_q_port3),
    .evalid   (zbt_evalid_port[3]),
    .t_ra     (zbt_t_ra_port3),
    .ra_o     (zbt_ra_o_port3),
    .t_rc     (zbt_t_rc_port3),
    .rc_o     (zbt_rc_o_port3),
    .t_rd     (zbt_t_rd_port3),
    .rd_o     (zbt_rd_o_port3),
    .rd_i     (zbt_rd_i3));

zbt_port #(addr_width, data_width, ctl_width, bank_group) zbt_port4(
    .clk      (clk),
    .rst      (rst),
    .sr       (1'b0),
    .pipeline (pipeline_reg[0]),
    .read     (zbt_readv[4]),
    .write    (zbt_writev[4]),
    .a        (ssram_addr[bank_order-1:0]),
    .d        (ld_i),
    .be       (lbe_i),
    .q        (zbt_q_port4),
    .evalid   (zbt_evalid_port[4]),
    .t_ra     (zbt_t_ra_port4),
    .ra_o     (zbt_ra_o_port4),
    .t_rc     (zbt_t_rc_port4),
    .rc_o     (zbt_rc_o_port4),
    .t_rd     (zbt_t_rd_port4),
    .rd_o     (zbt_rd_o_port4),
    .rd_i     (zbt_rd_i4));

zbt_port #(addr_width, data_width, ctl_width, bank_group) zbt_port5(
    .clk      (clk),
    .rst      (rst),
    .sr       (1'b0),
    .pipeline (pipeline_reg[0]),
    .read     (zbt_readv[5]),
    .write    (zbt_writev[5]),
    .a        (ssram_addr[bank_order-1:0]),
    .d        (ld_i),
    .be       (lbe_i),
    .q        (zbt_q_port5),
    .evalid   (zbt_evalid_port[5]),
    .t_ra     (zbt_t_ra_port5),
    .ra_o     (zbt_ra_o_port5),
    .t_rc     (zbt_t_rc_port5),
    .rc_o     (zbt_rc_o_port5),
    .t_rd     (zbt_t_rd_port5),
    .rd_o     (zbt_rd_o_port5),
    .rd_i     (zbt_rd_i5));

zbt_port #(addr_width, data_width, ctl_width, bank_group) zbt_port6(
    .clk      (clk),
    .rst      (rst),
    .sr       (1'b0),
    .pipeline (pipeline_reg[0]),
    .read     (zbt_readv[6]),
    .write    (zbt_writev[6]),
    .a        (ssram_addr[bank_order-1:0]),
    .d        (ld_i),
    .be       (lbe_i),
    .q        (zbt_q_port6),
    .evalid   (zbt_evalid_port[6]),
    .t_ra     (zbt_t_ra_port6),
    .ra_o     (zbt_ra_o_port6),
    .t_rc     (zbt_t_rc_port6),
    .rc_o     (zbt_rc_o_port6),
    .t_rd     (zbt_t_rd_port6),
    .rd_o     (zbt_rd_o_port6),
    .rd_i     (zbt_rd_i6));

zbt_port #(addr_width, data_width, ctl_width, bank_group) zbt_port7(
    .clk      (clk),
    .rst      (rst),
    .sr       (1'b0),
    .pipeline (pipeline_reg[0]),
    .read     (zbt_readv[7]),
    .write    (zbt_writev[7]),
    .a        (ssram_addr[bank_order-1:0]),
    .d        (ld_i),
    .be       (lbe_i),
    .q        (zbt_q_port7),
    .evalid   (zbt_evalid_port[7]),
    .t_ra     (zbt_t_ra_port7),
    .ra_o     (zbt_ra_o_port7),
    .t_rc     (zbt_t_rc_port7),
    .rc_o     (zbt_rc_o_port7),
    .t_rd     (zbt_t_rd_port7),
    .rd_o     (zbt_rd_o_port7),
    .rd_i     (zbt_rd_i7));

//
// For banks not present, generate dummy signals. For banks
// that are present, use the 'zbt_*_port' signals.
//
// This is done in a fairly cumbersome manner because XST does
// not appear to fully support arrays of vectors as of 5.1i+SP3.
//
assign zbt_q0 = (num_bank >= 1) ? zbt_q_port0 : 0;
assign zbt_q1 = (num_bank >= 2) ? zbt_q_port1 : 0;
assign zbt_q2 = (num_bank >= 3) ? zbt_q_port2 : 0;
assign zbt_q3 = (num_bank >= 4) ? zbt_q_port3 : 0;
assign zbt_q4 = (num_bank >= 5) ? zbt_q_port4 : 0;
assign zbt_q5 = (num_bank >= 6) ? zbt_q_port5 : 0;
assign zbt_q6 = (num_bank >= 7) ? zbt_q_port6 : 0;
assign zbt_q7 = (num_bank >= 8) ? zbt_q_port7 : 0;
assign zbt_t_ra0 = (num_bank >= 1) ? zbt_t_ra_port0 : { (max_bank*log_addr_width){1'b1} };
assign zbt_t_ra1 = (num_bank >= 2) ? zbt_t_ra_port1 : { (max_bank*log_addr_width){1'b1} };
assign zbt_t_ra2 = (num_bank >= 3) ? zbt_t_ra_port2 : { (max_bank*log_addr_width){1'b1} };
assign zbt_t_ra3 = (num_bank >= 4) ? zbt_t_ra_port3 : { (max_bank*log_addr_width){1'b1} };
assign zbt_t_ra4 = (num_bank >= 5) ? zbt_t_ra_port4 : { (max_bank*log_addr_width){1'b1} };
assign zbt_t_ra5 = (num_bank >= 6) ? zbt_t_ra_port5 : { (max_bank*log_addr_width){1'b1} };
assign zbt_t_ra6 = (num_bank >= 7) ? zbt_t_ra_port6 : { (max_bank*log_addr_width){1'b1} };
assign zbt_t_ra7 = (num_bank >= 8) ? zbt_t_ra_port7 : { (max_bank*log_addr_width){1'b1} };
assign zbt_ra_o0 = (num_bank >= 1) ? zbt_ra_o_port0 : 0;
assign zbt_ra_o1 = (num_bank >= 2) ? zbt_ra_o_port1 : 0;
assign zbt_ra_o2 = (num_bank >= 3) ? zbt_ra_o_port2 : 0;
assign zbt_ra_o3 = (num_bank >= 4) ? zbt_ra_o_port3 : 0;
assign zbt_ra_o4 = (num_bank >= 5) ? zbt_ra_o_port4 : 0;
assign zbt_ra_o5 = (num_bank >= 6) ? zbt_ra_o_port5 : 0;
assign zbt_ra_o6 = (num_bank >= 7) ? zbt_ra_o_port6 : 0;
assign zbt_ra_o7 = (num_bank >= 8) ? zbt_ra_o_port7 : 0;
assign zbt_t_rc0 = (num_bank >= 1) ? zbt_t_rc_port0 : { (max_bank*log_ctl_width){1'b1} };
assign zbt_t_rc1 = (num_bank >= 2) ? zbt_t_rc_port1 : { (max_bank*log_ctl_width){1'b1} };
assign zbt_t_rc2 = (num_bank >= 3) ? zbt_t_rc_port2 : { (max_bank*log_ctl_width){1'b1} };
assign zbt_t_rc3 = (num_bank >= 4) ? zbt_t_rc_port3 : { (max_bank*log_ctl_width){1'b1} };
assign zbt_t_rc4 = (num_bank >= 5) ? zbt_t_rc_port4 : { (max_bank*log_ctl_width){1'b1} };
assign zbt_t_rc5 = (num_bank >= 6) ? zbt_t_rc_port5 : { (max_bank*log_ctl_width){1'b1} };
assign zbt_t_rc6 = (num_bank >= 7) ? zbt_t_rc_port6 : { (max_bank*log_ctl_width){1'b1} };
assign zbt_t_rc7 = (num_bank >= 8) ? zbt_t_rc_port7 : { (max_bank*log_ctl_width){1'b1} };
assign zbt_rc_o0 = (num_bank >= 1) ? zbt_rc_o_port0 : 0;
assign zbt_rc_o1 = (num_bank >= 2) ? zbt_rc_o_port1 : 0;
assign zbt_rc_o2 = (num_bank >= 3) ? zbt_rc_o_port2 : 0;
assign zbt_rc_o3 = (num_bank >= 4) ? zbt_rc_o_port3 : 0;
assign zbt_rc_o4 = (num_bank >= 5) ? zbt_rc_o_port4 : 0;
assign zbt_rc_o5 = (num_bank >= 6) ? zbt_rc_o_port5 : 0;
assign zbt_rc_o6 = (num_bank >= 7) ? zbt_rc_o_port6 : 0;
assign zbt_rc_o7 = (num_bank >= 8) ? zbt_rc_o_port7 : 0;
assign zbt_t_rd0 = (num_bank >= 1) ? zbt_t_rd_port0 : { (max_bank*log_data_width){1'b1} };
assign zbt_t_rd1 = (num_bank >= 2) ? zbt_t_rd_port1 : { (max_bank*log_data_width){1'b1} };
assign zbt_t_rd2 = (num_bank >= 3) ? zbt_t_rd_port2 : { (max_bank*log_data_width){1'b1} };
assign zbt_t_rd3 = (num_bank >= 4) ? zbt_t_rd_port3 : { (max_bank*log_data_width){1'b1} };
assign zbt_t_rd4 = (num_bank >= 5) ? zbt_t_rd_port4 : { (max_bank*log_data_width){1'b1} };
assign zbt_t_rd5 = (num_bank >= 6) ? zbt_t_rd_port5 : { (max_bank*log_data_width){1'b1} };
assign zbt_t_rd6 = (num_bank >= 7) ? zbt_t_rd_port6 : { (max_bank*log_data_width){1'b1} };
assign zbt_t_rd7 = (num_bank >= 8) ? zbt_t_rd_port7 : { (max_bank*log_data_width){1'b1} };
assign zbt_rd_o0 = (num_bank >= 1) ? zbt_rd_o_port0 : 0;
assign zbt_rd_o1 = (num_bank >= 2) ? zbt_rd_o_port1 : 0;
assign zbt_rd_o2 = (num_bank >= 3) ? zbt_rd_o_port2 : 0;
assign zbt_rd_o3 = (num_bank >= 4) ? zbt_rd_o_port3 : 0;
assign zbt_rd_o4 = (num_bank >= 5) ? zbt_rd_o_port4 : 0;
assign zbt_rd_o5 = (num_bank >= 6) ? zbt_rd_o_port5 : 0;
assign zbt_rd_o6 = (num_bank >= 7) ? zbt_rd_o_port6 : 0;
assign zbt_rd_o7 = (num_bank >= 8) ? zbt_rd_o_port7 : 0;
assign zbt_evalid[0] = (num_bank >= 1) ? zbt_evalid_port[0] : read_enable;
assign zbt_evalid[1] = (num_bank >= 2) ? zbt_evalid_port[1] : read_enable;
assign zbt_evalid[2] = (num_bank >= 3) ? zbt_evalid_port[2] : read_enable;
assign zbt_evalid[3] = (num_bank >= 4) ? zbt_evalid_port[3] : read_enable;
assign zbt_evalid[4] = (num_bank >= 5) ? zbt_evalid_port[4] : read_enable;
assign zbt_evalid[5] = (num_bank >= 6) ? zbt_evalid_port[5] : read_enable;
assign zbt_evalid[6] = (num_bank >= 7) ? zbt_evalid_port[6] : read_enable;
assign zbt_evalid[7] = (num_bank >= 8) ? zbt_evalid_port[7] : read_enable;

//
// Instantiate the 'zbt_dpins' modules. These are used to give a
// drive strength of 16mA as recommended in a reference ZBT interface
// design from Xilinx.
//

zbt_dpins #(data_width*bank_group) iobufs0 (
    .i  (zbt_rd_i0),
    .o  (zbt_rd_o0),
    .t  (zbt_t_rd0),
    .io (rd0));

zbt_dpins #(data_width*bank_group) iobufs1 (
    .i  (zbt_rd_i1),
    .o  (zbt_rd_o1),
    .t  (zbt_t_rd1),
    .io (rd1));

zbt_dpins #(data_width*bank_group) iobufs2 (
    .i  (zbt_rd_i2),
    .o  (zbt_rd_o2),
    .t  (zbt_t_rd2),
    .io (rd2));

zbt_dpins #(data_width*bank_group) iobufs3 (
    .i  (zbt_rd_i3),
    .o  (zbt_rd_o3),
    .t  (zbt_t_rd3),
    .io (rd3));

zbt_dpins #(data_width*bank_group) iobufs4 (
    .i  (zbt_rd_i4),
    .o  (zbt_rd_o4),
    .t  (zbt_t_rd4),
    .io (rd4));

zbt_dpins #(data_width*bank_group) iobufs5 (
    .i  (zbt_rd_i5),
    .o  (zbt_rd_o5),
    .t  (zbt_t_rd5),
    .io (rd5));

zbt_dpins #(data_width*bank_group) iobufs6 (
    .i  (zbt_rd_i6),
    .o  (zbt_rd_o6),
    .t  (zbt_t_rd6),
    .io (rd6));

zbt_dpins #(data_width*bank_group) iobufs7 (
    .i  (zbt_rd_i7),
    .o  (zbt_rd_o7),
    .t  (zbt_t_rd7),
    .io (rd7));

//
// Output the address and control busses to the SSRAMs
//
always @ (
    zbt_t_ra0 or zbt_t_ra1 or zbt_t_ra2 or zbt_t_ra3 or
    zbt_t_ra4 or zbt_t_ra5 or zbt_t_ra6 or zbt_t_ra7 or
    zbt_ra_o0 or zbt_ra_o1 or zbt_ra_o2 or zbt_ra_o3 or
    zbt_ra_o4 or zbt_ra_o5 or zbt_ra_o6 or zbt_ra_o7)
begin
    for (i = 0; i < log_addr_width; i = i + 1) begin
        ra0[i] <= !zbt_t_ra0[i] ? zbt_ra_o0[i] : 1'bZ;
        ra1[i] <= !zbt_t_ra1[i] ? zbt_ra_o1[i] : 1'bZ;
        ra2[i] <= !zbt_t_ra2[i] ? zbt_ra_o2[i] : 1'bZ;
        ra3[i] <= !zbt_t_ra3[i] ? zbt_ra_o3[i] : 1'bZ;
        ra4[i] <= !zbt_t_ra4[i] ? zbt_ra_o4[i] : 1'bZ;
        ra5[i] <= !zbt_t_ra5[i] ? zbt_ra_o5[i] : 1'bZ;
        ra6[i] <= !zbt_t_ra6[i] ? zbt_ra_o6[i] : 1'bZ;
        ra7[i] <= !zbt_t_ra7[i] ? zbt_ra_o7[i] : 1'bZ;
    end
end

always @ (
    zbt_t_rc0 or zbt_t_rc1 or zbt_t_rc2 or zbt_t_rc3 or
    zbt_t_rc4 or zbt_t_rc5 or zbt_t_rc6 or zbt_t_rc7 or
    zbt_rc_o0 or zbt_rc_o1 or zbt_rc_o2 or zbt_rc_o3 or
    zbt_rc_o4 or zbt_rc_o5 or zbt_rc_o6 or zbt_rc_o7)
begin
    for (i = 0; i < log_ctl_width; i = i + 1) begin
        rc0[i] <= !zbt_t_rc0[i] ? zbt_rc_o0[i] : 1'bZ;
        rc1[i] <= !zbt_t_rc1[i] ? zbt_rc_o1[i] : 1'bZ;
        rc2[i] <= !zbt_t_rc2[i] ? zbt_rc_o2[i] : 1'bZ;
        rc3[i] <= !zbt_t_rc3[i] ? zbt_rc_o3[i] : 1'bZ;
        rc4[i] <= !zbt_t_rc4[i] ? zbt_rc_o4[i] : 1'bZ;
        rc5[i] <= !zbt_t_rc5[i] ? zbt_rc_o5[i] : 1'bZ;
        rc6[i] <= !zbt_t_rc6[i] ? zbt_rc_o6[i] : 1'bZ;
        rc7[i] <= !zbt_t_rc7[i] ? zbt_rc_o7[i] : 1'bZ;
    end
end

//
// Output global operating modes to SSRAMs
//
assign glob_lbo_l     = 1'b0;
assign glob_flothru_l = pipeline_reg[0];

//
// Instantiate the module that generates the SSRAM clocks
//		
clocks #(num_clock) clocks0(
    .lreset_l (lreset_l),
    .rst      (rst),
    .lclk     (lclk),
    .mclk     (mclk),
    .clk      (clk),
    .ramclko  (ramclko),
    .ramclki  (ramclki),
    .locked   (locked));
            
endmodule
