//
// zbt_port.v - ZBT interface and IO pins
//
// (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2003
//
// Example program for ADM-XRC: / ADM-XRC-P / ADM-XRCII-Lite / ADM-XRCII / ADM-XRCIIPro-Lite
//
// Module containing pins connected to a bank of ZBT SSRAM.
//
// The 'pipeline' input selects flowthrough ZBT SSRAM mode ('0') or
// pipelined ZBT SSRAM mode ('1').
// 
// The latency from 'write' and 'd' to 'rd_o' and 't_rd' valid is 3 clocks,
// in both pipelined and flowthrough mode.
//
// The latency from the 'read', 'write' and 'a' signals to 'ra' and 'rc'
// valid is 1 clock in pipelined mode and 2 clocks in flowthrough mode.
//
// The latency from 'read' to 'q' valid is 4 clocks in both pipelined and
// flowthrough mode.
//
// 'evalid' is an early valid signal, generated one cycle before 'q'
// becomes valid. The latency from 'read' to 'evalid' asserted is therefore
// 3 clocks in both pipelined and flowthrough mode.
//
// Note:
//
// 1. ADM-XRCII-L (ctl_width = 6) uses two physical banks of 16 bits,
//    combined into one 32-bit bank.
//
// 2. ADM-XRCIIP-L  (ctl_width = 14) has one 64 bit wide bank, of which
//    the lower 32 bits are used in this design.
//

`timescale 1ns / 100ps

module zbt_port(
    clk,        // Clock
    rst,        // Asychronous reset
    sr,         // Synchronous reset
    pipeline,   // Operating mode, see above
    read,       // Read command
    write,      // Write command
    a,          // Address
    d,          // Data
    be,         // Byte enables
    q,          // Data read from SSRAM
    evalid,     // Qualifies 'q' one cycle early
    t_ra,       // Tristate control for address pins
    ra_o,       // To SSRAM address pins
    t_rc,       // Tristate control for SSRAM control pins
    rc_o,       // To SSRAM control pins
    t_rd,       // Tristate control for data pins
    rd_o,       // To SSRAM data pins
    rd_i);      // From SSRAM data pins

parameter addr_width = 20;  // Number of address bits on SSRAM bank
parameter data_width = 32;  // Number of data bits on SSRAM bank
parameter ctl_width  = 9;   // Number of control bits on SSRAM bank
parameter bank_group = 1;   // Number of physical SSRAM banks to combine to make a logical bank

parameter be_width   = data_width / 8;

// One-hot states for state machine
parameter s1_idle    = 3'b001,
          s1_read0   = 3'b010,
          s1_read1   = 3'b100;

input                               clk;
input                               rst;
input                               sr;
input                               pipeline;
input                               read;
input                               write;
input  [addr_width-1:0]             a;
input  [31:0]                       d;
input  [3:0]                        be;
output [31:0]                       q;
output                              evalid;
output [addr_width*bank_group-1:0]  t_ra;
output [addr_width*bank_group-1:0]  ra_o;
output [ctl_width*bank_group-1:0]   t_rc;
output [ctl_width*bank_group-1:0]   rc_o;
output [data_width*bank_group-1:0 ] t_rd;
output [data_width*bank_group-1:0 ] rd_o;
input  [data_width*bank_group-1:0 ] rd_i;

wire                              clk;
wire                              rst;
wire                              sr;
wire                              pipeline;
wire                              read;
wire                              write;
wire [addr_width-1:0]             a;
wire [31:0]                       d;
wire [3:0]                        be;
reg  [31:0]                       q;
reg                               evalid;
wire [addr_width*bank_group-1:0]  t_ra;
reg  [addr_width*bank_group-1:0]  ra_o;
wire [ctl_width*bank_group-1:0]   t_rc;
reg  [ctl_width*bank_group-1:0]   rc_o;
reg  [data_width*bank_group-1:0 ] t_rd;
wire [data_width*bank_group-1:0 ] rd_o;
wire [data_width*bank_group-1:0 ] rd_i;

reg  [255:0]                      addr_out;
reg  [255:0]                      ra_o_d;
reg  [255:0]                      ctl_out;
reg  [255:0]                      rc_o_d;

reg  [2:0]                        state1, n_state1;
reg                               n_evalid;

reg  [addr_width-1:0]             a_q;
reg  [1:0]                        read_q;
reg                               write_q;
reg  [3:0]                        be_q;
reg  [data_width*bank_group-1:0]  d_q[2:0];
reg  [1:0]                        oe_rd;
reg  [2:0]                        ce_q;

integer i;

//
// XST synthesis constraints, required to prevent optimization of
// flip-flops that would otherwise prevent IOB packing.
//
// synthesis attribute equivalent_register_removal of ra_o is no;
// synthesis attribute equivalent_register_removal of rc_o is no;
// synthesis attribute equivalent_register_removal of rd_o is no;
// synthesis attribute equivalent_register_removal of rd_i is no;
// synthesis attribute iob of t_rd is true;
//

//
// The SSRAM address and control busses are always enabled onto the FPGA pins
//
assign t_ra = 0;
assign t_rc = 0;
assign rd_o = d_q[2];

//
// Drive 'rd' (SSRAM data bus) on writes
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        a_q     <= 0;
        be_q    <= 0;
        d_q[2]  <= 0;
        d_q[1]  <= 0;
        d_q[0]  <= 0;
        oe_rd   <= 0;
        t_rd    <= { (data_width*bank_group){1'b1} };
        read_q  <= 0;
        write_q <= 1'b0;
    end else begin
        if (sr) begin
            a_q     <= 0;
            be_q    <= 0;
            d_q[2]  <= 0;
            d_q[1]  <= 0;
            d_q[0]  <= 0;
            oe_rd   <= 0;
            t_rd    <= { (data_width*bank_group){1'b1} };
            read_q  <= 0;
            write_q <= 1'b0;
        end else begin
            a_q     <= a;
            be_q    <= be;
            d_q[2]  <= d_q[1];
            d_q[1]  <= d_q[0];
            d_q[0]  <= d;
            oe_rd   <= { oe_rd[0], write };
            t_rd    <= ~{ (data_width*bank_group){oe_rd[1]} };
            read_q  <= { read_q[0], read };
            write_q <= write;
        end
    end
end

//
// Generate 'q' (read SSRAM data bus on reads)
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        ce_q <= 0;
        q    <= 0;
    end else begin
        if (sr) begin
            ce_q <= 0;
            q    <= 0;
        end else begin
            ce_q <= { ce_q[1:0], read };
            if (ce_q[2]) begin
                q <= rd_i;
            end
        end
    end
end

//
// Drive 'ra' (SSRAM address bus) and 'rc' (SSRAM control bus)
//
// 'rc' pin designations:
//
//    bit   ADM-XRC/-P   ADM-XRCII-L   ADM-XRCII   ADM-XRCIIP-L
//
//      0   BWE0#        BWE0#         BWE0#       BWE0#
//      1   BWE1#        BWE1#         BWE1#       BWE1#
//      2   BWE2#        WE#           BWE2#       BWE2#
//      3   BWE3#        CE#           BWE3#       BWE3#
//      4   WE#          ADV#          WE#         BWE4#
//      5   CE#          OE#           CE#         BWE5#
//      6   ADV#         BWE2#         ADV#        BWE6#
//      7   OE#          BWE3#         OE#         BWE7#
//      8   CKE#         WE#           CKE#        WE#
//      9                CE#                       CE#
//     10                ADV#                      CE2#
//     11                OE#                       CE2
//     12                                          ADV#
//     13                                          OE#
//     14                                          CKE#
//
always @ (
    pipeline or
    a or a_q or
    be or be_q or
    read or read_q or
    write or write_q)
begin
    ra_o_d = 0;
    rc_o_d = 0;

    for (i = 0; i < bank_group; i = i + 1) begin
        ctl_out = 0;
        addr_out = 0;

        if (pipeline) begin
            addr_out[addr_width-1:0] = a;

            case (ctl_width)
                6: begin
                    // ADM-XRCII-Lite
                    ctl_out[5]    = ~read_q[1];
                    ctl_out[4]    = 1'b0;
                    ctl_out[3]    = ~(read | write);
                    ctl_out[2]    = ~write;
                    ctl_out[1:0]  = read ? 2'h0 : ~(be >> (2 * i));
                end

                9: begin
                    // ADM-XRC or ADM-XRC-P or ADM-XRCII
                    ctl_out[8]    = 1'b0;
                    ctl_out[7]    = ~read_q[1];
                    ctl_out[6]    = 1'b0;
                    ctl_out[5]    = ~(read | write);
                    ctl_out[4]    = ~write;
                    ctl_out[3:0]  = read ? 4'h0 : ~be;
                end

                15: begin
                    // ADM-XRCIIPro-Lite
                    ctl_out[14]    = 1'b0;
                    ctl_out[13]    = ~read_q[1];
                    ctl_out[12:10] = 3'b010;
                    ctl_out[9]     = ~(read | write);
                    ctl_out[8]     = ~write;
                    ctl_out[7:0]   = read_q ? 8'hf0 : ~{ 4'h0, be };
                end
            endcase
        end else begin
            addr_out[addr_width-1:0] = a_q;

            case (ctl_width)
                6: begin
                    // ADM-XRCII-Lite
                    ctl_out[5]    = ~read_q[1];
                    ctl_out[4]    = 1'b0;
                    ctl_out[3]    = ~(read_q[0] | write_q);
                    ctl_out[2]    = ~write_q;
                    ctl_out[1:0]  = read_q[0] ? 2'h0 : ~(be_q >> (2 * i));
                end

                9: begin
                    // ADM-XRC or ADM-XRC-P or ADM-XRCII
                    ctl_out[8]    = 1'b0;
                    ctl_out[7]    = ~read_q[1];
                    ctl_out[6]    = 1'b0;
                    ctl_out[5]    = ~(read_q[0] | write_q);
                    ctl_out[4]    = ~write_q;
                    ctl_out[3:0]  = read_q[0] ? 4'h0 : ~be_q;
                end

                15: begin
                    // ADM-XRCIIPro-Lite
                    ctl_out[14]    = 1'b0;
                    ctl_out[13]    = ~read_q[1];
                    ctl_out[12:10] = 3'b010;
                    ctl_out[9]     = ~(read_q[0] | write_q);
                    ctl_out[8]     = ~write_q;
                    ctl_out[7:0]   = read_q[0] ? 8'hf0 : ~{ 4'h0, be_q };
                end
            endcase
        end

        //
        // In XST, this seems to be the only way to get the equivalent of:
        //
        //     ra_o_d[addr_width*(i+1)-1:addr_width*i] = addr_out[addr_width-1:0];
        //     rc_o_d[ctl_width*(i+1)-1:ctl_width*i]   = ctl_out[ctl_width-1:0];
        //
        // Bit-select operations with a variable as an index do not appear to be
        // supported as of 5.1i+SP3.
        //
        ra_o_d = ra_o_d | (addr_out << (i * addr_width));
        rc_o_d = rc_o_d | (ctl_out << (i * ctl_width));
    end
end

always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        ra_o <= 0;
        rc_o <= { (ctl_width*bank_group){1'b1} };
    end else begin
        if (sr) begin
            ra_o <= 0;
            rc_o <= { (ctl_width*bank_group){1'b1} };
        end else begin
            ra_o <= ra_o_d;
            rc_o <= rc_o_d;
        end
    end
end

//
// Generate current state and 'evalid' signal
//
always @ (posedge rst or posedge clk)
begin
    if (rst) begin
        state1 <= s1_idle;
        evalid <= 1'b0;
    end else begin
        if (sr) begin
            state1 <= s1_idle;
            evalid <= 1'b0;
        end else begin
            state1 <= n_state1;
            evalid <= n_evalid;
        end
    end
end

//
// Generate next state
//
always @ (
    state1,
    read)
begin
    case (state1)
        s1_idle:
            if (read) begin
                n_state1 <= s1_read0;
                n_evalid <= 1'b0;
            end else begin
                n_state1 <= s1_idle;
                n_evalid <= 1'b0;
            end

        s1_read0:
            if (read) begin
                n_state1 <= s1_read1;
                n_evalid <= 1'b0;
            end else begin
                n_state1 <= s1_idle;
                n_evalid <= 1'b0;
            end

        s1_read1:
            if (read) begin
                n_state1 <= s1_read1;
                n_evalid <= 1'b1;
            end else begin
                n_state1 <= s1_idle;
                n_evalid <= 1'b0;
            end
    endcase
end

endmodule
