--
-- clock-xpl.vhd - Counts elapsed clock cycles on LCLK and MCLK inputs
--
-- (C) Copyright Alpha Data 2008
--
-- Example program for:
--
--   o ADM-XPL
--   o ADM-XP
--
-- This design implements the following registers in the FPGA space:
--
-- Local bus address  Type    Function
-- ---------------------------------------------------------------------------
--
--              0x00  RAX/WO  READ COUNT register; writing a 1 to bit 'i'
--                            issues a READ COUNT command to the corresponding
--                            cycle counter. This results in the updated value
--                            appearing at local bus address 0x80 + (i * 4).
--
--              0x04  RO/MBZ  READY register; a 1 in bit 'i' indicates that
--                            the corresponding cycle counter is ready to
--                            accept a new READ COUNT command.
--
--       0x0c - 0x7f          (reserved)
--
--       0x80 - 0xff  RO/MBZ  The cycle counter values appear in this region.
--                              - the LCLK cycle counter is at 0x80.
--                              - the MCLK cycle counter is at 0x84.
--                              - the REFCLK cycle counter is at 0x88.
--
--      0x100 onward          (reserved)
--
-- The technique used in this design accomodates up to 32 cycle counters, but
-- since there are only three clock inputs (LCLK, MCLK and REFCLK), only bits
-- 0, 1 and 2 of the READ COUNT and READY registers are useful.
-- Correspondingly, only locations 0x80, 0x84 and 0x88 are valid when reading
-- the cycle counter values.
--
-- Note that partial addess decoding is used extensively in this design, so
-- that writing to reserved locations (which is illegal) may result in the
-- READ COUNT register being written.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

library work;
use work.localbus.all;

entity clock is
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic; 
        lads_l           : in    std_logic;
        lbe_l            : in    std_logic_vector(3 downto 0);
        lblast_l         : in    std_logic;
        lbterm_l         : inout std_logic;
        lad              : inout std_logic_vector(31 downto 0);
        lready_l         : out   std_logic; 
        lwrite           : in    std_logic;
        fholda           : in    std_logic;
        mclk             : in    std_logic;
        refclk_n         : in    std_logic;
        refclk_p         : in    std_logic);
end entity;

architecture mixed of clock is

    component BUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;

    component IBUFGDS
        port(
            I  : in  std_logic;
            IB : in  std_logic;
            O  : out std_logic);
    end component;

    component cycle_counter
        generic(
            width            : in    natural);
        port(
            rst              : in    std_logic;                      -- Asynchronous reset
            clk              : in    std_logic;                      -- Clock whose cycles are to be counted
            lclk             : in    std_logic;                      -- LCLK
            -- The following signals are all sychronous to 'lclk'
            sr               : in    std_logic;                      -- Synchronous reset command
            read             : in    std_logic;                      -- Read count command
            done             : out   std_logic;                      -- Indicates command completed
            -- 'q' is valid after 'done' indicates a read count command has completed
            q                : out   std_logic_vector(width - 1 downto 0)); -- Cycle count
    end component;

    constant num_clock : natural := 3;

    signal refclk_ibufg : std_logic;
    signal refclk : std_logic;

    signal rst : std_logic;
    signal lblast_i : std_logic;
    signal lads_i : std_logic;
    signal lwrite_i : std_logic;
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal lbterm_i : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal la_i : std_logic_vector(23 downto 2);
    signal ld_o : std_logic_vector(31 downto 0);
    signal ld_iq : std_logic_vector(31 downto 0);
    signal eld_oe : std_logic;
    signal ld_oe_l : std_logic_vector(31 downto 0);
    signal lbe_iq_l : std_logic_vector(3 downto 0);

    signal qlads : std_logic;
    signal ds_xfer : std_logic;
    signal ds_decode : std_logic;    
    signal ds_write : std_logic;    

    signal la_q : std_logic_vector(23 downto 2);

    signal write_count_read : std_logic;
    signal count_read : std_logic_vector(31 downto 0);
    signal count_done : std_logic_vector(31 downto 0);
    type q_type is array(0 to num_clock - 1) of std_logic_vector(31 downto 0);
    signal count_q : q_type;
    signal ready_reg : std_logic_vector(31 downto 0);
    
    signal logic0, logic1 : std_logic;
    
    attribute IOB : string;
    attribute IOB of ld_oe_l : signal is "true";

begin

    logic0 <= '0';
    logic1 <= '1';

    --
    -- Convert the inputs to active high.
    --

    rst <= not lreset_l;
    lblast_i <= not lblast_l;
    lbterm_i <= not lbterm_l;
    lads_i <= not lads_l;
    lwrite_i <= lwrite;
    la_i <= lad(23 downto 2);
    
    --
    -- Register 'ld' and 'lbe_l' (should be absorbed into IOBs)
    --

    register_ld_lbe : process(lclk)
    begin
        if lclk'event and lclk = '1' then
            ld_iq <= lad;
            lbe_iq_l <= lbe_l;
        end if;
    end process;
    
    --
    -- Generate a qualified version of 'lads_l', which is
    -- asserted when the FPGA is addressed AND the FPGA is
    -- not the local bus master.
    --

    qlads <= lads_i and not la_i(23) and not fholda;

    --
    -- Latch the local bus address on the 'lads_l' pulse.
    --

    latch_addr : process(rst, lclk)
    begin
        if rst = '1' then
            la_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
            end if;
        end if;
    end process;
    
    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';
    
    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --

    lready_l <= lready_o_l when lready_oe_l = '0' else 'Z';
    
    --
    -- Drive the local data bus on a read.
    --

    gen_ld_oe_l : process(rst, lclk)
    begin
        if rst = '1' then
            ld_oe_l <= (others => '1');
        elsif lclk'event and lclk = '1' then
            ld_oe_l <= (others => not eld_oe);
        end if;
    end process;

    drive_lad : for i in lad'range generate
        lad(i) <= ld_o(i) when ld_oe_l(i) = '0' else 'Z';
    end generate;
    
    --
    -- If the current cycle is a write, generate the commands to the cycle
    -- counters.
    --

    update_regs : process(lclk, rst)
    begin
        if rst = '1' then
            ready_reg <= (others => '1');
            write_count_read <= '0';
            count_read <= (others => '0');
        elsif lclk'event and lclk = '1' then
            -- Implement READY register
            write_count_read <= ds_xfer and ds_write and not la_q(7);
            if write_count_read = '1' then
                -- Issuing a read command clears the corresponding bit in the READY register.
                for i in 0 to 3 loop
                    if lbe_iq_l(i) = '0' then
                       ready_reg(8 * (i + 1) - 1 downto 8 * i) <= ready_reg(8 * (i + 1) - 1 downto 8 * i) and
                                                                  not ld_iq(8 * (i + 1) - 1 downto 8 * i);
                       count_read(8 * (i + 1) - 1 downto 8 * i) <= ld_iq(8 * (i + 1) - 1 downto 8 * i);
                    end if;
                end loop;
            else
                -- The acknowledge from a cycle counter sets the corresponding bit in the
                -- READY register.
                ready_reg <= ready_reg or count_done;
                count_read <= (others => '0');
            end if;
        end if;
    end process;

    --
    -- Generate the 'ld_o' signal, outbound data for local bus. This is done according to
    -- which region of the FPGA space is being accessed:
    --
    --   o If 'la_q[7]' is 0, the command space is being accessed, and the only readable
    --     register in that region is the READY register.
    --
    --   o If 'la_q[7]' is 1, a cycle counter value is being read, 'la_q[6:2]' are used
    --     to index the 'count_q' array.
    --

    generate_ld_o : process(rst, lclk)
        variable data : std_logic_vector(31 downto 0);
    begin
        if rst = '1' then
            ld_o <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if la_q(7) = '1' then
                -- Select the element of 'count_q' indexed by 'la_q[6:2]'.
                data := (others => '0');
                for i in 0 to num_clock - 1 loop
                    if la_q(6 downto 2) = i then
                        data := data or count_q(i);
                    end if;
                end loop;
                ld_o <= data;
            else
                -- The READY register is the only register in the command region
                -- that can be read.
                ld_o <= ready_reg;
            end if;
        end if;
    end process;

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    dssm : plxdssm
        port map(
            clk => lclk,
            rst => rst,
            sr => logic0,
            qlads => qlads,
            lblast => lblast_i,
            lbterm => lbterm_i,
            lwrite => lwrite_i,
            eld_oe => eld_oe,
            ld_oe_l => open,
            lready_o_l  => lready_o_l,
            lready_oe_l => lready_oe_l,
            lbterm_o_l => lbterm_o_l,
            lbterm_oe_l => lbterm_oe_l,
            idle => open,
            transfer => ds_xfer,
            decode => ds_decode,
            write => ds_write,
            ready => logic1,
            stop => logic1);

    --
    -- Instantiate the cycle counter for LCLK
    --

    cycle_counter_lclk : cycle_counter
        generic map(
            width => 32)
        port map(
            rst => rst,
            lclk => lclk,
            clk => lclk,
            sr => logic0,
            read => count_read(0),
            done => count_done(0),
            q => count_q(0));

    --
    -- Instantiate the cycle counter for MCLK
    --

    cycle_counter_mclk : cycle_counter
        generic map(
            width => 32)
        port map(
            rst => rst,
            lclk => lclk,
            clk => mclk,
            sr => logic0,
            read => count_read(1),
            done => count_done(1),
            q => count_q(1));

    --
    -- Instantiate the cycle counter for REFCLK
    --

    ibufg_refclk : IBUFGDS
        port map(
            I  => refclk_p,
            IB => refclk_n,
            O  => refclk_ibufg);
    
    bufg_refclk : BUFG
        port map(
            I  => refclk_ibufg,
            O  => refclk);

    cycle_counter_refclk : cycle_counter
        generic map(
            width => 32)
        port map(
            rst => rst,
            lclk => lclk,
            clk => refclk,
            sr => logic0,
            read => count_read(2),
            done => count_done(2),
            q => count_q(2));

end mixed;
