--
-- cycle_counter.vhd - Cycle counter for arbitary clock, with logic to retime
--                     into LCLK domain.
--
-- (C) Copyright Alpha Data 2008
--
-- Usage:
--
--  1. This cycle counter accepts two commands, "read count" and "reset". These
--     must be single cycle pulses synchronous to 'lclk'.
--
--  2. The module outputs a pulse on 'done', also sychronous to 'lclk', in order
--     to indicate that a command has been completed.
--
--  3. It is illegal to issue another command until the 'done' acknowledgement
--     has been received for the current command.
--
-- This module is coded primarily for speed, as the target frequency may be
-- up to 500MHz on some architectures.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.localbus.all;

entity cycle_counter is
    generic(
        width            : in    natural);                       -- Must be in the range 1..32
    port(
        rst              : in    std_logic;                      -- Asynchronous reset
        clk              : in    std_logic;                      -- Clock whose cycles are to be counted
        lclk             : in    std_logic;                      -- LCLK
        -- The following signals are all sychronous to 'lclk'
        sr               : in    std_logic;                      -- Synchronous reset command
        read             : in    std_logic;                      -- Read count command
        done             : out   std_logic;                      -- Indicates command completed
        -- 'q' is valid after 'done' indicates a read count command has completed
        q                : out   std_logic_vector(width - 1 downto 0)); -- Cycle count
end entity;

architecture rtl of cycle_counter is

    component pulse_sync
        port(
            ar : in  std_logic;
            ik : in  std_logic;
            i  : in  std_logic;
            ok : in  std_logic;
            o  : out std_logic);
    end component;

    signal sr_int : std_logic;
    signal sr_ack : std_logic_vector(6 downto 0);
    signal sr_q_int0 : std_logic_vector(8 downto 0);
    signal sr_q_int1 : std_logic_vector(16 downto 0);
    signal sr_q_int2 : std_logic_vector(24 downto 0);
    signal sr_q_int3 : std_logic_vector(31 downto 0);
    signal q_int0 : std_logic_vector(8 downto 0) := (others => '0');
    signal q_int1 : std_logic_vector(16 downto 0) := (others => '0');
    signal q_int2 : std_logic_vector(24 downto 0) := (others => '0');
    signal q_int3 : std_logic_vector(31 downto 0) := (others => '0');
    signal q_int4 : std_logic_vector(31 downto 0);
    signal read_int, read_int_q : std_logic;
    signal read_intv : std_logic_vector(31 downto 0);
    signal done_int : std_logic;
    signal i_q : std_logic_vector(31 downto 0);

    attribute equivalent_register_removal : string;
    attribute equivalent_register_removal of sr_q_int0 : signal is "no";
    attribute equivalent_register_removal of sr_q_int1 : signal is "no";
    attribute equivalent_register_removal of sr_q_int2 : signal is "no";
    attribute equivalent_register_removal of sr_q_int3 : signal is "no";
    attribute equivalent_register_removal of q_int0 : signal is "no";
    attribute equivalent_register_removal of q_int1 : signal is "no";
    attribute equivalent_register_removal of q_int2 : signal is "no";
    attribute equivalent_register_removal of q_int3 : signal is "no";
    attribute equivalent_register_removal of q_int4 : signal is "no";

begin

    --
    -- Retime "read count" command and "reset" command pulses into 'clk' domain.
    --

    sync_sr : pulse_sync
        port map(
            ar => rst,
            ik => lclk,
            i => sr,
            ok => clk,
            o => sr_int);

    sync_read : pulse_sync
        port map(
            ar => rst,
            ik => lclk,
            i => read,
            ok => clk,
            o => read_int);

    --
    -- The cycle counter, counting 'clk' cycles. It is reset to zero on receipt
    -- of a "reset" command, which generates a pulse on 'sr_int'.
    --
    -- To make this run at a high frequency, it is pipelined by partitioning into
    -- four counters of length 8 bits.
    --

    -- Generate one reset signal per bit per pipeline stage, in order to minimise
    -- the fanout for reset nets and improve timing.
    resets : process(rst, clk)
    begin
        if rst = '1' then
            sr_q_int0 <= (others => '0');
            sr_q_int1 <= (others => '0');
            sr_q_int2 <= (others => '0');
            sr_q_int3 <= (others => '0');
        elsif clk'event and clk = '1' then
            sr_q_int0 <= (others => sr_int);
            sr_q_int1 <= (others => sr_int);
            sr_q_int2 <= (others => sr_int);
            sr_q_int3 <= (others => sr_int);
        end if;
    end process;

    counter : process(clk)
        -- Counter next value
        variable n_q0 : std_logic_vector(8 downto 0);
        variable n_q1 : std_logic_vector(16 downto 0);
        variable n_q2 : std_logic_vector(24 downto 0);
        variable n_q3 : std_logic_vector(31 downto 0);
    begin
        if clk'event and clk = '1' then
            -- Pipeline stage 0
            n_q0(8 downto 0) := ("0" & q_int0(7 downto 0)) + 1;

            -- Pipeline stage 1
            n_q1(7 downto 0) := q_int0(7 downto 0);
            n_q1(16 downto 8) := ("0" & q_int1(15 downto 8)) + q_int0(8);

            -- Pipeline stage 2
            n_q2(7 downto 0) := q_int1(7 downto 0);
            n_q2(15 downto 8) := q_int1(15 downto 8);
            n_q2(24 downto 16) := ("0" & q_int2(23 downto 16)) + q_int1(16);

            -- Pipeline stage 3
            n_q3(7 downto 0) := q_int2(7 downto 0);
            n_q3(15 downto 8) := q_int2(15 downto 8);
            n_q3(23 downto 16) := q_int2(23 downto 16);
            n_q3(31 downto 24) := q_int3(31 downto 24) + q_int2(24);

            -- Implement reset for pipeline stage 0
            for i in 8 downto 0 loop
                if sr_q_int0(i) = '1' then
                    q_int0(i) <= '0';
                else
                    q_int0(i) <= n_q0(i);
                end if;
            end loop;

            -- Implement reset for pipeline stage 1
            for i in 16 downto 0 loop
                if sr_q_int1(i) = '1' then
                    q_int1(i) <= '0';
                else
                    q_int1(i) <= n_q1(i);
                end if;
            end loop;

            -- Implement reset for pipeline stage 2
            for i in 24 downto 0 loop
                if sr_q_int2(i) = '1' then
                    q_int2(i) <= '0';
                else
                    q_int2(i) <= n_q2(i);
                end if;
            end loop;

            -- Implement reset for pipeline stage 3
            for i in 31 downto 0 loop
                if sr_q_int3(i) = '1' then
                    q_int3(i) <= '0';
                else
                    q_int3(i) <= n_q3(i);
                end if;
            end loop;

            -- Register counter output in order to help meet timing.
            q_int4 <= q_int3;
        end if;
    end process;

    --
    -- Get a version of 'sr_int' that is delayed by a number of 'clk' cycles.
    -- This will be used for generating 'done'.
    --

    gen_sr_ack : process(rst, clk)
    begin
        if rst = '1' then
            sr_ack <= (others => '0');
        elsif clk'event and clk = '1' then
            sr_ack(0) <= sr_int;
            sr_ack(sr_ack'high downto 1) <= sr_ack(sr_ack'high - 1 downto 0);
        end if;
    end process;

    --
    -- Latch counter output on receipt of a "read count" command, which is
    -- indicated by a pulse on 'read_int'.
    --

    capture_q : process(clk)
    begin
        if clk'event and clk = '1' then
            read_intv <= (others => read_int);

            for i in 0 to 31 loop
                if read_intv(i) = '1' then
                    i_q(i) <= q_int4(i);
                end if;
            end loop;
        end if;
    end process;

    q <= EXT(i_q, width);

    --
    -- Get a version of 'read_int' that is delayed by one 'clk' cycle.
    -- This will be used for generating 'done'.
    --

    ack_read : process(rst, clk)
    begin
        if rst = '1' then
            read_int_q <= '0';
        elsif clk'event and clk = '1' then
            read_int_q <= read_int;
        end if;
    end process;

    --
    -- Generate 'done' (acknowledge) for a "read count" command or a "reset" command.
    --

    gen_done_int : process(rst, clk)
    begin
        if rst = '1' then
            done_int <= '0';
        elsif clk'event and clk = '1' then
            done_int <= sr_ack(sr_ack'high) or read_int_q;
        end if;
    end process;

    sync_done : pulse_sync
        port map(
            ar => rst,
            ik => clk,
            i => done_int,
            ok => lclk,
            o => done);

end rtl;
