--
-- test_clock-xrce4fx.vhd - Testbench for entity 'clock'
--
-- (C) Copyright Alpha Data 2008
--
-- Example testbench for:
--
--   o ADPE-XRC-4FX
--
-- NOT SYNTHESIZABLE
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library std;
use std.textio.all;

library plxsim;
use plxsim.plxsim.all;

entity test_clock is
end entity;

architecture behav of test_clock is

    component clock
        port(
            lclk           : in    std_logic;
            lreset_l       : in    std_logic;
            lwrite         : in    std_logic;
            lads_l         : in    std_logic;
            lblast_l       : in    std_logic;
            lbterm_l       : inout std_logic;
            lad            : inout std_logic_vector(31 downto 0);
            lready_l       : out   std_logic;
            lbe_l          : in    std_logic_vector(3 downto 0);
            fholda         : in    std_logic;
            mclk_p         : in    std_logic;
            mclk_n         : in    std_logic;
            refclk200_p    : in    std_logic;
            refclk200_n    : in    std_logic;
            front_rx_p     : in    std_logic_vector(7 downto 0);
            front_rx_n     : in    std_logic_vector(7 downto 0);
            front_tx_p     : out   std_logic_vector(7 downto 0);
            front_tx_n     : out   std_logic_vector(7 downto 0);
            local_rx_p     : in    std_logic_vector(7 downto 0);
            local_rx_n     : in    std_logic_vector(7 downto 0);
            local_tx_p     : out   std_logic_vector(7 downto 0);
            local_tx_n     : out   std_logic_vector(7 downto 0));
    end component;

    type real_vector_t is array(natural range <>) of real;
    type time_vector_t is array(natural range <>) of time;

    procedure report_message(
        str : in string)
    is
        variable msg : line;
    begin
        write(msg, str);
        writeline(output, msg);
    end;

    function clock_name(
        input : in natural)
    return string is
    begin
        case input is
            when 1 => return "LCLK";
            when 2 => return "MCLK";
            when 3 => return "REFCLK200";
            when others =>
                assert false report "Unknown clock input " & conv_string(input) severity failure;
        end case;

        return "Unknown"; -- just to stop compiler complaining about lack of return value
    end;

    -- Convert vector of frequencies (in MHz) to vector of times (in microseconds)
    function to_periods(
        f : real_vector_t)
    return time_vector_t is
        variable p : time_vector_t(f'range);
    begin
        for i in f'range loop
            p(i) := (1.0 / f(i)) * 1.0 us;
        end loop;
        return p;
    end;
    
    -- Address of READ register
    constant read_reg_addr : std_logic_vector(31 downto 0) := X"00000000";
    -- Address of STATUS register
    constant status_reg_addr : std_logic_vector(31 downto 0) := X"00000004";
    -- Address of first cycle counter register
    constant count_reg_addr : std_logic_vector(31 downto 0) := X"00000080";

    constant lclk_freq : real := 60.0;
    constant mclk_freq : real := 325.0;
    constant refclk200_freq : real := 200.0; -- 200MHz crystal oscillator for IDELAY callibration

    constant frequencies : real_vector_t(1 to 3) := (
        lclk_freq,
        mclk_freq,
        refclk200_freq);

    constant periods : time_vector_t(frequencies'range) := to_periods(frequencies);

    constant priorities : integer_vector_t(0 to 0) := (others => 0);

    signal lclk : std_logic := '0';
    signal lreset_l : std_logic := '0';
    signal lbe_l : std_logic_vector(3 downto 0) := (others => 'Z');
    signal lad : std_logic_vector(31 downto 0) := (others => 'Z');
    signal lwrite : std_logic := 'Z';
    signal lads_l : std_logic := 'Z';
    signal lblast_l : std_logic := '1';
    signal lbterm_l : std_logic := 'Z';
    signal lready_l : std_logic := 'Z';
    signal lhold, lholda : std_logic_vector(0 downto 0);
    signal fholda : std_logic := '0';
    signal mclk_p : std_logic := '0';
    signal mclk_n : std_logic := '1';
    signal refclk200_p : std_logic := '0';
    signal refclk200_n : std_logic := '1';

    signal front_rx_p : std_logic_vector(7 downto 0) := (others => 'H');
    signal front_rx_n : std_logic_vector(7 downto 0) := (others => 'L');
    signal front_tx_p : std_logic_vector(7 downto 0);
    signal front_tx_n : std_logic_vector(7 downto 0);
    signal local_rx_p : std_logic_vector(7 downto 0) := (others => 'H');
    signal local_rx_n : std_logic_vector(7 downto 0) := (others => 'L');
    signal local_tx_p : std_logic_vector(7 downto 0);
    signal local_tx_n : std_logic_vector(7 downto 0);

    signal bus_in : locbus_in_t;
    signal bus_out : locbus_out_t := init_locbus_out;

    signal logic0, logic1 : std_logic;

begin

    logic0 <= '0';
    logic1 <= '1';

    lclk <= not lclk after periods(1) / 2;
    mclk_n <= not mclk_n after periods(2) / 2;
    mclk_p <= not mclk_p after periods(2) / 2;
    refclk200_n <= not refclk200_n after periods(3) / 2;
    refclk200_p <= not refclk200_p after periods(3) / 2;

    lreset_l <= transport '1' after 50 ns;

    lads_l <= 'H';
    lblast_l <= 'H';
    lready_l <= 'H';
    lbterm_l <= 'H';
    lwrite <= 'H';

    --
    -- This process iterates through the available cycle counters, measuring the
    -- approximate frequency at which they ar clocked.
    --

    stimulate: process
        variable index : natural;
        variable addr : std_logic_vector(31 downto 0);
        variable val32 : std_logic_vector(31 downto 0);
        variable data : byte_vector_t(0 to 3);
        variable be : byte_enable_t(0 to 3);
        variable n : natural;
        variable initial_count, final_count : natural;
        variable initial_time, final_time : time;
        variable frequency_mhz : real;
        variable elapsed_ns, elapsed_us : real;
        variable delta : natural;
    begin
        plxsim_wait_cycles(4, bus_in);

        be := "1111";

        plxsim_request_bus(true, bus_in, bus_out);

        for i in periods'range loop
            index := i - 1;

            report_message("Measuring clock input " & conv_string(i) & " (" & clock_name(i) & ")");

            -- Issue command to read cycle counter to READ register
            val32 := (others => '0');
            val32(index) := '1';
            data := conv_byte_vector(val32);
            plxsim_write(2, true, read_reg_addr, be, data, n, bus_in, bus_out);

            -- Poll STATUS register until read of cycle counter has been completed
            loop
                plxsim_read(2, true, status_reg_addr, be, data, n, bus_in, bus_out);
                val32 := conv_std_logic_vector(data);
                exit when val32(index) = '1';
            end loop;

            -- Read cycle counter register to get initial counter value
            addr := count_reg_addr + 4 * index;
            plxsim_read(2, true, addr, be, data, n, bus_in, bus_out);
            val32 := conv_std_logic_vector(data);
            initial_count := conv_integer(val32);
            initial_time := now;
            report_message("  Initial cycle counter value: " & conv_string(initial_count) & " at time " & conv_string(initial_time));

            -- Wait for some amount of time
            report_message("  Delaying for 10 us");
            wait for 10.0 us;
            plxsim_wait_cycles(1, bus_in);

            -- Issue command to read cycle counter to READ register
            val32 := (others => '0');
            val32(index) := '1';
            data := conv_byte_vector(val32);
            plxsim_write(2, true, read_reg_addr, be, data, n, bus_in, bus_out);

            -- Poll STATUS register until read of cycle counter has been completed
            loop
                plxsim_read(2, true, status_reg_addr, be, data, n, bus_in, bus_out);
                val32 := conv_std_logic_vector(data);
                exit when val32(index) = '1';
            end loop;

            -- Read cycle counter register to get final counter value
            addr := count_reg_addr + 4 * index;
            plxsim_read(2, true, addr, be, data, n, bus_in, bus_out);
            val32 := conv_std_logic_vector(data);
            final_count := conv_integer(val32);
            final_time := now;
            report_message("  Final cycle counter value: " & conv_string(final_count) & " at time " & conv_string(final_time));

            delta := final_count - initial_count;
            elapsed_ns := real((final_time - initial_time) / 1.0 ns);
            elapsed_us := elapsed_ns / 1.0e3;
            frequency_mhz := real(delta) / elapsed_us;

            report_message("  Measured frequency: " & conv_string(frequency_mhz) & " MHz, Actual frequency: " &
                conv_string(frequencies(i)) & " MHz");
        end loop;

        report_message("Simulation of CLOCK complete");
	assert false severity failure;

        wait;
    end process;

    UUT : clock
        port map(
            lclk           => lclk,
            lreset_l       => lreset_l,
            lwrite         => lwrite,
            lads_l         => lads_l,
            lblast_l       => lblast_l,
            lbterm_l       => lbterm_l,
            lad            => lad,
            lready_l       => lready_l,
            lbe_l          => lbe_l,
            fholda         => fholda,
            mclk_n         => mclk_n,
            mclk_p         => mclk_p,
            refclk200_n    => refclk200_n,
            refclk200_p    => refclk200_p,
            front_rx_p     => front_rx_p,
            front_rx_n     => front_rx_n,
            front_tx_p     => front_tx_p,
            front_tx_n     => front_tx_n,
            local_rx_p     => local_rx_p,
            local_rx_n     => local_rx_n,
            local_tx_p     => local_tx_p,
            local_tx_n     => local_tx_n);

    arb_0 : locbus_arb
        generic map(
            n_arb       => 1,
            priority    => priorities)
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lhold       => lhold,
            lholda      => lholda);

    agent_0 : locbus_agent_mux32
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lads_l      => lads_l,
            lbe_l       => lbe_l,
            lblast_l    => lblast_l,
            lad          => lad,
            lbterm_l    => lbterm_l,
            lready_l    => lready_l,
            lwrite      => lwrite,
            lhold       => lhold(0),
            lholda      => lholda(0),
            bus_in      => bus_in,
            bus_out     => bus_out);

    lbpcheck_0 : lbpcheck
        generic map(
            multiplexed => true,
            wide        => false)
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lads_l      => lads_l,
            l64_l       => lads_l,
            la          => lad(31 downto 2),
            lad_lo      => lad,
            lad_hi      => lad,
            lbe_lo_l    => lbe_l,
            lbe_hi_l    => lbe_l,
            lwrite      => lwrite,
            lblast_l    => lblast_l,
            lready_l    => lready_l,
            lbterm_l    => lbterm_l);

end behav;
