--
-- ddma-xrc.vhd - example demand-mode DMA design
--
-- (C) Copyright Alpha Data 2001-2008
--
-- Example program for:
--
--   o ADM-XRC
--   o ADM-XRC-P
--
-- Demand mode DMA is used to read data from the host via DMA channel 0,
-- into a FIFO. The data read from the FIFO and written back to the host via
-- DMA channel 1, also in demand-mode.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.localbus.all;

entity ddma is
    generic(
        la_top           : in    natural := 23);
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic;
        lwrite           : in    std_logic;
        lads_l           : in    std_logic;
        lblast_l         : in    std_logic;
        lbterm_l         : inout std_logic;
        ld               : inout std_logic_vector(31 downto 0);
        la               : in    std_logic_vector(la_top downto 2);
        lready_l         : inout std_logic;
        lbe_l            : in    std_logic_vector(3 downto 0);
        fholda           : in    std_logic;
        ldreq_l          : out   std_logic_vector(1 downto 0);
        ldack_l          : in    std_logic_vector(1 downto 0));
end ddma;

architecture mixed of ddma is

    --
    -- This value is the number of words / free locations that the design attempts to aggregate
    -- in the FIFO before attempting a demand-mode DMA burst. This is intended to increase the
    -- efficiency of the local bus by increasing the average burst length for demand-mode DMA
    -- transfers.
    --

    constant threshold : natural := 32;

    signal rst : std_logic;
    signal lads_i : std_logic;
    signal lblast_i : std_logic;
    signal lbe_iq_l : std_logic_vector(3 downto 0);
    signal lwrite_i : std_logic;
    signal ldack_i : std_logic_vector(1 downto 0);
    signal ldack_q : std_logic_vector(1 downto 0);
    signal ld_iq : std_logic_vector(31 downto 0);
    signal ld_o : std_logic_vector(31 downto 0);
    signal la_i : std_logic_vector(la_top downto 2);
    signal lready_i : std_logic;
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal lbterm_i : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal ldreq_o_l : std_logic_vector(1 downto 0);
    signal eld_oe : std_logic;
    signal ld_oe_l : std_logic_vector(31 downto 0);
    
    signal la_q : std_logic_vector(la_top downto 2);

    signal dd_qlads : std_logic_vector(1 downto 0);
    signal ds_qlads : std_logic;
    signal ds_idle : std_logic;
    signal ds_decode : std_logic;
    signal ds_write : std_logic;
    signal ds_xfer : std_logic;
    signal dd_idle : std_logic_vector(1 downto 0);
    signal dd_idle_q : std_logic_vector(1 downto 0);
    signal dd_idle_qq : std_logic_vector(1 downto 0);
    signal ds_ready : std_logic;
    signal ds_stop : std_logic;

    signal fifo_ne : std_logic;
    signal fifo_read : std_logic;
    signal fifo_radvt : std_logic;
    signal fifo_rev : std_logic;
    signal fifo_f : std_logic;
    signal fifo_nf : std_logic;
    signal fifo_write : std_logic;
    signal fifo_level : std_logic_vector(15 downto 0);
    signal fifo_q : std_logic_vector(31 downto 0);

    signal dd_req : std_logic_vector(1 downto 0);
    signal dd_burst : std_logic_vector(1 downto 0);

    signal write_icount : std_logic;
    signal icount : std_logic_vector(29 downto 0);
    signal icount0 : std_logic;
    signal icountn : std_logic;
    signal wrote_icount: std_logic;
    signal write_ocount : std_logic;
    signal ocount : std_logic_vector(29 downto 0);
    signal ocount0 : std_logic;
    signal ocountn : std_logic;
    signal wrote_ocount : std_logic;

    signal logic0, logic1 : std_logic;

    attribute iob : string;
    attribute iob of lbe_iq_l : signal is "true";
    attribute iob of ld_iq : signal is "true";
    attribute iob of ld_oe_l : signal is "true";

begin
    
    logic0 <= '0';
    logic1 <= '1';

    --
    -- Convert the inputs to active high.
    --

    rst        <= not lreset_l;
    lads_i     <= not lads_l;
    lblast_i   <= not lblast_l;
    lbterm_i   <= not lbterm_l;
    lready_i   <= not lready_l;
    lwrite_i   <= lwrite;
    la_i       <= la(la_top downto 2);
    ldack_i    <= not ldack_l;

    --
    -- Register LD and LBE# (should be mapped into IOBs)
    --

    gen_lbe_iq_l : process(rst, lclk)
    begin
        if rst = '1' then
            lbe_iq_l <= (others => '0');
        elsif lclk'event and lclk = '1' then
            lbe_iq_l <= lbe_l;
        end if;
    end process;

    gen_ld_iq : process(rst, lclk)
    begin
        if rst = '1' then
            ld_iq <= (others => '0');
        elsif lclk'event and lclk = '1' then
            ld_iq <= ld;
        end if;
    end process;

    --
    -- Output active low signals.
    --

    ldreq_l <= ldreq_o_l;

    --
    -- Generate qualified versions of 'lads_l'.
    --
    -- For direct slave access, this is when 'lads_l' is asserted AND the FPGA
    -- is being addressed AND the FPGA is not the local bus master AND the
    -- cycle is not demand-mode DMA.
    --
    -- For demand-mode DMA access, this is when 'lads_l' is asserted AND the
    -- FPGA is being addressed AND the FPGA is not the local bus master AND
    -- LDACK#<n> is asserted.
    --

    ds_qlads <= lads_i and not la_i(la_top) and not fholda;
    gen_dd_qlads : for i in ldreq_l'range generate
        dd_qlads(i) <= lads_i and not la_i(la_top) and not fholda and ldack_i(i);
    end generate;

    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';

    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --
 
    lready_l  <= lready_o_l when lready_oe_l = '0' else 'Z';

    --
    -- Drive the local data bus on a read.
    --
    
    gen_ld_oe_l : process(rst, lclk)
    begin
        if rst = '1' then
            ld_oe_l <= (others => '1');
        elsif lclk'event and lclk = '1' then
            ld_oe_l <= (others => not eld_oe);
        end if;
    end process;

    drive_ld : for i in ld'range generate
        ld(i) <= ld_o(i) when ld_oe_l(i) = '0' else 'Z';
    end generate;

    --
    -- Latch 'lad' on assertion of 'lads_l', in order to improve timing. There
    -- is no need to increment the address, because we don't allow bursting for
    -- register addresses, and DMA transfers that burst are expected to use
    -- constant local addressing.
    --

    latch_address : process(rst, lclk)
    begin
        if rst = '1' then
            la_q <= (others => '0');
            ldack_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
                ldack_q <= ldack_i;
            end if;
        end if;
    end process;

    --
    -- Generate 'ld_o', data to be output on local bus.
    --
    -- Depending on whether this is a demand-mode DMA burst or a normal direct
    -- slave burst, and the local bus address, we output one of
    --
    --   o FIFO data 'fifo_q'
    --   o 'icount' register value
    --   o 'ocount' register value
    --

    gen_ld_o : process(rst, lclk)
    begin
        if rst = '1' then
            ld_o <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if ldack_q(1) = '1' then
                ld_o <= fifo_q;
            else
                if la_q(2) = '1' then
                    ld_o <= ocount & ocountn & ocount0;
                else
                    ld_o <= icount & icountn & icount0;
                end if;
            end if;
        end if;
    end process;

    --
    -- This process implements the DMA transfer count registers and flags for
    -- each direction.
    --
    --    icount          number of words remaining in PCI-to-local DMA transfer
    --    icount0         '1' => 'icount' is zero
    --    icountn         '1' => 'icount' is less than or equal to 'threshold'
    --    ocount          number of words remaining in local-to-PCI DMA transfer
    --    ocount0         '1' => 'ocount' is zero
    --    ocountn         '1' => 'ocount' is less than or equal to 'threshold'
    --
    -- Prior to the host kicking off a demand-mode DMA transfer, the host must
    -- program the transfer count registers must be programmed with the length
    -- of the DMA transfer being performed.
    --

    gen_counts: process(rst, lclk)
    begin
        if rst = '1' then
            write_icount <= '0';
            icount  <= (others => '0');
            icountn <= '1';
            icount0 <= '1';
            wrote_icount <= '0';
            write_ocount <= '0';
            ocount  <= (others => '0');
            ocountn <= '1';
            ocount0 <= '1';
            wrote_ocount <= '0';
        elsif lclk'event and lclk = '1' then
            write_icount <= ds_xfer and ds_write and not la_q(2) and not OR_reduce(ldack_q);
            write_ocount <= ds_xfer and ds_write and la_q(2) and not OR_reduce(ldack_q);

            if write_icount = '1' then
                if lbe_iq_l(0) = '0' then
                    icount(5 downto 0) <= ld_iq(7 downto 2);
                end if;
                if lbe_iq_l(1) = '0' then 
                    icount(13 downto 6) <= ld_iq(15 downto 8);
                end if;
                if lbe_iq_l(2) = '0' then 
                    icount(21 downto 14) <= ld_iq(23 downto 16);
                end if;
                if lbe_iq_l(3) = '0' then 
                    icount(29 downto 22) <= ld_iq(31 downto 24);
                end if;
            else
                if fifo_write = '1' then
                    icount <= icount - 1;
                end if;
            end if;

            if write_ocount = '1' then
                if lbe_iq_l(0) = '0' then
                    ocount(5 downto 0) <= ld_iq(7 downto 2);
                end if;
                if lbe_iq_l(1) = '0' then
                    ocount(13 downto 6) <= ld_iq(15 downto 8);
                end if;
                if lbe_iq_l(2) = '0' then 
                    ocount(21 downto 14) <= ld_iq(23 downto 16);
                end if;
                if lbe_iq_l(3) = '0' then 
                    ocount(29 downto 22) <= ld_iq(31 downto 24);
                end if;
            else
                if fifo_read = '1' then
                    ocount <= ocount - 1;
                end if;
            end if;

            wrote_icount <= write_icount;
            if wrote_icount = '1' then
                if icount <= threshold then
                    icountn <= '1';
                else
                    icountn <= '0';
                end if;
                if icount = 0 then
                    icount0 <= '1';
                else
                    icount0 <= '0';
                end if;
            else
                if fifo_write = '1' then
                    if icount = (threshold + 1) then
                        icountn <= '1';
                    end if;
                    if icount = 1 then
                        icount0 <= '1';
                    end if;
                end if;
            end if;

            wrote_ocount <= write_ocount;
            if wrote_ocount = '1' then
                if ocount <= threshold then
                    ocountn <= '1';
                else
                    ocountn <= '0';
                end if;
                if ocount = 0 then
                    ocount0 <= '1';
                else
                    ocount0 <= '0';
                end if;
            else
                if fifo_read = '1' then
                    if ocount = (threshold + 1) then
                        ocountn <= '1';
                    end if;
                    if ocount = 1 then
                        ocount0 <= '1';
                    end if;
                end if;
            end if;
        end if;
    end process;

    --
    -- Delay the 'idle' signals from each of the demand-mode DMA state machines
    -- by one and two cycles.
    --

    gen_dd_idle_qq : process(rst, lclk)
    begin
        if rst = '1' then
            dd_idle_q <= (others => '1');
            dd_idle_qq <= (others => '1');
        elsif lclk'event and lclk = '1' then
            for i in ldreq_l'range loop
               dd_idle_q(i) <= dd_idle(i);
               dd_idle_qq(i) <= dd_idle_q(i);
            end loop;
        end if;
    end process;

    --
    -- For the PCI->local direction, we wait until the FIFO has space to hold a
    -- decent-sized burst before requesting a demand-mode DMA burst. We need to
    -- pause if the FIFO becomes nearly-full.
    --
    -- The 'dd_idle_qq(0)' term is necessary because the 'icount*' signals may
    -- not be properly updated until a cycle of two after the demand-mode DMA
    -- burst terminates. By preventing 'dd_req(0)' from being asserted until
    -- a couple of cycles after the burst terminates, a spurious assertion of
    -- LDREQ#<0> is avoided when 'icount0' is about to but has not yet
    -- transitioned to 1.
    --

    dd_req(0) <= '1' when icount0 = '0' and dd_idle_qq(0) = '1' and fifo_level <= (512 - threshold) else
                 '0';
    dd_burst(0) <= not fifo_nf;

    --
    -- For the local->PCI direction, we request a demand-mode DMA burst only if the
    -- DMA transfer is not complete ('ocount' is nonzero) and either:
    --
    --      (i) the FIFO contains enough data for a decent-sized burst, or
    --
    --     (ii) we're near the end of the DMA transfer AND the FIFO level
    --          is equal to or greater than the remaining transfer count.
    --
    -- The 'dd_idle_q(1)' term takes care of a corner case. At the end of a demand-
    -- mode DMA local bus cycle, it may take an extra clock cycle for the FIFO flags
    -- and level to be fully updated to reflect how much data has just been read out
    -- of the FIFO. Therefore, we force 'dd_req(1)' deasserted for one cycle at the
    -- end of each demand-mode DMA local bus cycle.
    --
    -- This could be optimized by only forcing 'dd_req(1)' deasserted when the FIFO
    -- level is known to be below a certain threshold, but for simplicity we do it
    -- unconditionally.
    --

    dd_req(1) <= '1' when (ocount0 = '0' and fifo_level >= threshold) or
                          (dd_idle_q(1) = '1' and ocount0 = '0' and ocountn = '1' and fifo_level >= ocount(15 downto 0))
                     else '0';
    dd_burst(1) <= not fifo_ne;

    --
    -- Writing into the FIFO only happens when data is written on DMA channel 0.
    -- Because the data and byte enables from the local bus are registered, the
    -- 'fifo_write' signal must also be registered.
    --

    gen_fifo_write : process(rst, lclk)
    begin
        if rst = '1' then
            fifo_write <= '0';
        elsif lclk'event and lclk = '1' then
            fifo_write <= ds_xfer and ds_write and ldack_q(0);
        end if;
    end process;

    --
    -- Advancing the FIFO permanent read pointer only happens when data is
    -- read on DMA channel 1.
    --

    fifo_read <= ds_xfer and not ds_write and ldack_q(1);

    --
    -- Advance the FIFO temporary read pointer during a burst. Because we are
    -- pipelining 'ld_o' we require the 'ds_decode' term in order to clock data
    -- through to 'ld_o' at the beginning of a read burst.
    --

    fifo_radvt <= (ds_xfer or ds_decode) and ldack_q(1);

    --
    -- We make the FIFO's temporary read pointer revert to its permanent
    -- read pointer at the end of a demand-mode DMA read local bus cycle.
    --

    fifo_rev <= dd_idle(1) and not dd_idle_q(1);

    --
    -- Generate 'ready' and 'stop' signals, which control the state machines
    -- that respond to local bus cycles.
    --

    ds_ready <= '1';                               -- We're always ready
    ds_stop <= not OR_reduce(ldack_q) or           -- Prevent bursting if not demand-mode DMA
               (ldack_q(0) and not dd_burst(0)) or -- Stop when demand-mode DMA and write FIFO is nearly full
               (ldack_q(1) and not dd_burst(1));   -- Stop when demand-mode DMA and read FIFO is nearly empty

    --
    -- Instantiate the state machines that control the LDREQ# lines.
    --

    gen_plxddsm2 : for i in ldreq_l'range generate
        U0 : plxddsm2
            port map(
                clk => lclk,
                rst => rst,
                sr => logic0,
                qlads => dd_qlads(i),
                lblast => lblast_i,
                lbterm => lbterm_i,
                lready => lready_i,
                request => dd_req(i),
                burst => dd_burst(i),
                ready => ds_ready, -- MUST be same as fed into 'plxdssm'
                idle => dd_idle(i),
                ldreq_o_l => ldreq_o_l(i));
    end generate;

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    plxdssm0: plxdssm
        port map(
            clk => lclk,
            rst => rst,
            sr => logic0,
            qlads => ds_qlads,
            lblast => lblast_i,
            lbterm => lbterm_i,
            lwrite => lwrite_i,
            eld_oe => eld_oe,
            ld_oe_l => open,
            lready_o_l => lready_o_l,
            lready_oe_l => lready_oe_l,
            lbterm_o_l => lbterm_o_l,
            lbterm_oe_l => lbterm_oe_l,
            idle => ds_idle,
            transfer => ds_xfer,
            decode => ds_decode,
            write => ds_write,
            ready => ds_ready, -- MUST be the same as fed into 'plxddsm2'
            stop => ds_stop);

    --
    -- 'rpel' = 2 => nearly-empty is 2 or less words present the FIFO.
    -- 'wpfl' = 6 => nearly-full is 6 or less free locations remaining in the FIFO.
    --
    -- These values are chosen to ensure that the FIFO cannot underflow or overflow.
    --

    sfifop0 : sfifop
	generic map(
            width      => 32,
            order      => 9,
            rpel       => 2,
            wpfl       => 6,
            family     => work.fifo.family_virtex,
            safety     => false)
        port map(
            rst        => rst,
            k          => lclk,
            wd         => ld_iq,
            wadv       => fifo_write,
            we         => open,
            wpf        => fifo_nf,
            wf         => fifo_f,
            wovf       => open,
            rq         => fifo_q,
            radv       => fifo_read,
            radvt      => fifo_radvt,
            rrev       => fifo_rev,
            rflush     => logic0,
            ret        => open,
            re         => open,
            rf         => open,
            rpet       => fifo_ne,
            rpe        => open,
            rudf       => open,
            rlevelt    => open,
            rlevel     => fifo_level);

end mixed;
