--
-- clocks_dcm.vhd - clocks multiplication module using DLLs for
--                  Virtex-II
--
-- (c) Alpha Data Parallel Systems Ltd. 2001
--
-- The local bus clock is multiplied by 2 using a Virtex-II DCM.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity clocks is
   port(
      rst              : in    std_logic;
      lclk             : in    std_logic;
      mclk             : in    std_logic;
      clk              : out   std_logic;
      multclk          : out   std_logic;
      locked           : out   std_logic_vector(31 downto 0));
end clocks;

architecture DCM of clocks is

    signal lclk_ibufg:   std_logic;
    signal lclk_bufg:    std_logic;
    signal clk0:         std_logic;
    signal i_clk:        std_logic;
    signal multclk0:     std_logic;
    
    signal logic0:       std_logic;
    signal logic1:       std_logic;

    signal dllrst:       std_logic;
    signal rstcnt:       std_logic_vector(2 downto 0);
    
    component IBUFG
        port(
            I:             in       std_logic;
            O:             out      std_logic);
    end component;
    
    component BUFG
        port(
            I:             in       std_logic;
            O:             out      std_logic);
    end component;
    
    component DCM
        port(
            CLKIN         : in  std_logic;
            CLKFB         : in  std_logic;
            DSSEN         : in  std_logic;
            PSINCDEC      : in  std_logic;
            PSEN          : in  std_logic;
            PSCLK         : in  std_logic;
            RST           : in  std_logic;
            CLK0          : out std_logic;
            CLK90         : out std_logic;
            CLK180        : out std_logic;
            CLK270        : out std_logic;
            CLK2X         : out std_logic;
            CLK2X180      : out std_logic;
            CLKDV         : out std_logic;
            CLKFX         : out std_logic;
            CLKFX180      : out std_logic;
            LOCKED        : out std_logic;
            PSDONE        : out std_logic;
            STATUS        : out std_logic_vector(7 downto 0));
    end component;

    attribute DLL_FREQUENCY_MODE : string;
    attribute DUTY_CYCLE_CORRECTION : string;
    attribute STARTUP_WAIT : string;

    attribute DLL_FREQUENCY_MODE of dll0 : label is "LOW";
    attribute DUTY_CYCLE_CORRECTION of dll0 : label is "TRUE";
    attribute STARTUP_WAIT of dll0 : label is "FALSE";
    
begin

    logic0 <= '0';
    logic1 <= '1';

    --
    -- Generate reset signal to DLLs/DCMs
    --

    gen_dllrst : process(rst, mclk)
    begin
        if rst = '1' then
            dllrst <= '1';
            rstcnt <= (others => '0');
        elsif mclk'event and mclk = '1' then
            if dllrst = '1' then
                rstcnt <= rstcnt + 1;
            end if;
            if AND_reduce(rstcnt) = '1' then
                dllrst <= '0';
            end if;
        end if;
    end process;
    
    clk <= i_clk;
    
    ibufg_lclk: IBUFG
        port map(
            I => lclk,
            O => lclk_ibufg);
    
    --
    -- The clock DCM both minimises skew and outputs a 2x clock.
    --

    dll0 : DCM
        port map(
            CLKIN    => lclk_ibufg,
            CLKFB    => i_clk,
            DSSEN    => logic0,
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => dllrst,
            CLK0     => clk0,
            CLK2X    => multclk0,
            LOCKED   => locked(0));

    locked(31 downto 1) <= (others => '0');

    bufg0: BUFG
        port map(
            I => clk0,
            O => i_clk);
    
    bufg1: BUFG
        port map(
            I => multclk0,
            O => multclk);

end DCM;
