--
-- clocks_dcm_v4.vhd - clock multiplication module using DCMs for
--                     Virtex-4
--
-- (c) Alpha Data Parallel Systems Ltd. 2006
--
-- The local bus clock is multiplied by 2 using a Virtex-II DCM.
--
-- Uses the DCM_STANDBY macro to workaround the NBTI issue that affects
-- early steppings of Virtex-4 devices. See Xilinx Answer Record 21127.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity clocks is
   port(
      rst              : in    std_logic;
      lclk             : in    std_logic;
      mclk             : in    std_logic;
      clk              : out   std_logic;
      multclk          : out   std_logic;
      locked           : out   std_logic_vector(31 downto 0));
end clocks;

architecture DCM of clocks is

    signal lclk_ibufg : std_logic;
    signal lclk_bufg : std_logic;
    signal clk0 : std_logic;
    signal i_clk : std_logic;
    signal multclk0 : std_logic;
    
    signal dllrst : std_logic;
    signal rstcnt : std_logic_vector(2 downto 0);
    
    signal logic0 : std_logic;
    signal logic1 : std_logic;

    component IBUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    component BUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    --
    -- DCM_STANDBY is a macro to work around the NBTI issue that affects
    -- early steppings of Virtex-4 devices. See Xilinx Answer Record 21127.
    --

    component DCM_STANDBY
        port(
            CLK0     : out std_logic;
            CLK180   : out std_logic;
            CLK270   : out std_logic;
            CLK2X    : out std_logic;
            CLK2X180 : out std_logic;
            CLK90    : out std_logic;
            CLKDV    : out std_logic;
            CLKFX    : out std_logic;
            CLKFX180 : out std_logic;
            DO       : out std_logic_vector(15 DOWNTO 0);
            LOCKED   : out std_logic;
            PSDONE   : out std_logic;
            CLKFB    : in  std_logic;
            CLKIN    : in  std_logic;
            PSCLK    : in  std_logic;
            PSEN     : in  std_logic;
            PSINCDEC : in  std_logic;
            RST      : in  std_logic);
    end component;
	 
    attribute DLL_FREQUENCY_MODE : string;
    attribute DUTY_CYCLE_CORRECTION : string;
    attribute STARTUP_WAIT : string;

--    attribute DLL_FREQUENCY_MODE of dll0 : label is "LOW";
--    attribute DUTY_CYCLE_CORRECTION of dll0 : label is "TRUE";
--    attribute STARTUP_WAIT of dll0 : label is "FALSE";
    
begin

    logic0 <= '0';
    logic1 <= '1';

    --
    -- Generate reset signal to DLLs/DCMs
    --

    gen_dllrst : process(rst, mclk)
    begin
        if rst = '1' then
            dllrst <= '1';
            rstcnt <= (others => '0');
        elsif mclk'event and mclk = '1' then
            if dllrst = '1' then
                rstcnt <= rstcnt + 1;
            end if;
            if AND_reduce(rstcnt) = '1' then
                dllrst <= '0';
            end if;
        end if;
    end process;
    
    clk <= i_clk;
    
    ibufg_lclk: IBUFG
        port map(
            I => lclk,
            O => lclk_ibufg);
    
    --
    -- The DCM produces a delay-compensated version of LCLK and a
    -- doubled version of LCLK, also delay-compensated.
    --

    dll0 : DCM_STANDBY
        port map(
            CLKIN    => lclk_ibufg,
            CLKFB    => i_clk,
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => dllrst,
            CLK0     => clk0,
            CLK2X    => multclk0,
            LOCKED   => locked(0));

    locked(31 downto 1) <= (others => '0');

    i_clk <= clk0; -- BUFG *not* required because DCM_STANDBY has a BUFGCTRL for CLK0

    bufg_multclk0 : BUFG
        port map(
            I => multclk0,
            O => multclk);

end DCM;
