--
-- dll-xrc4fx.vhd - Example FPGA design demonstrating clock multiplication
--
-- (C) Copyright Alpha Data 1999-2008
--
-- Example program for:
--
--   o ADM-XRC-4FX
--
-- The local bus clock is input and multiplied by 2 using a Virtex-IIPro DCM.
--
-- The 2x clock is used to clock a free running counter, which may be read
-- via by direct slave reads on the local bus (anywhere in the FPGA space).
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

library work;
use work.localbus.all;

entity dll is
    generic(
        la_top           : in    natural := 23);
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic;
        lwrite           : in    std_logic;
        lads_l           : in    std_logic;
        lblast_l         : in    std_logic;
        lbterm_l         : inout std_logic;
        lad              : inout std_logic_vector(31 downto 0);
        lready_l         : inout std_logic;
        lbe_l            : in    std_logic_vector(3 downto 0);
        fholda           : in    std_logic;
        refclk200_p      : in    std_logic;
        refclk200_n      : in    std_logic;
        -- MGT errata workaround
        pcie_rx_p        : in    std_logic_vector(7 downto 0);
        pcie_rx_n        : in    std_logic_vector(7 downto 0);
        pcie_tx_p        : out   std_logic_vector(7 downto 0);
        pcie_tx_n        : out   std_logic_vector(7 downto 0);
        front_rx_p       : in    std_logic_vector(7 downto 0);
        front_rx_n       : in    std_logic_vector(7 downto 0);
        front_tx_p       : out   std_logic_vector(7 downto 0);
        front_tx_n       : out   std_logic_vector(7 downto 0));
end dll;

architecture mixed of dll is
    
    component clocks
        port(
            rst              : in    std_logic;
            lclk             : in    std_logic;
            mclk             : in    std_logic;
            clk              : out   std_logic;
            multclk          : out   std_logic;
            locked           : out   std_logic_vector(31 downto 0));
    end component;

    component IBUFGDS
        port(
            I                : in    std_logic;
            IB               : in    std_logic;
            O                : out   std_logic);
    end component;

    component BUFG
        port(
            I                : in    std_logic;
            O                : out   std_logic);
    end component;

    component null_mgt
        port(
            clka             : in    std_logic;
            rxa_p            : in    std_logic_vector(7 downto 0);
            rxa_n            : in    std_logic_vector(7 downto 0);
            txa_p            : out   std_logic_vector(7 downto 0);
            txa_n            : out   std_logic_vector(7 downto 0);
            clkb             : in    std_logic;
            rxb_p            : in    std_logic_vector(7 downto 0);
            rxb_n            : in    std_logic_vector(7 downto 0);
            txb_p            : out   std_logic_vector(7 downto 0);
            txb_n            : out   std_logic_vector(7 downto 0));
    end component;

    signal rst : std_logic;
    signal lads_i : std_logic;
    signal lblast_i : std_logic;
    signal lbterm_i : std_logic;
    signal lwrite_i : std_logic;
    signal lbe_i : std_logic_vector(3 downto 0);
    signal ld_i : std_logic_vector(31 downto 0);
    signal la_i : std_logic_vector(la_top downto 2);   
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal ld_o : std_logic_vector(31 downto 0);
    signal ld_oe_l : std_logic;

    signal clk : std_logic;
    signal multclk : std_logic;
    signal locked : std_logic_vector(31 downto 0);
    
    signal qlads : std_logic;

    signal ds_xfer : std_logic;
    signal ds_decode : std_logic;
    signal ds_write : std_logic;

    signal la_q : std_logic_vector(23 downto 2);
    
    signal count : std_logic_vector(31 downto 0);
    signal count_q : std_logic_vector(31 downto 0);
    
    signal refclk200_ibufg : std_logic;
    signal refclk200 : std_logic;

    signal logic0, logic1 : std_logic;
    
begin

    logic0 <= '0';
    logic1 <= '1';

    --
    -- Convert the inputs to active high.
    --

    rst      <= not lreset_l;
    lblast_i <= not lblast_l;
    lbterm_i <= not lbterm_l;
    lads_i   <= not lads_l;
    lwrite_i <= lwrite;
    la_i     <= lad(la_top downto 2);
    ld_i     <= lad;
    lbe_i    <= not lbe_l;
    
    --
    -- Generate a qualified version of 'lads_l', which is
    -- asserted when the FPGA is addressed AND the FPGA is
    -- not the local bus master.
    --

    qlads <= lads_i and not la_i(la_top) and not fholda;

    --
    -- Latch the local bus address on the 'lads_l' pulse.
    --

    latch_addr : process(rst, clk)
    begin
        if rst = '1' then
            la_q <= (others => '0');
        elsif clk'event and clk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
            end if;
        end if;
    end process;
    
    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';
    
    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --

    lready_l <= lready_o_l when lready_oe_l = '0' else 'Z';
    
    --
    -- Drive the local data bus on a read.
    --

    gen_lad : for i in lad'range generate
        lad(i) <= ld_o(i) when ld_oe_l = '0' else 'Z';
    end generate;
    
    --
    -- Generate the 'ld_o' signal, outbound local bus data.
    --

    generate_ld_o : process(rst, clk)
    begin
        if rst = '1' then
            ld_o <= (others => '0');
        elsif clk'event and clk = '1' then
            if la_q(2) = '1' then
                ld_o <= locked;
            else
                ld_o <= count_q;
            end if;
        end if;
    end process;

    --
    -- Cycle counter, clocked by LCLK*2
    --

    cycle_counter : process(multclk, rst)
    begin
        if rst = '1' then
            count <= (others => '0');
        elsif multclk'event and multclk = '1' then
            if ds_xfer = '1' and ds_write = '1' then
                for i in 0 to 3 loop
                    if lbe_i(i) = '1' then
                        count(8 * (i + 1) - 1 downto 8 * i) <=
                            ld_i(8 * (i + 1) - 1 downto 8 * i);
                    end if;
                end loop;
            else
                count <= count + '1';
            end if;
        end if;
    end process;

    --
    -- Resample the LCLK*2 cycle counter to LCLK. Should be safe with
    -- regard to metastability because the two clock domains are
    -- phase-locked.
    --

    resample : process(clk, rst)
    begin
        if rst = '1' then
            count_q <= (others => '0');
        elsif clk'event and clk = '1' then
            count_q <= count;
        end if;
    end process;      

    --
    -- Instantiate the clocks module, which generates LCLK*2.
    --

    ibufg_refclk200 : IBUFGDS
        port map(
            I => refclk200_p,
            IB => refclk200_n,
            O => refclk200_ibufg);

    bufg_refclk200: BUFG
        port map(
            I => refclk200_ibufg,
            O => refclk200);

    clocks0 : clocks
        port map(
            rst              => rst,
            lclk             => lclk,
            mclk             => refclk200,
            clk              => clk,
            multclk          => multclk,
            locked           => locked);

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    plxdssm0 : plxdssm
        port map(
            clk              => clk,
            rst              => rst,
            sr               => logic0,
            qlads            => qlads,
            lblast           => lblast_i,
            lbterm           => lbterm_i,
            lwrite           => lwrite_i,
            eld_oe           => open,
            ld_oe_l          => ld_oe_l,
            lready_o_l       => lready_o_l,
            lready_oe_l      => lready_oe_l,
            lbterm_o_l       => lbterm_o_l,
            lbterm_oe_l      => lbterm_oe_l,
            idle             => open,
            transfer         => ds_xfer,
            decode           => ds_decode,
            write            => ds_write,
            ready            => logic1,
            stop             => logic1);
    
    --
    -- Instantiate NULL_PAIR instances, to workaround errata related to
    -- Virtex-4 MGTs.
    --

    null_mgt_0 : null_mgt
        port map(
            clka => clk,
            rxa_p => front_rx_p,
            rxa_n => front_rx_n,
            txa_p => front_tx_p,
            txa_n => front_tx_n,
            clkb => clk,
            rxb_p => pcie_rx_p,
            rxb_n => pcie_rx_n,
            txb_p => pcie_tx_p,
            txb_n => pcie_tx_n);

end mixed;
