--
-- test_dll-xrc4fx.vhd - Testbench for entity 'dll'
--
-- (C) Copyright Alpha Data 2007
--
-- Example testbench for:
--
--   o ADM-XRC-4FX
--
-- NOT SYNTHESIZABLE
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library std;
use std.textio.all;

library plxsim;
use plxsim.plxsim.all;

entity test_dll is
end test_dll;

architecture behav of test_dll is

    constant lclk_period : time := 30 ns;

    constant refclk200_period : time := 10 ns; -- not critical as only used for DCM reset

    constant priorities : integer_vector_t(0 to 0) := (others => 0);

    signal lclk : std_logic := '1';
    signal lreset_l : std_logic := '0';
    signal lbe_l : std_logic_vector(3 downto 0) := (others => 'Z');
    signal lad : std_logic_vector(31 downto 0) := (others => 'Z');
    signal lwrite : std_logic := '0';
    signal lads_l : std_logic := '1';
    signal lblast_l : std_logic := '1';
    signal lbterm_l : std_logic := 'Z';
    signal lready_l : std_logic := 'Z';
    signal lhold, lholda : std_logic_vector(0 downto 0);
    signal fholda : std_logic := '0';

    signal refclk200_p : std_logic := '0';
    signal refclk200_n : std_logic := '1';

    signal pcie_rx_p : std_logic_vector(7 downto 0) := (others => 'H');
    signal pcie_rx_n : std_logic_vector(7 downto 0) := (others => 'L');
    signal pcie_tx_p : std_logic_vector(7 downto 0);
    signal pcie_tx_n : std_logic_vector(7 downto 0);
    signal front_rx_p : std_logic_vector(7 downto 0) := (others => 'H');
    signal front_rx_n : std_logic_vector(7 downto 0) := (others => 'L');
    signal front_tx_p : std_logic_vector(7 downto 0);
    signal front_tx_n : std_logic_vector(7 downto 0);

    shared variable lclk_count : natural := 0;

    signal bus_in : locbus_in_t;
    signal bus_out : locbus_out_t := init_locbus_out;

    signal logic0, logic1 : std_logic;

    procedure report_message(
        str : in string)
    is
        variable msg : line;
    begin
        write(msg, str);
        writeline(output, msg);
    end;

    component dll
        port(
            lclk             : in    std_logic;
            lreset_l         : in    std_logic; 
            lwrite           : in    std_logic;
            lads_l           : in    std_logic;
            lblast_l         : in    std_logic;
            lbterm_l         : inout std_logic;
            lad              : inout std_logic_vector(31 downto 0);
            lready_l         : inout std_logic; 
            lbe_l            : in    std_logic_vector(3 downto 0);
            fholda           : in    std_logic;
            refclk200_p      : in    std_logic;
            refclk200_n      : in    std_logic;
            pcie_rx_p        : in    std_logic_vector(7 downto 0);
            pcie_rx_n        : in    std_logic_vector(7 downto 0);
            pcie_tx_p        : out   std_logic_vector(7 downto 0);
            pcie_tx_n        : out   std_logic_vector(7 downto 0);
            front_rx_p       : in    std_logic_vector(7 downto 0);
            front_rx_n       : in    std_logic_vector(7 downto 0);
            front_tx_p       : out   std_logic_vector(7 downto 0);
            front_tx_n       : out   std_logic_vector(7 downto 0));
    end component;

begin

    logic0 <= '0';
    logic1 <= '1';

    lclk <= not lclk after lclk_period / 2;

    refclk200_p <= not refclk200_p after refclk200_period / 2;
    refclk200_n <= not refclk200_n after refclk200_period / 2;

    lreset_l <= transport '1' after 50 ns;

    lads_l <= 'H';
    lblast_l <= 'H';
    lready_l <= 'H';
    lbterm_l <= 'H';
    lwrite <= 'H';

    stimulate : process
        variable data : byte_vector_t(0 to 3);
        variable be : byte_enable_t(0 to 3);
        variable n : natural;
        variable delta : std_logic_vector(31 downto 0);
        variable old_count : std_logic_vector(31 downto 0);
        variable count : std_logic_vector(31 downto 0);
        variable lclk_delta : natural;
        variable old_lclk_count : natural;
    begin
        --
        -- Wait for DLLs / DCMs to achieve lock
        -- With the DCM_STANDBY macro in use with Virtex-4, this takes a little
        -- longer in functional simulation compared to a normal DCM.
        --
        plxsim_wait_cycles(1000, bus_in);

        be := (others => '1');

        plxsim_request_bus(true, bus_in, bus_out);

        -- Initialize LCLK*2 cycle counter register
        data(0 to 3) := conv_byte_vector(X"00000000");
        plxsim_write(2, true, X"00000000", be, data, n, bus_in, bus_out);
        old_count := (others => '0');

        for i in 0 to 10 loop
            -- Read back the LCLK*2 cycle counter register
            plxsim_read(2, true, X"00000000", be, data, n, bus_in, bus_out);
            count := conv_std_logic_vector(data);
            delta := count - old_count;
            lclk_delta := lclk_count - old_lclk_count;
            if i > 0 then
                report_message("count=0x" & conv_string_hex(data(0 to 3)) & " delta=0x" & conv_string_hex(delta) &
                    " lclk_delta=0x" & conv_string_hex(lclk_delta));
            end if;
            old_count := count;
            old_lclk_count := lclk_count;
        end loop;

        plxsim_request_bus(false, bus_in, bus_out);

        report_message("Simulation of dll complete");

        wait;
    end process;

    count_lclk_cycles : process(lreset_l, lclk)
    begin
        if lreset_l = '0' then
            lclk_count := 0;
        elsif lclk'event and lclk = '1' then
            lclk_count := lclk_count + 1;
        end if;
    end process;

    UUT : dll
        port map(
            lclk           => lclk,
            lreset_l       => lreset_l,
            lwrite         => lwrite,
            lads_l         => lads_l,
            lblast_l       => lblast_l,
            lbterm_l       => lbterm_l,
            lad            => lad,
            lready_l       => lready_l,
            lbe_l          => lbe_l,
            fholda         => fholda,
            refclk200_p    => refclk200_p,
            refclk200_n    => refclk200_n,
            pcie_rx_p      => pcie_rx_p,
            pcie_rx_n      => pcie_rx_n,
            pcie_tx_p      => pcie_tx_p,
            pcie_tx_n      => pcie_tx_n,
            front_rx_p     => front_rx_p,
            front_rx_n     => front_rx_n,
            front_tx_p     => front_tx_p,
            front_tx_n     => front_tx_n);

    arb_0 : locbus_arb
        generic map(
            n_arb       => 1,
            priority    => priorities)
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lhold       => lhold,
            lholda      => lholda);

    agent_0 : locbus_agent_mux32
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lads_l      => lads_l,
            lad         => lad,
            lbe_l       => lbe_l,
            lblast_l    => lblast_l,
            lbterm_l    => lbterm_l,
            lready_l    => lready_l,
            lwrite      => lwrite,
            lhold       => lhold(0),
            lholda      => lholda(0),
            bus_in      => bus_in,
            bus_out     => bus_out);

    lbpcheck_0 : lbpcheck
        generic map(
            multiplexed => true,
            wide        => false)
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lads_l      => lads_l,
            l64_l       => lads_l,
            la          => lad(31 downto 2),
            lad_lo      => lad,
            lad_hi      => lad,
            lbe_lo_l    => lbe_l,
            lbe_hi_l    => lbe_l,
            lwrite      => lwrite,
            lblast_l    => lblast_l,
            lready_l    => lready_l,
            lbterm_l    => lbterm_l);

end behav;
