--
-- itest-xpl.vhd - FPGA design demonstrating interrupt generation
--
-- (C) Copyright Alpha Data 1999-2008
--
-- Example program for:
--
--   o ADM-XPL
--   o ADM-XP
--   o ADM-XRC-5LX
--   o ADM-XRC-5T1
--   o ADM-XRC-5T2
--
-- NOTES ON OPERATION
--
-- The logic in this FPGA design works in the way described
-- in the application note an-xrc06.pdf, in the doc/ directory
-- of this SDK.
--
-- 32 sources of interrupt are provided, each of which can be
-- triggered by writing a 1 to a corresponding bit of a write-
-- only register called TEST. When a source of interrupt has
-- is triggered, the corresponding bit of ISTAT becomes 1.
--
-- Each source of interrupt may be masked (disabled) by the
-- corresponding bit of IMASK being 1. The IMASK register resets
-- to 0xffffffff, ie. all interrupts disabled.
--
-- A write to byte 0 of the IARM register forces the FINTI_L signal
-- high (deasserted) for a single cycle, to "rearm" the FINTI_L signal.
-- The reason for including the IARM register is discussed in
-- an-xrc06.pdf .
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.localbus.all;

entity itest is
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic;
        lwrite           : in    std_logic;
        lads_l           : in    std_logic;
        lblast_l         : in    std_logic;
        lbterm_l         : inout std_logic;
        lad              : inout std_logic_vector(31 downto 0);
        lready_l         : inout std_logic;
        lbe_l            : in    std_logic_vector(3 downto 0);
        fholda           : in    std_logic;
        finti_l          : out   std_logic);
end itest;

architecture mixed of itest is

    signal rst : std_logic;
    signal lads_i : std_logic;
    signal lblast_i : std_logic;
    signal lbterm_i : std_logic;
    signal lbe_i : std_logic_vector(3 downto 0);
    signal lwrite_i : std_logic;
    signal la_i : std_logic_vector(23 downto 2);
    signal ld_i : std_logic_vector(31 downto 0);
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal ld_o : std_logic_vector(31 downto 0);
    signal ld_oe_l : std_logic;
    
    signal finti : std_logic;
    
    signal la_q : std_logic_vector(23 downto 2);

    signal qlads : std_logic;

    signal ds_xfer : std_logic;
    signal ds_decode : std_logic;
    signal ds_write : std_logic;
    
    signal istat : std_logic_vector(31 downto 0);
    signal imask : std_logic_vector(31 downto 0);
    signal iarm : std_logic;
    signal count : std_logic_vector(31 downto 0);

    signal logic0, logic1 : std_logic;
    
begin
    
    logic0 <= '0';
    logic1 <= '1';

    --
    -- Convert the inputs to active high.
    --	
    rst      <= not lreset_l;
    lblast_i <= not lblast_l;
    lbterm_i <= not lbterm_l;
    lads_i   <= not lads_l;
    lwrite_i <= lwrite;
    la_i     <= lad(23 downto 2);
    ld_i     <= lad;
    lbe_i    <= not lbe_l;
    
    --
    -- Generate a qualified version of 'lads_l', which is
    -- asserted when the FPGA is addressed AND the FPGA is
    -- not the local bus master.
    --

    qlads <= lads_i and not la_i(23) and not fholda;

    --
    -- Latch the local bus address on the 'lads_l' pulse.
    --

    latch_addr : process(rst, lclk)
    begin
        if rst = '1' then
            la_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
            end if;
        end if;
    end process latch_addr;
    
    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';
    
    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --

    lready_l <= lready_o_l when lready_oe_l = '0' else 'Z';
    
    --
    -- Drive the local data bus on a read.
    --

    gen_ld : for i in lad'range generate
        lad(i) <= ld_o(i) when ld_oe_l = '0' else 'Z';
    end generate;
    
    --
    -- Generate the 'ld_o' signal, outbound local bus data.
    --

    generate_ld_o : process(rst, lclk)
    begin
        if rst = '1' then
            ld_o <= (others => '0');
        elsif lclk'event and lclk = '1' then
            case la_q(4 downto 2) is
                when "000" =>
                    ld_o <= imask;
                when "001" =>
                    ld_o <= istat;
                when "100" =>
                    ld_o <= count;
                when others =>
                    ld_o <= (others => '-');
            end case;
        end if;
    end process;

    --
    -- generate 'finti_l' (FPGA interrupt)
    --

    gen_finti: process(lclk, rst)
    begin
        if rst = '1' then
            finti <= '0';
        elsif (lclk'event and lclk = '1') then
            if (not imask and istat) = EXT("0", 32) or iarm = '1' then
                finti <= '0';
            else
                finti <= '1';
            end if;
        end if;
    end process;

    finti_l <= not finti;

    --
    -- Implement the registers.
    --
    -- Register     Local addr      Type            Function
    -- -------------------------------------------------------------------
    -- IMASK        0x0             R/W             Interrupt mask
    -- ISTAT        0x4             R/W1C           Interrupt status
    -- IARM         0x8             WO/RAX          Interrupt rearm strobe
    -- TEST         0xC             WO              Sets ISTAT for testing
    -- COUNT        0x10            R/W             LCLK cycle counter
    --

    implement_registers : process(lclk, rst)
    begin
        if rst = '1' then
            imask <= (others => '1');
            istat <= (others => '0');
            iarm <= '0';
            count <= (others => '0');
        elsif (lclk'event and lclk = '1') then
            if ds_xfer = '1' and ds_write = '1' and la_q(4 downto 2) = "000" then
                -- write to IMASK
                if lbe_i(0) = '1' then
                    imask(7 downto 0) <= ld_i(7 downto 0);
                end if;
                if lbe_i(1) = '1' then
                    imask(15 downto 8) <= ld_i(15 downto 8);
                end if;
                if lbe_i(2) = '1' then
                    imask(23 downto 16) <= ld_i(23 downto 16);
                end if;
                if lbe_i(3) = '1' then
                    imask(31 downto 24) <= ld_i(31 downto 24);
                end if;
            end if;
            
            if ds_xfer = '1' and ds_write = '1' and la_q(4) = '0' and la_q(2) = '1' then
                if la_q(3) = '0' then                     
                    -- write to ISTAT - 1s clear corresponding bit
                    if lbe_i(0) = '1' then
                        istat(7 downto 0) <= istat(7 downto 0) and not ld_i(7 downto 0);
                    end if;
                    if lbe_i(1) = '1' then
                        istat(15 downto 8) <= istat(15 downto 8) and not ld_i(15 downto 8);
                    end if;
                    if lbe_i(2) = '1' then
                        istat(23 downto 16) <= istat(23 downto 16) and not ld_i(23 downto 16);
                    end if;
                    if lbe_i(3) = '1' then
                        istat(31 downto 24) <= istat(31 downto 24) and not ld_i(31 downto 24);
                    end if;
                else
                    -- write to TEST    
                    if lbe_i(0) = '1' then
                        istat(7 downto 0) <= istat(7 downto 0) or ld_i(7 downto 0);
                    end if;
                    if lbe_i(1) = '1' then
                        istat(15 downto 8) <= istat(15 downto 8) or ld_i(15 downto 8);
                    end if;
                    if lbe_i(2) = '1' then
                        istat(23 downto 16) <= istat(23 downto 16) or ld_i(23 downto 16);
                    end if;
                    if lbe_i(3) = '1' then
                        istat(31 downto 24) <= istat(31 downto 24) or ld_i(31 downto 24);
                    end if;
                end if;
            end if;
            
            -- write to IARM
            if ds_xfer = '1' and ds_write = '1' and la_q(3 downto 2) = "10" and lbe_i(0) = '1' then
                iarm <= '1';
            else
                iarm <= '0';
            end if;

            -- Cycle counter
            if istat(0) = '0' then
                -- Allow writes to cycle counter when ISTAT[0] is clear.
                if ds_xfer = '1' and ds_write = '1' and la_q(4 downto 2) = "100" then
                    -- write to COUNT
                    if lbe_i(0) = '1' then
                        count(7 downto 0) <= ld_i(7 downto 0);
                    end if;
                    if lbe_i(1) = '1' then
                        count(15 downto 8) <= ld_i(15 downto 8);
                    end if;
                    if lbe_i(2) = '1' then
                        count(23 downto 16) <= ld_i(23 downto 16);
                    end if;
                    if lbe_i(3) = '1' then
                        count(31 downto 24) <= ld_i(31 downto 24);
                    end if;
                end if;
            else
                -- Increment the cycle counter when ISTAT[0] is 1.
                count <= count + 1;
            end if;
        end if;
    end process;

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    plxdssm0 : plxdssm
        port map(
            clk              => lclk,
            rst              => rst,
            sr               => logic0,
            qlads            => qlads,
            lblast           => lblast_i,
            lbterm           => lbterm_i,
            lwrite           => lwrite_i,
            eld_oe           => open,
            ld_oe_l          => ld_oe_l,
            lready_o_l       => lready_o_l,
            lready_oe_l      => lready_oe_l,
            lbterm_o_l       => lbterm_o_l,
            lbterm_oe_l      => lbterm_oe_l,
            idle             => open,
            transfer         => ds_xfer,
            decode           => ds_decode,
            write            => ds_write,
            ready            => logic1,
            stop             => logic1);
    
end mixed;
