--
-- test_itest.vhd - Testbench for entity 'itest'
--
-- (C) Copyright Alpha Data 2004-2005
--
-- Example testbench for:
--
--   o ADM-XRC
--   o ADM-XRC-P
--   o ADM-XRC-II-Lite
--   o ADM-XRC-II
--   o ADM-XRC-4LX
--   o ADM-XRC-4SX
--
-- NOT SYNTHESIZABLE
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library std;
use std.textio.all;

library plxsim;
use plxsim.plxsim.all;

entity test_itest is
end test_itest;

architecture behav of test_itest is

    constant period:     time     := 30 ns;

    constant priorities : integer_vector_t(0 to 1) := (1 => 1, others => 0);

    signal lclk, mclk:         std_logic := '1';
    signal lreset_l:           std_logic := '0';
    signal la:                 std_logic_vector(31 downto 2) := (others => 'Z');
    signal lbe_l:              std_logic_vector(3 downto 0) := (others => 'Z');
    signal ld:                 std_logic_vector(31 downto 0) := (others => 'Z');
    signal lwrite:             std_logic := '0';
    signal lads_l:             std_logic := '1';
    signal lblast_l:           std_logic := '1';
    signal lbterm_l:           std_logic := 'Z';
    signal lready_l:           std_logic := 'Z';
    signal lhold, lholda:      std_logic_vector(1 downto 0);
    signal fholda:             std_logic := '0';
    signal finti_l:            std_logic;

    signal bus_in0, bus_in1 : locbus_in_t;
    signal bus_out0, bus_out1 : locbus_out_t := init_locbus_out;

    shared variable imask :    std_logic_vector(31 downto 0);

    shared variable finti_edge : std_logic;

    signal logic0, logic1 : std_logic;

    procedure report_message(
        str : in string)
    is
        variable msg : line;
    begin
        write(msg, str);
        writeline(output, msg);
    end;

    component itest
        port(
            lclk           : in    std_logic;
            lreset_l       : in    std_logic; 
            lwrite         : in    std_logic;
            lads_l         : in    std_logic;
            lblast_l       : in    std_logic;
            lbterm_l       : inout std_logic;
            ld             : inout std_logic_vector(31 downto 0);
            la             : in    std_logic_vector(23 downto 2);
            lready_l       : inout std_logic; 
            lbe_l          : in    std_logic_vector(3 downto 0);
            fholda         : in    std_logic;
            finti_l        : out   std_logic);
    end component;

begin

    logic0 <= '0';
    logic1 <= '1';

    lclk <= not lclk after period / 2;

    lreset_l <= transport '1' after 50 ns;

    lads_l <= 'H';
    lblast_l <= 'H';
    lready_l <= 'H';
    lbterm_l <= 'H';
    lwrite <= 'H';

    --
    -- This process simulates the interrupt logic in the CPLD on an ADM-XRC
    -- series card that triggersan interrupt on the host CPU.
    --
    finti_monitor : process
    begin
        loop
            -- Wait for a negative edge on 'finti_l'
            wait until finti_l'event and finti_l = '0';
            report_message("ITEST finti_monitor: detected edge on finti_l");
            finti_edge := '1';
        end loop;
    end process;

    --
    -- This process simulates the interrupt handler running on the host CPU.
    --
    interrupt_handler : process
        variable data : byte_vector_t(0 to 127);
        variable be : byte_enable_t(0 to 127);
        variable n : natural;
        variable istat : std_logic_vector(31 downto 0);
    begin
        be := (others => '1');

        plxsim_wait_cycles(5, bus_in0);

        -- Set interrupt mask, to enable interrupts 0 to 31
        imask := X"00000000";
        data(0 to 3) := conv_byte_vector(imask);
        plxsim_request_bus(true, bus_in1, bus_out1);
        plxsim_write(2, true, X"0", be(0 to 3), data(0 to 3), n, bus_in1, bus_out1);
        plxsim_request_bus(false, bus_in1, bus_out1);
 
        loop
            -- Wait for next edge on finti_l
            loop
                exit when finti_edge = '1';
                plxsim_wait_cycles(1, bus_in1);
            end loop;

            -- Interrupt response time of 15 cycles; unreastically short for purposes of simulation
            plxsim_wait_cycles(15, bus_in1);

            -- Check interrupt status register
            plxsim_request_bus(true, bus_in1, bus_out1);
            plxsim_read(2, true, X"4", be(0 to 3), data(0 to 3), n, bus_in1, bus_out1);
            plxsim_request_bus(false, bus_in1, bus_out1);
            istat := conv_std_logic_vector(data(0 to 3));

            -- Mask the ISTAT value just read with the current value of the IMASK register
            istat := istat and not imask;

            -- Clear the unmasked interrupts
            data(0 to 3) := conv_byte_vector(istat);
            plxsim_request_bus(true, bus_in1, bus_out1);
            plxsim_write(2, true, X"4", be(0 to 3), data(0 to 3), n, bus_in1, bus_out1);
            plxsim_request_bus(false, bus_in1, bus_out1);
            report_message("ITEST interrupt_handler: cleared interrupt(s), masked ISTAT = 0x" & conv_string_hex(istat));

            -- Rearm the interrupt logic
            finti_edge := '0';
            data(0 to 3) := conv_byte_vector(X"deadbeef"); -- actual value does not matter for IARM register
            plxsim_request_bus(true, bus_in1, bus_out1);
            plxsim_write(2, true, X"8", be(0 to 3), data(0 to 3), n, bus_in1, bus_out1);
            plxsim_request_bus(false, bus_in1, bus_out1);

        end loop;

        wait;
    end process;

    --
    -- This process simulates some logic within the FPGA that is generating
    -- interrupts asynchronously to the interrupt handler that runs on the
    -- host CPU.
    --
    stimulate: process
        variable data : byte_vector_t(0 to 127);
        variable be : byte_enable_t(0 to 127);
        variable n : natural;
        variable val32 : std_logic_vector(31 downto 0);
    begin
        plxsim_wait_cycles(20, bus_in0);

        be := (others => '1');

        --
        -- Generate interrupts as fast as possible; the interrupt handler process
        -- (see above) may end up seeing multiple interrupts per loop.
        --
        for i in 0 to 31 loop
            val32 := (others => '0');
            val32(i) := '1';
            data(0 to 3) := conv_byte_vector(val32);
            plxsim_request_bus(true, bus_in0, bus_out0);
            plxsim_write(2, true, X"c", be(0 to 3), data(0 to 3), n, bus_in0, bus_out0);
            plxsim_request_bus(false, bus_in0, bus_out0);
        end loop;

        report_message("ITEST stimulus process: completed");

        wait;
    end process;

    UUT : itest
        port map(
            lclk           => lclk,
            lreset_l       => lreset_l,
            lwrite         => lwrite,
            lads_l         => lads_l,
            lblast_l       => lblast_l,
            lbterm_l       => lbterm_l,
            ld             => ld,
            la             => la(23 downto 2),
            lready_l       => lready_l,
            lbe_l          => lbe_l,
            fholda         => fholda,
            finti_l        => finti_l);

    arb_0 : locbus_arb
        generic map(
            n_arb       => 2,
            priority    => priorities)
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lhold       => lhold,
            lholda      => lholda);

    agent_0 : locbus_agent_nonmux
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lads_l      => lads_l,
            la          => la,
            lbe_l       => lbe_l,
            lblast_l    => lblast_l,
            ld          => ld,
            lbterm_l    => lbterm_l,
            lready_l    => lready_l,
            lwrite      => lwrite,
            lhold       => lhold(0),
            lholda      => lholda(0),
            bus_in      => bus_in0,
            bus_out     => bus_out0);

    agent_1 : locbus_agent_nonmux
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lads_l      => lads_l,
            la          => la,
            lbe_l       => lbe_l,
            lblast_l    => lblast_l,
            ld          => ld,
            lbterm_l    => lbterm_l,
            lready_l    => lready_l,
            lwrite      => lwrite,
            lhold       => lhold(1),
            lholda      => lholda(1),
            bus_in      => bus_in1,
            bus_out     => bus_out1);

    lbpcheck_0 : lbpcheck
        generic map(
            multiplexed => false,
            wide        => false)
        port map(
            lclk        => lclk,
            lreset_l    => lreset_l,
            lads_l      => lads_l,
            l64_l       => lads_l,
            la          => la,
            lad_lo      => ld,
            lad_hi      => ld,
            lbe_lo_l    => lbe_l,
            lbe_hi_l    => lbe_l,
            lwrite      => lwrite,
            lblast_l    => lblast_l,
            lready_l    => lready_l,
            lbterm_l    => lbterm_l);

end behav;
