--
-- master-xrc2l.vhd - Direct Master example FPGA design
--
-- (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2007
--
-- Example program for:
--
--   o ADM-XRCII-Lite
--
-- NOTES ON OPERATION
-- ==================
--
-- LA[31:2] are connected to the FPGA.
--
-- This design implements the following registers, accessible by
-- direct slave tranfers:
--
--   Offset Type  Name     Function
--   ===================================================================
--
--   0x0    WO    ADDR     Local bus address for direct master cycles.
--
--   0x4    WO    WDATA    Data to be written in next direct master
--                         write cycle.
--
--   0x8    WO    CFG      Selects what type of direct master cycle to
--                         perform; bit fields as follows:
--
--                         [0] WRITE   1 => perform a write
--                                     0 => perform a read
--
--   0xC    RO    RDATA    Data read in last direct master read cycle.
--
--   0x10   RO    STAT     Status register; bit fields as follows:
--
--                         [0] IDLE    1 => no direct master cycle
--                                          in progress
--                                     0 => direct master cycle in
--                                          progress
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;

library work;
use work.localbus.all;

entity master is
    generic(
        la_top           : in    natural := 31);
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic;
        lwrite           : inout std_logic;
        lads_l           : inout std_logic;
        lblast_l         : inout std_logic;
        lbterm_l         : inout std_logic;
        ld               : inout std_logic_vector(31 downto 0);
        la               : inout std_logic_vector(la_top downto 2);
        lreadyi_l        : inout std_logic; 
        lbe_l            : inout std_logic_vector(3 downto 0);
        fhold            : out   std_logic; 
        fholda           : in    std_logic; 
        lreadyo_l        : in    std_logic;
        lbtermo_l        : in    std_logic);
end entity;

architecture mixed of master is
    
    signal rst : std_logic;
    signal lads_i : std_logic;
    signal lblast_i : std_logic;
    signal lbterm_i : std_logic;
    signal lbe_i : std_logic_vector(3 downto 0);
    signal lreadyo_i : std_logic;
    signal lbtermo_i : std_logic;
    signal lwrite_i : std_logic;
    signal ld_i : std_logic_vector(31 downto 0);
    signal ld_i_q : std_logic_vector(31 downto 0);
    signal la_i : std_logic_vector(la_top downto 2);
    signal lads_o_l : std_logic;
    signal lads_oe_l : std_logic;
    signal lbe_o_l : std_logic_vector(3 downto 0);
    signal lbe_oe_l : std_logic;
    signal lblast_o_l : std_logic;
    signal lblast_oe_l : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal lreadyi_o_l : std_logic;
    signal lreadyi_oe_l : std_logic;
    signal lwrite_o : std_logic;
    signal lwrite_oe_l : std_logic;
    
    signal la_oe_l : std_logic;
    signal la_o : std_logic_vector(la_top downto 2);
    signal ld_oe_l : std_logic_vector(31 downto 0);
    signal ld_o : std_logic_vector(31 downto 0);
    signal ld_o_ds : std_logic_vector(31 downto 0);
    signal ld_o_dm : std_logic_vector(31 downto 0);
    
    signal la_q : std_logic_vector(la_top downto 2);
    
    signal ds_qlads : std_logic;
    signal ds_xfer : std_logic;
    signal ds_decode : std_logic;
    signal ds_write : std_logic;
    signal ds_eld_oe : std_logic;
    
    signal reg_addr : std_logic_vector(31 downto 0);
    signal reg_wdata : std_logic_vector(31 downto 0);
    signal reg_cfg : std_logic_vector(31 downto 0);
    signal reg_rdata : std_logic_vector(31 downto 0);
    signal reg_stat : std_logic_vector(31 downto 0);

    signal dm_req : std_logic;
    signal dm_xfer : std_logic;
    signal dm_eld_oe : std_logic;
    signal dm_done : std_logic;

    signal logic0, logic1 : std_logic;
    
begin
    
    -- -------------------------------------------------------------------
    -- Common
    -- -------------------------------------------------------------------

    logic0 <= '0';
    logic1 <= '1';

    --    
    -- Convert inputs to active high.
    --

    rst <= not lreset_l;
    ld_i <= ld;

    --
    -- Register local bus data on the way in.
    --

    gen_ld_i_q: process(lclk, rst)
    begin
        if rst = '1' then
            ld_i_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            ld_i_q <= ld_i;
        end if;
    end process;

    --
    -- Generate 'ld_o', the data for the LD pins, depending on whether the
    -- current local bus transfer a direct slave transfer or a direct master
    -- transfer.
    --

    generate_ld_o : process(rst, lclk)
    begin
        if rst = '1' then
            ld_o <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if ds_eld_oe = '1' then
                ld_o <= ld_o_ds;
            else
                ld_o <= ld_o_dm;
            end if;
        end if;
    end process;

    --
    -- Drive the local data bus when responding to a direct slave read cycle
    -- or direct master write cycle.
    --

    gen_ld_oe_l : process(lclk, rst)
    begin
        if rst = '1' then
            ld_oe_l <= (others => '1');
        elsif lclk'event and lclk = '1' then
            ld_oe_l <= (others => not (ds_eld_oe or dm_eld_oe));
        end if;
    end process;

    gen_ld : for i in ld'range generate
        ld(i) <= ld_o(i) when ld_oe_l(i) = '0' else 'Z';
    end generate;

    -- -------------------------------------------------------------------
    -- Direct Slave
    -- -------------------------------------------------------------------

    --
    -- Convert inputs to active high.
    --

    la_i <= la;
    lads_i <= not lads_l;
    lbe_i <= not lbe_l;
    lblast_i <= not lblast_l;
    lbterm_i <= not lbterm_l;
    lwrite_i <= lwrite;

    --
    -- 'ds_qlads' is 'lads_l' qualified as a direct slave access to the FPGA.
    --

    ds_qlads <= lads_i and not la_i(23) and not fholda;

    --
    -- Latch local bus address on direct slave transfers.
    --

    latch_la : process(lclk, rst)
    begin
        if rst = '1' then
            la_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
            end if;
        end if;
    end process;

    --
    -- Drive the bus when responding to a direct slave cycle.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';
    lreadyi_l <= lreadyi_o_l when lreadyi_oe_l = '0' else 'Z';

    --
    -- Implement the registers that are writable via direct slave cycles.
    --

    implement_registers : process(lclk, rst)
    begin
        if rst = '1' then
            dm_req    <= '0';
            reg_addr  <= (others => '0');
            reg_wdata <= (others => '0');
            reg_cfg   <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if ds_xfer = '1' and ds_write = '1' and la_q(4 downto 2) = "000" then
                --
                -- Write to ADDR register
                --
                if lbe_i(0) = '1' then
                    reg_addr(7 downto 2) <= ld_i(7 downto 2);
                end if;
                if lbe_i(1) = '1' then
                    reg_addr(15 downto 8) <= ld_i(15 downto 8);
                end if;
                if lbe_i(2) = '1' then
                    reg_addr(23 downto 16) <= ld_i(23 downto 16);
                end if;
                if lbe_i(3) = '1' then
                    reg_addr(31 downto 24) <= ld_i(31 downto 24);
                end if;
            end if;

            --
            -- Writing to byte 3 of ADDR register also triggers a
            -- direct master cycle.
            --
            if ds_xfer = '1' and ds_write = '1' and la_q(4 downto 2) = "000" and lbe_i(3) = '1' then
                dm_req <= '1';
            else
                dm_req <= '0';
            end if;
            
            if ds_xfer = '1' and ds_write = '1' and la_q(4 downto 2) = "001" then
                --
                -- Write to WDATA register.
                --
                if lbe_i(0) = '1' then
                    reg_wdata(7 downto 0) <= ld_i(7 downto 0);
                end if;
                if lbe_i(1) = '1' then
                    reg_wdata(15 downto 8) <= ld_i(15 downto 8);
                end if;
                if lbe_i(2) = '1' then
                    reg_wdata(23 downto 16) <= ld_i(23 downto 16);
                end if;
                if lbe_i(3) = '1' then
                    reg_wdata(31 downto 24) <= ld_i(31 downto 24);
                end if;
            end if;
            
            if ds_xfer = '1' and ds_write = '1' and la_q(4 downto 2) = "010" then
                --
                -- Write to CFG register
                --
                if lbe_i(0) = '1' then
                    reg_cfg(0) <= ld_i(0);
                end if;
            end if;
        end if;
    end process;

    --
    -- Implement the registers that are readable via direct slave cycles.
    --

    implment_reg_rdata : process(rst, lclk)
    begin
        if rst = '1' then
            reg_rdata <= X"deadbeef";
        elsif lclk'event and lclk = '1' then
            if dm_xfer = '1' then
                reg_rdata <= ld_i_q;
            end if;
        end if;
    end process;

    reg_stat(0)           <= dm_done;
    reg_stat(31 downto 1) <= (others => '0');

    --
    -- Generate 'ld_o_ds', the data for the LD pins when current bus cycle
    -- is a direct slave transfer.
    --

    generate_ld_o_ds : process(
        la_q,
        reg_rdata,
        reg_stat)
    begin
        case la_q(4 downto 2) is
            when "011" =>
                ld_o_ds <= reg_rdata;
            when "100" =>
                ld_o_ds <= reg_stat;
            when others =>
                ld_o_ds <= (others => '-');
        end case;
    end process;

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    plxdssm0 : plxdssm
        port map(
            clk => lclk,
            rst => rst,
            sr => logic0,
            qlads => ds_qlads,
            lblast => lblast_i,
            lbterm => lbterm_i,
            lwrite => lwrite_i,
            eld_oe => ds_eld_oe,
            ld_oe_l => open,
            lready_o_l => lreadyi_o_l,
            lready_oe_l => lreadyi_oe_l,
            lbterm_o_l => lbterm_o_l,
            lbterm_oe_l => lbterm_oe_l,
            idle => open,
            transfer => ds_xfer,
            decode => ds_decode,
            write => ds_write,
            ready => logic1,
            stop => logic1); 

    -- -------------------------------------------------------------------
    -- Direct Master
    -- -------------------------------------------------------------------

    --
    -- Convert inputs to active high.
    --

    lreadyo_i <= not lreadyo_l;
    lbtermo_i <= not lbtermo_l;

    --
    -- Drive the bus when performing a direct master cycle.
    --

    la <= la_o when la_oe_l = '0' else (others => 'Z');
    lbe_l <= lbe_o_l when lbe_oe_l = '0' else (others => 'Z');
    lads_l <= lads_o_l when lads_oe_l = '0' else 'Z';
    lblast_l <= lblast_o_l when lblast_oe_l = '0' else 'Z';
    lwrite <= lwrite_o when lwrite_oe_l = '0' else 'Z';

    --
    -- Generate the address and byte enables to be output onto the local bus.
    --

    la_o  <= EXT("11111111" & reg_addr(23 downto 2), la_top - 1);
    lbe_o_l <= "0000";

    --
    -- There is currently only one source of data for direct master writes.
    --

    ld_o_dm <= reg_wdata;

    --
    -- Instantiate the direct master state machine; performs local bus
    -- cycles as a master.
    --

    plxdmsm0 : plxdmsm
        port map(
            clk => lclk,
            rst => rst,
            sr => logic0,
            request => dm_req,
            stop => logic1,
            transfer => dm_xfer,
            done => dm_done,
            write => reg_cfg(0),
            lready => lreadyo_i,
            lbterm => lbtermo_i,
            fholda => fholda,
            fhold => fhold,
            la_oe_l => la_oe_l,
            lads_o_l => lads_o_l,
            lads_oe_l => lads_oe_l,
            lbe_oe_l => lbe_oe_l,
            lblast_o_l => lblast_o_l,
            lblast_oe_l => lblast_oe_l,
            eld_oe => dm_eld_oe,
            ld_oe_l => open,
            lwrite => lwrite_o,
            lwrite_oe_l => lwrite_oe_l);

end mixed;
