--
-- user_app.vhd - user application code
--
-- This module is an implementation of the 'user_app' component that can be
-- commanded to perform a memory test of some or all of the memory banks.
--
-- SYNTHESIZABLE
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_arith.all;

library work;
use work.memif.all;
use work.memory_common.all;
use work.user_defs.all;

entity user_app is
    port(
        rst           : in    std_logic;   -- Reset from memory clock domain
        clk           : in    std_logic;   -- Clock from memory clock domain
        -- To/from local bus interface
        reg_in        : in    std_logic_vector(31 downto 0);   -- Incoming register write values
        reg_wr        : in    std_logic_vector(255 downto 0);  -- Byte write enables for 'reg_in'
        reg_out       : out   std_logic_vector(2047 downto 0); -- Outgoing register values
        -- To/from memory banks
        valid         : in    control_vector_t(max_num_bank - 1 downto 0);
        q             : in    data_vector_t(max_num_bank - 1 downto 0);
        qtag          : in    tag_vector_t(max_num_bank - 1 downto 0);
        ready         : in    control_vector_t(max_num_bank - 1 downto 0);
        req           : out   control_vector_t(max_num_bank - 1 downto 0);
        ce            : out   control_vector_t(max_num_bank - 1 downto 0);
        w             : out   control_vector_t(max_num_bank - 1 downto 0);
        a             : out   address_vector_t(max_num_bank - 1 downto 0);
        tag           : out   tag_vector_t(max_num_bank - 1 downto 0);
        d             : out   data_vector_t(max_num_bank - 1 downto 0);
        be            : out   be_vector_t(max_num_bank - 1 downto 0));
end entity;

architecture mixed of user_app is

    type user_vector_t is array(0 to 63) of std_logic_vector(31 downto 0);
    signal user_out : user_vector_t;

    type user_wr_vector_t is array(0 to 63) of std_logic_vector(3 downto 0);
    signal user_wr : user_wr_vector_t;

    type offset_vector_t is array(work.user_defs.bank_width'range) of std_logic_vector(31 downto 0);
    signal offset, eaddr : offset_vector_t;

    type length_vector_t is array(work.user_defs.bank_width'range) of std_logic_vector(31 downto 0);
    signal length : length_vector_t;

    type ephase_vector_t is array(0 to 15) of std_logic_vector(7 downto 0);
    signal ephase : ephase_vector_t;

    signal go : std_logic_vector(max_num_bank - 1 downto 0);
    signal done : std_logic_vector(max_num_bank - 1 downto 0);
    signal error : std_logic_vector(max_num_bank - 1 downto 0);

begin

    -- Divide 'reg_wr' into an array of 64 x 4-bit vectors for easy access
    gen_user_in : for i in 0 to 63 generate
        user_wr(i) <= reg_wr(4 * i + 3 downto 4 * i);
    end generate;

    gen_memory_test : for i in 0 to max_num_bank - 1 generate
        -- 'Go' command for bank 'i' is bit 'i' of USER63
        go(i) <= user_wr(63)(i / 8) and reg_in(i);

        -- Fora real memory bank that exists on the card, generate a 'memory_test' instance
        gen_real : if bank_present(i) generate
            -- Starting address registers are USER0-USER15
            offset_reg : process(rst, clk)
            begin
                if rst = '1' then
                    -- Default test offset is 0
                    offset(i) <= X"00000000";
                elsif clk'event and clk = '1' then
                    for j in 0 to 3 loop
                        if user_wr(i)(j) = '1' then
                            offset(i)(8 * j + 7 downto 8 * j) <= reg_in(8 * j + 7 downto 8 * j);
                        end if;
                    end loop;
                end if;
            end process;

            -- Length registers are USER16-USER31
            length_reg : process(rst, clk)
            begin
                if rst = '1' then
                    -- Default test length is 0x40 words
                    length(i) <= X"0000003f";
                elsif clk'event and clk = '1' then
                    for j in 0 to 3 loop
                        if user_wr(i + 16)(j) = '1' then
                            length(i)(8 * j + 7 downto 8 * j) <= reg_in(8 * j + 7 downto 8 * j);
                        end if;
                    end loop;
                end if;
            end process;

            U0 : memory_test
                generic map(
                    a_width => max_address_width,
                    d_width => bank_width(i),
                    tag_width => tag_width,
                    rd_width => rd_width(i))
                port map(
                    rst => rst,
                    clk => clk,
                    go => go(i),
                    offset => offset(i),
                    length => length(i),
                    done => done(i),
                    error => error(i),
                    eaddr => eaddr(i),
                    ephase => ephase(i),
                    valid => valid(i),
                    q => q(i)(bank_width(i) - 1 downto 0),
                    qtag => qtag(i),
                    ready => ready(i),
                    req => req(i),
                    ce => ce(i),
                    w => w(i),
                    a => a(i),
                    tag => tag(i),
                    d => d(i)(bank_width(i) - 1 downto 0),
                    be => be(i)(bank_be_width(i) - 1 downto 0));

            -- Tie off unused data bits, if necessary
            d(i)(max_data_width - 1 downto bank_width(i)) <= (others => '-');

            -- Tie off unused byte enables, if necessary
            be(i)(max_be_width - 1 downto bank_be_width(i)) <= (others => '-');

            -- Allow readback of 'offset' registers
            user_out(i) <= offset(i);

            -- Allow readback of 'length' registers
            user_out(i + 16) <= length(i);

            -- Allow read of 'eaddr' registers
            user_out(i + 32) <= eaddr(i);
        end generate;

        -- Generate dummy code for nonexistent memory banks
        gen_dummy : if not bank_present(i) generate
            req(i) <= '0';
            ce(i) <= '0';
            w(i) <= '-';
            a(i) <= (others => '-');
            tag(i) <= (others => '-');
            d(i) <= (others => '-');
            be(i) <= (others => '-');
            done(i) <= '1';
            error(i) <= '0';

            -- Reading 'offset' register returns undefined
            user_out(i) <= (others => '-');

            -- Reading 'length' register returns undefined
            user_out(i + 16) <= (others => '-');

            -- Reading 'eaddr' register returns undefined
            user_out(i + 32) <= (others => '-');

            -- 'ephase' is not meaningful for nonexistent banks
            ephase(i) <= (others => '-');
        end generate;
    end generate;

    -- USER[51:48] give 'ephase' for all banks
    output_ephase : for i in 0 to 3 generate
        word : for j in 0 to 3 generate
            user_out(48 + i)(8 * j + 7 downto 8 * j) <= ephase(4 * i + j);
        end generate;
    end generate;

    -- Allow status of memory tests to be read
    user_out(62) <= (others => '0');
    user_out(63) <= error & done;

    -- Unused regisers return undefined
    user_out(52 to 61) <= (others => (others => '-'));

    -- Combine the 64 32-bit registers of 'user_out' into the vector 'reg_out'
    gen_reg_out : for i in 0 to 63 generate
        reg_out(32 * i + 31 downto 32 * i) <= user_out(i);
    end generate;

end architecture;
