--
-- memclk-xpl.vhd - generate memory interface clocks, targetting ADM-XPL
--
-- This module produces 4 phases (0, 90, 180 and 270) of MCLK,
-- which is expected to be in the region of 66MHz, for use in
-- the memory ports.
--
-- It also generates a deskewed clock for the ZBT SRAM bank.
--
-- SYNTHESIZABLE
--
-- Targets models:
--
--   o ADM-XPL
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity memclk is
    port(
        rst           : in    std_logic;
        mclk          : in    std_logic;
        refclk125     : in    std_logic;
        memclk0       : out   std_logic;
        memclk90      : out   std_logic;
        memclk180     : out   std_logic;
        memclk270     : out   std_logic;
        ramclko       : out   std_logic_vector(0 downto 0);
        ramclki       : in    std_logic_vector(0 downto 0);
        locked        : out   std_logic_vector(1 downto 0);
        memrst        : out   std_logic);
end entity;

architecture struct of memclk is

    signal dllrst : std_logic;
    signal rstcnt : std_logic_vector(3 downto 0);

    signal umemclk0, bmemclk0 : std_logic;
    signal umemclk90, bmemclk90 : std_logic;
    signal umemclk180, bmemclk180 : std_logic;
    signal umemclk270, bmemclk270 : std_logic;
    signal locked_memclk : std_logic;

    signal rst_ramclk  : std_logic;
    signal locked_ramclk  : std_logic;
    signal ramclki_ibufg  : std_logic_vector(0 downto 0);

    signal memrst0 : std_logic_vector(7 downto 0);

    signal logic0, logic1 : std_logic;

    component DCM
        port(
            CLKIN         : in  std_logic;
            CLKFB         : in  std_logic;
            DSSEN         : in  std_logic;
            PSINCDEC      : in  std_logic;
            PSEN          : in  std_logic;
            PSCLK         : in  std_logic;
            RST           : in  std_logic;
            CLK0          : out std_logic;
            CLK90         : out std_logic;
            CLK180        : out std_logic;
            CLK270        : out std_logic;
            CLK2X         : out std_logic;
            CLK2X180      : out std_logic;
            CLKDV         : out std_logic;
            CLKFX         : out std_logic;
            CLKFX180      : out std_logic;
            LOCKED        : out std_logic;
            PSDONE        : out std_logic;
            STATUS        : out std_logic_vector(7 downto 0));
    end component;

    component BUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    component IBUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    attribute DLL_FREQUENCY_MODE : string;
    attribute DUTY_CYCLE_CORRECTION : string;
    attribute STARTUP_WAIT : string;
    attribute CLK_FEEDBACK : string;

    attribute DLL_FREQUENCY_MODE of dll_memclk : label is "LOW";
    attribute DUTY_CYCLE_CORRECTION of dll_memclk : label is "TRUE";
    attribute STARTUP_WAIT of dll_memclk : label is "FALSE";
    attribute CLK_FEEDBACK of dll_memclk : label is "1X";
    
    attribute DLL_FREQUENCY_MODE of dll_ramclk0 : label is "LOW";
    attribute DUTY_CYCLE_CORRECTION of dll_ramclk0 : label is "TRUE";
    attribute STARTUP_WAIT of dll_ramclk0 : label is "FALSE";
    attribute CLK_FEEDBACK of dll_ramclk0 : label is "1X";
    
    --
    -- If the synthesizer replicates an asynchronous reset signal due high fanout,
    -- this can prevent flip-flops being mapped into IOBs. We set the maximum
    -- fanout for such nets to a high enough value that replication never occurs.
    --
    attribute MAX_FANOUT : string;
    attribute MAX_FANOUT of memrst : signal is "1000000";

begin
    
    --
    -- Define constant values
    --
    logic0 <= '0';
    logic1 <= '1';

    locked <= locked_ramclk & locked_memclk;

    memclk0 <= bmemclk0;
    memclk90 <= bmemclk90;
    memclk180 <= bmemclk180;
    memclk270 <= bmemclk270;

    gen_memrst : process(rst, refclk125)
    begin
        if rst = '1' then
            memrst0 <= (others => '1');
            memrst <= '1';
        elsif refclk125'event and refclk125 = '1' then
            if locked_memclk = '0' or locked_ramclk = '0' then
                memrst0 <= (others => '1');
                memrst <= '1';
            else
                memrst0(0) <= '0';
                memrst0(memrst0'high downto 1) <= memrst0(memrst0'high - 1 downto 0);
                memrst <= OR_reduce(memrst0);
            end if;
        end if;
    end process;

    dll_memclk : DCM
        port map(
            CLKIN    => mclk,
            CLKFB    => bmemclk0,
            DSSEN    => logic0,
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => rst,
            CLK0     => umemclk0,
            CLK90    => umemclk90,
            CLK180   => umemclk180,
            CLK270   => umemclk270,
            LOCKED   => locked_memclk);

    bufg_memclk0 : BUFG
        port map(
            I => umemclk0,
            O => bmemclk0);

    bufg_memclk90 : BUFG
        port map(
            I => umemclk90,
            O => bmemclk90);

    bufg_memclk180 : BUFG
        port map(
            I => umemclk180,
            O => bmemclk180);

    bufg_memclk270 : BUFG
        port map(
            I => umemclk270,
            O => bmemclk270);

    --
    -- Generate the SSRAM clocks
    --
    
    rst_ramclk <= not locked_memclk;
    
    ibufg_ramclk0 : IBUFG
        port map(
            I => ramclki(0),
            O => ramclki_ibufg(0));
            
    dll_ramclk0 : DCM
        port map(
            CLKIN    => bmemclk0,
            CLKFB    => ramclki_ibufg(0),
            DSSEN    => logic0,
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => rst_ramclk,
            CLK0     => ramclko(0),
            LOCKED   => locked_ramclk);

end architecture;
