--
-- memclk-xrc2.vhd - generate memory interface clocks (Virtex-II)
--
-- This module doubles MCLK for use in the ZBT memory ports. Also performed
-- is deskew of the clocks driven to the ZBT SSRAM.
--
-- SYNTHESIZABLE
--
-- Targets models:
--
--   o ADM-XRC-II
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity memclk is
    port(
        rst           : in    std_logic;
        mclk_i        : in    std_logic;
        memclk0       : out   std_logic;
        ramclko       : out   std_logic_vector(1 downto 0);
        ramclki       : in    std_logic_vector(1 downto 0);
        locked        : out   std_logic_vector(2 downto 0);
        memrst        : out   std_logic);
end entity;

architecture struct of memclk is

    signal umclkx2, bmclkx2 : std_logic;

    signal umemclk0, bmemclk0 : std_logic;

    signal locked_mclkx2 : std_logic;

    signal ramclki_ibufg : std_logic_vector(1 downto 0);
    signal rst_ramclk : std_logic;
    signal locked_ramclk : std_logic_vector(1 downto 0);

    signal logic0, logic1 : std_logic;

    component DCM
        port(
            CLKIN         : in  std_logic;
            CLKFB         : in  std_logic;
            DSSEN         : in  std_logic;
            PSINCDEC      : in  std_logic;
            PSEN          : in  std_logic;
            PSCLK         : in  std_logic;
            RST           : in  std_logic;
            CLK0          : out std_logic;
            CLK90         : out std_logic;
            CLK180        : out std_logic;
            CLK270        : out std_logic;
            CLK2X         : out std_logic;
            CLK2X180      : out std_logic;
            CLKDV         : out std_logic;
            CLKFX         : out std_logic;
            CLKFX180      : out std_logic;
            LOCKED        : out std_logic;
            PSDONE        : out std_logic;
            STATUS        : out std_logic_vector(7 downto 0));
    end component;

    component BUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    component IBUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    attribute DLL_FREQUENCY_MODE : string;
    attribute DUTY_CYCLE_CORRECTION : string;
    attribute STARTUP_WAIT : string;
    attribute CLK_FEEDBACK : string;

    attribute DLL_FREQUENCY_MODE of dll_mclkx2 : label is "LOW";
    attribute DUTY_CYCLE_CORRECTION of dll_mclkx2 : label is "TRUE";
    attribute STARTUP_WAIT of dll_mclkx2 : label is "FALSE";
    attribute CLK_FEEDBACK of dll_mclkx2 : label is "2X";
    
    attribute DLL_FREQUENCY_MODE of dll_ramclk0 : label is "LOW";
    attribute DUTY_CYCLE_CORRECTION of dll_ramclk0 : label is "TRUE";
    attribute STARTUP_WAIT of dll_ramclk0 : label is "FALSE";
    attribute CLK_FEEDBACK of dll_ramclk0 : label is "1X";
    
    attribute DLL_FREQUENCY_MODE of dll_ramclk1 : label is "LOW";
    attribute DUTY_CYCLE_CORRECTION of dll_ramclk1 : label is "TRUE";
    attribute STARTUP_WAIT of dll_ramclk1 : label is "FALSE";
    attribute CLK_FEEDBACK of dll_ramclk1 : label is "1X";
    
    --
    -- If the synthesizer replicates an asynchronous reset signal due high fanout,
    -- this can prevent flip-flops being mapped into IOBs. We set the maximum
    -- fanout for such nets to a high enough value that replication never occurs.
    --
    attribute MAX_FANOUT : string;
    attribute MAX_FANOUT of memrst : signal is "1000000";

begin
    
    --
    -- Define constant values
    --
    logic0 <= '0';
    logic1 <= '1';

    memclk0 <= bmemclk0;

    locked <= locked_ramclk & locked_mclkx2;

    memrst <= not locked_mclkx2 or not AND_reduce(locked_ramclk);

    dll_mclkx2 : DCM
        port map(
            CLKIN    => mclk_i,
            CLKFB    => bmemclk0,
            DSSEN    => logic0,
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => rst,
            CLK0     => open,
            CLK90    => open,
            CLK180   => open,
            CLK270   => open,
            CLK2X    => umemclk0,
            CLK2x180 => open,
            LOCKED   => locked_mclkx2);

    bufg_mclkx2 : BUFG
        port map(
            I => umemclk0,
            O => bmemclk0);

    --
    -- Generate the SSRAM clocks
    --
    
    rst_ramclk <= not locked_mclkx2;
    
    ibufg_ramclk0 : IBUFG
        port map(
            I => ramclki(0),
            O => ramclki_ibufg(0));
            
    dll_ramclk0 : DCM
        port map(
            CLKIN    => bmemclk0,
            CLKFB    => ramclki_ibufg(0),
            DSSEN    => logic0,
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => rst_ramclk,
            CLK0     => ramclko(0),
            LOCKED   => locked_ramclk(0));

    ibufg_ramclk1 : IBUFG
        port map(
            I => ramclki(1),
            O => ramclki_ibufg(1));

    dll_ramclk1 : DCM
        port map(
            CLKIN    => bmemclk0,
            CLKFB    => ramclki_ibufg(1),
            DSSEN    => logic0,
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => rst_ramclk,
            CLK0     => ramclko(1),
            LOCKED   => locked_ramclk(1));

end architecture;
