--
-- memclk-xrc5tz.vhd - generate memory interface clocks (Virtex-5)
--
-- This module simply passes 'mclk' through to 'clk0', for driving the
-- SSRAM, and assumes that 'mclk' has already been buffered properly using
-- a BUFG type primitive.
--
-- It also generates 6 clocks, which are phase-matched copies of 'mclk',
-- for clocking the ZBT SSRAM devices on the card.
--
-- SYNTHESIZABLE
--
-- Targets models:
--
--   o ADM-XRC-5TZ
--
-- 'clk0' is baseline clock
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_arith.all;

library UNISIM;
use UNISIM.VCOMPONENTS.ALL;

library work;
use work.memif_def.all;

entity memclk is
    generic(
        num_clock     : in    natural);
    port(
        mclk              : in    std_logic;
        rst               : in    std_logic;
        refclk200         : in    std_logic;
        memclk0           : out   std_logic;
        ramclko           : out   std_logic_vector(num_clock - 1 downto 0);
        ramclki           : in    std_logic_vector(num_clock - 1 downto 0);
        locked            : out   std_logic_vector(2 downto 0);
        memrst            : out   std_logic);
end entity;

architecture mixed of memclk is

    component BUFG
        port(
            O : out std_ulogic;
            I : in  std_ulogic);
    end component;

    component DCM
        port(
            CLKIN         : in  std_logic;
            CLKFB         : in  std_logic;
            DSSEN         : in  std_logic;
            PSINCDEC      : in  std_logic;
            PSEN          : in  std_logic;
            PSCLK         : in  std_logic;
            RST           : in  std_logic;
            CLK0          : out std_logic;
            CLK90         : out std_logic;
            CLK180        : out std_logic;
            CLK270        : out std_logic;
            CLK2X         : out std_logic;
            CLK2X180      : out std_logic;
            CLKDV         : out std_logic;
            CLKFX         : out std_logic;
            CLKFX180      : out std_logic;
            LOCKED        : out std_logic;
            PSDONE        : out std_logic;
            STATUS        : out std_logic_vector(7 downto 0));
    end component;

    component IBUFG
        port(
            O : out std_ulogic;
            I : in  std_ulogic);
    end component;

    component ODDR
        generic(
            DDR_CLK_EDGE : in  string := "OPPOSITE_EDGE";
            INIT         : in  bit := '1';
            SRTYPE       : in  string := "SYNC");
        port(
            D1           : in  std_logic;
            D2           : in  std_logic;
            CE           : in  std_logic;
            C            : in  std_logic;
            S            : in  std_logic;
            R            : in  std_logic;
            Q            : out std_logic);
    end component;

    signal memrst0 : std_logic_vector(7 downto 0);

    signal idelay_rdy : std_logic;

    signal locked_ramclk : std_logic;
    signal locked_memclk : std_logic;
    signal ramclki_ibufg : std_logic_vector(0 downto 0);
    signal ramclk0 : std_logic;
    signal memclk0_0 : std_logic;
    signal memclk0_bufg : std_logic;
    signal ramclk0_bufg : std_logic;
    signal ramclk0_rst : std_logic;

    signal vcc : std_logic;
    signal gnd : std_logic;

    attribute DLL_FREQUENCY_MODE : string;
    attribute DUTY_CYCLE_CORRECTION : string;
    attribute STARTUP_WAIT : string;
    attribute CLK_FEEDBACK : string;
    attribute CLKIN_PERIOD : string;

    attribute DLL_FREQUENCY_MODE of dll_ramclk0 : label is "HIGH";
    attribute DUTY_CYCLE_CORRECTION of dll_ramclk0 : label is "TRUE";
    attribute STARTUP_WAIT of dll_ramclk0 : label is "FALSE";
    attribute CLK_FEEDBACK of dll_ramclk0 : label is "1X";
    attribute CLKIN_PERIOD of dll_ramclk0 : label is "5ns";

    attribute DLL_FREQUENCY_MODE of dll_memclk0 : label is "HIGH";
    attribute DUTY_CYCLE_CORRECTION of dll_memclk0 : label is "TRUE";
    attribute STARTUP_WAIT of dll_memclk0 : label is "FALSE";
    attribute CLK_FEEDBACK of dll_memclk0 : label is "1X";
--    attribute CLKIN_PERIOD of dll_memclk0 : label is "5ns";

    --
    -- If the synthesizer replicates an asynchronous reset signal due high fanout,
    -- this can prevent flip-flops being mapped into IOBs. We set the maximum
    -- fanout for such nets to a high enough value that replication never occurs.
    --
    attribute MAX_FANOUT : string;
    attribute MAX_FANOUT of memrst : signal is "1000000";

begin

    vcc <= '1';
    gnd <= '0';

    memclk0 <= memclk0_bufg;

    locked <= idelay_rdy & locked_ramclk & locked_memclk;
	    
    dll_memclk0 : DCM
        port map(
            CLKIN    => mclk,
            CLKFB    => memclk0_bufg,
            DSSEN    => gnd,
            PSINCDEC => gnd,
            PSEN     => gnd,
            PSCLK    => gnd,
            RST      => rst,
            CLK0     => memclk0_0,
            LOCKED   => locked_memclk);

    bufg_memclk0 : BUFG
        port map(
            I => memclk0_0,
            O => memclk0_bufg);

    gen_memrst : process(rst, refclk200)
    begin
        if rst = '1' then
            memrst0 <= (others => '1');
            memrst <= '1';
        elsif refclk200'event and refclk200 = '1' then
            if idelay_rdy = '0' or locked_ramclk = '0' or locked_memclk = '0' then
                memrst0 <= (others => '1');
                memrst <= '1';
            else
                memrst0(0) <= '0';
                memrst0(memrst0'high downto 1) <= memrst0(memrst0'high - 1 downto 0);
                memrst <= OR_reduce(memrst0);
            end if;
        end if;
    end process;

    --
    -- Generate the SSRAM clocks
    --
    
    ibufg_ramclk0 : IBUFG
        port map(
            I => ramclki(0),
            O => ramclki_ibufg(0));
            
    ramclk0_rst <= not locked_memclk;

    dll_ramclk0 : DCM
        port map(
            CLKIN    => mclk,
            CLKFB    => ramclki_ibufg(0),
            DSSEN    => gnd,
            PSINCDEC => gnd,
            PSEN     => gnd,
            PSCLK    => gnd,
            RST      => ramclk0_rst,
            CLK0     => ramclk0,
            LOCKED   => locked_ramclk);

    bufg_ramclk0 : BUFG
        port map(
            I => ramclk0,
            O => ramclk0_bufg);

    gen_ramclko : for i in 0 to num_clock - 1 generate
        U0 : ODDR
            port map(
                D1 => vcc,
                D2 => gnd,
                CE => vcc,
                C => ramclk0_bufg,
                R => gnd,
                S => gnd,
                Q => ramclko(i));
    end generate;

    --------------------------------------------------------------------------------------------------------------------
    -- IDELAYCTRL instantiation
    --------------------------------------------------------------------------------------------------------------------

    idelay_ctrl0 : IDELAYCTRL
        port map(
            REFCLK => refclk200,
            RST => rst,
            RDY => idelay_rdy);

end architecture;
