--
-- lclk_dll.vhd - generate local bus clock for distribution within FPGA
--                using a DLL. Also generates 2X copy of local bus clock
--                for memory interfaces.
--
-- SYNTHESIZABLE
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity lclk_dll is
    port(
        rst           : in    std_logic;
        lclk          : in    std_logic;
        mclk          : in    std_logic;
        clk           : out   std_logic;
        clk2x         : out   std_logic;
        locked        : out   std_logic);
end entity;

architecture struct of lclk_dll is

    signal count : std_logic_vector(4 downto 0);
    signal dllrst : std_logic;

    signal lclk_ibufg : std_logic;
    signal lclk_dll : std_logic;
    signal lclk_bufg : std_logic;
    signal clk2x_dll : std_logic;
    signal clk2x_bufg : std_logic;

    component BUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    component CLKDLL 
        port(
            CLKIN         : in  std_logic; 
            CLKFB         : in  std_logic; 
            RST           : in  std_logic; 
            CLK0          : out std_logic;
            CLK90         : out std_logic;
            CLK180        : out std_logic;
            CLK270        : out std_logic;
            CLK2X         : out std_logic;
            CLKDV         : out std_logic;
            LOCKED        : out std_logic);
    end component;
    
    component IBUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    attribute DLL_FREQUENCY_MODE : string;
    attribute DUTY_CYCLE_CORRECTION : string;
    attribute STARTUP_WAIT : string;
    attribute CLK_FEEDBACK : string;

    attribute DLL_FREQUENCY_MODE of dll_lclk : label is "LOW";
    attribute DUTY_CYCLE_CORRECTION of dll_lclk : label is "TRUE";
    attribute STARTUP_WAIT of dll_lclk : label is "FALSE";
    attribute CLK_FEEDBACK of dll_lclk : label is "1X";

begin
    
    clk <= lclk_bufg;
    clk2x <= clk2x_bufg;

    gen_count : process(rst, mclk)
    begin
        if rst = '1' then
            count <= (others => '0');
        elsif mclk'event and mclk = '1' then
            if count(count'high) = '0' then
                count <= count + 1;
            end if;
        end if;
    end process;

    dllrst <= count(count'high - 1);

    ibufg_lclk : IBUFG
        port map(
            I => lclk,
            O => lclk_ibufg);

    dll_lclk : CLKDLL
        port map(
            CLKIN  => lclk_ibufg, 
            CLKFB  => lclk_bufg,
            RST    => dllrst,
            CLK0   => lclk_dll,
            CLK90  => open,
            CLK180 => open,
            CLK270 => open,
            CLK2X  => clk2x_dll,
            CLKDV  => open,
            LOCKED => locked);

    bufg_lclk : BUFG
        port map(
            I => lclk_dll,
            O => lclk_bufg);

    bufg_clk2x : BUFG
        port map(
            I => clk2x_dll,
            O => clk2x_bufg);

end architecture;
