--
-- memclk-xrc4fc.vhd - generate memory interface clocks (Virtex-4)
--
-- This module produces 4 clocks to drive DDR-II SDRAM ports.
--
-- SYNTHESIZABLE
--
-- Targets models:
--
--   o ADM-XRC-4FX
--
-- 'clk0' is baseline clock ('clk2x' divided by 2)
-- 'clk45' is baseline clock phase shifted by +45 degrees ('clk2x90' divided by 2)
-- 'clk2x' is 'clk0' multiplied by 2
-- 'clk2x90' is 'clk2x' phase shifted by +90 degrees
--
-- NOTE: A problem with simulations of PMCD primitives may prevent this module
--       from producing clocks with the correct phase relationship if the UNISIM
--       library from an ISE version prior to 10.1i is used.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_arith.all;

library UNISIM;
use UNISIM.VCOMPONENTS.ALL;

library work;
use work.memif_def.all;

entity memclk is
port (
      mclk              : in  std_logic;
      rst               : in  std_logic;
      refclk200         : in  std_logic;
      clk0              : out std_logic;
      clk45             : out std_logic;
      clk2x0            : out std_logic;
      clk2x90           : out std_logic;
      locked            : out std_logic_vector(1 downto 0);
      memrst            : out std_logic);
end entity;

architecture mixed of memclk is

    component BUFG
        port(
            O : out std_ulogic;
            I : in  std_ulogic);
    end component;

    component DCM_BASE
        generic(
            DLL_FREQUENCY_MODE : string := "HIGH";
            CLKDV_DIVIDE : real := 2.0;
            DUTY_CYCLE_CORRECTION : boolean := TRUE;
            STARTUP_WAIT : boolean := FALSE);
        port(
            CLK0 : out std_ulogic;
            CLK180 : out std_ulogic;
            CLK270 : out std_ulogic;
            CLK2X : out std_ulogic;
            CLK2X180 : out std_ulogic;
            CLK90 : out std_ulogic;
            CLKDV : out std_ulogic;
            CLKFX : out std_ulogic;
            CLKFX180 : out std_ulogic;
            LOCKED : out std_ulogic;
            CLKFB : in std_ulogic;
            CLKIN : in std_ulogic;
            RST : in std_ulogic);
    end component;

    component PMCD
        generic(
            EN_REL : boolean := TRUE;
            RST_DEASSERT_CLK :string := "CLKA");
        port(
            CLKA1 : out std_ulogic;
            CLKA1D2 : out std_ulogic;
            CLKA1D4 : out std_ulogic;
            CLKA1D8 : out std_ulogic;
            CLKB1 : out std_ulogic;
            CLKC1 : out std_ulogic;
            CLKD1 :out std_ulogic;
            CLKA : in std_ulogic;
            CLKB : in std_ulogic;
            CLKC : in std_ulogic;
            CLKD : in std_ulogic;
            REL : in std_ulogic;
            RST : in std_ulogic);
    end component;

    signal memrst0 : std_logic_vector(7 downto 0);

    signal mclk0_dcm : std_logic;
    signal mclk90_dcm : std_logic;
    signal mclkdiv0_dcm : std_logic;

    signal mclk0_pmcd : std_logic;
    signal mclkdiv0_pmcd : std_logic;
    signal mclk90_pmcd : std_logic;
    signal mclkdiv90_pmcd : std_logic;

    signal mclk0_bufg : std_logic;
    signal mclkdiv0_bufg : std_logic;
    signal mclk90_bufg : std_logic;
    signal mclkdiv90_bufg : std_logic;

    signal dcm0_locked : std_logic;
    signal idelayctrl_rdy : std_logic;

    signal vcc : std_logic;
    signal gnd : std_logic;

    attribute DLL_FREQUENCY_MODE : string;
    attribute CLKDV_DIVIDE : real;
    attribute DUTY_CYCLE_CORRECTION : boolean;
    attribute STARTUP_WAIT : boolean;
    attribute DLL_FREQUENCY_MODE of dcm_0 : label is "HIGH";
    attribute CLKDV_DIVIDE of dcm_0 : label is 2.0;
    attribute DUTY_CYCLE_CORRECTION of dcm_0 : label is true;
    attribute STARTUP_WAIT of dcm_0 : label is false;

    attribute EN_REL : boolean;
    attribute RST_DEASSERT_CLK : string;
    attribute EN_REL of pmcd_0 : label is false;
    attribute RST_DEASSERT_CLK of pmcd_0 : label is "CLKB";

    --
    -- If the synthesizer replicates an asynchronous reset signal due high fanout,
    -- this can prevent flip-flops being mapped into IOBs. We set the maximum
    -- fanout for such nets to a high enough value that replication never occurs.
    --
    attribute MAX_FANOUT : string;
    attribute MAX_FANOUT of memrst : signal is "1000000";

begin

    vcc <= '1';
    gnd <= '0';

    -- Use the line immediately below with the UNISIM library from an ISE version prior to 10.1i
    -- clk0 <= not mclkdiv90_bufg when work.memif_def.memif_simulation else mclkdiv90_bufg;
    clk0 <= mclkdiv90_bufg;
    clk45 <= mclkdiv0_bufg;
    clk2x0 <= mclk90_bufg;
    clk2x90 <= mclk0_bufg;

    locked <= idelayctrl_rdy & dcm0_locked;

    gen_memrst : process(rst, refclk200)
    begin
        if rst = '1' then
            memrst0 <= (others => '1');
            memrst <= '1';
        elsif refclk200'event and refclk200 = '1' then
            if dcm0_locked = '0' or idelayctrl_rdy = '0' then
                memrst0 <= (others => '1');
                memrst <= '1';
            else
                memrst0(0) <= '0';
                memrst0(memrst0'high downto 1) <= memrst0(memrst0'high - 1 downto 0);
                memrst <= OR_reduce(memrst0);
            end if;
        end if;
    end process;

    dcm_0 : DCM_BASE
        generic map(
            DLL_FREQUENCY_MODE => "HIGH",
            CLKDV_DIVIDE => 2.0,
            DUTY_CYCLE_CORRECTION => TRUE,
            STARTUP_WAIT => FALSE)
        port map(
            CLK0  => mclk0_dcm,
            CLK90 => mclk90_dcm,
            CLK180 => open,
            CLK270 => open,
            CLK2X => open,
            CLK2X180 => open,
            CLKDV => mclkdiv0_dcm,
            CLKFX => open,
            CLKFX180 => open,
            LOCKED => dcm0_locked,
            CLKFB => mclk0_bufg,
            CLKIN => mclk,
            RST => rst);

    pmcd_0 : PMCD
        generic map(
            EN_REL => FALSE,
            RST_DEASSERT_CLK => "CLKB")
        port map(
            CLKA1   => mclk90_pmcd,
            CLKA1D2 => mclkdiv90_pmcd,
            CLKA1D4 => open,
            CLKA1D8 => open,
            CLKB1   => mclk0_pmcd,
            CLKC1   => mclkdiv0_pmcd,
            CLKD1   => open,
            CLKA    => mclk90_dcm,
            CLKB    => mclk0_dcm,
            CLKC    => mclkdiv0_dcm,
            CLKD    => gnd,
            REL     => dcm0_locked,
            RST     => rst);

    bufg_mclk0 : BUFG
        port map(
            I => mclk0_pmcd,
            O => mclk0_bufg);

    bufg_mclkdiv0 : BUFG
        port map(
            I => mclkdiv0_pmcd,
            O => mclkdiv0_bufg);

    bufg_mclk90 : BUFG
        port map(
            I => mclk90_pmcd,
            O => mclk90_bufg);

    bufg_mclkdiv90 : BUFG
        port map(
            I => mclkdiv90_pmcd,
            O => mclkdiv90_bufg);

    ---------------------------------------------------------------------------------------
    -- IDELAYCTRL instantiation
    ---------------------------------------------------------------------------------------

    idelay_ctrl0 : IDELAYCTRL
        port map(
            REFCLK => refclk200,
            RST => rst,
            RDY => idelayctrl_rdy);

end architecture;
