--
-- simple64-xrce4fx.vhd - Trivial design demonstrating how to implement
--                       registers in the FPGA, with 64-bit local bus.
--
-- (C) Copyright Alpha Data Parallel Systems Ltd. 2007
--
-- Example program for:
--
--   o ADPE-XRC-4FX
--
-- Implements two 64 bit registers, "reg0" and "reg1". When reg0 is read
-- back it, it returns nibble reversed data in its low 32-bits, and returns
-- normal data in its high 32 bits. When reg1 is read back, it returns
-- normal data.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

library work;
use work.localbus.all;

entity simple64 is
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic; 
        lwrite           : in    std_logic;
        lads_l           : in    std_logic;
        lblast_l         : in    std_logic;
        lbterm_l         : inout std_logic;
        lad              : inout std_logic_vector(63 downto 0);
        lready_l         : out   std_logic; 
        lbe_l            : in    std_logic_vector(7 downto 0);
        fholda           : in    std_logic;
        front_rx_p       : in    std_logic_vector(7 downto 0);
        front_rx_n       : in    std_logic_vector(7 downto 0);
        front_tx_p       : out   std_logic_vector(7 downto 0);
        front_tx_n       : out   std_logic_vector(7 downto 0);
        local_rx_p       : in    std_logic_vector(7 downto 0);
        local_rx_n       : in    std_logic_vector(7 downto 0);
        local_tx_p       : out   std_logic_vector(7 downto 0);
        local_tx_n       : out   std_logic_vector(7 downto 0));
end entity;

architecture mixed of simple64 is
    
    component IBUFGDS
        port(
            I                : in    std_logic;
            IB               : in    std_logic;
            O                : out   std_logic);
    end component;

    component BUFG
        port(
            I                : in    std_logic;
            O                : out   std_logic);
    end component;

    component null_mgt
        port(
            clka             : in    std_logic;
            rxa_p            : in    std_logic_vector(7 downto 0);
            rxa_n            : in    std_logic_vector(7 downto 0);
            txa_p            : out   std_logic_vector(7 downto 0);
            txa_n            : out   std_logic_vector(7 downto 0);
            clkb             : in    std_logic;
            rxb_p            : in    std_logic_vector(7 downto 0);
            rxb_n            : in    std_logic_vector(7 downto 0);
            txb_p            : out   std_logic_vector(7 downto 0);
            txb_n            : out   std_logic_vector(7 downto 0));
    end component;

    signal rst : std_logic;
    signal lblast_i : std_logic;
    signal lads_i : std_logic;
    signal lwrite_i : std_logic;
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal lbterm_i : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal la_i : std_logic_vector(23 downto 2);
    signal ld_o : std_logic_vector(63 downto 0);
    signal ld_i : std_logic_vector(63 downto 0);
    signal ld_oe_l : std_logic;
    signal lbe_i : std_logic_vector(7 downto 0);

    signal qlads : std_logic;
    signal ds_xfer : std_logic;
    signal ds_decode : std_logic;    
    signal ds_write : std_logic;    

    signal la_q : std_logic_vector(23 downto 2);
    
    signal reg0 : std_logic_vector(31 downto 0);
    signal reg0_rev : std_logic_vector(31 downto 0);
    signal reg1 : std_logic_vector(31 downto 0);
    
    signal mclk_ibufg : std_logic;
    signal mclk_bufg : std_logic;

    signal logic0, logic1 : std_logic;
    
begin

    logic0 <= '0';
    logic1 <= '1';

    --
    -- Convert the inputs to active high.
    --	

    rst <= not lreset_l;
    lblast_i <= not lblast_l;
    lbterm_i <= not lbterm_l;
    lads_i <= not lads_l;
    lwrite_i <= lwrite;
    la_i <= lad(la_i'range);
    ld_i <= lad;
    lbe_i <= not lbe_l;
    
    --
    -- Generate a qualified version of 'lads_l', which is
    -- asserted when the FPGA is addressed AND the FPGA is
    -- not the local bus master.
    --

    qlads <= lads_i and not la_i(23) and not fholda;

    --
    -- Latch the local bus address on the 'lads_l' pulse.
    --

    latch_addr : process(rst, lclk)
    begin
        if rst = '1' then
            la_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
            end if;
        end if;
    end process;
    
    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';
    
    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --

    lready_l <= lready_o_l when lready_oe_l = '0' else 'Z';
    
    --
    -- Drive the local data bus on a read.
    --

    lad <= ld_o when ld_oe_l = '0' else (others => 'Z');
    
    --
    -- If the current cycle is a write, update the registers
    --		

    update_regs : process(lclk, rst)
    begin
        if rst = '1' then
            reg0 <= (others => '0');
            reg1 <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if ds_xfer = '1' and ds_write = '1' then
                for i in 0 to 3 loop
                    if lbe_i(i) = '1' then
                        reg0(8 * (i + 1) - 1 downto 8 * i) <= ld_i(8 * (i + 1) - 1 downto 8 * i);
                    end if;
                end loop;
            end if;

            if ds_xfer = '1' and ds_write = '1' then
                for i in 0 to 3 loop
                    if lbe_i(i + 4) = '1' then
                        reg1(8 * (i + 1) - 1 downto 8 * i) <= ld_i(8 * (i + 5) - 1 downto 8 * (i + 4));
                    end if;
                end loop;
            end if;
        end if;
    end process;

    reg0_rev <= reg0(3 downto 0) &
                reg0(7 downto 4) &
                reg0(11 downto 8) &
                reg0(15 downto 12) &
                reg0(19 downto 16) &
                reg0(23 downto 20) &
                reg0(27 downto 24) &
                reg0(31 downto 28);

    --
    -- Generate the 'ld_o' signal, outbound data for local bus
    --

    generate_ld_o : process(rst, lclk)
    begin
        if rst = '1' then
            ld_o <= (others => '0');
        elsif lclk'event and lclk = '1' then
            ld_o <= reg1 & reg0_rev;
        end if;
    end process;

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    dssm : plxdssm
        port map(
            clk => lclk,
            rst => rst,
            sr => logic0,
            qlads => qlads,
            lblast => lblast_i,
            lbterm => lbterm_i,
            lwrite => lwrite_i,
            eld_oe => open,
            ld_oe_l => ld_oe_l,
            lready_o_l  => lready_o_l,
            lready_oe_l => lready_oe_l,
            lbterm_o_l => lbterm_o_l,
            lbterm_oe_l => lbterm_oe_l,
            idle => open,
            transfer => ds_xfer,
            decode => ds_decode,
            write => ds_write,
            ready => logic1,
            stop => logic1);

    --
    -- Instantiate NULL_PAIR instances, to workaround errata related to
    -- Virtex-4 MGTs.
    --

    null_mgt_0 : null_mgt
        port map(
            clka => lclk,
            rxa_p => front_rx_p,
            rxa_n => front_rx_n,
            txa_p => front_tx_p,
            txa_n => front_tx_n,
            clkb => lclk,
            rxb_p => local_rx_p,
            rxb_n => local_rx_n,
            txb_p => local_tx_p,
            txb_n => local_tx_n);

end mixed;
