--
-- clocks_dll2.vhd - clock generator module for ZBT design, using
--                   Virtex/Virtex-E/Virtex-EM DLLs
--
--                   Generates 2 SSRAM clocks.
--
-- (c) Alpha Data Parallel Systems Ltd. 1999-2001
--
-- Example program for ADM-XRC/ADM-XRC-P
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity clocks is
    generic(
        num_clock     : in    natural);
    port(
        lreset_l      : in    std_logic;
        rst           : out   std_logic;
        lclk          : in    std_logic;
        mclk          : in    std_logic;
        clk           : out   std_logic;
        ramclko       : out   std_logic_vector(num_clock - 1 downto 0);
        ramclki       : in    std_logic_vector(num_clock - 1 downto 0);
        locked        : out   std_logic_vector(31 downto 0));
end clocks;

architecture synthesis of clocks is

    signal lreset         : std_logic;
    signal i_rst          : std_logic;

    signal lclk0          : std_logic;
    signal lclk_i0        : std_logic;
    signal locked_lclk    : std_logic;
    signal locked_ramclk  : std_logic_vector(num_clock - 1 downto 0);
    signal ramclk_i       : std_logic_vector(num_clock - 1 downto 0);
    signal i_clk          : std_logic;
    
    signal logic0         : std_logic;
    signal logic1         : std_logic;

    signal dllrst         : std_logic;
    signal rstcnt         : std_logic_vector(3 downto 0);
    signal rst_ramclk     : std_logic;

    signal mclk_ibufg     : std_logic;
    signal mclk_bufg      : std_logic;

    component CLKDLL 
        port(
            CLKIN         : in  std_logic; 
            CLKFB         : in  std_logic; 
            RST           : in  std_logic; 
            CLK0          : out std_logic;
            CLK90         : out std_logic;
            CLK180        : out std_logic;
            CLK270        : out std_logic;
            CLK2X         : out std_logic;
            CLKDV         : out std_logic;
            LOCKED        : out std_logic);
    end component;
    
    component IBUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    component BUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    component OBUF
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;    

    attribute DUTY_CYCLE_CORRECTION : string;
    attribute STARTUP_WAIT : string;

    attribute DUTY_CYCLE_CORRECTION of dll_lclk : label is "TRUE";
    attribute STARTUP_WAIT of dll_lclk : label is "TRUE";
    attribute DUTY_CYCLE_CORRECTION of dll_0 : label is "TRUE";
    attribute STARTUP_WAIT of dll_0 : label is "TRUE";
    attribute DUTY_CYCLE_CORRECTION of dll_1 : label is "TRUE";
    attribute STARTUP_WAIT of dll_1 : label is "TRUE";

begin
    
    logic0 <= '0';
    logic1 <= '1';

    clk    <= i_clk;
    rst    <= i_rst;
    locked <= EXT(locked_ramclk & locked_lclk, 32);
    
    --
    -- Use a global buffer to distribute the reset. This avoids the
    -- problem of PAR being unable to pack registers into IOBs because
    -- various different buffered copies of the reset signal are used.
    --

    lreset <= not lreset_l;
    ibufg_reset : BUFG
        port map(
            I => lreset,
            O => i_rst);
    
    --
    -- Input MCLK
    --

    ibufg_mclk : IBUFG
        port map(
            I => mclk,
            O => mclk_ibufg);
    
    bufg_mclk : BUFG
        port map(
            I => mclk_ibufg,
            O => mclk_bufg);
    
    --
    -- Generate reset signal to DLLs/DCMs
    --
    gen_dllrst : process(i_rst, mclk_bufg)
    begin
        if i_rst = '1' then
            dllrst <= '1';
            rstcnt <= (others => '0');
        elsif mclk_bufg'event and mclk_bufg = '1' then
            if dllrst = '1' then
                rstcnt <= rstcnt + 1;
            end if;
            if AND_reduce(rstcnt) = '1' then
                dllrst <= '0';
            end if;
        end if;
    end process;
    
    --
    -- Input the local bus clock
    --

    ibufg_lclk : IBUFG
        port map(
            I => lclk,
            O => lclk0);
    
    dll_lclk : CLKDLL
        port map(
            CLKIN  => lclk0, 
            CLKFB  => i_clk,
            RST    => dllrst,
            CLK0   => lclk_i0,
            LOCKED => locked_lclk);

    bufg_lclk : BUFG
        port map(
            I => lclk_i0,
            O => i_clk);
    
    --
    -- Generate the SSRAM clocks
    --
    
    rst_ramclk <= not locked_lclk;
    
    ibufg_0 : IBUFG
        port map(
            I => ramclki(0),
            O => ramclk_i(0));

    dll_0 : CLKDLL
        port map(
            CLKIN  => i_clk, 
            CLKFB  => ramclk_i(0),
            RST    => rst_ramclk,
            CLK0   => ramclko(0),
            LOCKED => locked_ramclk(0));

    ibufg_1 : IBUFG
        port map(
            I => ramclki(1),
            O => ramclk_i(1));

    dll_1 : CLKDLL
        port map(
            CLKIN  => i_clk, 
            CLKFB  => ramclk_i(1),
            RST    => rst_ramclk,
            CLK0   => ramclko(1),
            LOCKED => locked_ramclk(1));

end synthesis;
