--
-- clocks_xrc4_sb.vhd - clock generator module for ZBT design, using DCMs
--                      and DDR flip-flops. Uses DCM_STANDBY workaround for
--                      Virtex-4 DCMs.
--
--                      Generates 'num_clock' SSRAM clocks.
--
-- (c) Alpha Data 2005-2007
--
-- Example program for ADM-XRC4LX/ADM-XRC4SX
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity clocks is
    generic(
        num_clock     : in    natural);
    port(
        lreset_l      : in    std_logic;
        rst           : out   std_logic;
        lclk          : in    std_logic;
        mclk          : in    std_logic;
        clk           : out   std_logic;
        ramclko       : out   std_logic_vector(num_clock - 1 downto 0);
        ramclki       : in    std_logic_vector(num_clock - 1 downto 0);
        locked        : out   std_logic_vector(31 downto 0));
end clocks;

architecture synthesis of clocks is

    signal i_rst          : std_logic;

    signal lclk_ibufg     : std_logic;
    signal locked_lclk    : std_logic;
    signal locked_ramclk  : std_logic_vector(0 downto 0);
    signal ramclki_ibufg  : std_logic_vector(0 downto 0);
    signal ramclk0        : std_logic;
    signal ramclk0_bufg   : std_logic;
    signal ramclk180_bufg : std_logic;
    signal i_ramclko      : std_logic_vector(num_clock - 1 downto 0);
    signal clk0           : std_logic;
    signal i_clk          : std_logic;
    
    signal logic0         : std_logic;
    signal logic1         : std_logic;

    signal dllrst         : std_logic;
    signal rstcnt         : std_logic_vector(3 downto 0);
    signal rst_ramclk     : std_logic;

    signal mclk_bufg      : std_logic;

    component BUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    component DCM
        port(
            CLKIN         : in  std_logic;
            CLKFB         : in  std_logic;
            DSSEN         : in  std_logic;
            PSINCDEC      : in  std_logic;
            PSEN          : in  std_logic;
            PSCLK         : in  std_logic;
            RST           : in  std_logic;
            CLK0          : out std_logic;
            CLK90         : out std_logic;
            CLK180        : out std_logic;
            CLK270        : out std_logic;
            CLK2X         : out std_logic;
            CLK2X180      : out std_logic;
            CLKDV         : out std_logic;
            CLKFX         : out std_logic;
            CLKFX180      : out std_logic;
            LOCKED        : out std_logic;
            PSDONE        : out std_logic;
            STATUS        : out std_logic_vector(7 downto 0));
    end component;

    component DCM_STANDBY
        port(
            CLK0          : OUT std_logic;
            CLK180        : OUT std_logic;
            CLK270        : OUT std_logic;
            CLK2X         : OUT std_logic;
            CLK2X180      : OUT std_logic;
            CLK90         : OUT std_logic;
            CLKDV         : OUT std_logic;
            CLKFX         : OUT std_logic;
            CLKFX180      : OUT std_logic;
            LOCKED        : OUT std_logic;
            DO            : OUT std_logic_vector(15 DOWNTO 0);
            PSDONE        : OUT std_logic;
            CLKFB         : IN  std_logic;
            CLKIN         : IN  std_logic;
            PSCLK         : IN  std_logic;
            PSEN          : IN  std_logic;
            PSINCDEC      : IN  std_logic;
            RST           : IN  std_logic);
    end component;

    component FDDRRSE
        port(
            D0 : in  std_logic;
            D1 : in  std_logic;
            C0 : in  std_logic;
            C1 : in  std_logic;
            CE : in  std_logic;
            R  : in  std_logic;
            S  : in  std_logic;
            Q  : out std_logic);
    end component;

    component IBUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;
    
    attribute DLL_FREQUENCY_MODE : string;
    attribute DUTY_CYCLE_CORRECTION : string;
    attribute STARTUP_WAIT : string;
    attribute CLK_FEEDBACK : string;
    
--    attribute DLL_FREQUENCY_MODE of dll_lclk : label is "LOW";
--    attribute DUTY_CYCLE_CORRECTION of dll_lclk : label is "TRUE";
--    attribute STARTUP_WAIT of dll_lclk : label is "FALSE";
--    attribute DLL_FREQUENCY_MODE of dll_lclk : label is "LOW";
--    attribute DUTY_CYCLE_CORRECTION of dll_0 : label is "TRUE";
--    attribute STARTUP_WAIT of dll_0 : label is "FALSE";
--    attribute CLK_FEEDBACK of dll_0 : label is "1X";
--    attribute DLL_FREQUENCY_MODE of dll_0 : label is "LOW";
    
begin
    
    --
    -- Define constant values
    --
    logic0 <= '0';
    logic1 <= '1';

    --
    -- Output signals
    --
    clk    <= i_clk;
    rst    <= i_rst;
    locked <= EXT(locked_ramclk & locked_lclk, 32);
    ramclko <= i_ramclko;

    i_rst  <= not lreset_l;
    
    --
    -- Input MCLK
    --

    bufg_mclk : BUFG
        port map(
            I => mclk,
            O => mclk_bufg);
    
    --
    -- Generate reset signal to DLLs/DCMs
    --
    gen_dllrst : process(i_rst, mclk_bufg)
    begin
        if i_rst = '1' then
            dllrst <= '1';
            rstcnt <= (others => '0');
        elsif mclk_bufg'event and mclk_bufg = '1' then
            if dllrst = '1' then
                rstcnt <= rstcnt + 1;
            end if;
            if AND_reduce(rstcnt) = '1' then
                dllrst <= '0';
            end if;
        end if;
    end process;
    
    --
    -- Input the local bus clock
    --

    ibufg_lclk : IBUFG
        port map(
            I => lclk,
            O => lclk_ibufg);

    dll_lclk : DCM_STANDBY
        port map(
            CLKIN    => lclk_ibufg,
            CLKFB    => i_clk,
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => dllrst,
            CLK0     => clk0,
            LOCKED   => locked_lclk);

    i_clk <= clk0;

    --
    -- Generate the SSRAM clocks
    --
    
    rst_ramclk <= not locked_lclk;
    
    ibufg_0 : IBUFG
        port map(
            I => ramclki(0),
            O => ramclki_ibufg(0));
            
    dll_0 : DCM_STANDBY
        port map(
            CLKIN    => i_clk,
            CLKFB    => ramclki_ibufg(0),
            PSINCDEC => logic0,
            PSEN     => logic0,
            PSCLK    => logic0,
            RST      => rst_ramclk,
            CLK0     => ramclk0,
            LOCKED   => locked_ramclk(0));

    bufg_ramclk0 : BUFG
        port map(
            I => ramclk0,
            O => ramclk0_bufg);

    ramclk180_bufg <= not ramclk0_bufg;

    gen_i_ramclko : for i in 0 to num_clock - 1 generate
        u0 : fddrrse
            port map(
                D0 => logic1,
                D1 => logic0,
                C0 => ramclk0_bufg,
                C1 => ramclk180_bufg,
                CE => logic1,
                R => logic0,
                S => logic0,
                Q => i_ramclko(i));
    end generate;

end synthesis;
