--
-- zbt_port.vhd - ZBT interface and IO pins
--
-- (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2003
--
-- Example program for: ADM-XPL
--
-- Module containing pins connected to a bank of ZBT SSRAM.
--
-- The 'pipeline' input selects flowthrough ZBT SSRAM mode ('0') or
-- pipelined ZBT SSRAM mode ('1').
-- 
-- The latency from 'write' and 'd' to 'rd_o' and 't_rd' valid is 3 clocks,
-- in both pipelined and flowthrough mode.
--
-- The latency from the 'read', 'write' and 'a' signals to 'ra' and 'rc'
-- valid is 1 clock in pipelined mode and 2 clocks in flowthrough mode.
--
-- The latency from 'read' to 'q' valid is 4 clocks in both pipelined and
-- flowthrough mode.
--
-- 'evalid' is an early valid signal, generated one cycle before 'q'
-- becomes valid. The latency from 'read' to 'evalid' asserted is therefore
-- 3 clocks in both pipelined and flowthrough mode.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity zbt_port is
    generic(
        addr_width : in    natural;   -- Number of address bits on SSRAM bank
        data_width : in    natural;   -- Number of data bits on SSRAM bank
        ctl_width  : in    natural;   -- Number of control bits on SSRAM bank
        bank_group : in    natural);  -- Number of physical SSRAM banks to combine to make a logical bank
    port(
        clk        : in    std_logic;                                               -- Clock
        rst        : in    std_logic;                                               -- Asychronous reset
        sr         : in    std_logic;                                               -- Synchronous reset
        pipeline   : in    std_logic;                                               -- Operating mode, see above
        read       : in    std_logic;                                               -- Read command
        write      : in    std_logic;                                               -- Write command
        a          : in    std_logic_vector(addr_width - 1 downto 0);               -- Address
        d          : in    std_logic_vector(63 downto 0);                           -- Data
        be         : in    std_logic_vector(7 downto 0);                            -- Byte enables
        q          : out   std_logic_vector(63 downto 0);                           -- Data read from SSRAM
        evalid     : out   std_logic;                                               -- Qualifies 'q' one cycle early
        t_ra       : out   std_logic_vector(addr_width * bank_group - 1 downto 0);  -- Tristate control for address pins
        ra_o       : out   std_logic_vector(addr_width * bank_group - 1 downto 0);  -- To SSRAM address pins
        t_rc       : out   std_logic_vector(ctl_width * bank_group - 1 downto 0);   -- Tristate control for SSRAM control pins
        rc_o       : out   std_logic_vector(ctl_width * bank_group - 1 downto 0);   -- To SSRAM control pins
        t_rd       : out   std_logic_vector(data_width * bank_group - 1 downto 0);  -- Tristate control for data pins
        rd_o       : out   std_logic_vector(data_width * bank_group - 1 downto 0);  -- To SSRAM data pins
        rd_i       : in    std_logic_vector(data_width * bank_group - 1 downto 0)); -- From SSRAM data pins
end zbt_port;

architecture mixed of zbt_port is

    constant be_width    : natural := data_width / 8;

    type state_type1 is (s1_idle, s1_read0, s1_read1);
    signal state1, n_state1 : state_type1;
    signal i_evalid, n_evalid  : std_logic;

    signal a_q     : std_logic_vector(addr_width - 1 downto 0);
    signal read_q  : std_logic_vector(1 downto 0);
    signal write_q : std_logic;

    signal be_q    : std_logic_vector(7 downto 0);
    type d_q_type is array(2 downto 0) of std_logic_vector(data_width * bank_group - 1 downto 0);
    signal d_q     : d_q_type;
    signal oe_rd   : std_logic_vector(1 downto 0);
    signal ce_q    : std_logic_vector(2 downto 0);

    --
    -- XST synthesis constraints, required to prevent optimization of
    -- flip-flops that would otherwise prevent IOB packing.
    --
    attribute equivalent_register_removal : string;
    attribute equivalent_register_removal of ra_o : signal is "no";
    attribute equivalent_register_removal of rc_o : signal is "no";
    attribute equivalent_register_removal of rd_o : signal is "no";
    attribute equivalent_register_removal of rd_i : signal is "no";
    attribute iob : string;
    attribute iob of t_rd : signal is "true";

begin

    assert (data_width * bank_group) = 64
        report "*** zbt_port: generic 'bank_group' or 'data_width' is invalid"
        severity failure;
    assert ctl_width = 15
        report "*** zbt_port: generic 'ctl_width' is invalid"
        severity failure;

    --
    -- The SSRAM address and control busses are always enabled onto the FPGA pins
    --
    t_ra   <= (others => '0');
    t_rc   <= (others => '0');
    rd_o   <= d_q(2);
    evalid <= i_evalid;

    --
    -- Drive 'rd' (SSRAM data bus) on writes
    --
    data_out: process(rst, clk)
    begin
        if rst = '1' then
            a_q     <= (others => '0');
            be_q    <= (others => '0');
            d_q     <= (others => (others => '0'));
            oe_rd   <= (others => '0');
            t_rd    <= (others => '1');
            read_q  <= (others => '0');
            write_q <= '0';
        elsif clk'event and clk = '1' then
            if sr = '1' then
                a_q     <= (others => '0');
                be_q    <= (others => '0');
                d_q     <= (others => (others => '0'));
                oe_rd   <= (others => '0');
                t_rd    <= (others => '1');
                read_q  <= (others => '0');
                write_q <= '0';
            else
                a_q     <= a;
                be_q    <= be;
                d_q     <= d_q(1 downto 0) & d;
                oe_rd   <= oe_rd(0 downto 0) & write;
                t_rd    <= (others => not oe_rd(1));
                read_q  <= read_q(0) & read;
                write_q <= write;
            end if;
        end if;
    end process;

    --
    -- Generate 'q' (read SSRAM data bus on reads)
    --
    data_in: process(rst, clk)
    begin
        if rst = '1' then
            ce_q <= (others => '0');
            q <= (others => '0');
        elsif clk'event and clk = '1' then
            if sr = '1' then
                ce_q <= (others => '0');
                q <= (others => '0');
            else
                ce_q <= ce_q(1 downto 0) & read;
                if ce_q(2) = '1' then
                    q <= rd_i;
                end if;
            end if;
        end if;
    end process;

    --
    -- Drive 'ra' (SSRAM address bus) and 'rc' (SSRAM control bus)
    --
    -- 'rc' pin designations:
    --
    --    bit   ADM-XPL
    --
    --      0   BWE0#
    --      1   BWE1#
    --      2   BWE2#
    --      3   BWE3#
    --      4   BWE4#
    --      5   BWE5#
    --      6   BWE6#
    --      7   BWE7#
    --      8   WE#
    --      9   CE#
    --     10   CE2#
    --     11   CE2
    --     12   ADV/LD#
    --     13   OE#
    --     14   CKE#
    --
    control : process(clk, rst)
        variable addr_out: std_logic_vector(addr_width - 1 downto 0);
        variable ctl_out:  std_logic_vector(ctl_width - 1 downto 0);
        variable be_out:  std_logic_vector(7 downto 0);
    begin
        if rst = '1' then
            ra_o <= (others => '0');
            rc_o <= (others => '1');
        elsif clk'event and clk = '1' then
            if sr = '1' then
                ra_o <= (others => '0');
                rc_o <= (others => '1');
            else
                for i in 0 to bank_group - 1 loop
                    if pipeline = '0' then
                        addr_out := a_q;

                        if write_q = '1' then
                            be_out := be_q;
                        else
                            be_out := (others => '0');
                        end if;

                        case ctl_width is
                            when 15 =>
                                -- ADM-XPL
                                ctl_out(14)           := '0';
                                ctl_out(13)           := not read_q(1);
                                ctl_out(12 downto 10) := "010";
                                ctl_out(9)            := not (read_q(0) or write_q);
                                ctl_out(8)            := not write_q;
                                ctl_out(7 downto 0)   := not be_out;

                            when others =>
                                null;
                        end case;
                    else
                        addr_out := a;

                        if write = '1' then
                            be_out := be;
                        else
                            be_out := (others => '0');
                        end if;

                        case ctl_width is
                            when 15 =>
                                -- ADM-XPL
                                ctl_out(14)           := '0';
                                ctl_out(13)           := not read_q(1);
                                ctl_out(12 downto 10) := "010";
                                ctl_out(9)            := not (read or write);
                                ctl_out(8)            := not write;
                                ctl_out(7 downto 0)   := not be_out;

                            when others =>
                                null;
                        end case;
                    end if;

                    ra_o(addr_width * (i + 1) - 1 downto addr_width * i) <= addr_out;
                    rc_o(ctl_width  * (i + 1) - 1 downto ctl_width  * i) <= ctl_out;
                end loop;

            end if;
        end if;
    end process;

    transition1: process(rst, clk)
    begin
        if rst = '1' then
            state1   <= s1_idle;
            i_evalid <= '0';
        elsif clk'event and clk = '1' then
            if sr = '1' then
                state1   <= s1_idle;
                i_evalid <= '0';
            else
                state1   <= n_state1;
                i_evalid <= n_evalid;
            end if;
        end if;
    end process;

    nextstate1: process(
        state1,
        read)
    begin
        case state1 is
            when s1_idle =>
                if read = '1' then
                    n_state1 <= s1_read0;
                    n_evalid <= '0';
                else
                    n_state1 <= s1_idle;
                    n_evalid <= '0';
                end if;

            when s1_read0 =>
                if read = '1' then
                    n_state1 <= s1_read1;
                    n_evalid <= '0';
                else
                    n_state1 <= s1_idle;
                    n_evalid <= '0';
                end if;

            when s1_read1 =>
                if read = '1' then
                    n_state1 <= s1_read1;
                    n_evalid <= '1';
                else
                    n_state1 <= s1_idle;
                    n_evalid <= '0';
                end if;
        end case;
    end process;

end mixed;
